unit LActions;

{*******************************************************}
{                                                       }
{       Light Windows                                   }
{       Action Logic                                    }
{                                                       }
{       1998..2004 YT                                   }
{                                                       }
{*******************************************************}

interface

uses
	SysUtils, SysConst,
	SysHacks,
	LObjects, LConsts;

type
	{ɂȂƂɎs}
	T0IdleWorker = class(T0Object, IObserver)
	private
		FIdentification: Cardinal;
		FGlobal: Boolean;
		procedure IObserver.Observe = Update;
	protected
		procedure Modify;
	public
		constructor Create(AGlobal: Boolean);
		procedure BeforeDestruction; override;
		procedure Update; virtual; abstract;
		property Global: Boolean read FGlobal;
		property Identification: Cardinal read FIdentification;
	end;

type
	T0Action = class;
	TActionItems = class;

	IActionItemData = interface
		['{7E7D36AE-B2DE-4ED9-A777-8FACE5C9B55A}']
		function GetData: Pointer;
		property Data: Pointer read GetData;
	end;

	TActionItem = class(T0CollectionItem, IActionItem, IActionItemData)
	private
		FText: WideString;
		FData: Pointer;
		FItems: TActionItems;
		FChecked: Boolean;
		function GetText: WideString;
		function GetChecked: Boolean;
		{IActionItemData}
		function GetData: Pointer;
	protected
		function GetItems: IActionItems; virtual;
	public
		constructor Create(ACollection: TActionItems; ANext: TActionItem);
		procedure BeforeDestruction; override;
		function Execute: Boolean;
		function ReadyItems: TActionItems;
		property Text: WideString read FText write FText;
		property Data: Pointer read FData write FData;
		property Checked: Boolean read FChecked write FChecked;
		property Items: IActionItems read GetItems;
	end;

	TActionItems = class(T0Collection, IActionItems)
	private
		FAction: T0Action;
		FTitle: WideString;
		FNeedUpdate: Boolean;
		function GetTitle: WideString;
	protected
		procedure Update;
	public
		constructor Create(AAction: T0Action);
		function Enumerate(Proc: CEachActionItem; Context: Pointer): IActionItem;
		procedure PostUpdate;
		property Action: T0Action read FAction;
		property Title: WideString read GetTitle write FTitle;
	end;

	T0Action = class(T0IdleWorker, IAction)
	private
		FName: String;
		FText: WideString;
		FHint: WideString;
		FEnabled: Boolean;
		FChecked: Boolean;
		function GetName: string;
		function GetText: WideString;
		function GetHint: WideString;
		function GetEnabled: Boolean;
		function GetChecked: Boolean;
	protected
		procedure DoExecute; virtual;
		procedure SetName(const Value: string);
		procedure SetText(const Value: WideString);
		procedure SetHint(const Value: WideString);
		procedure SetEnabled(Value: Boolean);
		procedure SetChecked(Value: Boolean);
		function GetItems: IActionItems; virtual;
		procedure DoUpdateItems(AItems: TActionItems); virtual;
		procedure DoExecuteItem(AItem: TActionItem); virtual;
	public
		constructor Create(AGlobal: Boolean);
		function Execute: Boolean;
		property Name: String read FName;
		property Text: WideString read FText;
		property Hint: WideString read FHint;
		property Enabled: Boolean read FEnabled;
		property Checked: Boolean read FChecked;
		property Items: IActionItems read GetItems;
	end;

	{sOɈϏANV}

	TTransferAction = class;
	TTransferOrder = (toSetup, toUpdate, toExecute);

	TActionHandler = procedure(Action: TTransferAction; Order: TTransferOrder) of object;

	TTransferAction = class(T0Action)
	private
		FHandler: TActionHandler;
	protected
		procedure DoExecute; override;
	public
		constructor Create(AGlobal: Boolean; AHandler: TActionHandler; Dummy: Pointer);
		procedure Update; override;
		procedure Setup(const AName: string; const AText, AHint: WideString);
		property Name write SetName;
		property Text write SetText;
		property Hint write SetHint;
		property Enabled write SetEnabled;
		property Checked write SetChecked;
	end;

type
	{ANV𑩂˂}
	T0ActionManager = class(T0Object, IActionManager)
	private
		FActions: array of IAction;
		FGlobal: Boolean;
		function GetActions(const AName: string): IAction;
		procedure Detach(const AAction: IAction);
	protected
		function Select(const AItems: IActionItems): IActionItem; virtual;
	public
		constructor Create(AGlobal: Boolean);
		function Add(AProc: TActionHandler): TTransferAction; overload;
		procedure Add(const AAction: IAction); overload;
		procedure IActionManager.Attach = Add;
		property Actions[const AName: string]: IAction read GetActions; default;
		property Global: Boolean read FGlobal;
	end;

type
	{OwnerɂĂΉɂȂƂŔjĂ}
	TGarbage = class(T0Object)
	private
		FLastProtected: T0Object;
	public
		procedure Collect;
		function Protect: T0Object;
		procedure Unprotect(Store: T0Object);
	end;

threadvar
	Garbage: TGarbage;

implementation

function GetObservable(Global: Boolean): IObservable;
var
	S: Pointer;
begin
	if Global then
		S := Pointer(Application)
	else
		S := Pointer(CurrentThread);

	Result := IInterface(S) as IObservable
end;

{ T0IdleWorker }

constructor T0IdleWorker.Create(AGlobal: Boolean);
begin
	inherited Create;
	FGlobal := AGlobal;
	GetObservable(Global).Attach(Self)
end;

procedure T0IdleWorker.BeforeDestruction;
begin
	GetObservable(Global).Detach(Self);
	inherited;
end;

procedure T0IdleWorker.Modify;
begin
	Inc(FIdentification);
end;

{ TActionItem }

procedure TActionItem.BeforeDestruction;
begin
	FItems.Free;
	inherited;
end;

constructor TActionItem.Create(ACollection: TActionItems; ANext: TActionItem);
begin
	inherited Create(ACollection, ANext)
end;

function TActionItem.Execute: Boolean;
var
	OwnerItems: TActionItems;
begin
	Assert(Owner is TActionItems);

	OwnerItems := TActionItems(Owner);

	if OwnerItems.Action <> nil then
		OwnerItems.Action.DoExecuteItem(Self);

	Result := True
end;

function TActionItem.GetChecked: Boolean;
begin
	Result := FChecked
end;

function TActionItem.GetData: Pointer;
begin
	Result := FData
end;

function TActionItem.GetItems: IActionItems;
begin
	Result := FItems
end;

function TActionItem.GetText: WideString;
begin
	Result := FText
end;

function TActionItem.ReadyItems: TActionItems;
begin
	{͒PȂ֗֐Ő[Ӗ͖ł}

	Assert(Owner is TActionItems);

	if FItems = nil then
		FItems := TActionItems.Create(TActionItems(Owner).Action);

	Result := FItems;
end;

{ TActionItems }

constructor TActionItems.Create(AAction: T0Action);
begin
	inherited Create;
	FAction := AAction;
	FNeedUpdate := True;
end;

function TActionItems.Enumerate(Proc: CEachActionItem; Context: Pointer): IActionItem;

	function EachItem(Item: T0Object): Boolean; cdecl;
	begin
		Assert(Item is TActionItem);
		Result := Proc(TActionItem(Item), Context)
	end;

begin
	Update;
	Result := TActionItem(
		inherited Enumerate(TActionItem, CEachObject(@EachItem), GetContext))
end;

function TActionItems.GetTitle: WideString;
begin
	Update;
	Result := FTitle
end;

procedure TActionItems.PostUpdate;
begin
  FNeedUpdate := True
end;

procedure TActionItems.Update;
begin
	if FNeedUpdate then
	begin
		FNeedUpdate := False; {ɂĂȂƁōċA}
		if Action <> nil then
			Action.DoUpdateItems(Self);
	end;
end;

{ T0Action }

constructor T0Action.Create(AGlobal: Boolean);
begin
	inherited Create(AGlobal);
	FEnabled := True;
end;

procedure T0Action.DoExecute;
var
	TheItems: IActionItems;
	Selected: IActionItem;
begin
	TheItems := GetItems;
	if TheItems <> nil then
	begin
		(* x: O[oANVSelectoro͍ڂ͓IɕωĂ͂Ȃ *)

		{TheItems.Update; //SelectC}
		Selected := (Application as IActionManager).Select(TheItems);
		if Selected <> nil then
			Selected.Execute
	end
end;

procedure T0Action.DoExecuteItem(AItem: TActionItem);
begin
end;

procedure T0Action.DoUpdateItems(AItems: TActionItems);
begin
end;

function T0Action.Execute: Boolean;

	procedure SyncUpdate;
	var
		Scheduler: IReadWriteSync;
	begin
		Scheduler := Application as IReadWriteSync;
		Scheduler.BeginWrite;
		try
			Update;
		finally
			Scheduler.EndWrite;
		end;
	end;

begin
	if Global then
		SyncUpdate
	else
		Update;

	Result := Enabled;
	if Result then DoExecute
end;

function T0Action.GetChecked: Boolean;
begin
	Result := FChecked
end;

function T0Action.GetEnabled: Boolean;
begin
	Result := FEnabled
end;

function T0Action.GetHint: WideString;
begin
	Result := FHint
end;

function T0Action.GetItems: IActionItems;
begin
	Result := nil
end;

function T0Action.GetName: string;
begin
	Result := FName
end;

function T0Action.GetText: WideString;
begin
	Result := FText
end;

procedure T0Action.SetChecked(Value: Boolean);
begin
	if FChecked <> Value then
	begin
		FChecked := Value;
		Modify
	end
end;

procedure T0Action.SetEnabled(Value: Boolean);
begin
	if FEnabled <> Value then
	begin
		FEnabled := Value;
		Modify
	end
end;

procedure T0Action.SetHint(const Value: WideString);
begin
	if FHint <> Value then
	begin
		FHint := Value;
		Modify
	end
end;

procedure T0Action.SetName(const Value: string);
begin
	Assert(FName = '');
	FName := Value;
end;

procedure T0Action.SetText(const Value: WideString);
begin
	if FText <> Value then
	begin
		FText := Value;
		Modify
	end
end;

{ TTransferAction }

constructor TTransferAction.
				Create(AGlobal: Boolean; AHandler: TActionHandler; Dummy: Pointer);
begin
	inherited Create(AGlobal);
	FHandler := AHandler;
	FHandler(Self, toSetup);
end;

procedure TTransferAction.DoExecute;
begin
	FHandler(Self, toExecute);
end;

procedure TTransferAction.Setup(const AName: string; const AText, AHint: WideString);
begin
	FName := AName;
	FText := AText;
	FHint := AHint;

	Modify;
end;

procedure TTransferAction.Update;
begin
	FHandler(Self, toUpdate);
end;

{ T0ActionManager }

constructor T0ActionManager.Create(AGlobal: Boolean);
begin
	inherited Create;
	FGlobal := AGlobal;
end;

function T0ActionManager.Add(AProc: TActionHandler): TTransferAction;
begin
	Result := TTransferAction.Create(Global, AProc, nil);
	Result.Owner := Self
end;

procedure T0ActionManager.Add(const AAction: IAction);
begin
	SetLength(FActions, Length(FActions) + 1);
	FActions[High(FActions)] := AAction;
end;

function T0ActionManager.GetActions(const AName: string): IAction;

	function Search(const AName: string): IAction;

		function SearchProc(Obj: T0Object): Boolean cdecl;
		begin
			Result := (Obj is T0Action) and
				(CompareText(T0Action(Obj).Name, AName) = 0)
		end;

	var
		I: Integer;
	begin
		for I := High(FActions) downto Low(FActions) do
			if CompareText(FActions[I].Name, AName) = 0 then
			begin
				Result := FActions[I];
				Exit;
			end;

		Result := T0Action(EnumerateObjects(CEachObject(@SearchProc), GetContext));
	end;

	function SearchGlobalAction(const A: string): IAction;
	var
		Man: IActionManager;
	begin
		Result := nil;
		if Supports(Application, IActionManager, Man) then
			Result := Man.Actions[A]
	end;

var
	Scheduler: IReadWriteSync;
begin
	if not Global then
	begin
		Result := Search(AName);
		if Result = nil then
			Result := SearchGlobalAction(AName)
	end
	else
	begin
		Scheduler := Application as IReadWriteSync;
		Scheduler.BeginRead;
		try
			Result := Search(AName)
		finally
      Scheduler.EndRead;
		end
	end;
	if Result = nil then
		raise EInvalidContainer.CreateFmt(SActionNotFound, [AName]);
end;

procedure T0ActionManager.Detach(const AAction: IAction);
var
	I, J: Integer;
begin
	for I := High(FActions) downto Low(FActions) do
		if FActions[I] = AAction then
		begin
			for J := I + 1 to High(FActions) do
				FActions[J - 1] := FActions[J];

			SetLength(FActions, Length(FActions) - 1);

			Exit;
		end;
end;

function T0ActionManager.Select(const AItems: IActionItems): IActionItem;
asm
	call System.@AbstractError
end;

{ TGarbage }

function TGarbage.Protect: T0Object;
begin
	Result := FLastProtected;
	FLastProtected := LastObject
end;

procedure TGarbage.Unprotect(Store: T0Object);
begin
	FLastProtected := Store
end;

procedure TGarbage.Collect;
var
	Deletable: Boolean;

	function EnumProc(Obj: T0Object): Boolean cdecl;
	begin
		if Deletable then
			Obj.Free
		else if Obj = FLastProtected then
			Deletable := True;

		Result := False;
	end;

begin
	Deletable := FLastProtected = nil;
	EnumerateObjects(CEachObject(@EnumProc), GetContext);
end;

end.

