unit LAddCtls;

{*******************************************************}
{                                                       }
{       LightWindows                                    }
{       Additional Controls                             }
{                                                       }
{       2001..2004 YT                                   }
{                                                       }
{*******************************************************}

(*
	LightListBox
		I[i[h[p̌yʃXg
	LightSwitch
		FSIXCb`
	SplitPane/SplitWindow
		EBhE
	LightButton
		{^
	Area
		PȂ̈
	Splitter
		
*)

interface

uses
	Windows, Messages, SysUtils, Types, CommCtrl,
	SysUtilsW, AlgUtils, WinTips,
	LObjects, LGraphics, LWindows, LActions, LStdCtls;

type
	TSimpleControlItem = class(T0ControlItem)
	private
		FText: WideString;
		FData: Pointer;
	protected
		function GetData: Pointer; override;
		procedure SetData(const Value: Pointer); override;
		function GetText: WideString; override;
		procedure SetText(const Value: WideString); override;
	end;

type
	TLightListBox = class(T0Window)
	public type
		TViewOption = (loEraseBkgnd, loDrawFocusRect);
		TViewOptions = set of TViewOption;
		TControlOption = (loMultiSelect, loAutoSelect);
		TControlOptions = set of TControlOption;
		TScrollBase = (lsbTop, lsbBottom);
	private const
		LLBM_RESETSCROLLBAR = WM_USER + 1;
	private
		FVertScroll: SCROLLINFO;
		FScrollBarManager: PScrollBarManager;
		FScrollFrom: Integer;
		FScrollBase: TScrollBase;
		FItemCount: Natural;
		FCaretIndex: Integer;
		FItemHeight: Integer;
		FAnchorIndex: Integer;
		FSelected: PByteArray;
		FSelectCount: Natural;
		FViewOptions: TViewOptions;
		FCtrlOptions: TControlOptions;
		procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
		procedure WMKillFocus(var Message: TWMKillFocus); message WM_KILLFOCUS;
		procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
		procedure WMLButtonUp(var Message: TWMLButtonUp); message WM_LBUTTONUP;
		procedure WMRButtonDown(var Message: TWMRButtonDown); message WM_RBUTTONDOWN;
		procedure WMVScroll(var Message: TWMScroll); message WM_VSCROLL;
		procedure WMSize(var Message: TWMSize); message WM_SIZE;
		procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
		procedure WMMouseWheel(var Message: TWMMouseWheel); message WM_MOUSEWHEEL;
		procedure LLBMResetScrollBar(var Message: TMessage); message LLBM_RESETSCROLLBAR;
		procedure SetViewOptions(Value: TViewOptions);
		procedure SetCtrlOptions(Value: TControlOptions);
		procedure SetItemCount(const Value: Natural);
		procedure SetCaretIndex(Value: Integer);
		procedure SetItemHeight(Value: Integer);
		function GetSelected(Index: Integer): Boolean;
		procedure SetSelected(Index: Integer; Value: Boolean);
	protected
		procedure Initialize; override;
		procedure PaintWindow(DC: HDC); override;
		procedure DrawItem(Index: Integer; DC: HDC; const Rect: TRect); virtual;
		function CanSelectItem(Index: Integer): Boolean; virtual;
		procedure DoSelectItem(Index: Integer; Value: Boolean); virtual;
		function KeyDown(ShiftState: TShiftKeys; Code: Byte): Boolean; override;
		procedure RepaintScroll(Y: Integer; const Clip: TRect);
		property ScrollBarManager: PScrollBarManager read FScrollBarManager write FScrollBarManager;
	public
		constructor Create(AParent: TAnyWindow; const APosition: TPoint; const ASize: TSize);
		procedure BeforeDestruction; override;
		procedure Move(Step: Integer; Group, Auto: Boolean);
		function ItemRect(AIndex: Integer): TRect;
		function ItemAtPos(Pos: TPoint): Integer;
		procedure SelectAll(ToSelect: Boolean);
		procedure Scroll(Base: TScrollBase; From: Integer); overload;
		procedure Scroll; overload;
		procedure ScrollUp;
		procedure ScrollDown;
		property ScrollBase: TScrollBase read FScrollBase;
		property ScrollFrom: Integer read FScrollFrom;
		property ItemCount: Natural read FItemCount write SetItemCount;
		property Selected[AIndex: Integer]: Boolean read GetSelected write SetSelected;
		property SelectCount: Natural read FSelectCount;
		property CaretIndex: Integer read FCaretIndex write SetCaretIndex;
		property ItemHeight: Integer read FItemHeight write SetItemHeight;
		property ViewOptions: TViewOptions read FViewOptions write SetViewOptions;
		property CtrlOptions: TControlOptions read FCtrlOptions write SetCtrlOptions;
		property Color;
	end;

type
	TLightSwitch = class(T0Window)
	private
		FItems: array of TSimpleControlItem;
		FItemWidth: Integer;
		FTextWidth: Integer;
		FSelected: T0ControlItem;
		procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
		procedure WMKillFocus(var Message: TWMKillFocus); message WM_KILLFOCUS;
		procedure WMLButtonDown(var Message: TWMLButtonDown);
						message WM_LBUTTONDOWN;
		procedure SetItemWidth(const Value: Integer);
		procedure SetTextWidth(const Value: Integer);
		function GetAt(AIndex: Integer): T0ControlItem;
	protected
		procedure Initialize; override;
		procedure PaintWindow(DC: HDC); override;
		procedure Finalize; override;
		function KeyDown(ShiftState: TShiftKeys; Code: Byte): Boolean; override;
	public
		property TextWidth: Integer read FTextWidth write SetTextWidth;
		property ItemWidth: Integer read FItemWidth write SetItemWidth;
		procedure Clear;
		function Add: T0ControlItem;
		//function Enumerator: IControlItemEnumerator;
		property Selected: T0ControlItem read FSelected write FSelected;
		function Count: Integer;
		property At[AIndex: Integer]: T0ControlItem read GetAt;
	end;

type
	T0SplitWindow = class;

	T0SplitPane = class(T0Window)
	private
		FHorzScrollBar: TScrollBar;
		FVertScrollBar: TScrollBar;
		function GetHost: T0SplitWindow;
	protected
		class procedure CreateClass(var WndClass: WNDCLASSEXW); override;
		property HorzScrollBar: TScrollBar read FHorzScrollBar;
		property VertScrollBar: TScrollBar read FVertScrollBar;
	public
		constructor Create(AParent: TAnyWindow; AHorz, AVert: TScrollBar);
		procedure BeforeDestruction; override;
		property Host: T0SplitWindow read GetHost;
	end;

	TSplitKind = TBarStyle;
	TSplitKinds = TBarStyles;
	TSplitPanes = array[0..1, 0..1] of T0SplitPane;

	T0SplitWindow = class(T0Window)
	private
		FPanes: TSplitPanes;
		FScrollBars: array[TBarStyle, 0..1] of TScrollBar;
		FSplitPer: array[TBarStyle] of Real;
		FScrollWidth: TSize;
		FSplitWidth: Integer;
		FBorderWidth: Integer;
		FSplitingDiff: Integer;
		FSplitingPos: Integer;
		FSplitingMax: Integer;
		FSplitable: TBarStyles;
		FSplited: TBarStyles;
		FSpliting: Boolean;
		FSplitingDir: TBarStyle;
		procedure WMNCCalcSize(var Message: TWMNCCalcSize); message WM_NCCALCSIZE;
		procedure WMNCPaint(var Message: TWMNCPaint); message WM_NCPAINT;
		procedure WMNCHitTest(var Message: TWMNCHitTest); message WM_NCHITTEST;
		procedure WMNCLButtonDown(var Message: TWMNCHitMessage); message WM_NCLBUTTONDOWN;
		procedure WMLButtonUp(var Message: TWMLButtonUp); message WM_LBUTTONUP;
		procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
		procedure WMSetCursor(var Message: TWMSetCursor); message WM_SETCURSOR;
		procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
		procedure WMVScroll(var Message: TWMScroll); message WM_VSCROLL;
		procedure WMHScroll(var Message: TWMScroll); message WM_HSCROLL;
		procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
		procedure WMLightWindows(var Message: TMessage); message WM_LIGHTWINDOWS;
	protected
		procedure PaintWindow(DC: HDC); override;
		procedure Initialize; override;
		function CreatePane(AHorz, AVert: TScrollBar): T0SplitPane; virtual;
		function CreateScrollBar(Dir: TBarStyle): TScrollBar;
		function SplitBox(Kind: TBarStyle): TRect;
		procedure DrawSpliting;
		procedure BeginSpliting(Dir: TBarStyle);
		procedure EndSpliting(Accept: Boolean);
		procedure MoveSpliting(Step: Integer);
		procedure DoSplit(Dir: TBarStyle; Per: Real); virtual;
		function KeyDown(ShiftState: TShiftKeys; Code: Byte): Boolean; override;
		property Panes: TSplitPanes read FPanes;
		property Splited: TBarStyles read FSplited;
		property Splitting: Boolean read FSpliting;
	public
		constructor Create(AParent: TAnyWindow;
						const APosition: TPoint; const ASize: TSize;
						const ASplitable: TBarStyles; const AHasScrollBars: TBarStyles;
						const AHasBorder: Boolean);
		procedure Realign; override;
		function CanFocus: Boolean; override;
	end;

type
	TLightButton = class(T0Window)
	private
		FHovering: Boolean;
		FTextColor: TColor;
		procedure WMMouseHover(var Message); message WM_MOUSEHOVER;
		procedure WMMouseLeave(var Message); message WM_MOUSELEAVE;
		procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
		procedure WMLButtonDown(var Message: TWMMouse); message WM_LBUTTONDOWN;
		procedure WMLButtonUp(var Message: TWMMouse); message WM_LBUTTONUP;
		procedure WMEnable(var Message: TWMEnable); message WM_ENABLE;
	protected
		procedure PaintWindow(DC: HDC); override;
	public
		constructor Create(AParent: TAnyWindow; const AText: WideString; const APosition: TPoint; const ASize: TSize);
		procedure AfterConstruction; override;
		property TextColor: TColor read FTextColor write FTextColor;
	end;

type
	TArea = class(T0Window)
	private
		procedure WMLButtonUp(var Message: TWMMouse); message WM_LBUTTONUP;
		procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
	protected
		procedure ColorChanged; override;
	public
		constructor Create(AParent: TAnyWindow;
			const APosition: TPoint; const ASize: TSize; Border: Boolean = False);
		property Color;
	end;

type
	TSplitter = class(T0Window)
	private
		FMousePos: TPoint;
		procedure WMSetCursor(var Message: TWMSetCursor); message WM_SETCURSOR;
		procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
		procedure WMLButtonDown(var Message: TWMMouse); message WM_LBUTTONDOWN;
		procedure WMLButtonUp(var Message: TWMMouse); message WM_LBUTTONUP;
	protected
		class function GetDefaultSize: tagSIZE; override;
	public
		constructor Create(AParent: TAnyWindow;
			const APosition: TPoint; const ASize: TSize; AAlign: TAlignment);
	end;

implementation

{ TSimpleControlItem }

function TSimpleControlItem.GetData: Pointer;
begin
	Result := FData;
end;

function TSimpleControlItem.GetText: WideString;
begin
	Result := FText
end;

procedure TSimpleControlItem.SetData(const Value: Pointer);
begin
	FData := Value
end;

procedure TSimpleControlItem.SetText(const Value: WideString);
begin
	FText := Value
end;

{ TLightListBox }

procedure TLightListBox.BeforeDestruction;
begin
	if FSelected <> nil then FreeMem(FSelected);
	inherited;
end;

function TLightListBox.CanSelectItem(Index: Integer): Boolean;
begin
	Result := True;
end;

constructor TLightListBox.Create(AParent: TAnyWindow; const APosition: TPoint; const ASize: TSize);
var
	CS: PCREATESTRUCTW;
begin
	inherited Create(AParent, '', APosition, ASize);

	CS := CreateStruct;
	CS^.style:= CS^.style or (WS_VSCROLL or WS_TABSTOP);
	CS^.dwExStyle:= CS^.dwExStyle or WS_EX_CLIENTEDGE;

	FCaretIndex := -1;

	FVertScroll.cbSize:= SizeOf(FVertScroll);
	FVertScroll.fMask := SIF_DISABLENOSCROLL or SIF_ALL;

	FScrollBarManager := DefaultScrollBarManager;

	FViewOptions:= [loEraseBkgnd, loDrawFocusRect];
	FCtrlOptions:= [loAutoSelect];

	FItemHeight:= Abs(ForClientFont.Height);

	Color := SystemColor(COLOR_WINDOW);
end;

procedure TLightListBox.DoSelectItem(Index: Integer; Value: Boolean);
begin
	Repaint(ItemRect(Index));
end;

procedure TLightListBox.DrawItem(Index: Integer; DC: HDC; const Rect: TRect);
begin
end;

function TLightListBox.GetSelected(Index: Integer): Boolean;
begin
	Assert((Index >= 0) and (Index < ItemCount));
	Result:= FSelected[Index div 8] and (1 shl (Index mod 8)) <> 0;
end;

procedure TLightListBox.Initialize;
begin
	inherited;
	ScrollBarManager.Initialize(Handle);
	ScrollBarManager.ShowScrollBar(Handle, SB_VERT, True);
	PostMessageW(LLBM_RESETSCROLLBAR, 0, 0);
end;

function TLightListBox.ItemAtPos(Pos: TPoint): Integer;
var
	ClientArea: TRect;
begin
	ClientArea := ClientRect;
	if (Pos.X < ClientArea.Left) or (Pos.X >= ClientArea.Right) then
		Result := -1
	else if FScrollBase = lsbTop then
	begin
		Result := FScrollFrom + (Pos.Y - ClientArea.Top) div ItemHeight;
		if Result >= ItemCount then Result := -1;
	end
	else
	begin
		Result := FScrollFrom - (ClientArea.Bottom - Pos.Y) div ItemHeight;
		if Result < 0 then Result := -1;
	end;
end;

function TLightListBox.ItemRect(AIndex: Integer): TRect;
var
	ClientArea: TRect;
begin
	ClientArea := ClientRect;
	Result.Left := ClientArea.Left;
	Result.Right := ClientArea.Right;
	if FScrollBase = lsbTop then
	begin
		Result.Top := ClientArea.Top + (AIndex - FScrollFrom) * FItemHeight;
		Result.Bottom := Result.Top + FItemHeight;
	end
	else
	begin
		Result.Bottom := ClientArea.Bottom - (FScrollFrom - AIndex) * FItemHeight;
		Result.Top := Result.Bottom - FItemHeight;
	end;
end;

function TLightListBox.KeyDown(ShiftState: TShiftKeys; Code: Byte): Boolean;
var
	G, A: Boolean;
begin
	if ssAlt in ShiftState then
		Result := inherited KeyDown(ShiftState, Code)
	else
	begin
		G := ssShift in ShiftState;
		A := (loAutoSelect in CtrlOptions) and not (ssControl in ShiftState);
		case Code of
		VK_UP:
			Move(-1, G, A);
		VK_DOWN:
			Move(+1, G, A);
		VK_PRIOR:
			Move(-(ClientSize.CY div ItemHeight), G, A);
		VK_NEXT:
			Move(ClientSize.CY div ItemHeight, G, A);
		VK_SPACE:
			if not G and not A then Selected[CaretIndex] := not Selected[CaretIndex];
		VK_SHIFT:
			FAnchorIndex := CaretIndex;
		VK_HOME:
			if ssControl in ShiftState then CaretIndex := 0;
		VK_END:
			if ssControl in ShiftState then CaretIndex := ItemCount - 1;
		else
			Result := inherited KeyDown(ShiftState, Code);
			Exit;
		end;
		Result := True;
	end	
end;

procedure TLightListBox.LLBMResetScrollBar(var Message: TMessage); {message}
var
	M: MSG;
begin
	if not PeekMessageW(M, Handle, LLBM_RESETSCROLLBAR, LLBM_RESETSCROLLBAR, PM_NOREMOVE) then
	begin
		FVertScroll.nMax := FItemCount - 1;
		if FScrollBase = lsbTop then
			FVertScroll.nPos := FScrollFrom
		else
			FVertScroll.nPos := FScrollFrom - Integer(FVertScroll.nPage) + 1;
		FVertScroll.nPage := ClientSize.CY div FItemHeight;
		ScrollBarManager.SetScrollInfo(Handle, SB_VERT, FVertScroll, True);
	end
end;

procedure TLightListBox.Move(Step: Integer; Group, Auto: Boolean);
var
	OldIndex, I: Integer;
begin
	if (Step <> 0) and (FItemCount > 0) then
	begin
		OldIndex := CaretIndex;
		FCaretIndex := OldIndex + Step;
		if FCaretIndex < 0 then
		begin
			FCaretIndex := 0
		end	
		else if FCaretIndex > ItemCount - 1 then
		begin
			FCaretIndex := ItemCount - 1;
		end;	
		if FCaretIndex <> OldIndex then
		begin
			if Group then
			begin
				if (OldIndex < FAnchorIndex) and (FCaretIndex > FAnchorIndex) then
				begin
					for I := OldIndex to FAnchorIndex - 1 do Selected[I] := False;
					for I := FAnchorIndex to FCaretIndex do Selected[I] := True;
				end
				else if (OldIndex > FAnchorIndex) and (FCaretIndex < FAnchorIndex) then
				begin
					for I:= OldIndex downto FAnchorIndex + 1 do Selected[I] := False;
					for I:= FAnchorIndex downto FCaretIndex do Selected[I] := True;
				end
				else if (OldIndex < FAnchorIndex) or (FCaretIndex < FAnchorIndex) then
				begin
					for I := OldIndex to FCaretIndex - 1 do Selected[I] := False;
					for I := OldIndex downto FCaretIndex do Selected[I] := True;
				end
				else
				begin
					for I := OldIndex downto FCaretIndex + 1 do Selected[I] := False;
					for I := OldIndex to FCaretIndex do Selected[I] := True;
				end
			end
			else if Auto then
			begin
				if (SelectCount > 1) or not Selected[OldIndex] then SelectAll(False);
				Selected[OldIndex] := False;
				Selected[FCaretIndex] := True;
			end;
			Scroll;
			Repaint(ItemRect(OldIndex));
			Repaint(ItemRect(CaretIndex));
		end;
	end
end;

procedure TLightListBox.PaintWindow(DC: HDC); {override}
var
	I: Integer;
	FirstIndex, LastIndex: Integer;
	ClipRect, DrawRect, ClientArea: TRect;
	Brush: TBrush;
begin
	GetClipBox(DC, ClipRect);
	ClientArea := ClientRect;

	if (loEraseBkgnd in ViewOptions) and (ClipRect.Top < ClipRect.Bottom) then
	begin
		Brush := TBrush.Create;
		Brush.Owner := Garbage;
		Brush.Color := Color.Value;
		FillRect(DC, ClipRect, Brush.Handle);
	end;

	if ItemCount > 0 then
	begin
		DrawRect.Left := 0;
		DrawRect.Right := ClientArea.Right;
		if FScrollBase = lsbTop then
		begin
			FirstIndex := FScrollFrom + ClipRect.Top div FItemHeight;
			if FirstIndex >= ItemCount then FirstIndex := ItemCount - 1;
			LastIndex := FScrollFrom + (ClipRect.Bottom + FItemHeight - 1) div FItemHeight;
			if LastIndex >= ItemCount then LastIndex := ItemCount - 1;
			DrawRect.Bottom := (FirstIndex - FScrollFrom) * FItemHeight;
		end
		else
		begin
			LastIndex := FScrollFrom - (ClientArea.Bottom - ClipRect.Bottom) div FItemHeight;
			if LastIndex < 0 then LastIndex := 0;
			FirstIndex := FScrollFrom - (ClientArea.Bottom - ClipRect.Top + FItemHeight - 1) div FItemHeight;
			if FirstIndex < 0 then FirstIndex := 0;
			DrawRect.Bottom := ClientArea.Bottom - (FScrollFrom - FirstIndex + 1) * FItemHeight;
		end;

		for I := FirstIndex to LastIndex do
		begin
			DrawRect.Top := DrawRect.Bottom;
			Inc(DrawRect.Bottom, FItemHeight);

			if Selected[I] then
			begin
				FillRect(DC, DrawRect, GetSysColorBrush(COLOR_HIGHLIGHT));
				SetBkColor(DC, GetSysColor(COLOR_HIGHLIGHT));
				SetTextColor(DC, GetSysColor(COLOR_HIGHLIGHTTEXT));
			end
			else
			begin
				SetBkColor(DC, Color.Value);
				SetTextColor(DC, 0);
			end;

			DrawItem(I, DC, DrawRect);

			if (loDrawFocusRect in ViewOptions) and (CaretIndex = I) and Focused then
			begin
				SetTextColor(DC, Color.Value xor $ffffff);
				DrawFocusRect(DC, DrawRect);
			end
		end;
	end	
end;

procedure TLightListBox.RepaintScroll(Y: Integer; const Clip: TRect);
begin
	if loEraseBkgnd in ViewOptions then
		ScrollWindowEx(Handle, 0, Y, nil, @Clip, 0, nil, SW_INVALIDATE)
	else
		Repaint
end;

procedure TLightListBox.Scroll(Base: TScrollBase; From: Integer);
var
	R1, R2: TRect;
	ClientSize: TSize;
begin
  ClientSize := Self.ClientSize;
	R1 := ItemRect(0);

	if From < 0 then
		From := 0
	else if From >= ItemCount then
		From := ItemCount - 1;

	if ClientSize.cy >= FItemHeight * ItemCount then
	begin
		Base := lsbTop;
		From := 0;
	end
	else if (FScrollBase = lsbTop) and
		(FScrollFrom + (ClientSize.CY + FItemHeight - 1) div FItemHeight - 1 > ItemCount) then
	begin
		Base := lsbBottom;
		From := ItemCount - 1;
	end
	else if (FScrollBase = lsbBottom) and
		(FScrollFrom - (ClientSize.CY + FItemHeight - 1) div FItemHeight + 1 < 0) then
	begin
		Base := lsbTop;
		From := 0;
	end;

	FScrollBase := Base;
	FScrollFrom := From;

	R2 := ItemRect(0);
	RepaintScroll(R2.Top - R1.Top, ClientRect);

	PostMessageW(LLBM_RESETSCROLLBAR, 0, 0);
end;

procedure TLightListBox.Scroll;
var
	R: TRect;
begin
	{Ȃ荓蔲}
	while True do
	begin
		R := ItemRect(CaretIndex);
		if R.Top < 0 then
			ScrollUp
		else if R.Bottom > ClientRect.Bottom then
			ScrollDown
		else
			Break
	end	
end;

procedure TLightListBox.ScrollDown;
var
	R1, R2: TRect;
begin
	if (ClientSize.CY + FItemHeight - 1) div FItemHeight <= ItemCount then
	begin
		R1 := ItemRect(0);
		if FScrollBase = lsbTop then
		begin
			FScrollBase := lsbBottom;
			FScrollFrom := FScrollFrom + (ClientSize.CY div ItemHeight);
		end
		else
		begin
			if FScrollFrom < ItemCount - 1 then Inc(FScrollFrom);
		end;
		R2 := ItemRect(0);
		RepaintScroll(R2.Top - R1.Top, ClientRect);
		PostMessageW(LLBM_RESETSCROLLBAR, 0, 0);
	end
end;

procedure TLightListBox.ScrollUp;
var
	R1, R2: TRect;
begin
	R1 := ItemRect(0);
	if FScrollBase = lsbTop then
	begin
		if FScrollFrom > 0 then Dec(FScrollFrom);
	end
	else
	begin
		FScrollBase := lsbTop;
		FScrollFrom := FScrollFrom - (ClientSize.CY div ItemHeight);
	end;
	R2 := ItemRect(0);
	RepaintScroll(R2.Top - R1.Top, ClientRect);
	PostMessageW(LLBM_RESETSCROLLBAR, 0, 0);
end;

procedure TLightListBox.SelectAll(ToSelect: Boolean);
var
	I: Integer;
begin
	for I:= 0 to ItemCount - 1 do
		Selected[I] := ToSelect;
end;

procedure TLightListBox.SetCaretIndex(Value: Integer);
var
	OldCaretIndex: Integer;
begin
	Assert((Value = 0) or (Value < FItemCount));
	if Value <> FCaretIndex then
	begin
		OldCaretIndex := CaretIndex;
		FCaretIndex := Value;
		Scroll;
		Repaint(ItemRect(OldCaretIndex));
		Repaint(ItemRect(CaretIndex));
	end;	
end;

procedure TLightListBox.SetCtrlOptions(Value: TControlOptions);
begin
	if FCtrlOptions <> Value then
	begin
		FCtrlOptions := Value;
	end
end;

procedure TLightListBox.SetItemCount(const Value: Natural);
var
	MemSize: Integer;
begin
	if FItemCount <> Value then
	begin
		FItemCount:= Value;

		//IԂL郁[m
		MemSize:= (FItemCount + 7) div 8;
		ReallocMem(FSelected, MemSize);
		FillChar(FSelected^, MemSize, 0);
		FSelectCount := 0;

		FCaretIndex := 0;

		if FItemCount = 0 then
			FCaretIndex := -1
		else if loAutoSelect in CtrlOptions then
			Selected[FCaretIndex] := True;

		FScrollBase := lsbTop;
		FScrollFrom := 0;

		Repaint;
		PostMessageW(LLBM_RESETSCROLLBAR, 0, 0);
	end
end;

procedure TLightListBox.SetItemHeight(Value: Integer);
begin
	Assert(Value > 0);

	if FItemHeight = Value then
		Exit;

	FItemHeight:= Value;

	Repaint;
	PostMessageW(LLBM_RESETSCROLLBAR, 0, 0);
end;

procedure TLightListBox.SetSelected(Index: Integer; Value: Boolean);
begin
	Assert((Index >= 0) and (Index < ItemCount));

	if not CanSelectItem(Index) then Exit;
	if Selected[Index] = Value then Exit;

	if Value and not (loMultiSelect in CtrlOptions) then
		SelectAll(False);

	if Value then
	begin
		if FSelected[Index div 8] and (1 shl (Index mod 8)) = 0 then Inc(FSelectCount);
		FSelected[Index div 8] := FSelected[Index div 8] or (1 shl (Index mod 8));
	end
	else
	begin
		if FSelected[Index div 8] and (1 shl (Index mod 8)) <> 0 then Dec(FSelectCount);
		FSelected[Index div 8] := FSelected[Index div 8] and not(1 shl(Index mod 8));
	end;

	DoSelectItem(Index, Value);
end;

procedure TLightListBox.SetViewOptions(Value: TViewOptions);
begin
	if FViewOptions <> Value then
	begin
		FViewOptions := Value;
		Repaint;
	end
end;

procedure TLightListBox.WMGetDlgCode(var Message: TWMGetDlgCode); {message}
begin
	Message.Result:= DLGC_WANTCHARS;
end;

procedure TLightListBox.WMKillFocus(var Message: TWMKillFocus); {message}
begin
	inherited;
	Repaint(ItemRect(CaretIndex));
end;

procedure TLightListBox.WMLButtonDown(var Message: TWMLButtonDown); {message}
var
	P: TPoint;
	I, Index: Integer;
begin
	SetFocus;

	P.X := Message.XPos;
	P.Y := Message.YPos;
	Index := ItemAtPos(P);
	if Index >= 0 then
	begin
		if (Message.Keys and (MK_SHIFT or MK_CONTROL) = MK_SHIFT) then
		begin
			SelectAll(False);
			if Index > CaretIndex then
				for I := CaretIndex to Index do Selected[I] := True
			else
				for I := Index to CaretIndex do Selected[I] := True;
		end;
		CaretIndex := Index;
	end;

	inherited;
end;

procedure TLightListBox.WMLButtonUp(var Message: TWMLButtonUp);
var
	P: TPoint;
	Index: Integer;
begin
	P.X := Message.XPos;
	P.Y := Message.YPos;
	Index := ItemAtPos(P);
	if Index >= 0 then
	begin
		if loAutoSelect in CtrlOptions then
		begin
			CaretIndex := Index;
			if Message.Keys and MK_CONTROL = 0 then SelectAll(False);
			Selected[CaretIndex]:= not Selected[CaretIndex];
		end
		else if Message.Keys and (MK_CONTROL or MK_SHIFT) = (MK_CONTROL or MK_SHIFT) then
		begin
			SelectAll(False);
		end
		else if Message.Keys and MK_CONTROL <> 0 then
		begin
			CaretIndex := Index;
			Selected[CaretIndex]:= not Selected[CaretIndex];
		end
	end;

	inherited
end;

procedure TLightListBox.WMMouseWheel(var Message: TWMMouseWheel);
var
	M: TWMVScroll;
begin
	M.Msg := WM_VSCROLL;
	M.ScrollBar := SB_VERT;

	if Message.WheelDelta > 0 then
	begin
		M.ScrollCode := SB_LINEUP;
		Dispatch(M);
	end
	else if Message.WheelDelta < 0 then
	begin
		M.ScrollCode := SB_LINEDOWN;
		Dispatch(M);
	end;
end;

procedure TLightListBox.WMRButtonDown(var Message: TWMRButtonDown);
var
	P: TPoint;
	I: Integer;
begin
	SetFocus;
	P.X := Message.XPos;
	P.Y := Message.YPos;
	I := ItemAtPos(P);
	if I >= 0 then CaretIndex := I;
	inherited;
end;

procedure TLightListBox.WMSetFocus(var Message: TWMSetFocus); {message}
begin
	inherited;
	Repaint(ItemRect(CaretIndex));
end;

procedure TLightListBox.WMSize(var Message: TWMSize); {message}
begin
	inherited;
	Scroll(FScrollBase, FScrollFrom);
	PostMessageW(LLBM_RESETSCROLLBAR, 0, 0);
end;

procedure TLightListBox.WMVScroll(var Message: TWMScroll); {message}
var
	S: Integer;
	R1, R2: TRect;
begin
	S := ProcessScroll(Message, ScrollBarManager, SB_VERT, FVertScroll);
	case Message.ScrollCode of
	SB_ENDSCROLL:
		;
	SB_LINEUP:
		ScrollUp;
	SB_LINEDOWN:
		ScrollDown;
	else
		if S <> 0 then
		begin
			R1 := ItemRect(0);
			if FVertScroll.nPos + Integer(FVertScroll.nPage) > FVertScroll.nMax then
			begin
				FScrollBase := lsbBottom;
				FScrollFrom := ItemCount - 1;
			end
			else
			begin
				FScrollBase := lsbTop;
				FScrollFrom := FVertScroll.nPos;
			end;
			R2 := ItemRect(0);
			RepaintScroll(R2.Top - R1.Top, ClientRect);
		end
	end
end;

{ TLightSwitch }

function TLightSwitch.Add: T0ControlItem;
begin
	SetLength(FItems, Length(FItems) + 1);
	FItems[High(FItems)] := TSimpleControlItem.Create;
	Result := FItems[High(FItems)];
end;

procedure TLightSwitch.Clear;
var
	I: Integer;
begin
	for I := Low(FItems) to High(FItems) do
  	FItems[I].Free;

	FItems := nil;
	FSelected := nil;
end;

function TLightSwitch.Count: Integer;
begin
	Result := Length(FItems)
end;

procedure TLightSwitch.Finalize;
begin
	Clear;
	inherited Finalize;
end;

function TLightSwitch.GetAt(AIndex: Integer): T0ControlItem;
begin
	Result := FItems[AIndex]
end;

procedure TLightSwitch.Initialize;
begin
  inherited;
  Font := ForLabelFont;
end;

function TLightSwitch.KeyDown(ShiftState: TShiftKeys; Code: Byte): Boolean;
var
	I: Integer;
begin
	case Code of
	VK_LEFT:
	begin
		for I := Low(FItems) + 1 to High(FItems) do
			if FItems[I] = Selected then
			begin
				FSelected := FItems[I - 1];
				Repaint;
			end;
		Result := True;
	end;
	VK_RIGHT:
	begin
		for I := High(FItems) - 1 downto Low(FItems) do
			if FItems[I] = Selected then
			begin
				FSelected := FItems[I + 1];
				Repaint;
			end;
		Result := True;
	end
	else
		Result := inherited KeyDown(ShiftState, Code)
	end;
end;

procedure TLightSwitch.PaintWindow(DC: HDC);
var
	I, X: Integer;
	S: WideString;
	R: TRect;
begin
	SelectObject(DC, Font.Handle);
	SetBkMode(DC, TRANSPARENT);
	S := Text;
	Windows.TextOutW(DC, 4, 4, PWideChar(S), Length(S));

	SetTextAlign(DC, TA_CENTER or TA_TOP or TA_NOUPDATECP);

	X := 4 + TextWidth;

	for I := Low(FItems) to High(FItems) do
	begin
		if FItems[I] = Selected then
		begin
			SetBkColor(DC, GetSysColor(COLOR_HIGHLIGHT));
			SetTextColor(DC, GetSysColor(COLOR_HIGHLIGHTTEXT));
		end
		else
		begin
			SetBkColor(DC, GetSysColor(COLOR_BTNFACE));
			SetTextColor(DC, GetSysColor(COLOR_WINDOWTEXT));
		end;

		R.Left := X;
		R.Top := 4;
		R.Right := X + ItemWidth;
		R.Bottom := ClientSize.CY - 4;

		S := FItems[I].Text;
		ExtTextOutW(DC, X + ItemWidth div 2, 4, ETO_OPAQUE, @R, PWideChar(S), Length(S), nil);

		X := X + ItemWidth;
	end;

	if Focused then
	begin
		SetBkColor(DC, Color.Value);
		SetTextColor(DC, Color.Value xor $ffffff);
		DrawFocusRect(DC, ClientRect);
	end;
end;

procedure TLightSwitch.SetItemWidth(const Value: Integer);
begin
	FItemWidth := Value;
	Repaint;
end;

procedure TLightSwitch.SetTextWidth(const Value: Integer);
begin
	FTextWidth := Value;
	Repaint;
end;

procedure TLightSwitch.WMKillFocus(var Message: TWMKillFocus);
begin
	inherited;
	Repaint;
end;

procedure TLightSwitch.WMLButtonDown(var Message: TWMLButtonDown);
var
	Pos: Integer;
begin
	{ div  0 ɋ߂Â + ItemWidth  1 }
	Pos := (Message.XPos - TextWidth + ItemWidth) div ItemWidth - 1;
	if Pos in [Low(FItems)..High(FItems)] then
	begin
		FSelected := FItems[Pos];
		Repaint;
	end;
	SetFocus;
end;

procedure TLightSwitch.WMSetFocus(var Message: TWMSetFocus);
begin
	inherited;
	Repaint;
end;

{ T0SplitPane }

procedure T0SplitPane.BeforeDestruction;
var
	I, J: Integer;
	H: T0SplitWindow;
begin
	H := Host;
	for I := Low(H.Panes) to High(H.Panes) do
		for J := Low(H.Panes[I]) to High(H.Panes[I]) do
			if H.Panes[I, J] = Self then
				H.FPanes[I, J] := nil;

	inherited;
end;

constructor T0SplitPane.Create(AParent: TAnyWindow; AHorz, AVert: TScrollBar);
begin
	inherited Create(AParent, '', ZeroPoint, ZeroSize);
	FHorzScrollBar := AHorz;
	FVertScrollBar := AVert;
	Owner := AParent;
	Color := SystemColor(COLOR_WINDOW);
end;

class procedure T0SplitPane.CreateClass(var WndClass: WNDCLASSEXW);
begin
	inherited CreateClass(WndClass);
	WndClass.hbrBackground := GetSysColorBrush(COLOR_WINDOW);
end;

function T0SplitPane.GetHost: T0SplitWindow;
begin
	Assert(Owner is T0SplitWindow);

	Result := T0SplitWindow(Owner)
end;

{ T0SplitWindow }

type
	TDimPoint = array[TBarStyle] of Integer;
	TDimSmallPoint = array[TBarStyle] of SmallInt;
	TDimSize = array[TBarStyle] of Integer;
	TDimRect = array[0..1, TBarStyle] of Integer;

const
	HTHSPLIT = 128;
	HTVSPLIT = 129;

function T0SplitWindow.CanFocus: Boolean;
begin
	Result := Enabled and Visible and ActiveControl.CanFocus
end;

constructor T0SplitWindow.Create(AParent: TAnyWindow; const APosition: TPoint;
				const ASize: TSize; const ASplitable: TBarStyles; const AHasScrollBars: TBarStyles;
				const AHasBorder: Boolean);
var
	CS: PCREATESTRUCTW;
begin
	inherited Create(AParent, '', APosition, ASize);
	FSplitable := ASplitable;
	if AHasBorder then FBorderWidth := 2;
	FSplitWidth := FBorderWidth * 2 + 3;
	if sbVertical in AHasScrollBars then FScrollWidth.cx := GetSystemMetrics(SM_CXVSCROLL);
	if sbHorizontal in AHasScrollBars then FScrollWidth.cy := GetSystemMetrics(SM_CYHSCROLL);

	CS := CreateStruct;
	CS^.style := CS^.style and not WS_TABSTOP;
end;

function T0SplitWindow.CreatePane(AHorz, AVert: TScrollBar): T0SplitPane;
begin
	Result := T0SplitPane.Create(Self, AHorz, AVert);
end;

procedure T0SplitWindow.Initialize;
begin
	inherited;

	FScrollBars[sbHorizontal, 0] := CreateScrollBar(sbHorizontal);
	FScrollBars[sbVertical, 0] := CreateScrollBar(sbVertical);

	FPanes[0, 0] := CreatePane(
					FScrollBars[sbHorizontal, 0], FScrollBars[sbVertical, 0]);

	ActiveControl := FPanes[0, 0];

	Realign;
end;

procedure T0SplitWindow.PaintWindow(DC: HDC);
var
	B, R, P: TRect;
	ParentWindow: T0TopWindow;
begin
	R := ClientRect;

	ParentWindow := TopParent;
	P := ParentWindow.ClientRect;
	B := BoundsRect;
	if (B.Right = P.Right) and (B.Bottom = P.Bottom) then
	begin
		R.Left := R.Right - FScrollWidth.CX;
		R.Top := R.Bottom - FScrollWidth.CY;
		DrawFrameControl(DC, R, DFC_SCROLL, DFCS_SCROLLSIZEGRIP);
	end
	else
	begin
		FillRect(DC, R, GetSysColorBrush(COLOR_BTNFACE))
	end;

	if R.Right > 0 then
		PostMessageW(WM_LIGHTWINDOWS, LW_NCPAINT, 0)
end;

function T0SplitWindow.SplitBox(Kind: TBarStyle): TRect;
begin
	if Kind = sbHorizontal then
	begin
		Result.Bottom := ClientRect.Bottom;
		if sbHorizontal in FSplited then
		begin
			Result.Top := 0;
			Result.Left := FPanes[0, 0].BoundsRect.Right;
			Result.Right := Result.Left + FSplitWidth;
		end
		else
		begin
			Result.Top := Result.Bottom - FScrollWidth.CY;
			Result.Left := 0;
			Result.Right := FSplitWidth;
		end
	end
	else
	begin
		Result.Right := ClientRect.Right;
		if sbVertical in FSplited then
		begin
			Result.Left := 0;
			Result.Top := FPanes[0, 0].BoundsRect.Bottom;
			Result.Bottom := Result.Top + FSplitWidth;
		end
		else
		begin
			Result.Left := Result.Right - FScrollWidth.CX;
			Result.Top := 0;
			Result.Bottom := FSplitWidth;
		end
	end
end;

procedure T0SplitWindow.WMNCCalcSize(var Message: TWMNCCalcSize);
begin
	inherited;
	Inc(Message.CalcSize_Params.rgrc[0].Left, FBorderWidth);
	Inc(Message.CalcSize_Params.rgrc[0].Top, FBorderWidth);
	Dec(Message.CalcSize_Params.rgrc[0].Right, FBorderWidth);
	Dec(Message.CalcSize_Params.rgrc[0].Bottom, FBorderWidth);
end;

procedure T0SplitWindow.WMNCHitTest(var Message: TWMNCHitTest);
var
	Area: TRect;
	Mouse: TPoint;

	function InBox(R: TRect): Boolean;
	begin
		OffsetRect(R, Area.Left + FBorderWidth, Area.Top + FBorderWidth);
		Result := PtInRect(R, Mouse);
	end;

begin
	GetWindowRect(Handle, Area);
	Mouse.X := Message.XPos;
	Mouse.Y := Message.YPos;

	if (Mouse.X >= Area.Right - FScrollWidth.CX) and
					(Mouse.Y >= Area.Bottom - FScrollWidth.CY) then
		Message.Result := HTBOTTOMRIGHT
	else if (Mouse.X < Area.Left + FBorderWidth) or
					(Mouse.Y < Area.Top + FBorderWidth) or
					(Mouse.X >= Area.Right - FBorderWidth) or
					(Mouse.Y >= Area.Bottom - FBorderWidth) then
		Message.Result := HTBORDER
	else if (sbHorizontal in FSplitable) and InBox(SplitBox(sbHorizontal)) then
		Message.Result := HTHSPLIT
	else if (sbVertical in FSplitable) and InBox(SplitBox(sbVertical)) then
		Message.Result := HTVSPLIT
	else
		Message.Result := HTCLIENT
end;

procedure T0SplitWindow.WMNCPaint(var Message: TWMNCPaint);
var
	DC: HDC;
	R: TRect;
	X, Y, S: Integer;
	I: TBarStyle;
begin
	if FPanes[0, 0].BoundsSize.CX = 0 then
	begin
{		if ClientSize.CX > FBorderWidth * 4 then
			PostMessageW(WM_LIGHTWINDOWS, LW_NCPAINT, 0);
}			
		Exit;
	end;

	DC := GetWindowDC(Handle);

	if FSplited <> [] then
	begin
		R := BoundsRect;
		OffsetRect(R, -R.Left, -R.Top);
		DrawEdge(DC, R, EDGE_BUMP, BF_RECT or BF_FLAT);
	end;

	S := FBorderWidth;
	for I := sbHorizontal to sbVertical do
	begin
		R := SplitBox(I);
		OffsetRect(R, FBorderWidth, FBorderWidth);
		if I in FSplited then
		begin
			InflateRect(R, -S, S);
			DrawEdge(DC, R, EDGE_BUMP, BF_MIDDLE or BF_FLAT)
		end
		else if I in FSplitable then
			DrawEdge(DC, R, EDGE_RAISED, BF_RECT or BF_MIDDLE);

		S := -S;
	end;

	for X := Ord(sbHorizontal in FSplited) downto 0 do
		for Y := Ord(sbVertical in FSplited) downto 0 do
		begin
			R := FPanes[X, Y].BoundsRect;
			OffsetRect(R, FBorderWidth, FBorderWidth);
			InflateRect(R, FBorderWidth, FBorderWidth);

			if X = Ord(sbHorizontal in FSplited) then
				Inc(R.Right, FScrollWidth.CX);

			if Y = Ord(sbVertical in FSplited) then
				Inc(R.Bottom, FScrollWidth.CX);

			DrawEdge(DC, R, EDGE_SUNKEN, BF_RECT);
		end;

	ReleaseDC(Handle, DC);
end;

procedure T0SplitWindow.WMNCLButtonDown(var Message: TWMNCHitMessage);
var
	Par: TAnyWindow;
begin
	case Message.HitTest of
	HTBOTTOMRIGHT:
	begin
		Par := TopParent;
		if Par <> nil then Par.Dispatch(Message)
	end;
	HTHSPLIT, HTVSPLIT:
	begin
		FSplitingDir := sbHorizontal;
		if Message.HitTest <> HTHSPLIT then
			FSplitingDir := sbVertical;

		BeginSpliting(FSplitingDir);
	end;
	else
		inherited
	end
end;

procedure T0SplitWindow.BeginSpliting(Dir: TBarStyle);
var
	Cur: TPoint;
begin
	FSpliting := True;
	FSplitingDir := Dir;

	FSplitingPos := 0;
	if FSplitingDir in FSplited then
		FSplitingPos := TDimSize(FPanes[0, 0].BoundsSize)[FSplitingDir];

	GetCursorPos(Cur);
	Cur := ScreenToClient(Cur);
	FSplitingDiff := FSplitingPos - TDimPoint(Cur)[FSplitingDir];
	FSplitingMax := TDimSize(ClientSize)[FSplitingDir] -
					TDimSize(FScrollWidth)[FSplitingDir] - FSplitWidth;

	Windows.SetFocus(Handle); {œ͂D}
  UpdateWindow(Handle);
	SetCapture(Handle);
	DrawSpliting;
end;

procedure T0SplitWindow.MoveSpliting(Step: Integer);
begin
	DrawSpliting;

	if (FSplitingPos = 0) and (Step > 0) then
		FSplitingPos := 20
	else if (FSplitingPos = FSplitingMax) and (Step < 0) then
		FSplitingPos := FSplitingMax - 21
	else
		Inc(FSplitingPos, Step);

	if FSplitingPos < 20 then
		FSplitingPos := 0
	else if FSplitingPos >= FSplitingMax - 20 then
		FSplitingPos := FSplitingMax;

	DrawSpliting;
end;

procedure T0SplitWindow.WMMouseMove(var Message: TWMMouseMove);
var
	AbsY, Step: Integer;
begin
	if FSpliting then
	begin
		AbsY := TDimSmallPoint(Message.Pos)[FSplitingDir] + FSplitingDiff;
		if AbsY < 20 then
			Step := -99999
		else if AbsY > FSplitingMax - 20 then
			Step := 9999
		else
			Step := AbsY - FSplitingPos;
		
		MoveSpliting(Step);
	end;
end;

function T0SplitWindow.KeyDown(ShiftState: TShiftKeys; Code: Byte): Boolean;
var
	Step: Integer;
begin
	if not FSpliting then
		Result := inherited KeyDown(ShiftState, Code)
	else
	begin
		Step := 8;
		if ssControl in ShiftState then
			Step := 1;

		case Code of
		VK_LEFT, VK_UP:
			MoveSpliting(-Step);
		VK_RIGHT, VK_DOWN:
			MoveSpliting(Step);
		VK_RETURN:
			EndSpliting(True);
		VK_ESCAPE:
			EndSpliting(False);
		end;
		Result := True;
	end;
end;

procedure T0SplitWindow.WMLButtonUp(var Message: TWMLButtonUp);
begin
	if FSpliting then
		EndSpliting(True);
end;

procedure T0SplitWindow.EndSpliting(Accept: Boolean);
type
	TSize2 = array[TBarStyle] of Integer;
var
	W1, W2: Integer;
	Closable: Boolean;
begin
	DrawSpliting;
	ReleaseCapture;

	if Accept then
	begin
		Closable := TSize2(FScrollWidth)[FSplitingDir] > 0;

		if FSplitingPos <= 0 then
		begin
			if (FSplitingDir in FSplited) and Closable then
			begin
				Exclude(FSplited, FSplitingDir);

				FScrollBars[FSplitingDir, 0].Free;
				FScrollBars[FSplitingDir, 0] := FScrollBars[FSplitingDir, 1];
				FScrollBars[FSplitingDir, 1] := nil;
				FPanes[0, 0].Free;
				if FSplitingDir = sbHorizontal then
				begin
					FPanes[0, 1].Free;

					FPanes[0, 0] := FPanes[1, 0];
					FPanes[1, 0] := nil;
					FPanes[0, 1] := FPanes[1, 1];
					FPanes[1, 1] := nil;
				end
				else
				begin
					FPanes[1, 0].Free;

					FPanes[0, 0] := FPanes[0, 1];
					FPanes[0, 1] := nil;
					FPanes[1, 0] := FPanes[1, 1];
					FPanes[1, 1] := nil;
				end;

				Realign;
			end
		end
		else if FSplitingPos >= FSplitingMax then
		begin
			if (FSplitingDir in FSplited) and Closable then
			begin
				Exclude(FSplited, FSplitingDir);

				FScrollBars[FSplitingDir, 1].Free;
				FScrollBars[FSplitingDir, 1] := nil;
				FPanes[1, 1].Free;
				FPanes[1, 1] := nil;
				if FSplitingDir = sbHorizontal then
				begin
					FPanes[1, 0].Free;
					FPanes[1, 0] := nil;
				end
				else
				begin
					FPanes[0, 1].Free;
					FPanes[0, 1] := nil;
				end;

				Realign;
			end
		end
		else
		begin
			W1 := FSplitingPos;
			W2 := TSize2(ClientSize)[FSplitingDir]
						- TSize2(FScrollWidth)[FSplitingDir] - FSplitWidth - FSplitingPos;

			DoSplit(FSplitingDir, W1 / (W1 + W2));
		end;
	end;

	FSpliting := False;

	if ActiveControl = nil then
		FPanes[0, 0].SetFocus
	else
		ActiveControl.SetFocus;
end;

procedure T0SplitWindow.WMSetCursor(var Message: TWMSetCursor);
begin
	case Message.HitTest of
	HTVSPLIT:
		SetCursor(LoadCursorW(0, PWideChar(IDC_SIZENS)));
	HTHSPLIT:
		SetCursor(LoadCursorW(0, PWideChar(IDC_SIZEWE)));
	else
		inherited
	end
end;

procedure T0SplitWindow.DrawSpliting;
var
	DC: HDC;
	R: TRect;
	X, Y: Integer;
begin
	GetClientRect(Handle, R);

	Inc(TDimRect(R)[0, FSplitingDir], FSplitingPos + FBorderWidth * 2);
	TDimRect(R)[1, FSplitingDir] :=
					TDimRect(R)[0, FSplitingDir] + FSplitWidth - FBorderWidth * 2;

	DC := GetWindowDC(Handle);

	if Win32MajorVersion >= 6 then
	begin
		BitBlt(DC, R.Left, R.Top, R.Right - R.Left, R.Bottom - R.Top,
			0, 0, 0, DSTINVERT)
	end
	else
	begin
		for X := R.Left to R.Right - 1 do
			for Y := R.Top to R.Bottom - 1 do
				if (X + Y) and 1 = 0 then
					SetPixelV(DC, X, Y, GetPixel(DC, X, Y) xor $ffffff)
	end;

	ReleaseDC(Handle, DC)
end;

procedure T0SplitWindow.Realign;
var
	CRect: TRect;
	X, Y: array[0..1] of record L, H: Integer end;
	CS, XI, YI, ScrX, ScrY: Integer;
begin
	CRect := ClientRect;

	X[0].L := CRect.Left;
	X[0].H := CRect.Right - FScrollWidth.CX;
	X[1] := X[0];

	if sbHorizontal in FSplited then
	begin
		CS := X[1].H - CRect.Left - FSplitWidth;
		X[0].H := X[0].L + Round(CS * FSplitPer[sbHorizontal]);
		X[1].L := X[0].H + FSplitWidth;
	end;

	Y[0].L := CRect.Top;
	Y[0].H := CRect.Bottom - FScrollWidth.CY;
	Y[1] := Y[0];

	if sbVertical in FSplited then
	begin
		CS := Y[1].H - CRect.Left - FSplitWidth;
		Y[0].H := Y[0].L + Round(CS * FSplitPer[sbVertical]);
		Y[1].L := Y[0].H + FSplitWidth;
	end;

	if sbHorizontal in FSplited then
		for XI := 0 to 1 do
			FScrollBars[sbHorizontal, XI].BoundsRect := Rect(X[XI].L, Y[1].H, X[XI].H, CRect.Bottom)
	else
	begin
		ScrX := X[0].L;
		if sbHorizontal in FSplitable then Inc(ScrX, FSplitWidth);
		FScrollBars[sbHorizontal, 0].BoundsRect := Rect(ScrX, Y[1].H, X[0].H, CRect.Bottom)
	end;

	if sbVertical in FSplited then
		for YI := 0 to 1 do
			FScrollBars[sbVertical, YI].BoundsRect := Rect(X[1].H, Y[YI].L, CRect.Right, Y[YI].H)
	else
	begin
		ScrY := Y[0].L;
		if sbVertical in FSplitable then Inc(ScrY, FSplitWidth);
		FScrollBars[sbVertical, 0].BoundsRect := Rect(X[1].H, ScrY, CRect.Right, Y[0].H)
	end;

	for YI := Ord(sbVertical in FSplited) downto 0 do
		for XI := Ord(sbHorizontal in FSplited) downto 0 do
			FPanes[XI, YI].BoundsRect := Rect(X[XI].L, Y[YI].L, X[XI].H, Y[YI].H);

	if CRect.Left < CRect.Right then
		PostMessageW(WM_LIGHTWINDOWS, LW_NCPAINT, 0)
end;

procedure T0SplitWindow.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin
	Message.Result := 0;
end;

procedure T0SplitWindow.WMHScroll(var Message: TWMScroll);
var
	X, Y: Integer;
begin
	inherited;

	for X := Ord(sbHorizontal in FSplited) downto 0 do
		if Message.ScrollBar = FScrollBars[sbHorizontal, X].Handle then
			for Y := Ord(sbVertical in FSplited) downto 0 do
				FPanes[X, Y].Dispatch(Message)
end;

procedure T0SplitWindow.WMVScroll(var Message: TWMScroll);
var
	X, Y: Integer;
begin
	inherited;

	for Y := Ord(sbVertical in FSplited) downto 0 do
		if Message.ScrollBar = FScrollBars[sbVertical, Y].Handle then
			for X := Ord(sbHorizontal in FSplited) downto 0 do
				FPanes[X, Y].Dispatch(Message)
end;

procedure T0SplitWindow.WMSetFocus(var Message: TWMSetFocus);
begin
	if not FSpliting then inherited
end;

procedure T0SplitWindow.WMLightWindows(var Message: TMessage);
begin
	inherited;
	if Message.WParam = LW_NCPAINT then
		RepaintFrame
end;

procedure T0SplitWindow.DoSplit(Dir: TBarStyle; Per: Real);
type
	TSize2 = array[TBarStyle] of Integer;
var
	S: TSize;
begin
	FSplitingDir := Dir;
	S := BoundsSize;
	FSplitingPos := TSize2(S)[Dir] div 2;
	FSplitPer[Dir] := Per;

	if not (FSplitingDir in FSplited) then
	begin
		if FSplitingDir = sbHorizontal then
		begin
			FScrollBars[FSplitingDir, 1] := CreateScrollBar(FSplitingDir);

			FPanes[1, 0] := CreatePane(FScrollBars[sbHorizontal, 1], FScrollBars[sbVertical, 0]);

			if sbVertical in FSplited then
				FPanes[1, 1] := CreatePane(FScrollBars[sbHorizontal, 1], FScrollBars[sbVertical, 1]);
		end
		else
		begin
			FScrollBars[sbVertical, 1] := FScrollBars[sbVertical, 0];
			FScrollBars[sbVertical, 0] := CreateScrollBar(FSplitingDir);

			FPanes[0, 1] := FPanes[0, 0];
			FPanes[0, 0] := CreatePane(FScrollBars[sbHorizontal, 0], FScrollBars[sbVertical, 0]);

			if sbHorizontal in FSplited then
			begin
				FPanes[1, 1] := FPanes[1, 0];
				FPanes[1, 0] := CreatePane(FScrollBars[sbHorizontal, 1], FScrollBars[sbVertical, 0]);
			end;
		end;
		Include(FSplited, FSplitingDir);
	end;
	Realign;
end;

function T0SplitWindow.CreateScrollBar(Dir: TBarStyle): TScrollBar;
type
	TSize2 = array[TBarStyle] of Integer;
begin
	Result := TScrollBar.Create(Self, ZeroPoint, ZeroSize, Dir);
	if TSize2(FScrollWidth)[Dir] = 0 then Result.Hide;
end;

{ TLightButton }

procedure TLightButton.AfterConstruction;
begin
	inherited;
	Font := ForLabelFont;
end;

constructor TLightButton.Create(AParent: TAnyWindow; const AText: WideString; const APosition: TPoint; const ASize: TSize);
var
	CS: PCreateStructW;
begin
	inherited;

	CS := CreateStruct;
	CS^.style := CS^.style and not WS_TABSTOP;

	FTextColor := SystemColor(COLOR_BTNTEXT);
end;

procedure TLightButton.PaintWindow(DC: HDC);
var
	R: TRect;
	W: WideString;
begin
	R := ClientRect;
	DrawFrameControl(DC, R, DFC_BUTTON, DFCS_BUTTONPUSH or DFCS_ADJUSTRECT or
		DFCS_FLAT * Ord(not FHovering) or
		DFCS_PUSHED * Ord(GetKeyState(VK_LBUTTON) and $8000 <> 0));
	W := Text;
	SelectObject(DC, Font.Handle);
	SetBkMode(DC, TRANSPARENT);

	if Enabled then
		SetTextColor(DC, FTextColor.Value)
	else
		SetTextColor(DC, GetSysColor(COLOR_GRAYTEXT));

	DrawTextW(DC, PWideChar(W), Length(W), R, DT_CENTER or DT_VCENTER or DT_SINGLELINE);
end;

procedure TLightButton.WMEnable(var Message: TWMEnable);
begin
	inherited;
	Repaint
end;

procedure TLightButton.WMLButtonDown(var Message: TWMMouse);
begin
	inherited;
	Repaint;
end;

procedure TLightButton.WMLButtonUp(var Message: TWMMouse);
var
	P: TAnyWindow;
begin
	inherited;
	P := Parent;
	if P <> nil then P.SendMessageW(WM_COMMAND, BN_CLICKED shl 16, Handle);
	Repaint;
end;

procedure TLightButton.WMMouseHover(var Message);
begin
	FHovering := True;
	inherited;
	Repaint
end;

procedure TLightButton.WMMouseLeave(var Message);
begin
	FHovering := False;
	inherited;
	Repaint;
end;

procedure TLightButton.WMMouseMove(var Message: TWMMouseMove);
var
	E: Windows.TTrackMouseEvent;
begin
	E.cbSize := SizeOf(E);
	E.dwFlags := TME_HOVER or TME_LEAVE;
	E.hwndTrack := Handle;
	E.dwHoverTime := HOVER_DEFAULT;
	TrackMouseEvent(E);
	inherited;
end;

{ TArea }

procedure TArea.ColorChanged;
begin
	inherited;
	if Handle <> 0 then Repaint;
end;

constructor TArea.Create(AParent: TAnyWindow;
	const APosition: TPoint; const ASize: TSize; Border: Boolean);
var
	CS: PCreateStructW;
begin
	inherited Create(AParent, '', APosition, ASize);
	CS := CreateStruct;
	CS^.style := CS^.style and not WS_TABSTOP;
	CS^.dwExStyle := CS^.dwExStyle or WS_EX_CONTROLPARENT;
	if Border then CS^.dwExStyle := CS^.dwExStyle or WS_EX_CLIENTEDGE;
	Color := SystemColor(COLOR_BTNFACE);
end;

procedure TArea.WMEraseBkgnd(var Message: TWMEraseBkgnd);
var
	Brush: TBrush;
begin
	if Color.SysColor = COLOR_BTNFACE then
		inherited
	else
	begin
		Brush := TBrush.Create;
		Brush.Owner := Garbage;
		Brush.Color := Color.Value;
		FillRect(Message.DC, ClientRect, Brush.Handle)
	end;
end;

procedure TArea.WMLButtonUp(var Message: TWMMouse);
var
	P: HWND;
	HDR: NMHDR;
begin
	P := GetParent(Handle);
	if P <> 0 then
	begin
		HDR.hwndFrom := Handle;
		HDR.idFrom := 0;
		HDR.code := NM_CLICK;
		Windows.SendMessageW(P, WM_NOTIFY, 0, LPARAM(@HDR));
	end;
end;

{ TSplitter }

constructor TSplitter.Create(AParent: TAnyWindow; const APosition: TPoint;
	const ASize: TSize; AAlign: TAlignment);
var
	CS: PCREATESTRUCTW;
begin
	inherited Create(AParent, '', APosition, ASize);
	Alignment := AAlign;
	CS := CreateStruct;
	CS^.style := CS^.style and not WS_TABSTOP;
end;

class function TSplitter.GetDefaultSize: tagSIZE;
begin
	Result.cx := 3;
	Result.cy := 3;
end;

procedure TSplitter.WMLButtonDown(var Message: TWMMouse);
begin
	GetCursorPos(FMousePos);
	SetCapture(Handle);
end;

procedure TSplitter.WMLButtonUp(var Message: TWMMouse);
begin
	ReleaseCapture
end;

procedure TSplitter.WMMouseMove(var Message: TWMMouseMove);
var
	Parent, Window: TAnyWindow;
	P: TPoint;
	R, PR: TRect;
	S: TSize;
begin
	if GetCapture = Handle then
	begin
		Parent := Self.Parent;
		Window := GetWnd(GetWindow(Handle, GW_HWNDPREV));
		if (Parent <> nil) and (Window <> nil) then
		begin
			PR := Parent.ClientRect;
			S := ClientSize;
			R := Window.BoundsRect;
			GetCursorPos(P);
			case Alignment of
			alLeft:
			begin
				Inc(R.Right, P.X - FMousePos.X);
				if R.Right < R.Left then R.Right := R.Left;
				if R.Right > PR.Right - S.CX then R.Right := PR.Right - S.CX;
			end;
			alBottom:
			begin
				Inc(R.Top, P.Y - FMousePos.Y);
				if R.Top > R.Bottom then R.Top := R.Bottom;
				if R.Top < S.CY then R.Top := S.CY;
			end;
			end;
			Window.BoundsRect := R;
			if Parent is T0Window then T0Window(Parent).Realign;
			FMousePos := P
		end;
	end
end;

procedure TSplitter.WMSetCursor(var Message: TWMSetCursor);
begin
	case Alignment of
	alLeft, alRight:
		SetCursor(LoadCursorW(0, PWideChar(IDC_SIZEWE)));
	else
		SetCursor(LoadCursorW(0, PWideChar(IDC_SIZENS)));
	end;
	Message.Result := 0
end;

end.

