unit LAddDlgs;

{*******************************************************}
{                                                       }
{       LightWindows                                    }
{       Additional Dialogs                              }
{                                                       }
{       2001..2006 YT                                   }
{                                                       }
{*******************************************************}

interface

uses
	Windows, CommCtrl, Messages, Types, SysUtils,
	SysHacks, WinTips,
	LObjects, LConsts, LActions, LWindows, LStdCtls, LCmmCtls, LCmmDlgs;

type
	TInputDialog = class(T0DialogWindow)
	private
		FEdit: T0Window;
		FOk: TButton;
		FCancel: TCloseButton;
		FCaretIndex: Integer;
		FSelectedLength: Integer;
		FHistory: Boolean;
		FSpaceHeight: Integer;
		procedure WMActivate(var Message: TWMActivate); message WM_ACTIVATE;
		function GetValue: WideString;
		procedure SetValue(const Value: WideString);
	protected
		procedure Initialize; override;
		procedure Setup; override;
		procedure HandleNotification(Sender: TAnyWindow; Info: PNMHDR; var Result: LRESULT); override;
		property OkButton: TButton read FOk;
	public
		constructor Create(AParent: T0TopWindow; const AText: WideString; AHistory: Boolean);
		property Value: WideString read GetValue write SetValue;
		property Edit: T0Window read FEdit;
		property CaretIndex: Integer read FCaretIndex write FCaretIndex;
		property SelectedLength: Integer read FSelectedLength write FSelectedLength;
		property History: Boolean read FHistory;
		function Add: T0ControlItem;
		procedure Clear;
		procedure SelectAll;
		property SpaceHeight: Integer read FSpaceHeight write FSpaceHeight;
		property OnQuery;
	end;

type
	CSelectionChange = procedure(Context: Pointer) cdecl;

	TSelectDialog = class(T0DialogWindow)
	private
		FList: TListBox;
		FOk: TButton;
		FCancel: TCloseButton;
		FSpaceSize: TSize;
		FOnSelectionChange: CSelectionChange;
		procedure SetSelected(const Value: T0ControlItem);
		function GetSelected: T0ControlItem;
		procedure SetSpaceSize(const Value: TSize);
		function GetSpaceRect: TRect;
	protected
		procedure Initialize; override;
		procedure Setup; override;
		procedure HandleNotification(Sender: TAnyWindow; Info: PNMHDR; var Result: LRESULT); override;
		procedure AdjustSize;
	public
		procedure Clear;
		function Add: T0ControlItem;
		function Enumerate(Proc: CEachItem; Context: Pointer): T0ControlItem;
		property List: TListBox read FList;
		property Selected: T0ControlItem read GetSelected write SetSelected;
		property SpaceSize: TSize read FSpaceSize write SetSpaceSize;
		property OnQuery;
		property OkButton: TButton read FOk;
		property SpaceRect: TRect read GetSpaceRect;
		property OnSelectionChange: CSelectionChange read FOnSelectionChange write FOnSelectionChange;
	end;

type
	TTreeDialog = class(T0DialogWindow)
	private
		FTree: TTreeView;
		FOk: TButton;
		FCancel: TCloseButton;
		FNilIsInvalid: Boolean;
		FParentIsInvalid: Boolean;
		function GetRoot: TTreeViewRoot;
		function GetSelected: TAnyTreeViewNode;
		procedure SetSelected(const Value: TAnyTreeViewNode);
	protected
		procedure Initialize; override;
		procedure Setup; override;
		procedure HandleNotification(Sender: TAnyWindow; Info: PNMHDR; var Result: LRESULT); override;
		procedure UpdateOkButton;
	public
		procedure Realign; override;
		property Root: TTreeViewRoot read GetRoot;
		property Selected: TAnyTreeViewNode read GetSelected write SetSelected;
		property NilIsInvalid: Boolean read FNilIsInvalid write FNilIsInvalid;
		property ParentIsInvalid: Boolean read FParentIsInvalid write FParentIsInvalid;
		property Tree: TTreeView read FTree;
	end;

type
	TVersionDialog = class(T0CommonDialog)
	protected
		function DoExecute: Boolean; override;
	public
		class function Execute(const ATitle: WideString): Boolean; overload;
	end;

function SelectActionItemsDialog(const AItems: IActionItems; CustomDraw: TTreeViewCustomDraw = nil): IActionItem;
procedure AlignButtons(Dialog: T0DialogWindow; const Buttons: array of TButton);

implementation

function SelectActionItemsDialog(const AItems: IActionItems; CustomDraw: TTreeViewCustomDraw = nil): IActionItem;
var
	Dialog: TTreeDialog;
	First: TAnyTreeViewNode;

	procedure AddItems(Tree: T0TreeViewNodes; Items: IActionItems);

		function EnumProc(const I: IActionItem): Boolean; cdecl;
		var
			T: TAnyTreeViewNode;
		begin
			T := Tree.Add;
			T.Text := I.Text;
			T.Data := Pointer(I);
			if First = nil then
				First := T;
			if I.Checked then
				Dialog.Selected := T;
			if I.Items <> nil then
			begin
				T.Data := nil;
				//I.Items.Update;
				AddItems(T.Children, I.Items);
			end;

			Result := False;
		end;

	begin
		Items.Enumerate(CEachActionItem(@EnumProc), GetContext)
	end;

begin
	//AItems.Update;

	Dialog := TTreeDialog.Create(CurrentThread.LeadWindow, AItems.Title);
	Dialog.Owner := Garbage;
	Dialog.NilIsInvalid := True;
	Dialog.Tree.CustomDraw := CustomDraw;

	First := nil;

	AddItems(Dialog.Root, AItems);

	if Dialog.Selected = nil then
		Dialog.Selected := First;

	Result := nil;
	if Dialog.Execute then
		Result := IActionItem(Dialog.Selected.Data);
end;

procedure AlignButtons(Dialog: T0DialogWindow; const Buttons: array of TButton);
var
	I, J: Integer;
	R: TRect;
begin
	R := Dialog.ClientRect;
	for I := High(Buttons) downto Low(Buttons) do begin
		if Buttons[I] <> nil then begin
			Dec(R.Bottom, Buttons[I].BoundsSize.CY + 3);
			if Dialog.WindowStyle[WS_THICKFRAME] then	begin
				Dec(R.Right, GetSystemMetrics(SM_CXVSCROLL));
			end;
			for J := High(Buttons) downto Low(Buttons) do begin
				if Buttons[J] <> nil then begin
					Dec(R.Right, Buttons[J].BoundsSize.CX);
					Buttons[J].Position := R.BottomRight;
					Dec(R.Right, 3);
				end;
			end;
			Break;
		end;
	end;
end;

const
	Point_4_4: TPoint = (X: 4; Y: 4);

{ TInputDialog }

function TInputDialog.Add: T0ControlItem;
begin
	Assert(FHistory);
	Result := TComboBox(FEdit).Add;
end;

procedure TInputDialog.Clear;
begin
	Assert(FHistory);
	TComboBox(FEdit).Clear;
end;

constructor TInputDialog.Create(AParent: T0TopWindow; const AText: WideString; AHistory: Boolean);
begin
	FHistory := AHistory;
	inherited Create(AParent, AText);
end;

function TInputDialog.GetValue: WideString;
begin
	Result := FEdit.Text
end;

procedure TInputDialog.HandleNotification(Sender: TAnyWindow; Info: PNMHDR; var Result: LRESULT);
begin
	inherited;
	if (Sender = FOK) and (Info.code = BN_CLICKED + NM_COMMAND) then
	begin
		if Valid then
		begin
			DialogResult := ID_OK;
			Close;
		end	
	end;
end;

procedure TInputDialog.Initialize;
const
	S: TSize = (CX: 320; CY: UseDefault);
begin
	inherited;

	if FHistory then
		FEdit := TComboBox.Create(Self, '', Point_4_4, S)
	else
		FEdit := TLineField.Create(Self, '', Point_4_4, S);

	FOK := TButton.Create(Self, SOK, ZeroPoint, DefaultSize);
	FOK.IsDefault := True;

	FCancel := TCloseButton.Create(Self, SCancel, ZeroPoint, DefaultSize);
end;

procedure TInputDialog.SelectAll;
begin
	FCaretIndex := 0;
	FSelectedLength := FEdit.SendMessageW(WM_GETTEXTLENGTH, 0, 0);
end;

procedure TInputDialog.Setup;
var
	Y: Integer;
begin
	Y := FEdit.BoundsRect.Bottom + FSpaceHeight + 4;

	FOk.Position := Point(8 + 320 - 160, Y);
	FCancel.Position := Point(8 + 320 - 82, Y);

	ClientSize := Size(8 + 320, Y + 29);

	ActiveControl := FEdit;

	inherited;
end;

procedure TInputDialog.SetValue(const Value: WideString);
begin
	FEdit.Text := Value;
	SelectAll
end;

procedure TInputDialog.WMActivate(var Message: TWMActivate);
begin
	inherited;
	if Message.Active <> 0 then
	begin
		if Edit is TLineField then
		begin
			TLineField(FEdit).CaretIndex := FCaretIndex;
			TLineField(FEdit).SelectedLength := FSelectedLength;
		end
		else
		begin
			TComboBox(FEdit).CaretIndex := FCaretIndex;
			TComboBox(FEdit).SelectedLength := FSelectedLength;
		end;
	end;
end;

{ TSelectDialog }

function TSelectDialog.Add: T0ControlItem;
begin
	Result := FList.Add
end;

procedure TSelectDialog.AdjustSize;
var
	DC: HDC;
	W: Integer;

	function EnumProc(I: T0ControlItem): Boolean cdecl;
	var
		TextSize: TSize;
		S: WideString;
	begin
		S := I.Text;
		GetTextExtentPoint32W(DC, PWideChar(S), Length(S), TextSize);
		if TextSize.CX > W then W := TextSize.CX;

		Result := False;
	end;

var
	Client, CS: TSize;
	P: TPoint;
	HeightCount: Integer;
begin
	W := 160;
	DC := GetDC(0);
	SaveDC(DC);
	SelectObject(DC, FList.Font.Handle);

	FList.Enumerate(CEachItem(@EnumProc), GetContext);

	RestoreDC(DC, -1);
	ReleaseDC(0, DC);

	HeightCount := FList.Count;
	if HeightCount > 8 then HeightCount := 8;

	CS.CX := W + 8;
	CS.CY := FList.ItemHeight * HeightCount;
	FList.ClientSize := CS;

	if SpaceSize.cy <= 25 then
	begin
		P.Y := FList.BoundsRect.Bottom + FSpaceSize.CY + 4;
		Client.CX := FList.BoundsSize.CX + 8;
	end
	else
	begin
		P.Y := FList.BoundsRect.Bottom;
		if P.Y < SpaceSize.CY then P.Y := SpaceSize.CY + 4;
		Inc(P.Y, 4);
		Client.CX := FList.BoundsSize.CX + SpaceSize.cx + 8;
	end;

	P.X := Client.CX - 160; FOk.Position := P;
	P.X := Client.CX -  82; FCancel.Position := P;

	Client.CY := FOk.BoundsRect.Bottom + 4;
	ClientSize := Client;
end;

procedure TSelectDialog.Clear;
begin
	FList.Clear
end;

function TSelectDialog.Enumerate(Proc: CEachItem; Context: Pointer): T0ControlItem;
begin
	Result := FList.Enumerate(Proc, Context)
end;

function TSelectDialog.GetSelected: T0ControlItem;
begin
	Result := FList.Selected
end;

function TSelectDialog.GetSpaceRect: TRect;
begin
	if SpaceSize.cy <= 25 then
	begin
		Result.Left := 4;
		Result.Top := List.BoundsRect.Bottom + 2;
		Result.Right := FSpaceSize.cx;
		Result.Bottom := Result.Top + FSpaceSize.cy;
	end
	else
	begin
		Result.Left := List.BoundsRect.Right + 2;
		Result.Top := 4;
		Result.Right := Result.Left + FSpaceSize.CX;
		Result.Bottom := FOk.BoundsRect.Top;
	end
end;

procedure TSelectDialog.HandleNotification(Sender: TAnyWindow; Info: PNMHDR; var Result: LRESULT);
begin
	if (Sender = FOK) and (Info.code = BN_CLICKED + NM_COMMAND) then
	begin
		if Valid then
		begin
			DialogResult := ID_OK;
			Close;
		end
	end
	else if (Sender = FList) and (Info.code = LBN_SELCHANGE + NM_COMMAND) then
	begin
		if Assigned(FOnSelectionChange) then FOnSelectionChange(Context);
	end;
	inherited;
end;

procedure TSelectDialog.Initialize;
begin
	inherited;

	ImeAllowed := False;

	FOK := TButton.Create(Self, SOK, ZeroPoint, DefaultSize);
	FOK.IsDefault := True;

	FCancel := TCloseButton.Create(Self, SCancel, ZeroPoint, DefaultSize);

	FList := TListBox.Create(Self, Point_4_4, ZeroSize);
end;

procedure TSelectDialog.SetSelected(const Value: T0ControlItem);
begin
	Assert(Value is TListBoxItem);
	FList.Selected := Value
end;

procedure TSelectDialog.SetSpaceSize(const Value: TSize);
begin
	FSpaceSize := Value;
	AdjustSize;
end;

procedure TSelectDialog.Setup;
begin
	if (FSpaceSize.cx = 0) and (FSpaceSize.cy = 0) then
	begin
		AdjustSize;
	end;

	if FList.SelectedIndex < 0 then
		FList.SelectedIndex := 0;

	ActiveControl := FList;

	inherited;
end;

{ TTreeDialog }

function TTreeDialog.GetRoot: TTreeViewRoot;
begin
	Result := Tree.Root
end;

function TTreeDialog.GetSelected: TAnyTreeViewNode;
begin
	Result := Tree.Selected
end;

procedure TTreeDialog.HandleNotification(Sender: TAnyWindow; Info: PNMHDR; var Result: LRESULT);
begin
	if (Sender = FOK) and (Info.code = BN_CLICKED + NM_COMMAND) then
	begin
		DialogResult := ID_OK;
		Close;
	end
	else if (Info.code = TVN_SELCHANGEDW) then
	begin
		UpdateOkButton;
	end;
	inherited;
end;

procedure TTreeDialog.Initialize;
const
	C: TSize = (CX: 320; CY: 0);
begin
	inherited;

	WindowStyle[WS_THICKFRAME] := True;

	FParentIsInvalid := True;
	ImeAllowed := False;

	ClientSize := C;

	FTree := TTreeView.Create(Self, Point_4_4, Size(313, 0),
		[tvsRoot, tvsButtons, tvsAutoExpand]);

	FOK := TButton.Create(Self, SOK, ZeroPoint, DefaultSize);
	FOK.IsDefault := True;

	FCancel := TCloseButton.Create(Self, SCancel, ZeroPoint, DefaultSize);
end;

procedure TTreeDialog.Realign;
var
	R: TRect;
begin
	if FTree <> nil then begin
		R := ClientRect;
		Inc(R.Top, 4);
		Dec(R.Right, 3);
		Inc(R.Left, 4);
		Dec(R.Bottom, 33);
		FTree.BoundsRect := R;
	end;
	AlignButtons(Self, [FOK, FCancel]);
end;

procedure TTreeDialog.SetSelected(const Value: TAnyTreeViewNode);
begin
	Tree.Selected := Value
end;

procedure TTreeDialog.Setup;
var
	W, H, MaxW: Integer;
	DC: HDC;
	TreeSize: TSize;

	procedure Enum(Node: T0TreeViewNodes; Level: Integer);

		function EnumProc(Item: TAnyTreeViewNode): Boolean; cdecl;
		var
			S: WideString;
			TextSize: TSize;
		begin
			S := Item.Text;
			GetTextExtentPoint32W(DC, PWideChar(S), Length(S), TextSize);
			Inc(TextSize.CX, Tree.Indent * (Level + 1));
			if TextSize.CX > W then W := TextSize.CX;
			Enum(TAnyTreeViewNode(Item).Children, Level + 1);

			Result := False
		end;

	begin
		Node.Enumerate(CEachTreeNode(@EnumProc), GetContext);
	end;

begin
	DC := GetDC(0);
	SaveDC(DC);
	SelectObject(DC, Tree.Font.Handle);

	W := 0;
	Enum(Tree.Root, 1);

	RestoreDC(DC, -1);
	ReleaseDC(0, DC);

	Inc(W, GetSystemMetrics(SM_CXEDGE) * 2 + GetSystemMetrics(SM_CXVSCROLL));
	{+ 4ł悤. XP΍cȂĂ̂H}

	if Tree.Count <= 10 then begin
		H := Tree.Count * Tree.ItemHeight + GetSystemMetrics(SM_CYEDGE) * 2
	end
	else begin
		H := 10 * Tree.ItemHeight + GetSystemMetrics(SM_CYEDGE) * 2;
	end;

	if W < 170 then begin
		W := 170
	end
	else begin
		MaxW := GetSystemMetrics(SM_CXSCREEN) * 4 div 5;
		if W > MaxW then begin
			W := MaxW;
			Inc(H, GetSystemMetrics(SM_CYHSCROLL));
		end;
	end;

	TreeSize.CX := W;
	TreeSize.CY := H;
	//Tree.BoundsSize := TreeSize;

	ClientSize := Size(TreeSize.CX + 7, TreeSize.CY + 37);

	ActiveControl := Tree;

	UpdateOkButton;

	inherited;
end;

procedure TTreeDialog.UpdateOkButton;
var
	S: TAnyTreeViewNode;
begin
	S := FTree.Selected;
	FOk.Enabled := (S <> nil) and
		(not ParentIsInvalid or (S.Children.Count = 0)) and
		(not NilIsInvalid or (S.Data <> nil));
end;

{ TVersionDialog }

function TVersionDialog.DoExecute: Boolean;
var
	ParWnd: HWND;
begin
	if Parent <> Nil then
		ParWnd:= Parent.Handle
	else
		ParWnd:= 0;

	Result:= VersionDialog(ParWnd, Title);
end;

class function TVersionDialog.Execute(const ATitle: WideString): Boolean;
var
	D: TVersionDialog;
begin
	D := TVersionDialog.Create(CurrentThread.LeadWindow, ATitle);
	D.Owner := Garbage;
	Result := D.Execute;
end;

end.

