unit LCmmCtls;

{*******************************************************}
{                                                       }
{       LightWindows                                    }
{       Common Controls                                 }
{                                                       }
{       2001..2008 YT                                   }
{                                                       }
{*******************************************************}

{$DEFINE IMAGELIST}
{$DEFINE TREEVIEW}
{$DEFINE LISTVIEW}
{$DEFINE TAB}
{$DEFINE STATUS}
{$DEFINE TOOLBAR}
{$DEFINE TOOLBAR_MENU}
{$DEFINE UPDOWN}
{$DEFINE PROGRESS}
{$DEFINE DATETIMEPICK}
{$DEFINE HEADER}

interface

uses
	Windows, Messages, SysUtils, Types, CommCtrl,
	SysUtilsW, SysHacks, AlgUtils, WinTips,
	LObjects, LActions, LMenus, LWindows, LStdCtls;

{$IFDEF IMAGELIST}
type
	TAnyImageList = class(T0Object)
	private
		FHandle: HIMAGELIST;
	public
		constructor Create(AHandle: HIMAGELIST);
		procedure Add(Icon: HICON);
		property Handle: HIMAGELIST read FHandle;
	end;

	TImageList = class(TAnyImageList)
	public
		constructor Create(AWidth, AHeight: Integer; AMask: Boolean);
		procedure BeforeDestruction; override;
	end;
{$ENDIF}

{$IFDEF TREEVIEW}
type
	TTreeView = class;
	TTreeViewNodes = class;
	T0TreeViewNodes = class;

	TAnyTreeViewNode = class(T0ControlItem)
	private
		FTreeView: TTreeView;
		FHandle: HTREEITEM;
		FChildren: TTreeViewNodes;
		function GetChildren: TTreeViewNodes;
		function GetImageIndex: Integer;
		procedure SetImageIndex(const Value: Integer);
		function GetHasChildren: Boolean;
		procedure SetHasChildren(const Value: Boolean);
		function GetExpanded: Boolean;
		procedure SetExpanded(const Value: Boolean);
		function GetLevel: Integer;
	protected
		function GetData: Pointer; override;
		procedure SetData(const Value: Pointer); override;
		function GetText: WideString; override;
		procedure SetText(const Value: WideString); override;
		function Successor: TAnyTreeViewNode;
	public
		constructor Create(ATreeView: TTreeView; AHandle: HTREEITEM);
		function NeedChildren: Boolean;
		function Brothers: T0TreeViewNodes;
		property TreeView: TTreeView read FTreeView;
		property Handle: HTREEITEM read FHandle;
		property ImageIndex: Integer read GetImageIndex write SetImageIndex;
		property Children: TTreeViewNodes read GetChildren;
		property HasChildren: Boolean read  GetHasChildren write SetHasChildren;
		property Expanded: Boolean read GetExpanded write SetExpanded;
		property Level: Integer read GetLevel;
	end;

	TTreeViewNode = class(TAnyTreeViewNode)
	private
		FData: Pointer;
	protected
		function GetData: Pointer; override;
		procedure SetData(const Value: Pointer); override;
	public
		constructor Create(ATreeNodes: T0TreeViewNodes);
	end;

	CEachTreeNode = function(Node: TAnyTreeViewNode; Context: Pointer): Boolean cdecl;
	CCompareTreeNode = function(X, Y: TTreeViewNode; Context: Pointer): Integer cdecl;

	T0TreeViewNodes = class(T0Object)
	private
		FTreeView: TTreeView;
	protected
		function FirstHandle: HTREEITEM; virtual; abstract;
		function ParentHandle: HTREEITEM; virtual; abstract;
	public
		constructor Create(ATreeView: TTreeView);
		function First: TAnyTreeViewNode;
		function Add: TAnyTreeViewNode;
		function Enumerate(Proc: CEachTreeNode; Context: Pointer): TAnyTreeViewNode;
		function Count: Integer;
		procedure Clear;
		procedure Sort(Compare: CCompareTreeNode; Context: Pointer);
		property TreeView: TTreeView read FTreeView;
	end;

	TTreeViewNodes = class(T0TreeViewNodes)
	private
		FParentHandle: HTREEITEM;
		function GetParent: TAnyTreeViewNode;
	protected
		function FirstHandle: HTREEITEM; override;
		function ParentHandle: HTREEITEM; override;
	public
		constructor Create(ATreeView: TTreeView; AParentHandle: HTREEITEM);
		property Parent: TAnyTreeViewNode read GetParent;
	end;

	TTreeViewRoot = class(T0TreeViewNodes)
	protected
		function FirstHandle: HTREEITEM; override;
		function ParentHandle: HTREEITEM; override;
	end;

	TTreeViewCustomDraw = function(TreeView: TTreeView; Node: TAnyTreeViewNode; Info: PNMTVCustomDraw): Boolean of object;

	TTreeView = class(T0SubClassWindow)
	type
		TTreeViewStyles = set of (
			tvsButtons     =  0 {TVS_HASBUTTONS      = $0001} ,
			tvsLines       =  1 {TVS_HASLINES        = $0002} ,
			tvsRoot        =  2 {TVS_LINESATROOT     = $0004} ,
			tvsEditable    =  3 {TVS_EDITLABELS      = $0008} ,
			_tvsNoDragDrop =  4 {TVS_DISABLEDRAGDROP = $0010} ,
			_tvsSelAlways  =  5 {TVS_SHOWSELALWAYS   = $0020} ,
			tvsRightToLeft =  6 {TVS_RTLREADING      = $0040} ,
			tvsNoToolTips  =  7 {TVS_NOTOOLTIPS      = $0080} ,
			tvsCheckBoxs   =  8 {TVS_CHECKBOXES      = $0100} ,
			tvsHotTrack    =  9 {TVS_TRACKSELECT     = $0200} ,
			tvsAutoExpand  = 10 {TVS_SINGLEEXPAND    = $0400} ,
			tvsInfoTip     = 11 {TVS_INFOTIP         = $0800} ,
			tvsRowSelect   = 12 {TVS_FULLROWSELECT   = $1000} ,
			tvsNoScroll    = 13 {TVS_NOSCROLL        = $2000} ,
			tvsOddHeight   = 14 {TVS_NONEVENHEIGHT   = $4000} ,
			_tvsHigh       = 31);
	public
  	type TItem = TAnyTreeViewNode;
	private
		FRoot: TTreeViewRoot;
		FCustomDraw: TTreeViewCustomDraw;
		function GetCount: Integer;
		function GetReadOnly: Boolean;
		procedure SetReadOnly(const Value: Boolean);
		function GetSelected: TAnyTreeViewNode;
		procedure SetSelected(const Value: TAnyTreeViewNode);
{$IFDEF IMAGELIST}
		procedure SetImageList(const Value: TAnyImageList);
{$ENDIF}
	protected
		class function SubClassName: PWideChar; override;
		class procedure CreateClass(var WndClass: WNDCLASSEXW); override;
		function GetLNode(ANode: HTREEITEM): TAnyTreeViewNode;
		function GetNode(ANode: HTREEITEM): TAnyTreeViewNode;
		procedure HandleNotification(Sender: TAnyWindow; Info: PNMHdr; var Result: Integer); override;
	public
		constructor Create(AParent: TAnyWindow; const APosition: TPoint;
						const ASize: TSize; const Styles: TTreeViewStyles);
		function ItemHeight: Integer;
		function Indent: Integer;
		function NodeFromPoint(const P: TPoint): TAnyTreeViewNode;
		procedure Clear;
		procedure SelectWithoutScroll(const Value: TAnyTreeViewNode);
		property Count: Integer read GetCount;
		property Selected: TAnyTreeViewNode read GetSelected write SetSelected;
		property Root: TTreeViewRoot read FRoot;
		property ReadOnly: Boolean read GetReadOnly write SetReadOnly;
		property CustomDraw: TTreeViewCustomDraw read FCustomDraw write FCustomDraw;
{$IFDEF IMAGELIST}
		property ImageList: TAnyImageList write SetImageList;
{$ENDIF}
	end;
{$ENDIF}

{$IFDEF LISTVIEW}
type
	TListView = class;

	TListViewColumn = class(T0Object)
	private
		FListView: TListView;
		FIndex: Integer;
		function GetWidth: Integer;
		procedure SetWidth(const Value: Integer);
	protected
		procedure SetText(const Value: WideString);
		function GetText: WideString;
	public
		constructor Create(AListView: TListView; AIndex: Integer);
		property Text: WideString read GetText write SetText;
		property Width: Integer read GetWidth write SetWidth;
	end;

	TListViewItem = class(T0ControlItem)
	private
		FListView: TListView;
		FIndex: Integer;
		function GetTexts(Index: Integer): WideString;
		procedure SetTexts(Index: Integer; const Value: WideString);
	protected
		function GetData: Pointer; override;
		procedure SetData(const Value: Pointer); override;
		function GetText: WideString; override;
		procedure SetText(const Value: WideString); override;
	public
		constructor Create(AListView: TListView; AIndex: Integer);
		property Texts[Index: Integer]: WideString read GetTexts write SetTexts;
	end;

	TListViewStyles = set of (vsMultiSelection);
	TListViewStyle = (vsReport = LVS_REPORT);

	TListView = class(T0SubClassWindow)
	public
  	type TItem = TListViewItem;
	private
		function GetFullRowSelection: Boolean;
		procedure SetFullRowSelection(const Value: Boolean);
		function GetAt(Index: Integer): TListViewItem;
		function GetSelected: TListViewItem;
		procedure SetSelected(const Value: TListViewItem);
	protected
		class function SubClassName: PWideChar; override;
		class procedure CreateClass(var WndClass: WNDCLASSEXW); override;
		function GetColumn(Index: Integer): TListViewColumn;
		function GetItem(Index: Integer): TListViewItem;
	public
		constructor Create(AParent: TAnyWindow; const APosition: TPoint; const ASize: TSize; AStyles: TListViewStyles; AStyle: TListViewStyle);
		function AddColumn: TListViewColumn;
		function ColumnCount: Integer;
		function Add: TListViewItem;
		function Count: Integer;
		function AtPos(const P: TPoint): TListViewItem;
		function Find(const AText: WideString): TListViewItem;
		procedure Reserve(ACount: Integer);
		property Selected: TListViewItem read GetSelected write SetSelected;
		property At[Index: Integer]: TListViewItem read GetAt;
		property FullRowSelection: Boolean read GetFullRowSelection write SetFullRowSelection;
	end;
{$ENDIF}

{$IFDEF TAB}
const
	TCN_MCLICK = TCN_LAST + 1; {邩c}
	TCN_RCLICK = TCN_LAST + 2; {邩c}
	TCN_DBLCLK = TCN_LAST + 3;

type
	TTabControl = class(T0SubClassWindow)
	public type
		TStyles = set of (
			tcsScrollOpposite =  0 {TCS_SCROLLOPPOSITE    = $0001} ,
			_tcsBottomRight   =  1 {TCS_BOTTOM            = $0002} ,
			tcsMultiSelect    =  2 {TCS_MULTISELECT       = $0004} ,
			tcsFlat           =  3 {TCS_FLATBUTTONS       = $0008} ,
			tcsForceIconLeft  =  4 {TCS_FORCEICONLEFT     = $0010} ,
			tcsForceLabelLeft =  5 {TCS_FORCELABELLEFT    = $0020} ,
			tcsHotTrack       =  6 {TCS_HOTTRACK          = $0040} ,
			_tcsVertical      =  7 {TCS_VERTICAL          = $0080} ,
			tcsButton         =  8 {TCS_BUTTONS           = $0100} ,
			tcsMultiLine      =  9 {TCS_MULTILINE         = $0200} ,
			tcsFixedWidth     = 10 {TCS_FIXEDWIDTH        = $0400} ,
			tcsRaggedRight    = 11 {TCS_RAGGEDRIGHT       = $0800} ,
			_tcsFocusOnClick  = 12 {TCS_FOCUSONBUTTONDOWN = $1000} ,
			tcsOwnerDrawFixed = 13 {TCS_OWNERDRAWFIXED    = $2000} ,
			tcsToolTip        = 14 {TCS_TOOLTIPS          = $4000} ,
			tcsNoFocus        = 15 {TCS_FOCUSNEVER        = $8000} ,
			_tcsHigh = 31);
		TItem = class(T0ControlItem)
		private
			FTabControl: TTabControl;
			FIndex: Integer;
		protected
			function GetData: Pointer; override;
			procedure SetData(const Value: Pointer); override;
			function GetText: WideString; override;
			procedure SetText(const Value: WideString); override;
		public
			constructor Create(ATabControl: TTabControl; AIndex: Integer);
		end;
	private
		function GetCount: Integer;
		function GetAt(AIndex: Integer): TItem;
		function GetSelectedIndex: Integer;
		procedure SetSelectedIndex(const Value: Integer);
		procedure TCMAdjustRect(var Message: TMessage); message TCM_ADJUSTRECT;
		procedure WMMButtonDown(var Message: TWMMButtonDown); message WM_MBUTTONDOWN;
		procedure WMMButtonUp(var Message: TWMMButtonUp); message WM_MBUTTONUP;
		procedure WMRButtonDown(var Message: TWMRButtonDown); message WM_RBUTTONDOWN;
		procedure WMRButtonUp(var Message: TWMRButtonUp); message WM_RBUTTONUP;
		procedure WMLButtonDblClk(var Message: TWMLButtonDblClk); message WM_LBUTTONDBLCLK;
	protected
		class function SubClassName: PWideChar; override;
		class procedure CreateClass(var WndClass: WNDCLASSEXW); override;
		procedure Initialize; override;
		procedure DialogKey(var Message: TWMKey); override;
		procedure ProcessMouseUp(var Message: TWMLButtonUp; Notify: Integer);
	public
		constructor Create(AParent: TAnyWindow; const APosition: TPoint;
			const ASize: TSize; ASide: TAlignment; const Styles: TStyles);
		function Add: TItem;
		procedure Remove(Item: T0ControlItem);
		procedure Clear;
		function DisplayRect: TRect; virtual;
		function DisplaySize: TSize;
		procedure Realign; override;
		function Enumerate(Proc: CEachItem; Context: Pointer): T0ControlItem;
		property SelectedIndex: Integer read GetSelectedIndex write SetSelectedIndex;
		property At[AIndex: Integer]: TItem read GetAt;
		property Count: Integer read GetCount;
		function OnSelectionChanged(Event: PNMHdr): Boolean; inline;
	end;
{$ENDIF}

{$IFDEF STATUS}
type
	TStatusPane = class;
	TStatusBar = class;

	TStatusPane = class(T0ControlItem)
	private
		FStatusBar: TStatusBar;
		FPred: TStatusPane;
		FSucc: TStatusPane;
		FText: WideString;
		FData: Pointer;
		FWidth: Integer;
		FBevel: TBevel;
		FAlignment: THorzAlignment;
		procedure SetBevel(const Value: TBevel);
		procedure SetWidth(const Value: Integer);
		procedure SetAlignment(const Value: THorzAlignment);
	protected
		function GetData: Pointer; override;
		procedure SetData(const Value: Pointer); override;
		function GetText: WideString; override;
		procedure SetText(const Value: WideString); override;
		procedure DoContextMenu(const P: TPoint); virtual;
	public
		constructor Create(AStatusBar: TStatusBar; ANext: TStatusPane);
		procedure BeforeDestruction; override;
		procedure Update; virtual;
		procedure AdjustSize;
		property Text: WideString read FText write SetText;
		property Bevel: TBevel read FBevel write SetBevel;
		property Width: Integer read FWidth write SetWidth;
		property Alignment: THorzAlignment read FAlignment write SetAlignment;
	end;

	TStatusBar = class(T0SubClassWindow)
	public
  	type TItem = TStatusPane;
	private
		FFirstPane: TStatusPane;
		FLastPane: TStatusPane;
		FHintWidth: Integer;
		FModified: Boolean;
		procedure WMSize(var Message: TWMSize); message WM_SIZE;
		procedure WMMove(var Message: TWMMove); message WM_MOVE;
		procedure WMSetText(var Message: TMessage); message WM_SETTEXT;
		procedure WMContextMenu(var Message: TWMContextMenu); message WM_CONTEXTMENU;
	protected
		class function SubClassName: PWideChar; override;
		class procedure CreateClass(var WndClass: WNDCLASSEXW); override;
		procedure Modify;
		procedure Update; override;
	public
		constructor Create(AParent: TAnyWindow);
		function PaneFromPoint(const P: TPoint): TStatusPane;
		procedure AfterConstruction; override;
		procedure BeforeDestruction; override;
		property HintWidth: Integer read FHintWidth;
	end;

type
	TActionStatusPane = class(TStatusPane)
	private
		FAction: IAction;
	protected
		procedure DoContextMenu(const P: TPoint); override;
	public
		constructor Create(AStatusBar: TStatusBar; ANext: TStatusPane; const AAction: IAction);
	end;

type
	TLockedKeyStatusPane = class(TStatusPane)
	private type
		TLockedKeyState = (lkUnlocked, lkCaps, lkKana);
		TLockedKeyWatcher = class(T0IdleWorker)
		private
			FText: WideString;
			FState: TLockedKeyState;
		public
			procedure Update; override;
			property Text: WideString read FText;
			property State: TLockedKeyState read FState;
			class function Instance: TLockedKeyWatcher;
		end;
	private
		FWatcher: TLockedKeyWatcher;
	public
		constructor Create(AStatusBar: TStatusBar; ANext: TStatusPane; AWatcher: TLockedKeyWatcher = nil);
		procedure Update; override;
	public const
		LockedKeyFaces: array[TLockedKeyState] of WideString = ('', 'CAPS', 'KANA');
	end;
{$ENDIF}

{$IFDEF TOOLBAR}
type
	T0ToolBarControl = class;

	T0ToolButton = class(T0ControlItem)
	private
		FToolBar: T0ToolBarControl;
		FID: Cardinal;
		function GetIndex: Integer;
		function GetState: DWORD;
		function GetStyle: DWORD;
		procedure SetState(const Value: DWORD);
		procedure SetStyle(const Value: DWORD);
	protected
		function GetText: WideString; override;
		procedure SetText(const Value: WideString); override;
		function GetSub: TAnyMenu; virtual; abstract;
		procedure SetSub(const Value: TAnyMenu); virtual; abstract;
		property Style: DWORD read GetStyle write SetStyle;
		property State: DWORD read GetState write SetState;
	public
		property ID: Cardinal read FID;
		property Index: Integer read GetIndex;
		property ToolBar: T0ToolBarControl read FToolBar;
		property Sub: TAnyMenu read GetSub write SetSub;
	end;

	TToolButton = class(T0ToolButton)
	private
		FData: Pointer;
		FSub: TAnyMenu;
	protected
		function GetData: Pointer; override;
		procedure SetData(const Value: Pointer); override;
		function GetSub: TAnyMenu; override;
		procedure SetSub(const Value: TAnyMenu); override;
		function TrackSubMenu: Boolean;
	public
		constructor Create(ABar: T0ToolBarControl; ANext: T0ToolButton);
		procedure BeforeDestruction; override;
	end;

	T0ToolBarControl = class(T0SubClassWindow)
	public
  	type TItem = T0ToolButton;
	private
		FActivedControl: T0Window;
		FDropping: Integer;
		FInTracking: Boolean;
		FAltPushed: Boolean;
		function GetAt(AIndex: Integer): T0ToolButton;
		function GetCount: Integer;
		procedure WMKillFocus(var Message: TWMKillFocus); message WM_KILLFOCUS;
	protected
		procedure Initialize; override;
		procedure HandleNotification(Sender: TAnyWindow; Info: PNMHDR; var Result: LRESULT); override;
		procedure SubMenuHook(var Message: MSG); virtual;
		class function SubClassName: PWideChar; override;
		class procedure CreateClass(var WndClass: TWNDCLASSEXW); override;
	public
		constructor Create(AParent: TAnyWindow);
		procedure Track;
		procedure Cancel;
		function ItemAtPos(const P: TPoint): T0ToolButton;
		function ItemIndexAtPos(const P: TPoint): Integer;
		property Count: Integer read GetCount;
		property At[AIndex: Integer]: T0ToolButton read GetAt; default;
	end;

{$IFDEF TOOLBAR_MENU}
type
	TMenuBar = class(T0ToolBarControl)
	private
		procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
	protected
		procedure Initialize; override;
		function KeyDown(ShiftState: TShiftKeys; Code: Byte): Boolean; override;
		function KeyUp(ShiftState: TShiftKeys; Code: Byte): Boolean; override;
		procedure SubMenuHook(var Message: MSG); override;
	public
		constructor Create(AParent: TAnyWindow);
		function ProcessDialogKey(var Message: TWMKey): Boolean;
	end;

function CreateMenuItem(Menu: TMenuBar; const Text: WideString): TToolButton; overload;
function CreateSubMenu(Item: TToolButton): TPopupMenu; overload;
{$ENDIF}
{$ENDIF}

{$IFDEF PROGRESS}
type
	TProgressBar = class(T0SubClassWindow)
	private
		function GetMax: Integer;
		function GetMin: Integer;
		function GetValue: Integer;
		procedure SetMax(const Value: Integer);
		procedure SetMin(const Value: Integer);
		procedure SetValue(const Value: Integer);
	protected
		class procedure CreateClass(var WndClass: TWNDCLASSEXW); override;
		class function SubClassName: PWideChar; override;
	public
		constructor Create(AParent: TAnyWindow; const APosition: TPoint; const ASize: TSize);
		property Min: Integer read GetMin write SetMin;
		property Max: Integer read GetMax write SetMax;
		property Value: Integer read GetValue write SetValue;
	end;
{$ENDIF}

{$IFDEF UPDOWN}
type
	TUpDown = class(T0SubClassWindow)
	private
		function GetMax: Integer;
		function GetMin: Integer;
		function GetValue: Integer;
		procedure SetMax(const Value: Integer);
		procedure SetMin(const Value: Integer);
		procedure SetValue(const Value: Integer);
	protected
		class procedure CreateClass(var WndClass: TWNDCLASSEXW); override;
		class function SubClassName: PWideChar; override;
		class function GetDefaultSize: tagSIZE; override;
	public
		constructor Create(AParent: TAnyWindow; const APosition: TPoint; const ASize: TSize);
		procedure Associate(Wnd: TAnyWindow);
		property Min: Integer read GetMin write SetMin;
		property Max: Integer read GetMax write SetMax;
		property Value: Integer read GetValue write SetValue;
	end;

type
	TNumberField = class(TLineField)
	private
		FUpDown: TUpDown;
		FValue: Integer;
		FMin: Integer;
		FMax: Integer;
	protected
		procedure Initialize; override;
	public
		constructor Create(AParent: TAnyWindow; Value, Min, Max: Integer; const APosition: TPoint; const ASize: TSize);
		property UpDown: TUpDown read FUpDown;
	end;
{$ENDIF}

{$IFDEF DATETIMEPICK}
type
	TDateTimePickerStyles = set of (
		dtsUpDown      = 0 {DTS_UPDOWN          = $0001} ,
		dtsCheckBoxs   = 1 {DTS_SHOWNONE        = $0002} ,
		dtsLongDate    = 2 {DTS_LONGDATEFORMAT  = $0004} ,
		dtsTime        = 3 {DTS_TIMEFORMAT      = $0009} ,
		dtsAppCanParse = 4 {DTS_APPCANPARSE     = $0010} ,
		dtsRightToLeft = 5 {DTS_RIGHTALIGN      = $0020} ,
		_dtsHigh = 31);

	TDateTimePicker = class(T0SubClassWindow)
	type
		TStyles = TDateTimePickerStyles;
	private
		FValue: TSystemTime;
		FChecked: Boolean;
		procedure SetFormat(const Value: WideString);
		procedure SetChecked(const Value: Boolean);
		procedure SetValue(const Value: TSystemTime);
	protected
		class procedure CreateClass(var WndClass: TWNDCLASSEXW); override;
		class function SubClassName: PWideChar; override;
		class function GetDefaultSize: TSize; override;
		procedure HandleNotification(Sender: TAnyWindow; Info: PNMHdr; var Result: Integer); override;
	public
		constructor Create(AParent: TAnyWindow; const APosition: TPoint; const ASize: TSize; const Styles: TDateTimePickerStyles);
		property Format: WideString write SetFormat;
		property Checked: Boolean read FChecked write SetChecked;
		property Value: TSystemTime read FValue write SetValue;
	end;
{$ENDIF}

{$IFDEF HEADER}
type
	THeaderControlStyles = set of (
		hdsButtons  =  1 {HDS_BUTTONS  = $00000002},
		hdsHotTrack =  2 {HDS_HOTTRACK = $00000004},
		hdsHidden   =  3 {HDS_HIDDEN   = $00000008},
		hdsDragDrop =  6 {HDS_DRAGDROP = $00000040},
		hdsFullDrag =  7 {HDS_FULLDRAG = $00000080},
		hdsNoSizing = 11 {HDS_NOSIZING = $00000800},
		_hdsHigh = 31);

	THeaderControl = class(T0SubClassWindow)
	type
		TStyles = THeaderControlStyles;
		TItem = class(T0ControlItem)
		private
			FHeader: THeaderControl;
			FIndex: Integer;
			procedure SetWidth(const Value: Integer);
			function GetWidth: Integer;
		protected
			function GetData: Pointer; override;
			procedure SetData(const Value: Pointer); override;
			function GetText: WideString; override;
			procedure SetText(const Value: WideString); override;
		public
			constructor Create(AHeaderControl: THeaderControl; AIndex: Integer);
			property Width: Integer read GetWidth write SetWidth;
		end;
	private
  	FEnableDrag: Boolean;
		function GetCount: Integer;
		procedure WMSetCursor(var Message: TWMSetCursor); message WM_SETCURSOR;
	protected
		class function SubClassName: PWideChar; override;
		class function GetDefaultSize: TSize; override;
    procedure HandleNotification(Sender: TAnyWindow; Info: PNMHdr;
      var Result: Integer); override;
	public
		constructor Create(AParent: TAnyWindow; const APosition: TPoint; const ASize: TSize; const Styles: TStyles);
		procedure AfterConstruction; override;
		function Add: TItem;
		property Count: Integer read GetCount;
		property EnableDrag: Boolean read FEnableDrag write FEnableDrag;
	end;
{$ENDIF}

{ Utility Routines }

function InitCommonControl(CC: Integer): Boolean;

implementation

function InitCommonControl(CC: Integer): Boolean;
var
	ICC: TINITCOMMONCONTROLSEX;
begin
	ICC.dwSize := SizeOf(TINITCOMMONCONTROLSEX);
	ICC.dwICC := CC;
	Result := InitCommonControlsEx(ICC);
	if not Result then InitCommonControls;
end;

{$IFDEF IMAGELIST}

{ TAnyImageList }

procedure TAnyImageList.Add(Icon: HICON);
begin
	{$IFOPT C+} Assert( {$ENDIF} ImageList_AddIcon(Handle, Icon) {$IFOPT C+} >= 0) {$ENDIF}
end;

constructor TAnyImageList.Create(AHandle: HIMAGELIST);
begin
	inherited Create;
	FHandle := AHandle
end;

{ TImageList }

procedure TImageList.BeforeDestruction;
begin
	ImageList_Destroy(Handle);
	inherited;
end;

constructor TImageList.Create(AWidth, AHeight: Integer; AMask: Boolean);
const
	M: array[Boolean] of DWORD = (ILC_COLORDDB, ILC_COLORDDB or ILC_MASK);
begin
	inherited Create(ImageList_Create(AWidth, AHeight, M[AMask], 0, 4));
end;

{$ENDIF}

{$IFDEF TREEVIEW}

{ TTreeViewNode }

constructor TAnyTreeViewNode.Create(ATreeView: TTreeView; AHandle: HTREEITEM);
begin
	inherited Create;
	FTreeView := ATreeView;
	FHandle := AHandle;
end;

function TAnyTreeViewNode.GetData: Pointer;
var
	Data: TV_ITEMW;
begin
	Data.mask := TVIF_PARAM;
	Data.hItem := Handle;

	TreeView.SendMessageW(TVM_GETITEMW, 0, LPARAM(@Data));

	Result := Pointer(Data.lParam);
end;

procedure TAnyTreeViewNode.SetData(const Value: Pointer);
var
	Data: TV_ITEMW;
begin
	Data.mask := TVIF_PARAM;
	Data.hItem := Handle;
	Data.lParam := LPARAM(Value);

	TreeView.SendMessageW(TVM_SETITEMW, 0, LPARAM(@Data));
end;

function TAnyTreeViewNode.Successor: TAnyTreeViewNode;
begin
	Result := TreeView.GetNode(HTREEITEM(
					TreeView.SendMessageW(TVM_GETNEXTITEM, TVGN_NEXT, LPARAM(Handle))))
end;

function TAnyTreeViewNode.GetChildren: TTreeViewNodes;
begin
	if FChildren = nil then
	begin
		FChildren := TTreeViewNodes.Create(TreeView, Handle);
		FChildren.Owner := Self;
	end;
	Result := FChildren;
end;

function TAnyTreeViewNode.GetText: WideString;
var
	Data: TV_ITEMW;
	Buffer: array[0..1023] of WideChar;
begin
	Data.mask := TVIF_TEXT;
	Data.hItem := Handle;
	Data.pszText := Buffer;
	Data.cchTextMax := Length(Buffer);

	TreeView.SendMessageW(TVM_GETITEMW, 0, LPARAM(@Data));
	Result := Buffer;
end;

procedure TAnyTreeViewNode.SetText(const Value: WideString);
var
	Data: TV_ITEMW;
begin
	Data.mask := TVIF_TEXT;
	Data.hItem := Handle;
	Data.pszText := PWideChar(Value);

	TreeView.SendMessageW(TVM_SETITEMW, 0, LPARAM(@Data));
end;

function TAnyTreeViewNode.GetImageIndex: Integer;
var
	Data: TV_ITEMW;
begin
	Data.mask := TVIF_IMAGE;
	Data.hItem := Handle;
	TreeView.SendMessageW(TVM_GETITEMW, 0, LPARAM(@Data));
	Result := Data.iImage;
end;

procedure TAnyTreeViewNode.SetImageIndex(const Value: Integer);
var
	Data: TV_ITEMW;
begin
	Data.mask := TVIF_IMAGE or TVIF_SELECTEDIMAGE;
	Data.hItem := Handle;
	Data.iImage := Value;
	Data.iSelectedImage := Value;
	TreeView.SendMessageW(TVM_SETITEMW, 0, LPARAM(@Data));
end;

function TAnyTreeViewNode.GetHasChildren: Boolean;
var
	Data: TV_ITEMW;
begin
	Data.mask := TVIF_CHILDREN;
	Data.hItem := Handle;
	TreeView.SendMessageW(TVM_GETITEMW, 0, LPARAM(@Data));
	Result := Data.cChildren <> 0
end;

procedure TAnyTreeViewNode.SetHasChildren(const Value: Boolean);
var
	Data: TV_ITEMW;
begin
	Data.mask := TVIF_CHILDREN;
	Data.hItem := Handle;
	Data.cChildren := Ord(Value);
	TreeView.SendMessageW(TVM_SETITEMW, 0, LPARAM(@Data));
end;

function TAnyTreeViewNode.NeedChildren: Boolean;
begin
	Result := HasChildren and
		(TreeView.SendMessageW(TVM_GETNEXTITEM, TVGN_CHILD, LPARAM(Handle)) = 0)
end;

function TAnyTreeViewNode.GetExpanded: Boolean;
var
	Item: TTVItem;
begin
	Result := False;
	Item.mask := TVIF_STATE;
	Item.hItem := Handle;
	if TreeView.SendMessageW(TVM_GETITEMW, 0, LPARAM(@Item)) <> 0 then
		Result := Item.state and TVIS_EXPANDED <> 0;
end;

procedure TAnyTreeViewNode.SetExpanded(const Value: Boolean);
var
	Item: TTVItem;
begin
	Item.mask := TVIF_STATE;
	Item.hItem := Handle;
	Item.stateMask := TVIS_EXPANDED;
	Item.state := TVIS_EXPANDED and not (Ord(Value) - 1);
	TreeView.SendMessageW(TVM_SETITEMW, 0, LPARAM(@Item))
end;

function TAnyTreeViewNode.Brothers: T0TreeViewNodes;
var
	Node: TAnyTreeViewNode;
begin
	Node := TreeView.GetNode(HTREEITEM(TreeView.SendMessageW(TVM_GETNEXTITEM, TVGN_PARENT, LPARAM(Handle))));
	if Node = nil then
		Result := TreeView.Root
	else
		Result := Node.Children
end;

function TAnyTreeViewNode.GetLevel: Integer;
var
	Node: HTREEITEM;
begin
	Result := -1;
	Node := Handle;
	repeat
		Node := HTREEITEM(TreeView.SendMessageW(TVM_GETNEXTITEM, TVGN_PARENT, LPARAM(Node)));
		Inc(Result);
	until Node = nil;
end;

{ TTreeViewNode }

constructor TTreeViewNode.Create(ATreeNodes: T0TreeViewNodes);
var
	Data: TV_INSERTSTRUCTW;
	H: HTREEITEM;
begin
	Data.hParent := ATreeNodes.ParentHandle;
	Data.hInsertAfter := nil;
	Data.item.mask := TVIF_PARAM;
	Data.item.lParam := LPARAM(Self);
	H := HTREEITEM(ATreeNodes.TreeView.SendMessageW(TVM_INSERTITEMW, 0, LPARAM(@Data)));
	inherited Create(ATreeNodes.TreeView, H);
end;

function TTreeViewNode.GetData: Pointer;
begin
	Result := FData
end;

procedure TTreeViewNode.SetData(const Value: Pointer);
begin
	FData := Value
end;

{ T0TreeViewNodes }

function T0TreeViewNodes.Add: TAnyTreeViewNode;
begin
	Result := TTreeViewNode.Create(Self);
end;

procedure T0TreeViewNodes.Clear;
var
	F: HTREEITEM;
begin
	repeat
		F := FirstHandle;
		if F = nil then Exit;
		TreeView.SendMessageW(TVM_DELETEITEM, 0, LPARAM(F));
	until False;
end;

function T0TreeViewNodes.Count: Integer;
var
	Node: HTREEITEM;
begin
	Result := 0;
	Node := FirstHandle;
	while Node <> nil do
	begin
		Inc(Result);
		Node := HTREEITEM(TreeView.SendMessageW(TVM_GETNEXTITEM, TVGN_NEXT, LPARAM(Node)))
	end;
end;

constructor T0TreeViewNodes.Create(ATreeView: TTreeView);
begin
	inherited Create;
	FTreeView := ATreeView;
end;

function T0TreeViewNodes.Enumerate(Proc: CEachTreeNode; Context: Pointer): TAnyTreeViewNode;
var
	Next: TAnyTreeViewNode;
begin
	Result := First;
	while Result <> nil do
	begin
		Next := Result.Successor;
		if Proc(Result, Context) then Exit;
		Result := Next
	end;
end;

function T0TreeViewNodes.First: TAnyTreeViewNode;
begin
	Result := TreeView.GetNode(FirstHandle)
end;

type
	PTreeNodeCompareParam = ^TTreeNodeCompareParam;
	TTreeNodeCompareParam = record
		Self: T0TreeViewNodes;
		Compare: CCompareTreeNode;
		Context: Pointer;
	end;

function TreeNodeCompare(lParam1, lParam2, lParamSort: Longint): Integer stdcall;
var
	Param: PTreeNodeCompareParam;
	X, Y: TTreeViewNode;
begin
	Param := PTreeNodeCompareParam(lParamSort);
	try
		X := TObject(lParam1) as TTreeViewNode;
		Y := TObject(lParam2) as TTreeViewNode;
		Result := Param.Compare(X, Y, Param.Context);
	except
  	Result := 0
  end;
end;

procedure T0TreeViewNodes.Sort(Compare: CCompareTreeNode; Context: Pointer);
var
	Param: TTreeNodeCompareParam;
	CB: TTVSortCB;
begin
	Param.Self := Self;
	Param.Compare := Compare;
	Param.Context := Context;
	CB.hParent := ParentHandle;
	CB.lpfnCompare := TreeNodeCompare;
	CB.lParam := LPARAM(@Param);
	TreeView.SendMessageW(TVM_SORTCHILDRENCB, 0, LPARAM(@CB));	
end;

{ TTreeViewNodes }

constructor TTreeViewNodes.Create(ATreeView: TTreeView; AParentHandle: HTREEITEM);
begin
	inherited Create(ATreeView);
	FParentHandle := AParentHandle;
end;

function TTreeViewNodes.FirstHandle: HTREEITEM;
begin
	Result := HTREEITEM(TreeView.SendMessageW(TVM_GETNEXTITEM, TVGN_CHILD, LPARAM(FParentHandle)))
end;

function TTreeViewNodes.GetParent: TAnyTreeViewNode;
begin
	Result := TreeView.GetNode(FParentHandle)
end;

function TTreeViewNodes.ParentHandle: HTREEITEM;
begin
	Result := FParentHandle
end;

{ TTreeViewRoot }

function TTreeViewRoot.FirstHandle: HTREEITEM;
begin
	Result := HTREEITEM(TreeView.SendMessageW(TVM_GETNEXTITEM, TVGN_ROOT, 0));
end;

function TTreeViewRoot.ParentHandle: HTREEITEM;
begin
	Result := nil
end;

{ TTreeView }

procedure TTreeView.Clear;
begin
	SendMessageW(TVM_DELETEITEM, 0, LPARAM(TVI_ROOT));
end;

constructor TTreeView.Create(AParent: TAnyWindow; const APosition: TPoint;
				const ASize: TSize; const Styles: TTreeViewStyles);
var
	CS: PCreateStructW;
begin
	inherited Create(AParent, '', APosition, ASize);

	CS := CreateStruct;

	DWORD(CS^.Style) := DWORD(CS^.Style) or
			(WS_TABSTOP or TVS_DISABLEDRAGDROP or TVS_SHOWSELALWAYS) or
			DWORD(Styles);

	CS^.dwExStyle := CS^.dwExStyle or WS_EX_CLIENTEDGE;

	FRoot := TTreeViewRoot.Create(Self);
	FRoot.Owner := Self;
end;

class procedure TTreeView.CreateClass(var WndClass: WNDCLASSEXW);
begin
	InitCommonControl(ICC_TREEVIEW_CLASSES);
	inherited;
end;

function TTreeView.GetCount: Integer;
begin
	Result := SendMessageW(TVM_GETCOUNT, 0, 0);
end;

function TTreeView.GetLNode(ANode: HTREEITEM): TAnyTreeViewNode;
var
	Data: TV_ITEMW;
begin
	if ANode = nil then
		Result := nil
	else
	begin
		try
			Data.mask := TVIF_PARAM;
			Data.hItem := ANode;
			SendMessageW(TVM_GETITEMW, 0, LPARAM(@Data));
			Result := TObject(Data.lParam) as TTreeViewNode;
		except
			Result := nil
		end
	end
end;

function TTreeView.GetNode(ANode: HTREEITEM): TAnyTreeViewNode;
begin
	if ANode = nil then
		Result := nil
	else
	begin
		Result := GetLNode(ANode);
		if Result = nil then
		begin
			Result := TAnyTreeViewNode.Create(Self, ANode);
			Result.Owner := Garbage;
		end
	end
end;

function TTreeView.GetReadOnly: Boolean;
begin
	Result := not WindowStyle[TVS_EDITLABELS]
end;

function TTreeView.GetSelected: TAnyTreeViewNode;
begin
	Result := GetNode(HTREEITEM(SendMessageW(TVM_GETNEXTITEM, TVGN_CARET, 0)))
end;

procedure TTreeView.HandleNotification(Sender: TAnyWindow; Info: PNMHdr; var Result: Integer);
var
	Node: TAnyTreeViewNode;
	CustomDrawInfo: PNMTVCustomDraw;
	TreeViewInfo: PNMTreeViewW;
begin
	if Sender = Self then
	begin
		case Info^.code of
		TVN_DELETEITEMW:
			begin
				TreeViewInfo := PNMTreeViewW(Info);
				Node := GetLNode(TreeViewInfo.itemOld.hItem);
				if Node <> nil then Node.Free;
			end;
		NM_CUSTOMDRAW:
			begin
				if @FCustomDraw <> nil then
				begin
					CustomDrawInfo := PNMTVCUSTOMDRAW(Info);

					//OutputDebugString(PChar(IntToStr(CustomDrawInfo^.nmcd.dwDrawStage)));

					case CustomDrawInfo^.nmcd.dwDrawStage of
					CDDS_PREPAINT:
					begin
						Result := CDRF_NOTIFYITEMDRAW;
					end;
					CDDS_ITEMPREPAINT:
					begin
						if CustomDrawInfo^.nmcd.dwDrawStage and CDDS_ITEM <> 0 then
						begin
							Node := GetNode(HTREEITEM(CustomDrawInfo^.nmcd.dwItemSpec));
							Result := CDRF_DODEFAULT;
							if FCustomDraw(Self, Node, CustomDrawInfo) then
								Result := CDRF_SKIPDEFAULT;
						end;
					end;
					end;
				end;
			end;
		else
			inherited;
		end;
	end
	else
		inherited;
end;

function TTreeView.Indent: Integer;
begin
	Result := SendMessageW(TVM_GETINDENT, 0, 0);
end;

function TTreeView.ItemHeight: Integer;
var
	R: TRect;
begin
	R.Left := Integer(Root.FirstHandle);
	SendMessageW(TVM_GETITEMRECT, 1, LPARAM(@R));

	Result := R.Bottom - R.Top;
end;

function TTreeView.NodeFromPoint(const P: TPoint): TAnyTreeViewNode;
var
	HitTest: TTVHitTestInfo;
begin
	HitTest.pt.X := P.X;
	HitTest.pt.Y := P.Y;
	Result := nil;
	if SendMessageW(TVM_HITTEST, 0, LPARAM(@HitTest)) <> 0 then
		Result := GetNode(HitTest.hItem)
end;

procedure TTreeView.SelectWithoutScroll(const Value: TAnyTreeViewNode);
var
	H: HTREEITEM;
begin
	H := nil;
	if Value <> nil then
		H := Value.Handle;

	SendMessageW(TVM_SELECTITEM, TVGN_CARET, LPARAM(H));
end;

{$IFDEF IMAGELIST}
procedure TTreeView.SetImageList(const Value: TAnyImageList);
var
	H: HIMAGELIST;
begin
	H := 0;
	if Value <> nil then H := Value.Handle;
	SendMessageW(TVM_SETIMAGELIST, TVSIL_NORMAL, H)
end;
{$ENDIF}

procedure TTreeView.SetReadOnly(const Value: Boolean);
begin
	WindowStyle[TVS_EDITLABELS] := not Value
end;

procedure TTreeView.SetSelected(const Value: TAnyTreeViewNode);
var
	H, Pre: HTREEITEM;
begin
	H := nil;
	if Value <> nil then
		H := Value.Handle;

	SendMessageW(TVM_SELECTITEM, TVGN_CARET, LPARAM(H));
	if H <> nil then
	begin
		Pre := HTREEITEM(SendMessageW(TVM_GETNEXTITEM, TVGN_PREVIOUSVISIBLE, LPARAM(H)));
		if Pre = nil then Pre := H;
		SendMessageW(TVM_SELECTITEM, TVGN_FIRSTVISIBLE, LPARAM(Pre));
	end;
end;

class function TTreeView.SubClassName: PWideChar;
begin
	Result := WC_TREEVIEW
end;

{$ENDIF}

{$IFDEF LISTVIEW}

{ TListViewColumn }

constructor TListViewColumn.Create(AListView: TListView; AIndex: Integer);
begin
	inherited Create;
	FListView := AListView;
	FIndex := AIndex;
end;

function TListViewColumn.GetText: WideString;
var
	Col: LV_COLUMNW;
	Buffer: array[0..1023] of WideChar;
begin
	Col.mask := LVCF_TEXT;
	Col.pszText := Buffer;
	Col.cchTextMax := Length(Buffer);
	FListView.SendMessageW(LVM_GETCOLUMNW, FIndex, LPARAM(@Col));
	Result := Buffer;
end;

function TListViewColumn.GetWidth: Integer;
var
	Col: LV_COLUMNW;
begin
	Col.mask := LVCF_WIDTH;
	FListView.SendMessageW(LVM_GETCOLUMNW, FIndex, LPARAM(@Col));
	Result := Col.cx;
end;

procedure TListViewColumn.SetText(const Value: WideString);
var
	Col: LV_COLUMNW;
begin
	Col.mask := LVCF_TEXT;
	Col.pszText := PWideChar(Value);
	FListView.SendMessageW(LVM_SETCOLUMNW, FIndex, LPARAM(@Col))
end;

procedure TListViewColumn.SetWidth(const Value: Integer);
var
	Col: LV_COLUMNW;
begin
	Col.mask := LVCF_WIDTH;
	Col.cx := Value;
	FListView.SendMessageW(LVM_SETCOLUMNW, FIndex, LPARAM(@Col))
end;

{ TListViewItem }

constructor TListViewItem.Create(AListView: TListView; AIndex: Integer);
begin
	inherited Create;
	FListView := AListView;
	FIndex := AIndex;
end;

function TListViewItem.GetData: Pointer;
var
	Item: LV_ITEMW;
begin
	Item.mask := LVIF_PARAM;
	Item.iItem := FIndex;
	FListView.SendMessageW(LVM_GETITEMW, 0, LPARAM(@Item));
	Result := Pointer(Item.lParam);
end;

function TListViewItem.GetText: WideString;
begin
	Result := GetTexts(0)
end;

function TListViewItem.GetTexts(Index: Integer): WideString;
var
	Item: LV_ITEMW;
	Buffer: array[0..1024] of WideChar;
begin
	Item.mask := LVIF_TEXT;
	Item.iItem := FIndex;
	Item.iSubItem := Index;
	Item.pszText := Buffer;
	Item.cchTextMax := Length(Buffer);
	FListView.SendMessageW(LVM_GETITEMW, 0, LPARAM(@Item));
	Result := Buffer
end;

procedure TListViewItem.SetData(const Value: Pointer);
var
	Item: LV_ITEMW;
begin
	Item.mask := LVIF_PARAM;
	Item.iItem := FIndex;
	Item.lParam := LPARAM(Value);
	FListView.SendMessageW(LVM_SETITEMW, 0, LPARAM(@Item));
end;

procedure TListViewItem.SetText(const Value: WideString);
begin
	SetTexts(0, Value);
end;

procedure TListViewItem.SetTexts(Index: Integer; const Value: WideString);
var
	Item: LV_ITEMW;
begin
	Item.mask := LVIF_TEXT;
	Item.iItem := FIndex;
	Item.iSubItem := Index;
	Item.pszText := PWideChar(Value);
	FListView.SendMessageW(LVM_SETITEMW, 0, LPARAM(@Item));
end;

{ TListView }

function TListView.Add: TListViewItem;
var
	Item: LV_ITEMW;
begin
	Item.mask := LVIF_TEXT;
	Item.iItem := Count;
	Item.iSubItem := 0;
	Item.pszText := '';
	Result := GetItem(SendMessageW(LVM_INSERTITEMW, 0, LPARAM(@Item)));
end;

function TListView.AddColumn: TListViewColumn;
var
	Col: LV_COLUMNW;
begin
	Col.mask := LVCF_FMT or LVCF_TEXT;
	Col.fmt := LVCFMT_LEFT;
	Col.pszText := '';
	Result := GetColumn(SendMessageW(LVM_INSERTCOLUMNW, ColumnCount, LPARAM(@Col)))
end;

function TListView.AtPos(const P: TPoint): TListViewItem;
var
	I: Integer;
	R: TRect;
begin
	for I := 0 to Count - 1 do
	begin
		R.Left := LVIR_BOUNDS;
		SendMessageW(LVM_GETITEMRECT, I, LPARAM(@R));
		if PtInRect(R, P) then
		begin
			Result := At[I];
			Exit;
		end
	end;
	Result := nil
end;

function TListView.ColumnCount: Integer;
var
	Col: LV_COLUMNW;
begin
	Col.mask := LVCF_FMT;
	Result := 0;
	while SendMessageW(LVM_GETCOLUMNW, Result, LPARAM(@Col)) <> 0 do
		Inc(Result);
end;

function TListView.Count: Integer;
begin
	Result := SendMessageW(LVM_GETITEMCOUNT, 0, 0)
end;

constructor TListView.Create(AParent: TAnyWindow; const APosition: TPoint; const ASize: TSize; AStyles: TListViewStyles; AStyle: TListViewStyle);
const
	S: array[Boolean] of Integer = (LVS_SINGLESEL, 0);
var
	CS: PCreateStructW;
begin
	inherited Create(AParent, '', APosition, ASize);

	CS := CreateStruct;

	CS^.Style := CS^.Style or (WS_TABSTOP or LVS_SHOWSELALWAYS) or
		Ord(AStyle) or
		S[vsMultiSelection in AStyles];
	CS^.dwExStyle := CS^.dwExStyle or WS_EX_CLIENTEDGE;
end;

class procedure TListView.CreateClass(var WndClass: WNDCLASSEXW);
begin
	InitCommonControl(ICC_LISTVIEW_CLASSES);
	inherited;
end;

function TListView.Find(const AText: WideString): TListViewItem;
var
	Info: TLVFindInfoW;
	R: Integer;
begin
	Info.flags := LVFI_STRING;
	Info.psz := PWideChar(AText);
	R := SendMessageW(LVM_FINDITEMW, -1, LPARAM(@Info));
	if R < 0 then Result := nil else Result := GetAt(R);
end;

function TListView.GetAt(Index: Integer): TListViewItem;

	function FindProc(Obj: T0Object): Boolean cdecl;
	begin
		Result := (Obj is TListViewItem) and
						(TListViewItem(Obj).FListView = Self) and
						(TListViewItem(Obj).FIndex = Index)
	end;

begin
	{S~}
	Result := TListViewItem(Garbage.EnumerateObjects(CEachObject(@FindProc), GetContext));
	if Result <> nil then Exit;

	{S~ɖ΍쐬}
	Result := TListViewItem.Create(Self, Index);
	Result.Owner := Garbage;
end;

function TListView.GetColumn(Index: Integer): TListViewColumn;
begin
	if Index >= 0 then
	begin
		Result := TListViewColumn.Create(Self, Index);
		Result.Owner := Garbage;
	end
	else
		Result := nil
end;

function TListView.GetFullRowSelection: Boolean;
begin
	Result := SendMessageW(LVM_GETEXTENDEDLISTVIEWSTYLE, 0, 0) and
		LVS_EX_FULLROWSELECT <> 0
end;

function TListView.GetItem(Index: Integer): TListViewItem;
begin
	if Index >= 0 then
	begin
		Result := TListViewItem.Create(Self, Index);
		Result.Owner := Garbage;
	end
	else
		Result := nil
end;

function TListView.GetSelected: TListViewItem;
begin
	Result := At[SendMessageW(LVM_GETNEXTITEM, -1, LVNI_SELECTED)]
end;

procedure TListView.Reserve(ACount: Integer);
begin
	SendMessageW(LVM_SETITEMCOUNT, ACount, 0)
end;

procedure TListView.SetFullRowSelection(const Value: Boolean);
begin
	SendMessageW(LVM_SETEXTENDEDLISTVIEWSTYLE,
			LVS_EX_FULLROWSELECT,
			LVS_EX_FULLROWSELECT);
end;

procedure TListView.SetSelected(const Value: TListViewItem);
var
	Item: TLVItem;
	Sel: Integer;
begin
	if Value <> nil then
	begin
		Item.stateMask := LVIS_SELECTED;
		Item.state := LVIS_SELECTED;
		SendMessageW(LVM_SETITEMSTATE, Value.FIndex, LPARAM(@Item))
	end
	else
	begin
		Sel := SendMessageW(LVM_GETNEXTITEM, -1, LVNI_SELECTED);
		Item.stateMask := LVIS_SELECTED;
		Item.state := 0;
		SendMessageW(LVM_SETITEMSTATE, Sel, LPARAM(@Item))
	end
end;

class function TListView.SubClassName: PWideChar;
begin
	Result := WC_LISTVIEW
end;

{$ENDIF}

{$IFDEF TAB}

{ TTabControl.TItem }

constructor TTabControl.TItem.Create(ATabControl: TTabControl; AIndex: Integer);
begin
	inherited Create;
	FTabControl := ATabControl;
	FIndex := AIndex;
end;

function TTabControl.TItem.GetData: Pointer;
var
	Item: TC_ITEMW;
begin
	Item.mask := TCIF_PARAM;
	FTabControl.SendMessageW(TCM_GETITEMW, FIndex, LPARAM(@Item));
	Result := Pointer(Item.lParam);
end;

function TTabControl.TItem.GetText: WideString;
var
	Item: TC_ITEMW;
	Buffer: array[0..1023] of WideChar;
begin
	Item.mask := TCIF_TEXT;
	Item.pszText := Buffer;
	Item.cchTextMax := Length(Result);
	FTabControl.SendMessageW(TCM_GETITEMW, FIndex, LPARAM(@Item));
	Result := Buffer
end;

procedure TTabControl.TItem.SetData(const Value: Pointer);
var
	Item: TC_ITEMW;
begin
	Item.mask := TCIF_PARAM;
	Item.lParam := LPARAM(Value);
	FTabControl.SendMessageW(TCM_SETITEMW, FIndex, LPARAM(@Item));
end;

procedure TTabControl.TItem.SetText(const Value: WideString);
var
	Item: TC_ITEMW;
begin
	Item.mask := TCIF_TEXT;
	Item.pszText := PWideChar(Value);
	Item.cchTextMax := Length(Value);
	FTabControl.SendMessageW(TCM_SETITEMW, FIndex, LPARAM(@Item));
end;

{ TTabControl }

function TTabControl.Add: TItem;
var
	Item: TC_ITEMW;
	I: Integer;
begin
	Item.mask := TCIF_TEXT;
	Item.pszText := '';
	Item.cchTextMax := 0;

	I := SendMessageW(TCM_INSERTITEMW, Count, LPARAM(@Item));
	Result := TItem.Create(Self, I);
	Result.Owner := Garbage;
end;

procedure TTabControl.Clear;
begin
	SendMessageW(TCM_DELETEALLITEMS, 0, 0)
end;

constructor TTabControl.Create(AParent: TAnyWindow; const APosition: TPoint;
				const ASize: TSize; ASide: TAlignment; const Styles: TStyles);
const
//	M: array[Boolean] of DWORD = (TCS_SINGLELINE, TCS_MULTILINE);
	S: array[TAlignment] of DWORD =
					(0, TCS_VERTICAL, 0, TCS_VERTICAL or TCS_RIGHT, 0, TCS_BOTTOM);
var
	CS: PCreateStructW;
begin
	inherited Create(AParent, '', APosition, ASize);
	CS := CreateStruct;
	DWORD(CS^.style) := DWORD(CS^.style) or S[ASide] or DWORD(Styles);
	if tcsNoFocus in Styles then
	begin
		DWORD(CS^.style) := DWORD(CS^.style) and not WS_TABSTOP;
	end;
	CS^.dwExStyle := CS^.dwExStyle or WS_EX_CONTROLPARENT;
end;

class procedure TTabControl.CreateClass(var WndClass: WNDCLASSEXW);
begin
	InitCommonControl(ICC_TAB_CLASSES);
	inherited;
end;

procedure TTabControl.DialogKey(var Message: TWMKey);
var
	Shift: TShiftKeys;
	Index: Integer;
begin
	if Message.CharCode = 9 then
	begin
		Shift := GetShiftState;
		if ssControl in Shift then
		begin
			Message.Result := 1;
			Index := SelectedIndex;
			if ssShift in Shift then
			begin
				Dec(Index);
				if Index < 0 then Index := Count - 1;
			end
			else
			begin
				Inc(Index);
				if Index >= Count then Index := 0;
			end;
			SelectedIndex := Index;
      Notify(TCN_SELCHANGE);
			Exit;
		end	
	end;
	inherited;
end;

function TTabControl.DisplayRect: TRect;
begin
	Result := ClientRect;
	SendMessageW(TCM_ADJUSTRECT, 0, LPARAM(@Result));
end;

function TTabControl.DisplaySize: TSize;
var
	R: TRect;
begin
	R := DisplayRect;
	Result.cx := R.Right - R.Left;
	Result.cy := R.Bottom - R.Top
end;

function TTabControl.Enumerate(Proc: CEachItem; Context: Pointer): T0ControlItem;
var
	I: Integer;
begin
	for I := 0 to Count - 1 do
	begin
		Result := GetAt(I);
		if Proc(Result, Context) then Exit;
	end;
	Result := nil;
end;

function TTabControl.GetAt(AIndex: Integer): TItem;
begin
	Assert((AIndex >= 0) and (AIndex < Count));

	Result := TItem.Create(Self, AIndex);
	Result.Owner := Garbage;
end;

function TTabControl.GetCount: Integer;
begin
	Result := SendMessageW(TCM_GETITEMCOUNT, 0, 0)
end;

function TTabControl.GetSelectedIndex: Integer;
begin
	Result := SendMessageW(TCM_GETCURSEL, 0, 0)
end;

procedure TTabControl.Initialize;
begin
	inherited;
	Font := ForLabelFont
end;

function TTabControl.OnSelectionChanged(Event: PNMHdr): Boolean;
begin
	Result := (Self <> nil) and (Event.hwndFrom = Handle)
  	and (Event.code = TCN_SELCHANGE);
end;

procedure TTabControl.ProcessMouseUp(var Message: TWMLButtonUp; Notify: Integer);
var
	P: HWND;
	N: TNMHdr;
	SelectedIndex: Integer;
	SelectedRect: TRect;
	CursorPoint: TPoint;
begin
	SelectedIndex := Self.SelectedIndex;
	if SendMessageW(TCM_GETITEMRECT, SelectedIndex, LPARAM(@SelectedRect)) <> 0 then
	begin
		CursorPoint.X := Message.XPos;
		CursorPoint.Y := Message.YPos;
		if PtInRect(SelectedRect, CursorPoint) then
		begin
			P := GetParent(Handle);
			if P <> 0 then
			begin
				N.hwndFrom := Handle;
				N.idFrom := 0;
				N.code := Notify;
				Windows.SendMessageW(P, WM_NOTIFY, 0, LPARAM(@N));
			end
		end
	end
end;

procedure TTabControl.Realign;
var
	R: TRect;
begin
	R := DisplayRect;
	LWindows.Realign(Handle, R);
end;

procedure TTabControl.Remove(Item: T0ControlItem);
begin
	Assert(Item is TItem);
	SendMessageW(TCM_DELETEITEM, TItem(Item).FIndex, 0)
end;

procedure TTabControl.SetSelectedIndex(const Value: Integer);
begin
	SendMessageW(TCM_SETCURSEL, Value, 0);
end;

class function TTabControl.SubClassName: PWideChar;
begin
	Result := WC_TABCONTROL
end;

procedure TTabControl.TCMAdjustRect(var Message: TMessage);
var
	Saved: TClass;
begin
	Saved := ExceptionClass;
	ExceptionClass := EAbort;
	try
		inherited;
	except
		PRect(Message.LParam)^ := ClientRect
	end;
	ExceptionClass := Saved;
end;

procedure TTabControl.WMLButtonDblClk(var Message: TWMLButtonDblClk);
var
	P: HWND;
	N: TNMHdr;
	SelectedIndex: Integer;
	SelectedRect: TRect;
	CursorPoint: TPoint;
begin
	inherited;

	SelectedIndex := Self.SelectedIndex;
	if SendMessageW(TCM_GETITEMRECT, SelectedIndex, LPARAM(@SelectedRect)) <> 0 then
	begin
		CursorPoint.X := Message.XPos;
		CursorPoint.Y := Message.YPos;
		if PtInRect(SelectedRect, CursorPoint) then
		begin
			P := GetParent(Handle);
			if P <> 0 then
			begin
				N.hwndFrom := Handle;
				N.idFrom := 0;
				N.code := TCN_DBLCLK;
				Windows.SendMessageW(P, WM_NOTIFY, 0, LPARAM(@N));
			end
		end
	end
end;

procedure TTabControl.WMMButtonDown(var Message: TWMMButtonDown);
begin
	if not WindowStyle[TCS_FOCUSNEVER] then
		inherited;

	SendMessageW(WM_LBUTTONDOWN, TMessage(Message).WParam, TMessage(Message).LParam);
	SendMessageW(WM_LBUTTONUP, TMessage(Message).WParam, TMessage(Message).LParam);
end;

procedure TTabControl.WMMButtonUp(var Message: TWMMButtonUp);
begin
	inherited;
	ProcessMouseUp(Message, TCN_MCLICK);
end;

procedure TTabControl.WMRButtonDown(var Message: TWMRButtonDown);
begin
	if not WindowStyle[TCS_FOCUSNEVER] then
		inherited;

	SendMessageW(WM_LBUTTONDOWN, TMessage(Message).WParam, TMessage(Message).LParam);
	SendMessageW(WM_LBUTTONUP, TMessage(Message).WParam, TMessage(Message).LParam);
end;

procedure TTabControl.WMRButtonUp(var Message: TWMRButtonUp);
begin
	inherited;
	ProcessMouseUp(Message, TCN_RCLICK);
end;

{$ENDIF}

{$IFDEF STATUS}

{ TStatusPane }

constructor TStatusPane.Create(AStatusBar: TStatusBar; ANext: TStatusPane);
begin
	inherited Create;

	FAlignment := alLeft; // Ord(alLeft) <> 0 

	AddToLinkList(AStatusBar.FFirstPane, AStatusBar.FLastPane, Self, FPred, FSucc, ANext);
	FStatusBar := AStatusBar;

	FStatusBar.FModified := True {FStatusBar.Modify}
end;

procedure TStatusPane.BeforeDestruction;
begin
	FStatusBar.FModified := True {FStatusBar.Modify};
	RemoveFromLinkList(FStatusBar.FFirstPane, FStatusBar.FLastPane, Self, FPred, FSucc);
end;

procedure TStatusPane.Update;
begin
end;

procedure TStatusPane.SetText(const Value: WideString);
begin
	if FText = Value then Exit;
	FText := Value;
	FStatusBar.FModified:= True {FStatusBar.Modify}
end;

procedure TStatusPane.SetBevel(const Value: TBevel);
begin
	if FBevel = Value then Exit;
	FBevel:= Value;
	FStatusBar.FModified:= True {FStatusBar.Modify}
end;

procedure TStatusPane.SetWidth(const Value: Integer);
begin
	if FWidth = Value then Exit;
	FWidth:= Value;
	FStatusBar.FModified:= True {FStatusBar.Modify}
end;

procedure TStatusPane.SetAlignment(const Value: THorzAlignment);
begin
	if FAlignment = Value then Exit;
	FAlignment:= Value;
	FStatusBar.FModified:= True {FStatusBar.Modify}
end;

function TStatusPane.GetData: Pointer;
begin
	Result := FData
end;

function TStatusPane.GetText: WideString;
begin
	Result := FText
end;

procedure TStatusPane.SetData(const Value: Pointer);
begin
	FData := Value
end;

procedure TStatusPane.AdjustSize;
begin
	Width := GetFontSize(FStatusBar.SendMessageW(WM_GETFONT, 0, 0), Text).CX + 16;
end;

procedure TStatusPane.DoContextMenu(const P: TPoint);
begin
end;

{ TStatusBar }

constructor TStatusBar.Create(AParent: TAnyWindow);
var
	CS: PCreateStructW;
begin
	Alignment := alBottom; //ꏊł̂ ?

	inherited Create(AParent, '', ZeroPoint, ZeroSize);

	CS:= CreateStruct;
	CS^.style:= CS^.style and not WS_TABSTOP;
end;

procedure TStatusBar.AfterConstruction; {override}
var
	M: TMessage;
begin
	inherited;

	//Height ݒ. ڃIWi Window Procedure Ăł.
	M.Msg:= WM_SIZE;
	DefaultHandler(M);

	//Text ƘAĂ擪Xe[^X̘g
	SendMessageW(SB_SETTEXTW, SBT_NOBORDERS, 0);
end;

class function TStatusBar.SubClassName: PWideChar;
begin
	Result := STATUSCLASSNAME;
end;

class procedure TStatusBar.CreateClass(var WndClass: WNDCLASSEXW);
begin
	InitCommonControl(ICC_BAR_CLASSES);
	inherited;
end;

procedure TStatusBar.Modify;
begin
	FModified:= True;
end;

procedure TStatusBar.Update;
var
	I: Integer;
	E: TStatusPane;
	Count, Position, ParentStyle: Integer;
	Widths: array[0..255] of Integer;
	AlignedText: WideString;
	Par: TAnyWindow;
begin
	inherited;

	Par := Parent;

	Widths[0] := 0;
	Count := 1; // XVłɐ𐔂 Width Zbg肷

	E := FFirstPane;
	while E <> nil do
	begin
		Assert(E is TStatusPane);

		E.Update;
		Widths[Count] := E.Width + Widths[Count -1];
		Inc(Count);

		E := E.FSucc;
	end;

	if Self.FModified then
	begin
		FModified:= False;
		// \XV
		//  & 
		Widths[0] := BoundsSize.CX - Widths[Count - 1];

		if Par <> nil then
		begin
			ParentStyle := Par.WindowLong[GWL_STYLE];
			if (ParentStyle and WS_THICKFRAME <> 0) and
							(ParentStyle and WS_MAXIMIZE = 0) then
				Dec(Widths[0], 16);
				//  16 ɍ͂Ȃ. SizeGrip 擾悤 API 
				// Ȃ̘b.
				// SB_GETRECT ō𓾂ē]p̂ ?
		end;

		FHintWidth := Widths[0]; {Xe[^Xo[ɉuɎĝŊoĂ}

		for I:= 1 to Count - 1 do
			Inc(Widths[I], Widths[0]);
		SendMessageW(SB_SETPARTS, Count, LPARAM(@Widths));
		// eLXg
		Position:= 1;

		E := FFirstPane;
		while E <> nil do
		begin
			AlignedText := #9#9 + TStatusPane(E).Text;

			SendMessageW(SB_SETTEXTW, Position, LPARAM(
							PWideChar(Pointer(AlignedText)) +
							(2 + Ord(alLeft)) - Ord(TStatusPane(E).Alignment)));

			// alLeft, alCenter, alRight ̏ŕł邩gZ

			Inc(Position);

			E := E.FSucc;
		end;
	end;

	// Hint ̎\

	if Par is T0TopWindow then
		if T0TopWindow(Par).Hint <> Text then
			Text := T0TopWindow(Par).Hint;
end;

procedure TStatusBar.WMSize(var Message: TWMSize); {message}
//pȂ
begin
	(* inherited; *)
	FModified:= True;
	Update; //thbOɑΉ邽ߑɍXV
end;

procedure TStatusBar.WMMove(var Message: TWMMove); {message}
{
	eK WM_SIZE 𑗂ĂƂ͌Ȃ.
	Realign ł͈ړƂ.
	ړ̏ꍇ̓TCYςȂ̂ōč\͕sv.
}
begin
	Repaint;
	UpdateWindow(Handle); {Update ȂƂ܂Ȃ悤}
end;

procedure TStatusBar.WMSetText(var Message: TMessage); {message}
begin
	if PWideChar(Message.LParam) = Text then Exit;

	//inherited ̎_œŕύXčĕ`悳̂ł~߂
	SendMessageW(WM_SETREDRAW, 0, 0);
	inherited;
	SendMessageW(WM_SETREDRAW, 1, 0);

	SendMessageW(SB_SETTEXTW, SBT_NOBORDERS, Message.LParam);
end;

procedure TStatusBar.BeforeDestruction;
begin
	while FLastPane <> nil do FLastPane.Free;
	inherited;
end;

procedure TStatusBar.WMContextMenu(var Message: TWMContextMenu);
var
	P: TPoint;
	Pane: TStatusPane;
begin
	P.X := Message.XPos;
	P.Y := Message.YPos;
	Pane := PaneFromPoint(ScreenToClient(P));
	if Pane <> nil then Pane.DoContextMenu(P);
end;

function TStatusBar.PaneFromPoint(const P: TPoint): TStatusPane;
var
	I: TStatusPane;
	R: TRect;
	Index: Integer;
begin
	I := FFirstPane;
	Index := 0;
	while I <> nil do
	begin
		Inc(Index);
		SendMessageW(SB_GETRECT, Index, LPARAM(@R));
		if PtInRect(R, P) then
		begin
			Result := I;
			Exit;
		end;
		I := I.FSucc;
	end;
	Result := nil
end;

{ TActionStatusPane }

constructor TActionStatusPane.Create(AStatusBar: TStatusBar; ANext: TStatusPane; const AAction: IAction);
begin
	FAction := AAction;
	inherited Create(AStatusBar, ANext);
end;

procedure TActionStatusPane.DoContextMenu(const P: TPoint);
var
	Items: IActionItems;
	Selected: IActionItem;
begin
	if FAction.Enabled then
	begin
		Items := FAction.Items;
		if Items <> nil then
		begin
			Selected := SelectActionItemsPopup(Items, P);
			if Selected <> nil then
				Selected.Execute;
		end
	end
end;

{ TLockedKeyWatcher }

var
	SingleInstance: TLockedKeyStatusPane.TLockedKeyWatcher;

procedure TLockedKeyStatusPane.TLockedKeyWatcher.Update;
var
	NewState: TLockedKeyState;
begin
	if GetKeyState(VK_KANA) and 1 <> 0 then
		NewState:= lkKana
	else if GetKeyState(VK_CAPITAL) and 1 <> 0 then
		NewState:= lkCaps
	else
		NewState:= lkUnlocked;

	if FState <> NewState then
	begin
		FState:= NewState;
		FText:= LockedKeyFaces[NewState];
		Modify;
	end;
end;

class function TLockedKeyStatusPane.TLockedKeyWatcher.Instance: TLockedKeyWatcher;
begin
	Application.BeginWrite;
	try
		if SingleInstance = nil then
		begin
			SingleInstance := TLockedKeyWatcher.Create(True);
			SingleInstance.Owner := Application;
		end;
	finally
		Application.EndWrite;
	end;
	Result := SingleInstance;
end;

{ TLockedKeyStatusPane }

constructor TLockedKeyStatusPane.Create(AStatusBar: TStatusBar;
				ANext: TStatusPane; AWatcher: TLockedKeyWatcher);
begin
	inherited Create(AStatusBar, ANext);

	if AWatcher <> nil then
		FWatcher:= AWatcher
	else
		FWatcher:= TLockedKeyWatcher.Instance;

	Width:= 40;
	Alignment:= alCenter;
end;

procedure TLockedKeyStatusPane.Update; {override}
begin
	inherited;

	if FWatcher.Text <> Text then
		Text := FWatcher.Text;
end;

{$ENDIF}

{$IFDEF TOOLBAR}

{ T0ToolButton }

function T0ToolButton.GetIndex: Integer;
var
	Info: TBBUTTONINFOW;
begin
	Info.cbSize := SizeOf(Info);
	Info.dwMask := 0;
	Result := ToolBar.SendMessageW(TB_GETBUTTONINFOW, ID, LPARAM(@Info))
end;

function T0ToolButton.GetState: DWORD;
var
	Info: TBBUTTONINFOW;
begin
	Info.cbSize := SizeOf(Info);
	Info.dwMask := TBIF_STATE;
	ToolBar.SendMessageW(TB_GETBUTTONINFOW, ID, LPARAM(@Info));
	Result := Info.fsState;
end;

function T0ToolButton.GetStyle: DWORD;
var
	Info: TBBUTTONINFOW;
begin
	Info.cbSize := SizeOf(Info);
	Info.dwMask := TBIF_STYLE;
	ToolBar.SendMessageW(TB_GETBUTTONINFOW, ID, LPARAM(@Info));
	Result := Info.fsStyle;
end;

function T0ToolButton.GetText: WideString;
var
	Info: TBBUTTONINFOW;
	Buffer: array[0..1023] of WideChar;
begin
	Info.cbSize := SizeOf(Info);
	Info.dwMask := TBIF_TEXT;
	Info.pszText := Buffer;
	Info.cchText := Length(Buffer);
	ToolBar.SendMessageW(TB_GETBUTTONINFOW, ID, LPARAM(@Info));
	Result := Buffer
end;

procedure T0ToolButton.SetState(const Value: DWORD);
var
	Info: TBBUTTONINFOW;
begin
	Info.cbSize := SizeOf(Info);
	Info.dwMask := TBIF_STATE;
	Info.fsState := Value;
	ToolBar.SendMessageW(TB_SETBUTTONINFOW, ID, LPARAM(@Info));
end;

procedure T0ToolButton.SetStyle(const Value: DWORD);
var
	Info: TBBUTTONINFOW;
begin
	Info.cbSize := SizeOf(Info);
	Info.dwMask := TBIF_STYLE;
	Info.fsStyle := Value;
	ToolBar.SendMessageW(TB_SETBUTTONINFOW, ID, LPARAM(@Info));
end;

procedure T0ToolButton.SetText(const Value: WideString);
var
	Info: TBBUTTONINFOW;
begin
	Info.cbSize := SizeOf(Info);
	Info.dwMask := TBIF_TEXT;
	Info.pszText := PWideChar(Value);
	Info.cchText := Length(Value);
	if ToolBar.SendMessageW(TB_SETBUTTONINFOW, ID, LPARAM(@Info)) = 0 then
		RaiseLastOSError
end;

{ TToolButton }

function GetMsgProc(nCode: Integer; wP: WPARAM; lP: LPARAM): LRESULT; stdcall;
var
	C: T0Window;
begin
	if (nCode >= 0) and (wP = PM_REMOVE) then
	begin
		C := GetLWnd(PMSG(lp)^.hwnd);
		if C <> nil then
		begin
			C := C.TopParent.ActiveControl;
			if C is T0ToolBarControl then
				T0ToolBarControl(C).SubMenuHook(PMSG(lp)^);
		end
	end;
	Result := CallNextHookEx(0, nCode, wP, lP);
end;

procedure TToolButton.BeforeDestruction;
begin
	(CurrentThread as ICommandManager).UnregisterCommand(ID);
	inherited;
end;

constructor TToolButton.Create(ABar: T0ToolBarControl; ANext: T0ToolButton);
const
	S: array[0..1] of WideChar = (#32, #0);
	Template: TTBButton = (
					iBitmap: -2;
					fsState: TBSTATE_ENABLED;
					fsStyle: TBSTYLE_AUTOSIZE);
var
	Button: TTBButton;
	R: LRESULT;
begin
	inherited Create;

	FToolBar := ABar;
	FID := (CurrentThread as ICommandManager).RegisterCommand(Self);

	Button := Template;
	Button.dwData := DWORD(Self);
	Button.iString := Integer(S);
	Button.idCommand := FID;

	if ANext = nil then
		R := ABar.SendMessageW(TB_ADDBUTTONSW, 1, LPARAM(@Button))
	else
		R := ABar.SendMessageW(TB_INSERTBUTTONW, ANext.Index, LPARAM(@Button));

	if R = 0 then
		RaiseLastOSError
end;

function TToolButton.GetData: Pointer;
begin
	Result := FData
end;

function TToolButton.GetSub: TAnyMenu;
begin
	Result := FSub
end;

procedure TToolButton.SetData(const Value: Pointer);
begin
	FData := Value
end;

procedure TToolButton.SetSub(const Value: TAnyMenu);
begin
	FSub := Value;
	Style := Style or TBSTYLE_DROPDOWN;
end;

function TToolButton.TrackSubMenu: Boolean;
var
	P: TPoint;
	R: TRect;
	C: DWORD;
	H: HHOOK;
	Top: T0TopWindow;
	I: Integer;
begin
	State := State or TBSTATE_PRESSED;

	I := Index;
	ToolBar.FDropping := I;
	ToolBar.SendMessageW(TB_GETITEMRECT, I, LPARAM(@R));
	P.X := R.Left;
	P.Y := R.Bottom;
	P := ToolBar.ClientToScreen(P);

	Top := ToolBar.TopParent;

	H := SetWindowsHookExW(WH_GETMESSAGE, GetMsgProc, hInstance, GetCurrentThreadID);

	keybd_event(VK_DOWN, 0, 0, 0);
	keybd_event(VK_DOWN, 0, KEYEVENTF_KEYUP, 0);

	C := DWORD(TrackPopupMenu(Sub.Handle, TPM_RETURNCMD or TPM_LEFTBUTTON,
					P.X, P.Y, 0, Top.Handle, nil));

	UnhookWindowsHookEx(H);

	ToolBar.FDropping := -1;

	Top.PostMessageW(WM_COMMAND, C, 0);

	State := State and not TBSTATE_PRESSED;

	Result := C > 0;
end;

{ T0ToolBarControl }

procedure T0ToolBarControl.Cancel;
var
	W: T0TopWindow;
begin
	if not FInTracking then Exit;

	W := TopParent;
	W.EndTracking;
	W.ActiveControl := FActivedControl;

	FInTracking := False;
end;

constructor T0ToolBarControl.Create(AParent: TAnyWindow);
var
	CS: PCreateStructW;
begin
	Alignment := alTop; //ꏊł̂?

	inherited Create(AParent, '', ZeroPoint, ZeroSize);

	CS := CreateStruct;
	CS^.style := CS^.style and not WS_TABSTOP or CCS_NODIVIDER;
end;

class procedure T0ToolBarControl.CreateClass(var WndClass: TWNDCLASSEXW);
begin
	InitCommonControl(ICC_BAR_CLASSES);
	inherited;
end;

function T0ToolBarControl.GetAt(AIndex: Integer): T0ToolButton;
var
	Button: TTBButton;
begin
	SendMessageW(TB_GETBUTTON, AIndex, LPARAM(@Button));

	Result := nil;

	try
		if TObject(Button.dwData) is T0ToolButton then
			Result := T0ToolButton(Button.dwData)
	except
	end;

	{ExternalToolButtonKvȂc}
end;

function T0ToolBarControl.GetCount: Integer;
begin
	Result := SendMessageW(TB_BUTTONCOUNT, 0, 0)
end;

procedure T0ToolBarControl.HandleNotification(Sender: TAnyWindow; Info: PNMHDR; var Result: LRESULT);
var
	Button: T0ToolButton;
begin
	if (Sender = Self) and (Info.code = TBN_DROPDOWN) then
	begin
		Track;
		Button := GetObjectFromCommand(PNMTOOLBARW(Info)^.iItem) as T0ToolButton;
		if Button.Sub <> nil then
		begin
			if TToolButton(Button).TrackSubMenu then Cancel;
		end
	end;
	inherited;
end;

procedure T0ToolBarControl.Initialize;
begin
	inherited;

	//WindowLong[GWL_STYLE] := WindowLong[GWL_STYLE] or (CCS_NOMOVEY or CCS_NOPARENTALIGN or CCS_NORESIZE);
	WindowStyle[CCS_NOMOVEY or CCS_NOPARENTALIGN or CCS_NORESIZE] := True;

	SendMessageW(TB_BUTTONSTRUCTSIZE, SizeOf(TTBButton), 0)
end;

function T0ToolBarControl.ItemAtPos(const P: TPoint): T0ToolButton;
var
	I: Integer;
begin
	I := ItemIndexAtPos(P);
	if I < 0 then
		Result := nil
	else
		Result := At[I];
end;

function T0ToolBarControl.ItemIndexAtPos(const P: TPoint): Integer;
var
	I: Integer;
	R: TRect;
begin
	for I := Count - 1 downto 0 do
	begin
		SendMessageW(TB_GETITEMRECT, I, LPARAM(@R));
		if PtInRect(R, P) then
		begin
			Result := I;
			Exit;
		end;
	end;
	Result := -1;
end;

class function T0ToolBarControl.SubClassName: PWideChar;
begin
	Result := TOOLBARCLASSNAME
end;

procedure T0ToolBarControl.SubMenuHook(var Message: MSG);
begin
end;

procedure T0ToolBarControl.Track;
var
	W: T0TopWindow;
begin
	if FInTracking then Exit;

	FInTracking := True;
  FAltPushed := GetKeyState(VK_MENU) and $80 <> 0;

	W := TopParent;
	W.PostMessageW(WM_CANCELMODE, 0, 0);
	FActivedControl := W.ActiveControl;
	W.BeginTracking;

	Windows.SetFocus(Handle);

	if SendMessageW(TB_GETHOTITEM, 0, 0) < 0 then
		SendMessageW(TB_SETHOTITEM, 0, 0)
	else if FAltPushed then
		PostMessageW(WM_KEYDOWN, VK_DOWN, 0)
end;

procedure T0ToolBarControl.WMKillFocus(var Message: TWMKillFocus);
begin
	inherited;
	Cancel;
end;

{$IFDEF TOOLBAR_MENU}

function CreateMenuItem(Menu: TMenuBar; const Text: WideString): TToolButton; overload;
begin
	Result := TToolButton.Create(Menu, nil);
	Result.Owner := Menu;
	Result.Text := Text
end;

function CreateSubMenu(Item: TToolButton): TPopupMenu; overload;
begin
	Result := TPopupMenu.Create;
	Result.Owner := Item;
	Item.Sub := Result;
end;

{ TMenuBar }

constructor TMenuBar.Create(AParent: TAnyWindow);
var
	CS: PCreateStructW;
begin
	inherited Create(AParent);

	CS := CreateStruct;
	CS^.style := CS^.style or (TBSTYLE_FLAT or TBSTYLE_LIST);
end;

procedure TMenuBar.Initialize;
var
	H: Integer;
	S: TSize;
begin
	inherited;

	H := GetSystemMetrics(SM_CYMENUSIZE);

	S := ClientSize;
	S.CY := H + 6;
	ClientSize := S;

	SendMessageW(TB_SETBUTTONSIZE, 0, MakeLong(24, H));
end;

function TMenuBar.KeyDown(ShiftState: TShiftKeys; Code: Byte): Boolean;
var
	I: Integer;
	B: T0ToolButton;
begin
	Result := True;
	case Code of
	VK_LEFT, VK_RIGHT:
	begin
		I := SendMessageW(TB_GETHOTITEM, 0, 0);
		if Code = VK_LEFT then
			I := I + Count - 1
		else
			I := I + 1;

		I := I mod Count;
		SendMessageW(TB_SETHOTITEM, I, 0);
	end;
	VK_RETURN, VK_DOWN, VK_UP:
	begin
		I := SendMessageW(TB_GETHOTITEM, 0, 0);
		if I >= 0 then
		begin
			B := At[I];
			if B.Sub <> nil then
				if TToolButton(B).TrackSubMenu then Cancel
		end;	
	end;
	VK_ESCAPE:
		Cancel;
	else
		Result := inherited KeyDown(ShiftState, Code)
	end
end;

function TMenuBar.KeyUp(ShiftState: TShiftKeys; Code: Byte): Boolean;
begin
	Result := True;
	case Code of
	VK_MENU, VK_F10:
		if FAltPushed then
			FAltPushed := False
		else
			Cancel;
	else
		Result := inherited KeyDown(ShiftState, Code)
	end
end;

function TMenuBar.ProcessDialogKey(var Message: TWMKey): Boolean;
var
	C: WideChar;
	S: WideString;
	I: Integer;
begin
	Result := Visible and (Message.Msg = WM_SYSKEYUP) and ((Message.CharCode = VK_F10) or (Message.CharCode = VK_MENU));
	if not Result then
		if Message.Msg = WM_SYSCHAR then
		begin
			C := WideUpCase(WideChar(Message.CharCode));
			if (C >= 'A') and (C <= 'Z') then
			begin
				S := WideString('&') + C;
				for I := Count - 1 downto 0 do
				begin
					if Pos(S, WideUpperCase(At[I].Text)) > 0 then
					begin
						SendMessageW(TB_SETHOTITEM, I, 0);
						Result := True;
						Exit;
					end;
				end;
			end;
		end;
end;

procedure TMenuBar.SubMenuHook(var Message: MSG);
var
	Index, I: Integer;
	P: TPoint;
begin
	case Message.message of
	WM_KEYDOWN:
		if Message.wParam in [VK_LEFT, VK_RIGHT] then
		begin
			Index := SendMessageW(TB_GETHOTITEM, 0, 0);
			if Message.wParam = VK_LEFT then
				Index := Index + Count - 1
			else
				Index := Index + 1;

			Index := Index mod Count;

			SendMessageW(TB_SETHOTITEM, Index, 0);

			TopParent.PostMessageW(WM_CANCELMODE, 0, 0);
			PostMessageW(WM_KEYDOWN, VK_DOWN, 0);
		end;
	WM_CHAR:
	begin
		if Message.wParam = 13 then Message.message := WM_NULL;
	end;
	WM_MOUSEMOVE:
	begin
		GetCursorPos(P);
		P := ScreenToClient(P);
		I := ItemIndexAtPos(P);
		if (I >= 0) and (I <> FDropping) then
		begin
			SendMessageW(TB_SETHOTITEM, I, 0);
			TopParent.PostMessageW(WM_CANCELMODE, 0, 0);
			PostMessageW(WM_KEYDOWN, VK_DOWN, 0);
		end;
	end
	end;
end;

procedure TMenuBar.WMLButtonDown(var Message: TWMLButtonDown);
var
	P: TPoint;
begin
	inherited;
	P.X := Message.XPos;
	P.Y := Message.YPos;
	if ItemIndexAtPos(P) < 0 then Cancel;
end;

{$ENDIF}
{$ENDIF}

{$IFDEF PROGRESS}

{ TProgressBar }

constructor TProgressBar.Create(AParent: TAnyWindow; const APosition: TPoint;
	const ASize: TSize);
var
	CS: PCreateStructW;
begin
	inherited Create(AParent, '', APosition, ASize);
	CS := CreateStruct;
	CS^.style := CS^.style and not WS_TABSTOP or PBS_SMOOTH
end;

class procedure TProgressBar.CreateClass(var WndClass: TWNDCLASSEXW);
begin
	InitCommonControl(ICC_PROGRESS_CLASS);
	inherited;
end;

function TProgressBar.GetMax: Integer;
begin
	Result := SendMessageW(PBM_GETRANGE, 0, 0)
end;

function TProgressBar.GetMin: Integer;
begin
	Result := SendMessageW(PBM_GETRANGE, 1, 0)
end;

function TProgressBar.GetValue: Integer;
begin
	Result := SendMessageW(PBM_GETPOS, 0, 0)
end;

procedure TProgressBar.SetMax(const Value: Integer);
begin
	SendMessageW(PBM_SETRANGE32, GetMin, Value)
end;

procedure TProgressBar.SetMin(const Value: Integer);
begin
	SendMessageW(PBM_SETRANGE32, Value, GetMax)
end;

procedure TProgressBar.SetValue(const Value: Integer);
begin
	SendMessageW(PBM_SETPOS, Value, 0)
end;

class function TProgressBar.SubClassName: PWideChar;
begin
	Result := PROGRESS_CLASS
end;

{$ENDIF}

{$IFDEF UPDOWN}

{ TUpDown }

procedure TUpDown.Associate(Wnd: TAnyWindow);
begin
	SendMessageW(UDM_SETBUDDY, Wnd.Handle, 0);
end;

constructor TUpDown.Create(AParent: TAnyWindow; const APosition: TPoint; const ASize: TSize);
var
	CS: PCreateStructW;
begin
	inherited Create(AParent, '', APosition, ASize);
	CS := CreateStruct;
	CS^.style := CS^.style and not WS_TABSTOP
		or (UDS_ALIGNRIGHT	or UDS_SETBUDDYINT or UDS_ARROWKEYS);
end;

class procedure TUpDown.CreateClass(var WndClass: TWNDCLASSEXW);
begin
	InitCommonControl(ICC_UPDOWN_CLASS);
	inherited;
end;

class function TUpDown.GetDefaultSize: TSize;
begin
	Result.CX := GetSystemMetrics(SM_CXVSCROLL);
	Result.CY := GetSystemMetrics(SM_CYVSCROLL) * 3 div 2;
end;

function TUpDown.GetMax: Integer;
begin
	SendMessageW(UDM_GETRANGE32, 0, LPARAM(@Result))
end;

function TUpDown.GetMin: Integer;
begin
	SendMessageW(UDM_GETRANGE32, WPARAM(@Result), 0)
end;

function TUpDown.GetValue: Integer;
begin
	Result := SendMessageW(UDM_GETPOS32, 0, 0)
end;

procedure TUpDown.SetMax(const Value: Integer);
begin
	SendMessageW(UDM_SETRANGE32, GetMin, Value)
end;

procedure TUpDown.SetMin(const Value: Integer);
begin
	SendMessageW(UDM_SETRANGE32, Value, GetMax)
end;

procedure TUpDown.SetValue(const Value: Integer);
begin
	SendMessageW(UDM_SETPOS32, 0, Value)
end;

class function TUpDown.SubClassName: PWideChar;
begin
	Result := UPDOWN_CLASS
end;

{ TNumberField }

constructor TNumberField.Create(AParent: TAnyWindow; Value, Min, Max: Integer; const APosition: TPoint; const ASize: TSize);
var
	CS: PCreateStructW;
begin
	FValue := Value;
	FMin := Min;
	FMax := Max;
	inherited Create(AParent, '', APosition, ASize);
	CS := CreateStruct;
	CS^.style := CS^.style or (ES_NUMBER or ES_RIGHT);
end;

procedure TNumberField.Initialize;
begin
	inherited;
	FUpDown := TUpDown.Create(Parent, ZeroPoint, DefaultSize);
	FUpDown.Min := FMin;
	FUpDown.Max := FMax;
	FUpDown.Associate(Self);
	FUpDown.Value := FValue;
end;

{$ENDIF}

{$IFDEF DATETIMEPICK}

{ TDateTimePicker }

constructor TDateTimePicker.Create(AParent: TAnyWindow; const APosition: TPoint; const ASize: TSize;  const Styles: TDateTimePickerStyles);
var
	CS: PCreateStructW;
begin
	FChecked := True;
	inherited Create(AParent, '', APosition, ASize);
	CS := CreateStruct;
	CS^.style := CS^.style or Integer(Styles);
end;

class procedure TDateTimePicker.CreateClass(var WndClass: TWNDCLASSEXW);
begin
	InitCommonControl(ICC_DATE_CLASSES);
	inherited;
end;

class function TDateTimePicker.GetDefaultSize: TSize;
begin
	Result := OneLineDefaultSize
end;

procedure TDateTimePicker.HandleNotification(Sender: TAnyWindow; Info: PNMHdr; var Result: Integer);
begin
	if (Sender = Self) and (Info.code = DTN_DATETIMECHANGE) then
	begin
		FChecked := PNMDATETIMECHANGE(Info).dwFlags = GDT_VALID;
		FValue := PNMDATETIMECHANGE(Info).st;
	end
	else
		inherited;
end;

procedure TDateTimePicker.SetChecked(const Value: Boolean);
begin
	SendMessageW(DTM_SETSYSTEMTIME, Ord(not Value), LPARAM(@FValue));
end;

procedure TDateTimePicker.SetFormat(const Value: WideString);
begin
	SendMessageW(DTM_SETFORMATW, 0, LPARAM(PWideChar(Value)));
end;

procedure TDateTimePicker.SetValue(const Value: TSystemTime);
begin
	FValue := Value;
	SendMessageW(DTM_SETSYSTEMTIME, GDT_VALID, LPARAM(@Value));
end;

class function TDateTimePicker.SubClassName: PWideChar;
begin
	Result := DATETIMEPICK_CLASS
end;

{$ENDIF}

{$IFDEF HEADER}

{ THeaderControl.TItem }

constructor THeaderControl.TItem.Create(AHeaderControl: THeaderControl; AIndex: Integer);
begin
	FHeader := AHeaderControl;
  FIndex := AIndex;
	inherited Create;
end;

function THeaderControl.TItem.GetData: Pointer;
var
	Item: HD_ITEMW;
begin
	Item.Mask := HDI_LPARAM;
	FHeader.SendMessageW(HDM_GETITEMW, FIndex, LPARAM(@Item));
	Result := Pointer(Item.lParam);
end;

function THeaderControl.TItem.GetText: WideString;
var
	Item: HD_ITEMW;
begin
	SetLength(Result, 1024);
	Item.Mask := HDI_TEXT;
	Item.pszText := Pointer(Result);
	Item.cchTextMax := 1024;
	FHeader.SendMessageW(HDM_GETITEMW, FIndex, LPARAM(@Item));
	SetLength(Result, lstrlenW(Pointer(Result)));
end;

function THeaderControl.TItem.GetWidth: Integer;
var
	Item: HD_ITEMW;
begin
	Item.Mask := HDI_WIDTH;
	FHeader.SendMessageW(HDM_GETITEMW, FIndex, LPARAM(@Item));
	Result := Item.cxy;
end;

procedure THeaderControl.TItem.SetData(const Value: Pointer);
var
	Item: HD_ITEMW;
begin
	Item.Mask := HDI_LPARAM;
	Item.lParam := Integer(Value);
	FHeader.SendMessageW(HDM_SETITEMW, FIndex, LPARAM(@Item));
end;

procedure THeaderControl.TItem.SetText(const Value: WideString);
var
	Item: HD_ITEMW;
begin
	Item.Mask := HDI_TEXT;
	Item.pszText := PWideChar(Value);
	Item.cchTextMax := Length(Value) + 1;
	FHeader.SendMessageW(HDM_SETITEMW, FIndex, LPARAM(@Item));
end;

procedure THeaderControl.TItem.SetWidth(const Value: Integer);
var
	Item: HD_ITEMW;
begin
	Item.Mask := HDI_WIDTH;
	Item.cxy := Value;
	FHeader.SendMessageW(HDM_SETITEMW, FIndex, LPARAM(@Item));
end;

{ THeader }

function THeaderControl.Add: TItem;
var
	Item: HD_ITEMW;
	NewIndex: Integer;
begin
	Item.Mask := HDI_WIDTH or HDI_FORMAT or HDI_TEXT;
	Item.cxy := 8;
	Item.fmt := HDF_LEFT or HDF_STRING;
	Item.pszText := '';
	Item.cchTextMax := 1;
	NewIndex := SendMessageW(HDM_INSERTITEMW, Count, LPARAM(@Item));
	Result := TItem.Create(Self, NewIndex);
	Result.Owner := Garbage;
end;

procedure THeaderControl.AfterConstruction;
begin
  inherited;
	Font := StatusFont;
end;

constructor THeaderControl.Create(AParent: TAnyWindow; const APosition: TPoint;
	const ASize: TSize; const Styles: TStyles);
var
	CS: PCreateStructW;
begin
	inherited Create(AParent, '', APosition, ASize);
	CS := CreateStruct;
	CS^.style := CS^.style or Integer(Styles);
end;

function THeaderControl.GetCount: Integer;
begin
	Result := Header_GetItemCount(Handle)
end;

class function THeaderControl.GetDefaultSize: TSize;
begin
	Result := OneLineDefaultSize
end;

procedure THeaderControl.HandleNotification(Sender: TAnyWindow; Info: PNMHdr;
	var Result: Integer);
begin
	if (Sender = Self) and (Info.code = HDN_BEGINTRACKW) then
  	Result := 1
	else
		inherited;
end;

class function THeaderControl.SubClassName: PWideChar;
begin
	Result := WC_HEADER;
end;

procedure THeaderControl.WMSetCursor(var Message: TWMSetCursor);
begin
	if EnableDrag then
		inherited
	else
		SetCursor(LoadCursorW(0, PWideChar(IDC_ARROW)));
end;

{$ENDIF}

end.

