unit LCmmDlgs;

{*******************************************************}
{                                                       }
{       LightWindows                                    }
{       Common Dialogs                                  }
{                                                       }
{       2001..2004, 2008 YT                             }
{                                                       }
{*******************************************************}

interface

uses
	Windows, Messages, Types, CommDlg, ShlObj, ActiveX, SysUtils, Dlgs,
	SysUtilsW, ShlTips, 
	LConsts, LObjects, LActions, LWindows, LGraphics;

type
	T0CommonDialog = class(T0Object)
	private
		FParent: T0TopWindow;
		FTitle: WideString;
	protected
		function DoExecute: Boolean; virtual; abstract;
	public
		constructor Create(AParent: T0TopWindow; const ATitle: WideString);
		function Execute: Boolean;
		property Parent: T0TopWindow read FParent;
		property Title: WideString read FTitle write FTitle;
	end;

type
	TMessageBox = class(T0CommonDialog)
	private
		FText: WideString;
		FStyle: DWORD;
		FResult: DWORD;
	protected
		function DoExecute: Boolean; override;
	public
		class function Execute(Wnd: T0TopWindow; const Text, Title: WideString; Style: DWORD): DWORD; overload;
		class function Execute(const Text, Title: WideString; Style: DWORD): DWORD; overload;
		class function Execute(const Text: WideString; Style: DWORD): DWORD; overload;
		property Text: WideString read FText write FText;
		property Style: DWORD read FStyle write FStyle;
		property Result: DWORD read FResult;
	end;

type
	TOpenFileStyles = set of (
		ofReadOnly             =  0 {OFN_READONLY             = $00000001} ,
		ofOverwritePrompt      =  1 {OFN_OVERWRITEPROMPT      = $00000002} ,
		ofHideReadOnly         =  2 {OFN_HIDEREADONLY         = $00000004} ,
		ofNoChangeDir          =  3 {OFN_NOCHANGEDIR          = $00000008} ,
		ofShowHelp             =  4 {OFN_SHOWHELP             = $00000010} ,
		_ofEnableHook          =  5 {OFN_ENABLEHOOK           = $00000020} ,
		ofEnableTemplate       =  6 {OFN_ENABLETEMPLATE       = $00000040} ,
		ofEnableTemplateHandle =  7 {OFN_ENABLETEMPLATEHANDLE = $00000080} ,
		ofNoValidate           =  8 {OFN_NOVALIDATE           = $00000100} ,
		ofAllowMultiSelect     =  9 {OFN_ALLOWMULTISELECT     = $00000200} ,
		ofExtensionDifferent   = 10 {OFN_EXTENSIONDIFFERENT   = $00000400} ,
		ofPathMustExist        = 11 {OFN_PATHMUSTEXIST        = $00000800} ,
		ofFileMustExist        = 12 {OFN_FILEMUSTEXIST        = $00001000} ,
		ofCreatePrompt         = 13 {OFN_CREATEPROMPT         = $00002000} ,
		ofShareAware           = 14 {OFN_SHAREAWARE           = $00004000} ,
		ofNoReadOnlyReturn     = 15 {OFN_NOREADONLYRETURN     = $00008000} ,
		ofNoTestFileCreate     = 16 {OFN_NOTESTFILECREATE     = $00010000} ,
		ofNoNetworkButton      = 17 {OFN_NONETWORKBUTTON      = $00020000} ,
		ofNoLongNames          = 18 {OFN_NOLONGNAMES          = $00040000} ,
		_ofExplorer            = 19 {OFN_EXPLORER             = $00080000} ,
		ofNoDereferenceLinks   = 20 {OFN_NODEREFERENCELINKS   = $00100000} ,
		ofLongNames            = 21 {OFN_LONGNAMES            = $00200000} ,
		ofEnableIncludeNotify  = 22 {OFN_ENABLEINCLUDENOTIFY  = $00400000} ,
		ofEnableSizing         = 23 {OFN_ENABLESIZING         = $00800000} ,
		{24͌}
		ofDontAddToRecent      = 25 {OFN_DONTADDTORECENT      = $02000000} ,
		{26,27͌}
		ofForceShowHidden      = 28 {OFN_FORCESHOWHIDDEN      = $10000000} ,
		_ofHigh = 31);

	TOpenFileDialog = class(T0CommonDialog)
	private
		FDefWndProc: Pointer;
		FHandle: HWnd;
		FTemplate: Pointer;
		FHistoryList: TWideStringDynArray;
		FStyles: TOpenFileStyles;
		FFilter: WideString;
		FFilterIndex: Integer;
		FCurrentFilterIndex: Integer;
		FInitialDir: WideString;
		FDefaultExt: WideString;
		FFileName: TWideFileName;
		FFiles: TWideStringDynArray;
		procedure WMDestroy(var Message: TWMDestroy); message WM_DESTROY;
		procedure WMNCDestroy(var Message: TWMNCDestroy); message WM_NCDESTROY;
		procedure WMNotify(var Message: TWMNotify); message WM_NOTIFY;
		function GetNowFileName: TWideFileName;
	protected
		function CanClose(var OpenFileName: TOpenFileNameW): Boolean;
		function DoCanClose: Boolean; dynamic;
		function DoExecute: Boolean; override;
		procedure DoSelectionChange; dynamic;
		procedure DoFolderChange; dynamic;
		procedure DoTypeChange; dynamic;
		procedure GetFileNames(var OpenFileName: TOpenFileNameW);
		function GetStaticRect: TRect; virtual;
		procedure DoClose; dynamic;
		procedure DoShow; dynamic;
		function WindowProc(Msg: UINT; WP: WParam; LP: LParam): LRESULT;
		property Template: Pointer read FTemplate write FTemplate;
		function DoAPI(var Param: TOpenFilenameW): Boolean; virtual;
	public
		procedure AfterConstruction; override;
		procedure DefaultHandler(var Message); override;
		property Files: TWideStringDynArray read FFiles;
		property HistoryList: TWideStringDynArray read FHistoryList write FHistoryList;
		property Handle: HWnd read FHandle;
		property DefaultExt: WideString read FDefaultExt write FDefaultExt;
		property FileName: TWideFileName read FFileName write FFileName;
		property Filter: WideString read FFilter write FFilter;
		property FilterIndex: Integer read FFilterIndex write FFilterIndex;
		property InitialDirectory: WideString read FInitialDir write FInitialDir;
		property Styles: TOpenFileStyles read FStyles write FStyles;
		property NowFileName: TWideFileName read GetNowFileName;
		property NowFilterIndex: Integer read FCurrentFilterIndex;
	end;

	TSaveFileDialog = class(TOpenFileDialog)
	protected
		function DoAPI(var Param: tagOFNW): Boolean; override;
	public
		procedure AfterConstruction; override;
	end;

type
	TColorDialog = class(T0CommonDialog)
	private
		FCustomColors: array[0..15] of COLORREF;
		FColor: TColor;
	protected
		function DoExecute: Boolean; override;
	public
		property Color: TColor read FColor write FColor;
	end;

type
	TFontDialogStyles = set of (
		_cfScreenFonts         =  0 {CF_SCREENFONTS          = $00000001} ,
		cfPrinterFonts         =  1 {CF_PRINTERFONTS         = $00000002} ,
		cfShowHelp             =  2 {CF_SHOWHELP             = $00000004} ,
		cfEnableHook           =  3 {CF_ENABLEHOOK           = $00000008} ,
		cfEnableTemplate       =  4 {CF_ENABLETEMPLATE       = $00000010} ,
		cfEnableTemplateHandle =  5 {CF_ENABLETEMPLATEHANDLE = $00000020} ,
		_cfInitToLogFontStruct =  6 {CF_INITTOLOGFONTSTRUCT  = $00000040} ,
		cfUseStyle             =  7 {CF_USESTYLE             = $00000080} ,
		cfEffects              =  8 {CF_EFFECTS              = $00000100} ,
		cfApply                =  9 {CF_APPLY                = $00000200} ,
		cfAnsiOnly             = 10 {CF_ANSIONLY             = $00000400} ,
		cfNoVectorFonts        = 11 {CF_NOVECTORFONTS        = $00000800} ,
		cfNoSimulations        = 12 {CF_NOSIMULATIONS        = $00001000} ,
		cfLimitSize            = 13 {CF_LIMITSIZE            = $00002000} ,
		cfFixedPitchOnly       = 14 {CF_FIXEDPITCHONLY       = $00004000} ,
		cfWISIWYG              = 15 {CF_WYSIWYG              = $00008000} ,
		cfForceFontExist       = 16 {CF_FORCEFONTEXIST       = $00010000} ,
		cfScalableOnly         = 17 {CF_SCALABLEONLY         = $00020000} ,
		cfTTOnly               = 18 {CF_TTONLY               = $00040000} ,
		cfNoFaceSel            = 19 {CF_NOFACESEL            = $00080000} ,
		cfNoStyleSel           = 20 {CF_NOSTYLESEL           = $00100000} ,
		cfNoSizeSel            = 21 {CF_NOSIZESEL            = $00200000} ,
		cfSelectScript         = 22 {CF_SELECTSCRIPT         = $00400000} ,
		cfNoScriptSel          = 23 {CF_NOSCRIPTSEL          = $00800000} ,
		cfNoVertFonts          = 24 {CF_NOVERTFONTS          = $01000000} ,
		_cfHigh = 31);

	TFontDialog = class(T0CommonDialog)
	private
		FFont: TFont;
		FStyles: TFontDialogStyles;
	protected
		function DoExecute: Boolean; override;
	public
		procedure AfterConstruction; override;
		property Font: TFont read FFont;
		property Styles: TFontDialogStyles read FStyles write FStyles;
	end;

type
	TBrowseFolderStyles = set of (
		bfOnlyDir        =  0 {BIF_RETURNONLYFSDIRS   = $0001} ,
		bfNoNetwork      =  1 {BIF_DONTGOBELOWDOMAIN  = $0002} ,
		bfShowText       =  2 {BIF_STATUSTEXT         = $0004} ,
		bfOnlyFSAncestor =  3 {BIF_RETURNFSANCESTORS  = $0008} ,
		bfEditBox        =  4 {BIF_EDITBOX            = $0010} ,
		bfValidate       =  5 {BIF_VALIDATE           = $0020} ,
		bfOnlyComputer   = 12 {BIF_BROWSEFORCOMPUTER  = $1000} ,
		bfOnlyPrinter    = 13 {BIF_BROWSEFORPRINTER   = $2000} ,
		bfIncludeFiles   = 14 {BIF_BROWSEINCLUDEFILES = $4000} ,
		_bifHigh = 31);

	TBrowseFolderDialog = class(T0CommonDialog)
	private
		FItemIDList: PItemIDList;
		FNowFolder: PItemIDList;
		FRootDir: TWideFileName;
		FText: WideString;
		FHandle: HWND;
		FStyles: TBrowseFolderStyles;
		FAutoShowPath: Boolean;
		procedure SetDirectory(const Value: TWideFileName);
		function GetDirectory: TWideFileName;
		procedure SetItemIDList(const Value: PItemIDList);
	protected
		function DoExecute: Boolean; override;
		procedure DoSelectionChange(var ShowText: WideString); virtual;
		procedure DoShow; virtual;
		procedure SetNowFolder(Value: PItemIDList);
	public
		constructor Create(AParent: T0TopWindow; const ATitle: WideString);
		procedure BeforeDestruction; override;
		procedure AcceptOK(AEnabled: Boolean);
		property Handle: HWND read FHandle;
		property NowFolder: PItemIDList read FNowFolder;
		property Text: WideString read FText write FText;
		property Directory: TWideFileName read GetDirectory write SetDirectory;
		property ItemIDList: PItemIDList read FItemIDList write SetItemIDList;
		property RootDir: TWideFileName read FRootDir write FRootDir;
		property Styles: TBrowseFolderStyles read FStyles write FStyles;
		property AutoShowPath: Boolean read FAutoShowPath write FAutoShowPath;
	end;

implementation

threadvar
	CreationDialog: T0CommonDialog;

{ T0CommonDialog }

constructor T0CommonDialog.Create(AParent: T0TopWindow; const ATitle: WideString);
begin
	inherited Create;
	FParent := AParent;
	FTitle := ATitle
end;

function T0CommonDialog.Execute: Boolean;
var
	FPUControlWord: Word;
begin
	asm
		// Avoid FPU control word change in NETRAP.dll, NETAPI32.dll, etc
		FNSTCW  FPUControlWord
	end;
  Application.Async;
	try
		Result := DoExecute
	finally
		Application.RestoreAsync;
		asm
			FNCLEX
			FLDCW FPUControlWord
		end;
	end;
end;

{ TMessageBox }

function TMessageBox.DoExecute: Boolean;
var
	W: TAnyWindow;
	HW: HWND;
begin
	W := FParent;
	if W = nil then HW := GetActiveWindow else HW := W.Handle;
	FResult := MessageBoxW(HW, PWideChar(Text), PWideChar(Title), Style);

	Result := FResult <> ID_CANCEL;
end;

class function TMessageBox.Execute(Wnd: T0TopWindow; const Text, Title: WideString; Style: DWORD): DWORD;
var
	Dialog: TMessageBox;
begin
	Dialog := Create(Wnd, Title);
	try
		Dialog.Text := Text;
		Dialog.Style := Style;
		Dialog.Execute;
		Result := Dialog.Result;
	finally
		Dialog.Free;
	end
end;

class function TMessageBox.Execute(const Text, Title: WideString; Style: DWORD): DWORD;
var
	Window: T0TopWindow;
begin
	Window := nil;
	if CurrentThread <> nil then Window := CurrentThread.LeadWindow;
	Result := Execute(Window, Text, Title, Style)
end;

class function TMessageBox.Execute(const Text: WideString; Style: DWORD): DWORD;
begin
	Result := Execute(Text, Application.Title, Style)
end;

{ TOpenFileDialog }

function WindowProc(Wnd: HWND; Msg: UINT; WP: WParam; LP: LParam): LRESULT; stdcall;
var
	Dialog: TOpenFileDialog;
begin
	Dialog := Pointer(GetPropW(Wnd, LWPropName));
	Result := Dialog.WindowProc(Msg, WP, LP);
end;

function ExplorerHook(Wnd: HWnd; Msg: UINT; WParam: WPARAM; LParam: LPARAM): UINT; stdcall;
begin
	Result := 0;
	if Msg = WM_INITDIALOG then
	begin
		TOpenFileDialog(CreationDialog).FHandle := Wnd;
		SetPropW(Wnd, LWPropName, DWORD(CreationDialog));

		TOpenFileDialog(CreationDialog).FDefWndProc :=
						Pointer(SetWindowLongW(Wnd, GWL_WNDPROC, Longint(@WindowProc)));

		CreationDialog := nil;
	end
end;

procedure TOpenFileDialog.AfterConstruction;
begin
	inherited;

	FStyles := [ofHideReadOnly, ofEnableSizing, ofPathMustExist, ofFileMustExist];
	FFilterIndex := 1;
end;

function TOpenFileDialog.CanClose(var OpenFileName: TOpenFileNameW): Boolean;
begin
	GetFileNames(OpenFileName);
	Result := DoCanClose;
	FFiles := nil;
end;

procedure TOpenFileDialog.DefaultHandler(var Message);
begin
	if FHandle <> 0 then
		with TMessage(Message) do
			Result := CallWindowProcW(FDefWndProc, FHandle, Msg, WParam, LParam)
end;

function TOpenFileDialog.DoCanClose: Boolean;
begin
	Result := True
end;

procedure TOpenFileDialog.DoClose;
begin
end;

function TOpenFileDialog.DoExecute: Boolean;

	function AllocFilterStr(const S: WideString): WideString;
	var
		I: Integer;
	begin
		Result := S + #0;
		for I := Length(S) downto 1 do if Result[I] = '|' then  Result[I] := #0;
	end;

const
	MultiSelectBufferSize = High(Word) - 16;
var
	OpenFilename: TOpenFilenameW;
	TempFilter, TempFilename: WideString;
begin
	CreationDialog := Self;

	FFiles := nil;
	FillChar(OpenFileName, SizeOf(OpenFileName), 0);
	with OpenFilename do
	begin
		lStructSize := SizeOf(TOpenFilename);

		hInstance := SysInit.HInstance;
		TempFilter := AllocFilterStr(FFilter);
		lpstrFilter := PWideChar(TempFilter);
		nFilterIndex := FFilterIndex;
		FCurrentFilterIndex := FFilterIndex;
		if ofAllowMultiSelect in FStyles then
			nMaxFile := MultiSelectBufferSize else
			nMaxFile := MAX_PATH;
		SetLength(TempFilename, nMaxFile + 2);
		lpstrFile := PWideChar(TempFilename);
		FillChar(lpstrFile^, nMaxFile + 2, 0);
		lstrcpynW(lpstrFile, PWideChar(FFileName), nMaxFile);
		lpstrInitialDir := PWideChar(FInitialDir);
		lpstrTitle := PWideChar(FTitle);

		Flags := (OFN_ENABLEHOOK or OFN_EXPLORER) or DWORD(FStyles);
		FlagsEx := 0; {OFN_EX_NOPLACESBAR͂܂A}

		lpstrDefExt := PWideChar(FDefaultExt);
		if lpstrDefExt[0] = '.' then Inc(lpstrDefExt);

		lpfnHook := ExplorerHook;

		if Template <> nil then
		begin
			Flags := Flags or OFN_ENABLETEMPLATE;
			lpTemplateName := Template;
		end;
		hWndOwner := FParent.Handle;
		Result := DoAPI(OpenFileName);
		if Result then
		begin
			GetFileNames(OpenFilename);
			if (Flags and OFN_EXTENSIONDIFFERENT) <> 0 then
				Include(FStyles, ofExtensionDifferent) else
				Exclude(FStyles, ofExtensionDifferent);
			if (Flags and OFN_READONLY) <> 0 then
				Include(FStyles, ofReadOnly) else
				Exclude(FStyles, ofReadOnly);
			FFilterIndex := nFilterIndex;
		end;
	end;
end;

procedure TOpenFileDialog.DoFolderChange;
begin
end;

procedure TOpenFileDialog.DoSelectionChange;
begin
end;

procedure TOpenFileDialog.DoShow;
begin
end;

procedure TOpenFileDialog.DoTypeChange;
begin
end;

procedure TOpenFileDialog.GetFileNames(var OpenFileName: TOpenFileNameW);

	function ExtractFileName(P: PWideChar; var S: TWideFilename): PWideChar;
	begin
		Result := WideStrScan(P, WideChar(#0));
		if Result = nil then
		begin
			S := P;
			Result := WideStrEnd(P);
		end
		else
		begin
			SetString(S, P, Result - P);
			Inc(Result);
		end;
	end;

	procedure ExtractFileNames(P: PWideChar);
	var
		DirName, FileName: TWideFilename;
	begin
		P := ExtractFileName(P, DirName);
		P := ExtractFileName(P, FileName);
		if FileName = '' then
		begin
			SetLength(FFiles, Length(FFiles) + 1);
			FFiles[High(FFiles)] := DirName
		end
		else
		begin
			if AnsiLastChar(DirName)^ <> '\' then
				DirName := DirName + '\';
			repeat
				if (FileName[1] <> '\') and ((Length(FileName) <= 3) or
					(FileName[2] <> ':') or (FileName[3] <> '\')) then
					FileName := DirName + FileName;

				SetLength(FFiles, Length(FFiles) + 1);
				FFiles[High(FFiles)] := FileName;

				P := ExtractFileName(P, FileName);
			until FileName = '';
		end;
	end;

begin
	with OpenFileName do
	begin
		if ofAllowMultiSelect in FStyles then
		begin
			ExtractFileNames(lpstrFile);
			FFileName := FFiles[0];
		end
		else
		begin
			ExtractFileName(lpstrFile, FFileName);
			SetLength(FFiles, Length(FFiles) + 1);
			FFiles[High(FFiles)] := FileName;
		end;
	end;
end;

function TOpenFileDialog.GetStaticRect: TRect;
begin
	Result := ZeroRect;
	if FHandle <> 0 then
	begin
		GetWindowRect(GetDlgItem(FHandle, stc32), Result);
		MapWindowPoints(0, FHandle, Result, 2);
	end;
end;

function TOpenFileDialog.WindowProc(Msg: UINT; WP: WParam; LP: LParam): LRESULT;
var
	Message: TMessage;
begin
	Message.Msg := Msg;
	Message.WParam := WP;
	Message.LParam := LP;
	Message.Result := 0;
	try
		Dispatch(Message);
	except
		ShowException(ExceptObject, ExceptAddr);
	end;
	Result := Message.Result;
end;

procedure TOpenFileDialog.WMDestroy(var Message: TWMDestroy);
begin
	DoClose;
	inherited
end;

procedure TOpenFileDialog.WMNCDestroy(var Message: TWMNCDestroy);
begin
	inherited;
	RemovePropW(FHandle, LWPropName);
	FHandle := 0;
end;

function TOpenFileDialog.GetNowFileName: TWideFileName;
var
	Path: array[0..MAX_PATH] of WideChar;
begin
	Assert(FHandle <> 0);

	SendMessageW(GetParent(FHandle), CDM_GETFILEPATH, SizeOf(Path), LPARAM(@Path));
	Result := Path;
end;

procedure TOpenFileDialog.WMNotify(var Message: TWMNotify);
var
	Index: Integer;
	{Include: Boolean;}
begin
	case Message.NMHdr^.code of
		CDN_FILEOK:
			if not CanClose(POFNotifyW(Message.NMHdr)^.lpOFN^) then
			begin
				Message.Result := 1;
				SetWindowLongW(Handle, DWL_MSGRESULT, Message.Result);
				Exit;
			end;
		CDN_INITDONE: DoShow;
		CDN_SELCHANGE: DoSelectionChange;
		CDN_FOLDERCHANGE: DoFolderChange;
		CDN_TYPECHANGE:
			begin
				Index := POFNotifyW(Message.NMHdr)^.lpOFN^.nFilterIndex;
				if Index <> FCurrentFilterIndex then
				begin
					FCurrentFilterIndex := Index;
					DoTypeChange;
				end;
			end;
{		CDN_INCLUDEITEM:
			if Message.NMHdr <> nil then
			begin
				Include := True;
				DoIncludeItem(TOFNotifyExW(POFNotifyEx(Message.NMHdr)^), Include);
				Message.Result := Byte(Include);
			end; }
	end;

	inherited;
end;

function TOpenFileDialog.DoAPI(var Param: TOpenFilenameW): Boolean;
begin
	Result := GetOpenFileNameW(Param)
end;

{ TSaveFileDialog }

procedure TSaveFileDialog.AfterConstruction;
begin
	inherited;
	FStyles := [ofHideReadOnly, ofEnableSizing, ofOverwritePrompt];
end;

function TSaveFileDialog.DoAPI(var Param: tagOFNW): Boolean;
begin
	Result := GetSaveFileNameW(Param)
end;

{ TColorDialog }

function TColorDialog.DoExecute: Boolean;
var
	R: TChooseColorW;
begin
	R.lStructSize  := SizeOf(R);
	R.hWndOwner    := Parent.Handle;
	R.hInstance    := SysInit.HInstance;
	R.rgbResult    := Color.Value;
	R.lpCustColors := @FCustomColors[0];
	R.Flags        := CC_FULLOPEN or CC_RGBINIT;
	R.lCustData    := LPARAM(Self);
	R.lpfnHook     := nil;
	R.lpTemplateName := nil;
	Result := ChooseColorW(R);
	if Result then
		Color := RGBColor(R.rgbResult);
end;

{ TFontDialog }

procedure TFontDialog.AfterConstruction;
begin
	inherited;
	FFont := TFont.Create;
	FFont.Owner := Self;
end;

function TFontDialog.DoExecute: Boolean;
var
	R: TChooseFontW;
	LogFont: LOGFONTW;
begin
	LogFont := FFont.Log;

	FillChar(R, SizeOf(R), 0);
	R.lStructSize := SizeOf(R);
	R.hWndOwner   := Parent.Handle;
	R.lpLogFont   := @LogFont;
	R.Flags       := CF_SCREENFONTS or CF_INITTOLOGFONTSTRUCT or DWORD(Styles);
	R.lCustData   := LPARAM(Self);
	R.hInstance   := SysInit.HInstance;

	Result := ChooseFontW(R);
	if Result then
		FFont.Assign(LogFont);
end;

{ TBrowseFolderDialog }

function BrowseCallbackProc(Wnd: HWND; Msg: UINT; LP, Data: LPARAM): Integer; stdcall;
var
	Self: TBrowseFolderDialog absolute Data;
var
	ShowText: WideString;
	R, SR: TRect;
begin
	Assert(TObject(Data) is TBrowseFolderDialog);

	ShowText:= Self.FText;

	case Msg of
	BFFM_SELCHANGED:
	begin
		Self.SetNowFolder(PItemIDList(LP));
	end;
	BFFM_INITIALIZED:
	begin
		//nh
		Self.FHandle := Wnd;

		// Dir I
		if Self.ItemIDList <> nil then
		begin
			SendMessageW(Wnd, BFFM_SETSELECTIONW, 0, LPARAM(Self.ItemIDList));
			Self.SetNowFolder(Self.ItemIDList);
		end;

		if Self.Parent <> nil then
		begin
			{e̒}
			GetWindowRect(Wnd, R);
			GetWindowRect(Self.Parent.Handle, SR);
			SetWindowPos(Wnd, 0,
							SR.Left + (SR.Right - SR.Left - R.Right + R.Left) div 2,
							SR.Top + (SR.Bottom - SR.Top - R.Bottom + R.Top) div 2, 0, 0,
							SWP_NOSIZE or SWP_NOZORDER);
		end;

		//Cxg
		Self.DoShow;
	end;
	end;

	//ICxg
	Self.DoSelectionChange(ShowText);

	//eLXg𑗂
	if bfShowText in Self.Styles then
		SendMessageW(Wnd, BFFM_SETSTATUSTEXTW, 0, LPARAM(PWideChar(ShowText)));

	Result := 0;
end;

constructor TBrowseFolderDialog.Create(AParent: T0TopWindow; const ATitle: WideString);
begin
	inherited;

	FStyles := [bfOnlyDir];
end;

procedure TBrowseFolderDialog.BeforeDestruction;
begin
	if FItemIDList <> nil then CoTaskMemFree(FItemIDList);
	inherited;
end;

procedure TBrowseFolderDialog.DoShow;
begin
end;

procedure TBrowseFolderDialog.DoSelectionChange(var ShowText: WideString);
var
	Desktop: IShellFolder;
begin
	if AutoShowPath then
	begin
		SHGetDesktopFolder(Desktop);
		ShowText := GetDisplayName(Desktop, NowFolder, SHGDN_FORPARSING);
	end;
end;

function TBrowseFolderDialog.DoExecute: Boolean; {override}
var
	BrowseInfo: TBrowseInfoW;
	pidlBrowse: PItemIDList;
	DisplayNameBuffer: array[0..MAX_PATH] of WideChar;
begin
	FillChar(BrowseInfo, SizeOf(BrowseInfo), 0);

	BrowseInfo.hwndOwner := Parent.Handle;
	BrowseInfo.pidlRoot := nil;
	BrowseInfo.pszDisplayName := DisplayNameBuffer;
	BrowseInfo.lpszTitle := PWideChar(FTitle);
	BrowseInfo.ulFlags := DWORD(Styles);
	BrowseInfo.lpfn:= BrowseCallbackProc;
	BrowseInfo.lParam:= Longint(Self);

	//s
	try
		pidlBrowse := SHBrowseForFolderW(BrowseInfo);
	finally
		FHandle := 0;
		SetNowFolder(nil);
	end;

	Result := pidlBrowse <> nil;
	if Result then
		SetItemIDList(pidlBrowse);
end;

procedure TBrowseFolderDialog.AcceptOK(AEnabled: Boolean);
begin
	{ OK {^I^sɂ. }
	Assert(Handle <> 0);
	SendMessageW(Handle, BFFM_ENABLEOK, 0, Longint(AEnabled));
end;

procedure TBrowseFolderDialog.SetDirectory(const Value: TWideFileName);
var
	Desktop: IShellFolder;
	W: WideString;
	E, A: DWORD;
	New: PItemIDList;
begin
	W := Value;
	SHGetDesktopFolder(Desktop);
	if Desktop.ParseDisplayName(0, nil, PWideChar(W), E, New, A) = S_OK then
		SetItemIDList(New);
end;

function TBrowseFolderDialog.GetDirectory: TWideFileName;
var
	Buffer: array[0..1023] of WideChar;
begin
	if FItemIDList <> nil then
	begin
		SHGetPathFromIDListW(FItemIDList, Buffer);
		if (Buffer[0] = ':') and (Buffer[1] = ':') then Buffer[0] := #0;
		Result := Buffer;
	end
end;

procedure TBrowseFolderDialog.SetItemIDList(const Value: PItemIDList);
begin
	if FItemIDList <> nil then
		CoTaskMemFree(FItemIDList);

	if Value = nil then
		FItemIDList := nil
	else
		FItemIDList := CopyItemIDList(Value);
end;

procedure TBrowseFolderDialog.SetNowFolder(Value: PItemIDList);
begin
	if FNowFolder <> nil then
		CoTaskMemFree(FNowFolder);

	if Value = nil then
		FNowFolder := nil
	else
		FNowFolder := CopyItemIDList(Value);
end;

end.

