unit LDDEMan;

{*******************************************************}
{                                                       }
{       Light Windows                                   }
{       DDE Manager                                     }
{                                                       }
{       2000..2004, 2008 YT                             }
{                                                       }
{*******************************************************}

interface

uses
	Windows, DDEML, SysUtils, SysConst,
	SysHacks, AlgUtils,
	LConsts, LObjects;

type
	TDDEMessage = record
		CallType: UINT;
		Fmt: UINT;
		Conv: HCONV;
		sz1: HSZ;
		sz2: HSZ;
		Data: HDDEDATA;
		Data1: DWORD;
		Data2: DWORD;
		Result: HDDEDATA;
	end;

	T0DDETopic = class;

	T0DDEItem = class(T0Object)
	private
		FPredDDEItem: T0DDEItem;
		FSuccDDEItem: T0DDEItem;
		FTopic: T0DDETopic;
		FName: HSZ;
		function GetName: WideString;
	public
		constructor Create(ATopic: T0DDETopic; const AName: WideString);
		procedure AfterConstruction; override;
		procedure BeforeDestruction; override;
		property Name: WideString read GetName;
		property Topic: T0DDETopic read FTopic;
	end;

	T0DDEConversation = class(T0Object)
	private
		FConversation: HCONV;
	protected
		property Conversation: HCONV read FConversation;
	end;

	CEachDDEItem = function(Item: T0DDEItem; Context: Pointer): Boolean cdecl;

	T0DDETopic = class(T0DDEConversation)
	private
		FFirstDDEItem: T0DDEItem;
		FLastDDEItem: T0DDEItem;
		FTopic: HSZ;
		function GetTopic: WideString;
		function GetLinking: Boolean;
		procedure XTYPDisconnect(var Message: TDDEMessage); message XTYP_DISCONNECT;
	protected
		procedure AdviseStartItem(Item: T0DDEItem); virtual;
		procedure AdviseStopItem(Item: T0DDEItem); virtual;
		procedure SetTopic(const Value: WideString);
		procedure Disconnect;
		procedure DoOpen; virtual;
		procedure DoClose; virtual;
		function IndexOfName(const Item: WideString): T0DDEItem;
		function IndexOfHSZName(const Item: HSZ): T0DDEItem;
	public
		constructor Create;
		destructor Destroy; override;
		procedure CloseLink; virtual; abstract;
		function EnumerateItems(Proc: CEachDDEItem; Context: Pointer): T0DDEItem;
		property Linking: Boolean read GetLinking;
		property Topic: WideString read GetTopic;
		procedure DefaultHandler(var Message); override;
	end;

	T0DDEServer = class;

	TDDEServerItem = class(T0DDEItem)
	protected
		procedure DoPoke(const Data: WideString); virtual;
		procedure DoRequest(var Data: WideString); virtual;
	public
		constructor Create(ATopic: T0DDEServer; const AName: WideString);
	end;

	T0DDEServer = class(T0DDETopic)
	private
		procedure XTYPConnect(var Message: TDDEMessage); message XTYP_CONNECT;
		procedure XTYPConnectConfirm(var Message: TDDEMessage); message XTYP_CONNECT_CONFIRM;
		procedure XTYPExecute(var Message: TDDEMessage); message XTYP_EXECUTE;
		procedure XTYPAdvStart(var Message: TDDEMessage); message XTYP_ADVSTART;
		procedure XTYPAdvStop(var Message: TDDEMessage); message XTYP_ADVSTOP;
		procedure XTYPAdvReq(var Message: TDDEMessage); message XTYP_ADVREQ;
		procedure XTYPRequest(var Message: TDDEMessage); message XTYP_REQUEST;
		procedure XTYPPoke(var Message: TDDEMessage); message XTYP_POKE;
	protected
		procedure Poke(Item: HSZ; hData: HDDEDATA);
		procedure Request(Item: HSZ; var hData: HDDEDATA);
		procedure DoExecute(const Command: WideString); virtual;
		procedure DoPoke(Item: T0DDEItem; const Data: WideString); virtual;
		procedure DoRequest(Item: T0DDEItem; var Data: WideString); virtual;
	public
		constructor Create(const ATopic: WideString);
		destructor Destroy; override;
		procedure CloseLink; override;
		procedure PostAdvise(const Item: WideString);
	end;

	TDDEClient = class;

	TDDEClientItem = class(T0DDEItem)
	protected
		procedure DoAdviseData(const Data: String); virtual;
	public
		constructor Create(ATopic: TDDEClient; const AName: String);
	end;

	TDDEClient = class(T0DDETopic)
	private
		FService: HSZ;
		FTimeOut: Cardinal;
		function GetService: String;
		procedure XTYPAdvData(var Message: TDDEMessage); message XTYP_ADVDATA;
	protected
		procedure AdviseStartItem(Item: T0DDEItem); override;
		procedure AdviseStopItem(Item: T0DDEItem); override;
		procedure SetService(Value: String);
		procedure AdvData(Item: HSZ; hData: HDDEDATA);
		procedure DoAdviseData(Item: T0DDEItem; const Data: String); virtual;
	public
		constructor Create(const AService, ATopic: String);
		destructor Destroy; override;
		function RequestData(Item: T0DDEItem): String;
		procedure Request(Item: T0DDEItem);
		function Poke(Item: T0DDEItem; const Data: String): Boolean;
		function Execute(const Command: WideString): Boolean;
		procedure OpenLink;
		procedure CloseLink; override;
		property Service: String read GetService;
		property TimeOut: Cardinal read FTimeOut write FTimeOut {default 60000};
	end;

type
	IDDEExecuter = interface
		['{61F67EF6-1ED8-426B-8943-FB4716A6709E}']
		function Execute(const Command: WideString): Boolean;
	end;

type
	EDDEError = class(Exception);

const
	DMLErrorMessages: array[DMLERR_FIRST..DMLERR_LAST] of string = (
					SDMLERR_ADVACKTIMEOUT,       SDMLERR_BUSY,
					SDMLERR_DATAACKTIMEOUT,      SDMLERR_DLL_NOT_INITIALIZED,
					SDMLERR_DLL_USAGE,           SDMLERR_EXECACKTIMEOUT,
					SDMLERR_INVALIDPARAMETER,    SDMLERR_LOW_MEMORY,
					SDMLERR_MEMORY_ERROR,        SDMLERR_NOTPROCESSED,
					SDMLERR_NO_CONV_ESTABLISHED, SDMLERR_POKEACKTIMEOUT,
					SDMLERR_POSTMSG_FAILED,      SDMLERR_REENTRANCY,
					SDMLERR_SERVER_DIED,         SDMLERR_SYS_ERROR,
					SDMLERR_UNADVACKTIMEOUT,     SDMLERR_UNFOUND_QUEUE_ID);

procedure EnableDDE(Server: Boolean);
procedure DisableDDE(Server: Boolean);
function SendDDECommand(const Service, Topic, Command: WideString): Boolean;

implementation

type
	TDDETopicArray = array[0..32767] of T0DDEConversation;

var
	DDEMLInstance: Integer;
	DDEServerName: HSZ;
	DDETopics: ^TDDETopicArray;
	DDETopicCount: Integer;
	DDEEnabledCount: Integer;
	DDEServerCount: Integer;

function DMLErrorMessage: String;
var
	E: Integer;
begin
	E := DdeGetLastError(DDEMLInstance);
	if (E >= DMLERR_FIRST) and (E <= DMLERR_LAST) then
		Result := DMLErrorMessages[E];
end;

function DDECreateString(const S: WideString): HSZ;
begin
	Assert(DDEMLInstance <> 0);

	if S = '' then
		Result := 0
	else
	begin
		Result := DdeCreateStringHandleW(DDEMLInstance, PWideChar(S), CP_WINUNICODE);
		if Result = 0 then
			raise EDDEError.Create(DMLErrorMessage);
	end;
end;

function DDEExtractString(SZ: HSZ): WideString;
var
	Size: Integer;
begin
	Assert(DDEMLInstance <> 0);

	if SZ <> 0 then
	begin
		Size := DdeQueryStringW(DDEMLInstance, SZ, nil, 0, CP_WINUNICODE);
		SetLength(Result, Size);
		DdeQueryStringW(DDEMLInstance, SZ, PWideChar(Result), Size + 1, CP_WINUNICODE);
	end;
end;

procedure DDEFreeString(SZ: HSZ);
begin
	Assert(DDEMLInstance <> 0);

	if SZ <> 0 then
		if not DdeFreeStringHandle(DDEMLInstance, SZ) then
			raise EDdeError.Create(DMLErrorMessage);
end;

function DdeDataToStr(Data: HDDEDATA): WideString;
var
	Size: Integer;
begin
	Assert(DDEMLInstance <> 0);

	if Data <> 0 then
	begin
		Size := DdeGetData(Data, nil, 0, 0);
		SetLength(Result, Size div 2);
		DdeGetData(Data, PWideChar(Result), Size, 0);
		Result := PWideChar(Result);
	end;
end;

function DDEExecute(const Cmd: WideString): Boolean;
var
	Executer: IDDEExecuter;
begin
	Result := False;
	if Supports(Application, IDDEExecuter, Executer) then
		try
			Result := Executer.Execute(Cmd);
		except
			ShowException(ExceptObject, ExceptAddr);
		end;
end;

function DdeCallBack(CallType, Fmt: UINT; Conv: HCONV; sz1, sz2: HSZ;
				Data: HDDEDATA; Data1, Data2: DWORD): HDDEDATA; stdcall;
var
	Message: TDDEMessage;
	I: Integer;
begin
	Message.CallType := CallType;
	Message.Fmt      := Fmt;
	Message.Conv     := Conv;
	Message.sz1      := sz1;
	Message.sz2      := sz2;
	Message.Data     := Data;
	Message.Data1    := Data1;
	Message.Data2    := Data2;
	Message.Result   := 0;

	case CallType of
	XTYP_CONNECT:
	begin
		if DdeCmpStringHandles(sz2, DDEServerName) = 0 then
			if CompareText(DdeExtractString(sz1), 'SYSTEM') = 0 then
				Message.Result := 1
			else
				for I := 0 to DDETopicCount - 1 do
				begin
					DDETopics[I].Dispatch(Message);
					if Message.Result <> 0 then
						Break;
				end;
	end;
	XTYP_CONNECT_CONFIRM:
	begin
		if DdeCmpStringHandles(sz2, DDEServerName) = 0 then
			for I := 0 to DDETopicCount - 1 do
				DDETopics[I].Dispatch(Message);
	end;
	XTYP_ERROR:
	begin
		MessageBoxA(0, Pointer(DMLErrorMessages[Data1]), PChar(ParamStr(0)), MB_ICONERROR);
	end;
	else 
		for I := 0 to DDETopicCount - 1 do
			if Message.Conv = DDETopics[I].Conversation then
				DDETopics[I].Dispatch(Message);
	end;

	if CallType = XTYP_EXECUTE then
	begin
		if (Message.Result = 0) and (CompareText(DDEExtractString(sz1), 'SYSTEM') = 0) then
			if DDEExecute(DdeDataToStr(Data)) then
				Message.Result := DDE_FACK;
	end;

	Result := Message.Result;
end;

procedure EnableDDE(Server: Boolean);
begin
	if DDEEnabledCount = 0 then
	begin
		if DdeInitializeW(DDEMLInstance, DdeCallBack, 0, 0) <> DMLERR_NO_ERROR then
		begin
			DDEMLInstance := 0;
			raise EDDEError.Create(DMLErrorMessage);
		end;
	end;
	Inc(DDEEnabledCount);

	if Server then
	begin
		if DDEServerCount = 0 then
		begin
			DDEServerName := DDECreateString(
							ChangeFileExt(ExtractFileName(ParamStr(0)), ''));

			DdeNameService(DDEMLInstance, DDEServerName, 0, DNS_REGISTER);
		end;
		Inc(DDEServerCount);
	end;
end;

procedure DisableDDE(Server: Boolean);
begin
	if Server then
	begin
		Dec(DDEServerCount);
		if (DDEServerCount = 0) and (DDEServerName <> 0) then
		begin
			DdeNameService(DDEMLInstance, DDEServerName, 0, DNS_UNREGISTER);
			DdeFreeString(DDEServerName);
			DDEServerName := 0;
		end
	end;

	Dec(DDEEnabledCount);
	if (DDEEnabledCount = 0) and (DDEMLInstance <> 0) then
	begin
		DdeUninitialize(DDEMLInstance);
		DDEMLInstance := 0;
	end;
end;

function SendDDECommand(const Service, Topic, Command: WideString): Boolean;
var
	ServiceHandle, TopicHandle: HSZ;
	Conv: HCONV;
begin
	ServiceHandle := DdeCreateString(Service);
	TopicHandle := DdeCreateString(Topic);
	try
		Conv := DdeConnect(DDEMLInstance, ServiceHandle, TopicHandle, nil);
		Result := Conv <> 0 ;
		if not Result then exit;
		try
			Result := DdeClientTransaction(PWideChar(Command), (Length(Command) + 1) * SizeOf(WideChar),
							Conv, 0, 0, XTYP_EXECUTE, TIMEOUT_ASYNC, nil) <> 0;
		finally
			DdeDisconnect(Conv);
		end;
	finally
		DdeFreeString(TopicHandle);
		DdeFreeString(ServiceHandle);
	end;
end;

{ T0DDEItem }

constructor T0DDEItem.Create(ATopic: T0DDETopic; const AName: WideString);
begin
	inherited Create;

	FTopic := ATopic;

	FName := DDECreateString(AName);
end;

procedure T0DDEItem.AfterConstruction;
begin
	inherited;

	AddToLinkList(Topic.FFirstDDEItem, Topic.FLastDDEItem,
					Self, FPredDDEItem, FSuccDDEItem);

	if FTopic.Linking then
		FTopic.AdviseStartItem(Self);
end;

procedure T0DDEItem.BeforeDestruction;
begin
	if FTopic.Linking then
		FTopic.AdviseStopItem(Self);

	RemoveFromLinkList(Topic.FFirstDDEItem, Topic.FLastDDEItem,
					Self, FPredDDEItem, FSuccDDEItem);

	DDEFreeString(FName);

	inherited;
end;

function T0DDEItem.GetName: WideString;
begin
	Result := DDEExtractString(FName);
end;

{ T0DDETopic }

constructor T0DDETopic.Create;
begin
	inherited Create;

	Inc(DDETopicCount);
	ReallocMem(DDETopics, SizeOf(T0DDETopic) * DDETopicCount);
	DDETopics[DDETopicCount - 1] := Self;

	EnableDDE(Self is T0DDEServer);
end;

destructor T0DDETopic.Destroy;
var
	I, J: Integer;
begin
	DDEFreeString(FTopic);

	for I := 0 to DDETopicCount - 1 do
		if DDETopics[I] = Self then
		begin
			for J := I to DDETopicCount - 2 do
				DDETopics[J] := DDETopics[J + 1];

			Dec(DDETopicCount);
			ReallocMem(DDETopics, SizeOf(T0DDETopic) * DDETopicCount);

			Break;
		end;

	DisableDDE(Self is T0DDEServer);

	inherited;

	Assert((FFirstDDEItem = nil) and (FLastDDEItem = nil));
end;

procedure T0DDETopic.SetTopic(const Value: WideString);
var
	N: HSZ;
begin
	if Conversation <> 0 then
		raise EDdeError.Create(SDDEAlreadyLinking);

	N := DDECreateString(Value);
	if DdeCmpStringHandles(N, FTopic) = 0 then
		DDEFreeString(N)
	else
	begin
		DdeFreeString(FTopic);
		FTopic := N;
	end;
end;

function T0DDETopic.GetTopic: WideString;
begin
	Result := DDEExtractString(FTopic);
end;

function T0DDETopic.GetLinking: Boolean;
begin
	Result := Conversation <> 0
end;

procedure T0DDETopic.AdviseStartItem(Item: T0DDEItem);
begin
end;

procedure T0DDETopic.AdviseStopItem(Item: T0DDEItem);
begin
end;

procedure T0DDETopic.Disconnect;
begin
	try
		FConversation := 0;
		DoClose;
	except
		ShowException(ExceptObject, ExceptAddr);
	end;
end;

procedure T0DDETopic.DoOpen;
begin
end;

procedure T0DDETopic.DoClose;
begin
end;

function T0DDETopic.IndexOfHSZName(const Item: HSZ): T0DDEItem;

	function EnumProc(I: T0DDEItem): Boolean; cdecl;
	begin
		Result := DdeCmpStringHandles(I.FName, Item) = 0
	end;

begin
	Result := EnumerateItems(CEachDDEItem(@EnumProc), GetContext)
end;

function T0DDETopic.IndexOfName(const Item: WideString): T0DDEItem;
var
	N: HSZ;
begin
	N := DDECreateString(Item);
	try
		Result := IndexOfHSZName(N);
	finally
		DDEFreeString(N);
	end;
end;

procedure T0DDETopic.DefaultHandler(var Message);
begin
	TDDEMessage(Message).Result := 0
end;

procedure T0DDETopic.XTYPDisconnect(var Message: TDDEMessage);
begin
	Disconnect
end;

function T0DDETopic.EnumerateItems(Proc: CEachDDEItem; Context: Pointer): T0DDEItem;
var
	Next: T0DDEItem;
begin
	Result := FFirstDDEItem;
	while Result <> nil do
	begin
		Next := Result.FSuccDDEItem;
		if Proc(Result, Context) then Exit;
		Result := Next;
	end;
	Result := nil
end;

{ TDDEServerItem }

constructor TDDEServerItem.Create(ATopic: T0DDEServer; const AName: WideString);
begin
	inherited Create(ATopic, AName);
end;

procedure TDDEServerItem.DoPoke(const Data: WideString);
begin
end;

procedure TDDEServerItem.DoRequest(var Data: WideString);
begin
end;

{ T0DDEServer }

constructor T0DDEServer.Create(const ATopic: WideString);
begin
	inherited Create;

	SetTopic(ATopic);
end;

destructor T0DDEServer.Destroy;
begin
	//I[vȂI
	if Linking then
		CloseLink;

	inherited;
end;

procedure T0DDEServer.CloseLink;
begin
	if FConversation = 0 then
		raise EDdeError.Create(SDDENoLinking);

	if not DdeDisconnect(FConversation) then
		raise EDdeError.Create(DMLErrorMessage);

	FConversation := 0;
	DoClose;
end;

procedure T0DDEServer.PostAdvise(const Item: WideString);
var
	hItem: HSZ;
begin
	hItem:= DdeCreateString(Item);
	DdePostAdvise(DDEMLInstance, FTopic, hItem);
	DdeFreeString(hItem);
end;

procedure T0DDEServer.Request(Item: HSZ; var hData: HDDEDATA);
var
	Data: WideString;
	It: T0DDEItem;
begin
	hData:= 0;
	try
		It := IndexOfHSZName(Item);
		if It <> nil then
		begin
			DoRequest(It, Data);
			hData := DdeCreateDataHandle(DDEMLInstance,
							PWideChar(Data), (Length(Data) + 1) * SizeOf(WideChar),
							0, Item, CF_UNICODETEXT, 0);
		end;
	except
		ShowException(ExceptObject, ExceptAddr)
	end;
end;

procedure T0DDEServer.DoRequest(Item: T0DDEItem; var Data: WideString);
begin
	TDDEServerItem(Item).DoRequest(Data);
end;

procedure T0DDEServer.Poke(Item: HSZ; hData: HDDEDATA);
var
	It: T0DDEItem;
begin
	try
		It := IndexOfHSZName(Item);
		if It <> nil then
			DoPoke(It, DdeDataToStr(hData));
	except
		ShowException(ExceptObject, ExceptAddr);
	end;
end;

procedure T0DDEServer.DoPoke(Item: T0DDEItem; const Data: WideString);
begin
	TDDEServerItem(Item).DoPoke(Data);
end;

procedure T0DDEServer.DoExecute(const Command: WideString);
begin
end;

procedure T0DDEServer.XTYPConnect(var Message: TDDEMessage);
begin
	if (Conversation = 0) and (DdeCmpStringHandles(Message.sz1, FTopic) = 0) then
		Message.Result := 1
end;

procedure T0DDEServer.XTYPConnectConfirm(var Message: TDDEMessage);
begin
	if Message.sz1 = FTopic then
		try
			FConversation := Message.Conv;
			DoOpen;
		except
			ShowException(ExceptObject, ExceptAddr)
		end;
end;

procedure T0DDEServer.XTYPExecute(var Message: TDDEMessage);
var
	Cmd: string;
begin
	Message.Result := DDE_FACK;
	Cmd := DDEDataToStr(Message.Data);
	try
		DoExecute(Cmd);
	except
		ShowException(ExceptObject, ExceptAddr)
	end;
end;

procedure T0DDEServer.XTYPAdvReq(var Message: TDDEMessage);
begin
	Request(Message.sz2, Message.Result);
end;

procedure T0DDEServer.XTYPRequest(var Message: TDDEMessage);
begin
	Request(Message.sz2, Message.Result);
end;

procedure T0DDEServer.XTYPPoke(var Message: TDDEMessage);
begin
	Poke(Message.sz2, Message.Data);
end;

procedure T0DDEServer.XTYPAdvStart(var Message: TDDEMessage);
begin
	if IndexOfHSZName(Message.sz2) <> nil then
		Message.Result := 1;
end;

procedure T0DDEServer.XTYPAdvStop(var Message: TDDEMessage);
begin
end;

{ TDDEClientItem }

constructor TDDEClientItem.Create(ATopic: TDDEClient; const AName: String);
begin
	inherited Create(ATopic, AName);
end;

procedure TDDEClientItem.DoAdviseData(const Data: String);
begin
end;

{ TDDEClient }

constructor TDDEClient.Create(const AService, ATopic: String);
begin
	inherited Create;
	FTimeOut := 60000;
end;

destructor TDDEClient.Destroy;
begin
	if Conversation <> 0 then
		CloseLink;

	DdeFreeString(FService);

	inherited;
end;

function TDDEClient.GetService: String;
begin
	Result := DDEExtractString(FService);
end;

procedure TDDEClient.SetService(Value: String);
var
	N: HSZ;
begin
	if Conversation <> 0 then
		raise EDdeError.Create(SDDEAlreadyLinking);

	N := DDECreateString(Value);

	if DdeCmpStringHandles(N, FService) = 0 then
		DDEFreeString(N)
	else
	begin
		DDEFreeString(FService);
		FService := N;
	end
end;

function TDDEClient.RequestData(Item: T0DDEItem): String;
var
	hData: HDDEDATA;
begin
	hData := DdeClientTransaction(nil, 0, Conversation,
					Item.FName, CF_TEXT, XTYP_REQUEST, TimeOut, nil);

	if hData = 0 then
		Raise EDdeError.Create(DMLErrorMessage);

	Result := DdeDataToStr(hData);

	DdeFreeDataHandle(hData);
end;

procedure TDDEClient.Request(Item: T0DDEItem);
begin
	DoAdviseData(Item, RequestData(Item));
end;

function TDDEClient.Poke(Item: T0DDEItem; const Data: String): Boolean;
var
	Buf: String;
begin
	if Conversation = 0 then
		raise EDdeError.Create(SDDENoLinking);

	Buf := Data;
	UniqueString(Buf);

	Result := DdeClientTransaction(PChar(Buf), Length(Buf) + 1, Conversation,
		Item.FName, CF_TEXT, XTYP_POKE, TimeOut, nil) = DDE_FACK;
end;

function TDDEClient.Execute(const Command: WideString): Boolean;
begin
	if Conversation <> 0 then
		Result := DdeClientTransaction(PWideChar(Command), (Length(Command) + 1) * SizeOf(WideChar),
			Conversation, 0, 0, XTYP_EXECUTE, TIMEOUT_ASYNC, nil) <> 0
	else
		Result := False
end;

procedure TDDEClient.OpenLink;
var
	I: T0DDEItem;
begin
	if Conversation <> 0 then
		raise EDdeError.Create(SDDEAlreadyLinking);

	FConversation := DdeConnect(DDEMLInstance, FService, FTopic, nil);

	if FConversation = 0 then
		raise EDdeError.Create(DMLErrorMessage);

	I := FFirstDDEItem;
	while I <> nil do
	begin
		AdviseStartItem(I);
		I := I.FSuccDDEItem;
	end;

	DoOpen;
end;

procedure TDDEClient.CloseLink;
var
	I: T0DDEItem;
begin
	if Conversation = 0 then
		raise EDdeError.Create(SDDENoLinking);

	I := FFirstDDEItem;
	while I <> nil do
	begin
		AdviseStopItem(I);
		I := I.FSuccDDEItem;
	end;

	if not DdeDisconnect(Conversation) then
		raise EDdeError.Create(DMLErrorMessage);

	FConversation := 0;
	DoClose;
end;

procedure TDDEClient.AdviseStartItem(Item: T0DDEItem);
begin
	if DdeClientTransaction(nil, 0, Conversation, Item.FName, CF_TEXT,
					XTYP_ADVSTART or XTYPF_ACKREQ, TimeOut, nil) = 0 then
		raise EDdeError.Create(DMLErrorMessage);
end;

procedure TDDEClient.AdviseStopItem(Item: T0DDEItem);
begin
	if DdeClientTransaction(Nil, 0, Conversation, Item.FName, CF_TEXT,
					XTYP_ADVSTOP, TimeOut, Nil) = 0 then
		raise EDdeError.Create(DMLErrorMessage);
end;

procedure TDDEClient.AdvData(Item: HSZ; hData: HDDEDATA);
var
	It: T0DDEItem;
	Data: String;
begin
	try
		It := IndexOfHSZName(Item);
		if It <> nil then
		begin
			if hData = 0 then
				Data:= RequestData(It)
			else
				Data:= DdeDataToStr(hData);

			DoAdviseData(It, Data);
		end;
	except
		ShowException(ExceptObject, ExceptAddr)
	end;
end;

procedure TDDEClient.DoAdviseData(Item: T0DDEItem; const Data: String);
begin
	TDDEClientItem(Item).DoAdviseData(Data);
end;

procedure TDDEClient.XTYPAdvData(var Message: TDDEMessage);
begin
	AdvData(Message.sz2, Message.Data);
	Message.Result := DDE_FACK;
end;

end.

