unit LDXAudio;

interface

uses
	Windows, SysUtils, ActiveX,
	DirectSound, DirectMusic,
  ShlTips,
	LObjects;

type
	IAudioPath = IDirectMusicAudioPath8;

type
	TAudioManager = class(T0Collection)
	private
		FDirectMusicLoader: IDirectMusicLoader8;
		FDirectMusicPerformance: IDirectMusicPerformance8;
    procedure SetVolume(const Value: Integer);
	public
		constructor Create;
		destructor Destroy; override;
		function CreateAudioPath: IDirectMusicAudioPath8;
		property DirectMusicLoader: IDirectMusicLoader8 read FDirectMusicLoader;
		property DirectMusicPerformance: IDirectMusicPerformance8 read FDirectMusicPerformance;
		property Volume: Integer write SetVolume;
	end;

	TMusicKind = (mkWave, mkMIDI, mkMP3);

	TMusic = class(T0CollectionItem)
	private
		FDirectMusicSegment: IDirectMusicSegment8;
		FFileName: WideString;
		FKind: TMusicKind;
		FPrimary: Boolean;
		function GetManager: TAudioManager;
    function GetLength: Integer;
	public
		constructor Create(AManager: TAudioManager; const ASegment: IDirectMusicSegment8;
						AKind: TMusicKind); overload;
		constructor Create(AManager: TAudioManager; const AFileName: WideString); overload;
		constructor Create(Source: TMusic); overload;
		procedure BeforeDestruction; override;
		procedure Stop;
		function Play(Repeats: Integer = 1; const APath: IDirectMusicAudioPath8 = nil): Boolean;
		property Manager: TAudioManager read GetManager;
		property DirectMusicSegment: IDirectMusicSegment8 read FDirectMusicSegment;
		property Kind: TMusicKind read FKind;
		property Primary: Boolean read FPrimary write FPrimary;
		property Length: Integer read GetLength;
	end;

const
	InfiniteRepeats = Integer(DMUS_SEG_REPEAT_INFINITE);

implementation

function DMErrorString(Value: HResult): string;
begin
  case Value of
    HResult(DMUS_S_PARTIALLOAD)                    : Result := 'The object could only load partially. This can happen if some components, such as embedded tracks and tools, '+'are not registered properly. It can also happen if some content is missing; for example, if a segment uses a DLS collection that is not in the loader''s current search directory.';
    HResult(DMUS_S_PARTIALDOWNLOAD)                : Result := 'Return value from IDirectMusicBand::Download() which indicates that some of the instruments safely downloaded, but others failed. This usually occurs when some instruments are on PChannels not supported by the performance or port.';
    HResult(DMUS_S_REQUEUE)                        : Result := 'Return value from IDirectMusicTool::ProcessPMsg() which indicates to the performance that it should cue the PMsg again automatically.';
    HResult(DMUS_S_FREE)                           : Result := 'Return value from IDirectMusicTool::ProcessPMsg() which indicates to the performance that it should free the PMsg automatically.';
    HResult(DMUS_S_END)                            : Result := 'Return value from IDirectMusicTrack::Play() which indicates to the segment that the track has no more data after mtEnd.';
    HResult(DMUS_S_STRING_TRUNCATED)               : Result := 'Returned string has been truncated to fit the buffer size.';
    HResult(DMUS_S_LAST_TOOL)                      : Result := 'Returned from IDirectMusicGraph::StampPMsg(), this indicates that the PMsg is already stamped with the last tool in the graph. The returned PMsg''s tool pointer is now NULL.';
    HResult(DMUS_S_OVER_CHORD)                     : Result := 'Returned from IDirectMusicPerformance::MusicToMIDI(), this indicates that no note has been calculated because the music value has the note at a position higher than the top note of the chord. '+'This applies only to DMUS_PLAYMODE_NORMALCHORD play mode. This success code indicates that the caller should not do anything with the note. It is not meant to be played against this chord.';
    HResult(DMUS_S_UP_OCTAVE)                      : Result := 'Returned from IDirectMusicPerformance::MIDIToMusic(),  and IDirectMusicPerformance::MusicToMIDI(), this indicates that the note conversion generated a note value that is below 0, '+'so it has been bumped up one or more octaves to be in the proper MIDI range of 0 through 127. Note that this is valid for MIDIToMusic() '+'when using play modes DMUS_PLAYMODE_FIXEDTOCHORD and DMUS_PLAYMODE_FIXEDTOKEY, both of which store MIDI values in wMusicValue. With MusicToMIDI(), it is valid for all play modes. Ofcourse, DMUS_PLAYMODE_FIXED will never return this success code.';
    HResult(DMUS_S_DOWN_OCTAVE)                    : Result := 'Returned from IDirectMusicPerformance::MIDIToMusic(),  and IDirectMusicPerformance::MusicToMIDI(), this indicates that the note conversion generated a note value that is above 127, '+'so it has been bumped down one or more octaves to be in the proper MIDI range of 0 through 127. Note that this is valid for MIDIToMusic() '+'when using play modes DMUS_PLAYMODE_FIXEDTOCHORD and DMUS_PLAYMODE_FIXEDTOKEY, both of which store MIDI values in wMusicValue. With MusicToMIDI(), it is valid for all play modes. Ofcourse, DMUS_PLAYMODE_FIXED will never return this success code.';
    HResult(DMUS_S_NOBUFFERCONTROL)                : Result := 'Although the audio output from the port will be routed to the same device as the given DirectSound buffer, buffer controls such as pan and volume will not affect the output.';
    HResult(DMUS_S_GARBAGE_COLLECTED)              : Result := 'The requested operation was not performed because during CollectGarbage the loader determined that the object had been released.';
    HResult(DMUS_E_DRIVER_FAILED)                  : Result := 'An unexpected error was returned from a device driver, indicating possible failure of the driver or hardware.';
    HResult(DMUS_E_PORTS_OPEN)                     : Result := 'The requested operation cannot be performed while there are instantiated ports in any process in the system.';
    HResult(DMUS_E_DEVICE_IN_USE)                  : Result := 'The requested device is already in use (possibly by a non-DirectMusic client) and cannot be opened again.';
		HResult(DMUS_E_INSUFFICIENTBUFFER)             : Result := 'Buffer is not large enough for requested operation.';
    HResult(DMUS_E_BUFFERNOTSET)                   : Result := 'No buffer was prepared for the download data.';
    HResult(DMUS_E_BUFFERNOTAVAILABLE)             : Result := 'Download failed due to inability to access or create download buffer.';
    HResult(DMUS_E_NOTADLSCOL)                     : Result := 'Error parsing DLS collection. File is corrupt.';
    HResult(DMUS_E_INVALIDOFFSET)                  : Result := 'Wave chunks in DLS collection file are at incorrect offsets.';
    HResult(DMUS_E_ALREADY_LOADED)                 : Result := 'Second attempt to load a DLS collection that is currently open. ';
    HResult(DMUS_E_INVALIDPOS)                     : Result := 'Error reading wave data from DLS collection. Indicates bad file.';
    HResult(DMUS_E_INVALIDPATCH)                   : Result := 'There is no instrument in the collection that matches patch number.';
    HResult(DMUS_E_CANNOTSEEK)                     : Result := 'The IStream* doesn''t support Seek().';
    HResult(DMUS_E_CANNOTWRITE)                    : Result := 'The IStream* doesn''t support Write().';
    HResult(DMUS_E_CHUNKNOTFOUND)                  : Result := 'The RIFF parser doesn''t contain a required chunk while parsing file.';
    HResult(DMUS_E_INVALID_DOWNLOADID)             : Result := 'Invalid download id was used in the process of creating a download buffer.';
    HResult(DMUS_E_NOT_DOWNLOADED_TO_PORT)         : Result := 'Tried to unload an object that was not downloaded or previously unloaded.';
    HResult(DMUS_E_ALREADY_DOWNLOADED)             : Result := 'Buffer was already downloaded to synth.';
    HResult(DMUS_E_UNKNOWN_PROPERTY)               : Result := 'The specified property item was not recognized by the target object.';
    HResult(DMUS_E_SET_UNSUPPORTED)                : Result := 'The specified property item may not be set on the target object.';
    HResult(DMUS_E_GET_UNSUPPORTED)                : Result := 'The specified property item may not be retrieved from the target object.';
    HResult(DMUS_E_NOTMONO)                        : Result := 'Wave chunk has more than one interleaved channel. DLS format requires MONO.';
    HResult(DMUS_E_BADARTICULATION)                : Result := 'Invalid articulation chunk in DLS collection.';
    HResult(DMUS_E_BADINSTRUMENT)                  : Result := 'Invalid instrument chunk in DLS collection.';
    HResult(DMUS_E_BADWAVELINK)                    : Result := 'Wavelink chunk in DLS collection points to invalid wave.';
    HResult(DMUS_E_NOARTICULATION)                 : Result := 'Articulation missing from instrument in DLS collection.';
    HResult(DMUS_E_NOTPCM)                         : Result := 'Downoaded DLS wave is not in PCM format. ';
    HResult(DMUS_E_BADWAVE)                        : Result := 'Bad wave chunk in DLS collection';
    HResult(DMUS_E_BADOFFSETTABLE)                 : Result := 'Offset Table for download buffer has errors. ';
    HResult(DMUS_E_UNKNOWNDOWNLOAD)                : Result := 'Attempted to download unknown data type.';
    HResult(DMUS_E_NOSYNTHSINK)                    : Result := 'The operation could not be completed because no sink was connected to the synthesizer.';
    HResult(DMUS_E_ALREADYOPEN)                    : Result := 'An attempt was made to open the software synthesizer while it was already open. ASSERT?';
    HResult(DMUS_E_ALREADYCLOSED)                  : Result := 'An attempt was made to close the software synthesizer while it was already open. ASSERT?';
    HResult(DMUS_E_SYNTHNOTCONFIGURED)             : Result := 'The operation could not be completed because the software synth has not yet been fully configured. ASSERT?';
    HResult(DMUS_E_SYNTHACTIVE)                    : Result := 'The operation cannot be carried out while the synthesizer is active.';
    HResult(DMUS_E_CANNOTREAD)                     : Result := 'An error occurred while attempting to read from the IStream* object.';
    HResult(DMUS_E_DMUSIC_RELEASED)                : Result := 'The operation cannot be performed because the final instance of the DirectMusic object was released. Ports cannot be used after final release of the DirectMusic object.';
    HResult(DMUS_E_BUFFER_EMPTY)                   : Result := 'There was no data in the referenced buffer.';
    HResult(DMUS_E_BUFFER_FULL)                    : Result := 'There is insufficient space to insert the given event into the buffer.';
		HResult(DMUS_E_PORT_NOT_CAPTURE)               : Result := 'The given operation could not be carried out because the port is a capture port.';
    HResult(DMUS_E_PORT_NOT_RENDER)                : Result := 'The given operation could not be carried out because the port is a render port.';
    HResult(DMUS_E_DSOUND_NOT_SET)                 : Result := 'The port could not be created because no DirectSound has been specified. Specify a DirectSound interface via the IDirectMusic::SetDirectSound method; pass NULL to have DirectMusic manage usage of DirectSound.';
    HResult(DMUS_E_ALREADY_ACTIVATED)              : Result := 'The operation cannot be carried out while the port is active.';
    HResult(DMUS_E_INVALIDBUFFER)                  : Result := 'Invalid DirectSound buffer was handed to port. ';
    HResult(DMUS_E_WAVEFORMATNOTSUPPORTED)         : Result := 'Invalid buffer format was handed to the synth sink.';
    HResult(DMUS_E_SYNTHINACTIVE)                  : Result := 'The operation cannot be carried out while the synthesizer is inactive.';
    HResult(DMUS_E_DSOUND_ALREADY_SET)             : Result := 'IDirectMusic::SetDirectSound has already been called. It may not be changed while in use.';
    HResult(DMUS_E_INVALID_EVENT)                  : Result := 'The given event is invalid (either it is not a valid MIDI message or it makes use of running status). The event cannot be packed into the buffer.';
    HResult(DMUS_E_UNSUPPORTED_STREAM)             : Result := 'The IStream* object does not contain data supported by the loading object.';
    HResult(DMUS_E_ALREADY_INITED)                 : Result := 'The object has already been initialized.';
    HResult(DMUS_E_INVALID_BAND)                   : Result := 'The file does not contain a valid band.';
    HResult(DMUS_E_TRACK_HDR_NOT_FIRST_CK)         : Result := 'The IStream* object''s data does not have a track header as the first chunk, and therefore can not be read by the segment object.';
    HResult(DMUS_E_TOOL_HDR_NOT_FIRST_CK)          : Result := 'The IStream* object''s data does not have a tool header as the first chunk, and therefore can not be read by the graph object.';
    HResult(DMUS_E_INVALID_TRACK_HDR)              : Result := 'The IStream* object''s data contains an invalid track header (ckid is 0 and fccType is NULL,) and therefore can not be read by the segment object.';
    HResult(DMUS_E_INVALID_TOOL_HDR)               : Result := 'The IStream* object''s data contains an invalid tool header (ckid is 0 and fccType is NULL,) and therefore can not be read by the graph object.';
    HResult(DMUS_E_ALL_TOOLS_FAILED)               : Result := 'The graph object was unable to load all tools from the IStream* object data. This may be due to errors in the stream, or the tools being incorrectly registered on the client.';
    HResult(DMUS_E_ALL_TRACKS_FAILED)              : Result := 'The segment object was unable to load all tracks from the IStream* object data. This may be due to errors in the stream, or the tracks being incorrectly registered on the client.';
    HResult(DMUS_E_NOT_FOUND)                      : Result := 'The requested item was not contained by the object.';
    HResult(DMUS_E_NOT_INIT)                       : Result := 'A required object is not initialized or failed to initialize.';
    HResult(DMUS_E_TYPE_DISABLED)                  : Result := 'The requested parameter type is currently disabled. Parameter types may be enabled and disabled by certain calls to SetParam().';
    HResult(DMUS_E_TYPE_UNSUPPORTED)               : Result := 'The requested parameter type is not supported on the object.';
    HResult(DMUS_E_TIME_PAST)                      : Result := 'The time is in the past, and the operation can not succeed.';
    HResult(DMUS_E_TRACK_NOT_FOUND)                : Result := 'The requested track is not contained by the segment.';
    HResult(DMUS_E_TRACK_NO_CLOCKTIME_SUPPORT)     : Result := 'The track does not support clock time playback or getparam.';
    HResult(DMUS_E_NO_MASTER_CLOCK)                : Result := 'There is no master clock in the performance. Be sure to call IDirectMusicPerformance::Init().';
    HResult(DMUS_E_LOADER_NOCLASSID)               : Result := 'The class id field is required and missing in the DMUS_OBJECTDESC.';
    HResult(DMUS_E_LOADER_BADPATH)                 : Result := 'The requested file path is invalid.';
    HResult(DMUS_E_LOADER_FAILEDOPEN)              : Result := 'File open failed - either file doesn''t exist or is locked.';
    HResult(DMUS_E_LOADER_FORMATNOTSUPPORTED)      : Result := 'Search data type is not supported.';
    HResult(DMUS_E_LOADER_FAILEDCREATE)            : Result := 'Unable to find or create object.';
    HResult(DMUS_E_LOADER_OBJECTNOTFOUND)          : Result := 'Object was not found.';
    HResult(DMUS_E_LOADER_NOFILENAME)              : Result := 'The file name is missing from the DMUS_OBJECTDESC.';
    HResult(DMUS_E_INVALIDFILE)                    : Result := 'The file requested is not a valid file.';
    HResult(DMUS_E_ALREADY_EXISTS)                 : Result := 'The tool is already contained in the graph. Create a new instance.';
		HResult(DMUS_E_OUT_OF_RANGE)                   : Result := 'Value is out of range, for instance the requested length is longer than the segment.';
    HResult(DMUS_E_SEGMENT_INIT_FAILED)            : Result := 'Segment initialization failed, most likely due to a critical memory situation.';
    HResult(DMUS_E_ALREADY_SENT)                   : Result := 'The DMUS_PMSG has already been sent to the performance object via IDirectMusicPerformance::SendPMsg().';
    HResult(DMUS_E_CANNOT_FREE)                    : Result := 'The DMUS_PMSG was either not allocated by the performance via IDirectMusicPerformance::AllocPMsg(), or it was already freed via IDirectMusicPerformance::FreePMsg().';
    HResult(DMUS_E_CANNOT_OPEN_PORT)               : Result := 'The default system port could not be opened.';
    HResult(DMUS_E_CANNOT_CONVERT)                 : Result := 'A call to MIDIToMusic() or MusicToMIDI() resulted in an error because the requested conversion could not happen. This usually occurs when the provided DMUS_CHORD_KEY structure has an invalid chord or scale pattern.';
    HResult(DMUS_E_DESCEND_CHUNK_FAIL)             : Result := 'The end of the file was reached before the desired chunk was found.';
    HResult(DMUS_E_NOT_LOADED)                     : Result := 'An attempt to use this object failed because it first needs to be loaded.';
    HResult(DMUS_E_SCRIPT_LANGUAGE_INCOMPATIBLE)   : Result := 'The activeX scripting engine for the script''s language is not compatible with DirectMusic.';
    HResult(DMUS_E_SCRIPT_UNSUPPORTED_VARTYPE)     : Result := 'A varient was used that had a type that is not supported by DirectMusic.';
    HResult(DMUS_E_SCRIPT_ERROR_IN_SCRIPT)         : Result := 'An error was encountered while parsing or executing the script. The pErrorInfo parameter (if supplied) was filled with information about the error.';
    HResult(DMUS_E_SCRIPT_CANTLOAD_OLEAUT32)       : Result := 'Loading of oleaut32.dll failed.  VBScript and other activeX scripting languages require use of oleaut32.dll.  On platforms where oleaut32.dll is not present, only the DirectMusicScript language, which doesn''t require oleaut32.dll can be used.';
    HResult(DMUS_E_SCRIPT_LOADSCRIPT_ERROR)        : Result := 'An error occured while parsing a script loaded using LoadScript.  The script that was loaded contains an error.';
    HResult(DMUS_E_SCRIPT_INVALID_FILE)            : Result := 'The script file is invalid.';
    HResult(DMUS_E_INVALID_SCRIPTTRACK)            : Result := 'The file contains an invalid script track.';
    HResult(DMUS_E_SCRIPT_VARIABLE_NOT_FOUND)      : Result := 'The script does not contain a variable with the specified name.';
    HResult(DMUS_E_SCRIPT_ROUTINE_NOT_FOUND)       : Result := 'The script does not contain a routine with the specified name.';
    HResult(DMUS_E_SCRIPT_CONTENT_READONLY)        : Result := 'Scripts variables for content referenced or embedded in a script cannot be set.';
    HResult(DMUS_E_SCRIPT_NOT_A_REFERENCE)         : Result := 'Attempt was made to set a script''s variable by reference to a value that was not an object type.';
    HResult(DMUS_E_SCRIPT_VALUE_NOT_SUPPORTED)     : Result := 'Attempt was made to set a script''s variable by value to an object that does not support a default value property.';
    HResult(DMUS_E_INVALID_SEGMENTTRIGGERTRACK)    : Result := 'The file contains an invalid segment trigger track.';
    HResult(DMUS_E_INVALID_LYRICSTRACK)            : Result := 'The file contains an invalid lyrics track.';
    HResult(DMUS_E_INVALID_PARAMCONTROLTRACK)      : Result := 'The file contains an invalid parameter control track.';
    HResult(DMUS_E_AUDIOVBSCRIPT_SYNTAXERROR)      : Result := 'A script written in AudioVBScript could not be read because it contained a statement that is not allowed by the AudioVBScript language.';
    HResult(DMUS_E_AUDIOVBSCRIPT_RUNTIMEERROR)     : Result := 'A script routine written in AudioVBScript failed because an invalid operation occurred.  For example, adding the number 3 to a segment object would produce this error.  So would attempting to call a routine that doesn''t exist.';
    HResult(DMUS_E_AUDIOVBSCRIPT_OPERATIONFAILURE) : Result := 'A script routine written in AudioVBScript failed because a function outside of a script failed to complete. For example, a call to PlaySegment that fails to play because of low memory would return this error.';
    HResult(DMUS_E_AUDIOPATHS_NOT_VALID)           : Result := 'The Performance has set up some PChannels using the AssignPChannel command, which makes it not capable of supporting audio paths.';
    HResult(DMUS_E_AUDIOPATHS_IN_USE)              : Result := 'This is the inverse of the previous error. The Performance has set up some audio paths, which makes is incompatible with the calls to allocate pchannels, etc. ';
    HResult(DMUS_E_NO_AUDIOPATH_CONFIG)            : Result := 'A segment or song was asked for its embedded audio path configuration, but there isn''t any.';
    HResult(DMUS_E_AUDIOPATH_INACTIVE)             : Result := 'An audiopath is inactive, perhaps because closedown was called.';
    HResult(DMUS_E_AUDIOPATH_NOBUFFER)             : Result := 'An audiopath failed to create because a requested buffer could not be created.';
    HResult(DMUS_E_AUDIOPATH_NOPORT)               : Result := 'An audiopath could not be used for playback because it lacked port assignments.';
    HResult(DMUS_E_NO_AUDIOPATH)                   : Result := 'Attempt was made to play segment in audiopath mode and there was no audiopath.';
    HResult(DMUS_E_INVALIDCHUNK)                   : Result := 'Invalid data was found in a RIFF file chunk.';
    HResult(DMUS_E_AUDIOPATH_NOGLOBALFXBUFFER)     : Result := 'Attempt was made to create an audiopath that sends to a global effects buffer which did not exist.';
		HResult(DMUS_E_INVALID_CONTAINER_OBJECT)       : Result := 'The file does not contain a valid container object.';
		else Result := 'DirectMusic unrecognized error(' + IntToStr(Value) + ')';
  end;
end;

{ TAudioManager }

constructor TAudioManager.Create;
begin
	inherited;

	FDirectMusicLoader := CreateComObject(CLSID_DirectMusicLoader) as IDirectMusicLoader8;
	FDirectMusicPerformance := CreateComObject(CLSID_DirectMusicPerformance) as IDirectMusicPerformance8;
	FDirectMusicPerformance.InitAudio(nil, nil, 0, DMUS_APATH_DYNAMIC_STEREO, 64, DMUS_AUDIOF_ALL, nil);
end;

function TAudioManager.CreateAudioPath: IDirectMusicAudioPath8;
begin
	DirectMusicPerformance.CreateStandardAudioPath(
					DMUS_APATH_DYNAMIC_STEREO, 128, True, Result);
end;

destructor TAudioManager.Destroy;
begin
	Clear;
	if FDirectMusicPerformance <> nil then
	begin
		FDirectMusicPerformance.Stop(nil, nil, 0, 0);
		FDirectMusicPerformance.CloseDown;
	end;
	inherited;
end;

procedure TAudioManager.SetVolume(const Value: Integer);
var
	Path: IDirectMusicAudioPath;
begin
	DirectMusicPerformance.GetDefaultAudioPath(Path);
	Path.SetVolume(Value, 0);
end;

{ TMusic }

procedure TMusic.BeforeDestruction;
begin
	if FDirectMusicSegment <> nil then Stop;
	inherited;
end;

constructor TMusic.Create(AManager: TAudioManager; const ASegment: IDirectMusicSegment8; AKind: TMusicKind);
begin
	inherited Create(AManager);
	FKind := AKind;
	FDirectMusicSegment := ASegment;

	if Kind = mkMIDI then
		DirectMusicSegment.SetParam(GUID_StandardMIDIFile, $FFFFFFFF, 0, 0, nil);
end;

constructor TMusic.Create(AManager: TAudioManager; const AFileName: WideString);
var
	Segment: IDirectMusicSegment8;
	HR: HRESULT;
	E: EOSError;
	Kind: TMusicKind;
begin
  FFileName := AFileName;

	HR := AManager.DirectMusicLoader.LoadObjectFromFile(CLSID_DirectMusicSegment,
					IID_IDirectMusicSegment8, PWideChar(AFileName), Segment);
	if FAILED(HR) then
	begin
		E := EOSError.Create(DMErrorString(HR));
		E.ErrorCode := HR;
		raise E;
	end;

	if lstrcmpiW(PWideChar(AFileName) + System.Length(AFileName) - 8, '.mp3.wav') = 0 then
		Kind := mkMP3
	else if lstrcmpiW(PWideChar(AFileName) + System.Length(AFileName) - 4, '.mid') = 0 then
		Kind := mkMIDI
	else
		Kind := mkWave;

	Create(AManager, Segment, Kind);
end;

constructor TMusic.Create(Source: TMusic);
var
	Segment: IDirectMusicSegment;
begin
	Source.DirectMusicSegment.Clone(0, 0, Segment);
	Create(Source.Manager, Segment as IDirectMusicSegment8, Source.Kind);
end;

function TMusic.GetLength: Integer;
begin
	FDirectMusicSegment.GetLength(Result)
end;

function TMusic.GetManager: TAudioManager;
begin
	Assert(Owner is TAudioManager);
	Result := TAudioManager(Owner)
end;

function TMusic.Play(Repeats: Integer = 1; const APath: IDirectMusicAudioPath8 = nil): Boolean;
var
	State: IDirectMusicSegmentState;
	Flag: DWORD;
begin
	DirectMusicSegment.SetRepeats(Repeats);
	DirectMusicSegment.Download(Manager.DirectMusicPerformance);

	Flag := 0;
	if not Primary then Flag := DMUS_SEGF_SECONDARY;

	Result :=	Succeeded(Manager.DirectMusicPerformance.PlaySegmentEx(
					DirectMusicSegment, nil, nil, Flag, 0, @State, nil, APath));
end;

procedure TMusic.Stop;
begin
	Manager.DirectMusicPerformance.StopEx(DirectMusicSegment, 0, 0);
end;

end.
