unit LEditor;

interface

uses
	Windows, Messages, Types, SysUtils, Imm,
	SysHacks, WinTips,
	BinaryBuffer, BinaryDecoder, BinaryEditor,
	LObjects, LActions, LGraphics, LWindows, LStdCtls, LAddCtls;

type
	TEditorColor = record
		Foreground: COLORREF;
		Background: COLORREF;
	end;

	TStateColors = array[TState] of TEditorColor;

	TEditorLayout = class(TLayout)
	private
		FStateColors: TStateColors;
		FSelectedColor: TEditorColor;
		FLineColor: COLORREF;
		FSymbolColor: COLORREF;
		FNumberColor: COLORREF;
		FWidthCache: array[WideChar] of Byte;
		FFontSize: TSize;
		FFont: TAnyFont;
		FFixedFont: TAnyFont;
		FSecondFont: TAnyFont;
		FSecondFontLetters: WideString;
		FHalfWidthFont: TAnyFont;
		FHalfWidthFontShift: Integer;
		FLineMargin: Integer;
		FDC: HDC;
		FLineCount: Integer;
		FTabCount: Integer;
    function GetImeFont: TAnyFont;
	protected
		procedure ReadyDC;
	public
		constructor Create(
			ADocument: TDocument;
			ADecoder: TDecoder;
			AFont: TAnyFont;
			ASecondFont: TAnyFont;
			const ASecondFontLetters: WideString;
			AHalfWidthFont: TAnyFont;
			AFixedFont: TAnyFont;
			ALineCount: Integer;
			ATabCount: Integer;
			AHanging: WideString;
			AExpelling: WideString;
			AIndenting: WideString;
			AWordWrap: Boolean;
			ALineMargin: Integer;
			AStateColors: TStateColors;
			ASelectedColor: TEditorColor;
			ALineColor: COLORREF;
			ASymbolColor: COLORREF;
			ANumberColor: COLORREF);
		procedure BeforeDestruction; override;
		procedure Release;
		function LetterWidth(Letter: WideChar): Integer; override;
		function StandardLetterWidth: Integer; override;
		property Font: TAnyFont read FFont;
		property FixedFont: TAnyFont read FFixedFont;
		property SecondFont: TAnyFont read FSecondFont;
		property SecondFontLetters: WideString read FSecondFontLetters;
		property HalfWidthFont: TAnyFont read FHalfWidthFont;
		property HalfWidthFontShift: Integer read FHalfWidthFontShift;
		property ImeFont: TAnyFont read GetImeFont;
		property LineMargin: Integer read FLineMargin;
		property StateColors: TStateColors read FStateColors;
		property SelectedColor: TEditorColor read FSelectedColor;
		property LineColor: COLORREF read FLineColor;
		property SymbolColor: COLORREF read FSymbolColor;
		property NumberColor: COLORREF read FNumberColor;
		property LineCount: Integer read FLineCount;
		property TabCount: Integer read FTabCount;
	end;

	TScrollDirection = (sdUp, sdDown);
	TEditorShowOptions = set of (esoLineNumber, esoRuler, esoRulerColmn, esoDoubleBuffering);

	TEditorDraggingState = (edsNone, edsDown, edsDragOrClick);

	TEditorRecordingCommand = (ercInputText, ercInputData,
		ercInputTab, ercInputLineFeed,
		ercDelete, ercBackspace,
		ercCursorLeft, ercCursorUp, ercCursorRight, ercCursorDown,
		ercCursorLineTop, ercCursorLineEnd, ercWordLeft, ercWordRight,
		ercStartSelection, ercClearSelection);

	{XN[ʒuۂꍇ̊ʒu}
	TScrollingLinePosition = record
		Base: TScrollBase;
  	Gap : Integer;
	end;

	{ҏWEBhE}
	TEditorPane = class(T0SplitPane, IViewUI)
	private
		FDraggingPosition: TPoint;
		FView: TView;
		FLayout: TEditorLayout;
		FLeft: Integer;
		FCaretPosition: Natural;
		FCaretLine: Integer;
		FCaretColumn: Integer;
		FAnchorPosition: Natural;
		FUnderLineY: Integer;
		FCaretPreferableX: Integer;
		FAutoScrolling: DWORD;
		FScrollFrom: Natural;
		FScrollBase: TScrollBase;
		FShowOptions: TEditorShowOptions;
		FLineFeed: TLineFeedCode;
		FSemiFreeMoving: Boolean;
		FKeepCaretYPosition: Boolean;
		FHexMode: Boolean;
		FHexInputing: Byte;
		FInsertMode: Boolean;
		FSelected: Boolean;
		FUpdateScrollBar: Boolean;
		FAdjustStartLine: Boolean;
		FMouseRepeating: Boolean;
		FDraggingState: TEditorDraggingState;
		function GetSelected: Boolean;
		function GetSelectedLength: Natural;
		function GetSelectionFirstPosition: Natural;
		function GetSelectionLastPosition: Natural;
		function GetSelectedText: WideString;
		procedure SetSelectedText(const Value: WideString);
		function GetSelectedData: string;
		procedure SetSelectedData(const Value: string);
		procedure SetCaretPosition(const Value: Natural);
		procedure SetCaretLine(const Value: Integer);
		procedure SetHexMode(const Value: Boolean);
		procedure SetInsertMode(const Value: Boolean);
		procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
		procedure WMLButtonUp(var Message: TWMLButtonUp); message WM_LBUTTONUP;
		procedure WMLButtonDblClk(var Message: TWMLButtonDblClk); message WM_LBUTTONDBLCLK;
		procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
		procedure WMMouseWheel(var Message: TWMMouseWheel); message WM_MOUSEWHEEL;
		procedure WMImeRequest(var Message: TMessage); message WM_IME_REQUEST;
		procedure WMCopy(var Message: TWMCopy); message WM_COPY;
		procedure WMHScroll(var Message: TWMHScroll); message WM_HSCROLL;
		procedure WMVScroll(var Message: TWMVScroll); message WM_VSCROLL;
		procedure WMKillFocus(var Message: TWMKillFocus); message WM_KILLFOCUS;
		procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
		procedure WMSize(var Message: TWMSize); message WM_SIZE;
		procedure WMTimer(var Message: TWMTimer); message WM_TIMER;
		procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
		procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
		{IViewUI}
		procedure Changed(First, OldLast, NewLast, RepaintingFirst, RepaintingLast: Natural);
		function ScrollBase: TScrollBase;
	protected
		procedure Initialize; override;
		procedure Finalize; override;
		procedure Update; override;
		function KeyDown(ShiftState: TShiftKeys; Code: Byte): Boolean; override;
		function KeyInput(Code: WideChar): Boolean; override;
		function KeyUp(ShiftState: TShiftKeys; Code: Byte): Boolean; override;
		function ReplaceLength(Bytes: Natural): Natural;
		procedure ApplyLayout; virtual;
		procedure RepaintArea(First, Last: Natural);
		procedure RepaintCaret;
		procedure RepaintScroll(X, Y: Integer);
		procedure PaintWindow(DC: HDC); override;
		procedure UpdateScrollBar;
		procedure AdjustForBottom;
		procedure SetCaretPositionInternal;
		function ViewRect: TRect;
		function HexWidth: Integer;
		procedure BeginDrag; virtual;
		procedure RecordCommand(Command: TEditorRecordingCommand; Text: WideChar = #0; Data: AnsiChar = #0); virtual;
	public
		constructor Create(AParent: TAnyWindow; AHorz, AVert: TScrollBar; ALayout: TLayout; AViewClass: TViewClass);
		procedure ReleaseView;
		procedure SetLayout(ALayout: TLayout; AViewClass: TViewClass);
		function CanUndo: Boolean;
		function CanRedo: Boolean;
		procedure Undo;
		procedure Redo;
		procedure InputText(const Text: WideString);
		procedure InputData(const Data: string);
		procedure FeedLine;
		procedure FeedForm;
		procedure Backspace;
		procedure Delete;
		procedure DeleteSelected;
		procedure CopyToClipboard;
		procedure CutToClipboard;
		procedure PasteFromClipboard(AsText: Boolean = False);
		function CanPaste: Boolean;
		procedure StartSelection;
		procedure ClearSelection;
		procedure IndentSelection;
		procedure UnindentSelection;
		procedure SelectAll;
		procedure CursorLeft;
		procedure CursorRight;
		procedure CursorUp;
		procedure CursorDown;
		procedure CursorFirst;
		procedure CursorLast;
		procedure CursorLineTop(Smart: Boolean = True);
		procedure CursorLineEnd;
		procedure CursorScrollUp;
		procedure CursorScrollDown;
		procedure WordLeft;
		procedure WordRight;
		procedure SelectWord(Dir: Integer);
		procedure ScrollUp;
		procedure ScrollDown;
		procedure AutoScroll(Dir: TScrollDirection; Repeating: Boolean);
		procedure StopAutoScroll;
		function ScrollingLinePositon: TScrollingLinePosition;
		procedure Scroll(const Position: TScrollingLinePosition); overload;
		procedure Scroll; overload;
		function ClientToPosition(const P: TPoint): Natural;
		function ClientToHexMode(const P: TPoint): Boolean;
		property CaretPosition: Natural read FCaretPosition write SetCaretPosition;
		property CaretLine: Integer read FCaretLine write SetCaretLine;
		property CaretColumn: Integer read FCaretColumn;
		property Selected: Boolean read GetSelected;
		property SelectionFirstPosition: Natural read GetSelectionFirstPosition;
		property SelectionLastPosition: Natural read GetSelectionLastPosition;
		property SelectedText: WideString read GetSelectedText write SetSelectedText;
		property SelectedData: string read GetSelectedData write SetSelectedData;
		property SelectedLength: Natural read GetSelectedLength;
		property HexMode: Boolean read FHexMode write SetHexMode;
		property InsertMode: Boolean read FInsertMode write SetInsertMode;
		property LineFeed: TLineFeedCode read FLineFeed write FLineFeed;
		property ShowOptions: TEditorShowOptions read FShowOptions write FShowOptions;
		property SemiFreeMoving: Boolean read FSemiFreeMoving write FSemiFreeMoving;
		property KeepCaretYPosition: Boolean read FKeepCaretYPosition write FKeepCaretYPosition;
		property View: TView read FView;
		property Layout: TEditorLayout read FLayout;
	end;

	TEditor = class(T0SplitWindow)
	private
		FDocument: TDocument;
		FMarker: TBufferMarker;
		FUnderLine: Boolean;
		FLostFocus: Boolean;
		procedure SetUnderLine(const Value: Boolean);
		function GetMarker: TBufferMarker;
		function GetMarked: Boolean;
	protected
		function CreatePane(AHorz, AVert: TScrollBar): T0SplitPane; override;
		function ReadyLayout: TLayout;
		function ReadyViewClass: TViewClass;
		function GetReferencePane: TEditorPane;
		procedure DoSplit(Dir: TBarStyle; Per: Real); override;
		function CreateDefaultLayout: TEditorLayout; virtual;
		procedure ReleaseMarker;
	public
		constructor Create(AParent: TAnyWindow;
			const APosition: TPoint; const ASize: TSize; ASplitable: TBarStyles);
		destructor Destroy; override;
		procedure Split;
		procedure Switch;
		procedure RepaintPanes;
		property Splited;
		property Document: TDocument read FDocument;
		property UnderLine: Boolean read FUnderLine write SetUnderLine;
		property LostFocus: Boolean read FLostFocus write FLostFocus;
		property Marker: TBufferMarker read GetMarker;
		property Marked: Boolean read GetMarked;
	end;

implementation

const
	PowerWitchFormat = 'PWtR Clipboard Format';
	AutoScrollUpTimerID = $A01;
	AutoScrollDownTimerID = $A02;
	MouseScrollTimerID = $A03;
	TimerElapse = 25;
	ModifiedColor =  $7f7fff;
	UnderLineColor = $7f7fff;
	BomMark: WideChar = WideChar($2620);
	RulerHeight = 8;

type
	TDigit = 0..15;

procedure PaintDigit(DC: HDC; X, Y: Integer; Color: COLORREF; Digit: TDigit);
const
	Digits: array[TDigit] of Cardinal = (
		{0} $25552, {1} $22222, {2} $25127, {3} $61216,
		{4} $55711, {5} $74616, {6} $24652, {7} $71222,
		{8} $25252, {9} $75311, {A} $25575, {B} $65656,
		{C} $25452, {D} $65556, {E} $74747, {F} $74744);
var
	XI, YI: Integer;
	Bit, Data: Cardinal;
begin
	Data := Digits[Digit];
	Bit := 1;
	for YI := 4 downto 0 do
	begin
		for XI := 2 downto 0 do
		begin
			if Data and Bit <> 0 then SetPixel(DC, X + XI, Y + YI, Color);
			Bit := Bit shl 1;
		end;
		Bit := Bit shl 1;
	end;
end;

{ TEditorLayout }

procedure TEditorLayout.BeforeDestruction;
begin
	Release;
	FFont.Free;
	FFixedFont.Free;
	FSecondFont.Free;
	FHalfWidthFont.Free;
	inherited;
end;

constructor TEditorLayout.Create(ADocument: TDocument; ADecoder: TDecoder;
	AFont, ASecondFont: TAnyFont; const ASecondFontLetters: WideString;
	AHalfWidthFont, AFixedFont: TAnyFont; ALineCount, ATabCount: Integer;
	AHanging, AExpelling, AIndenting: WideString; AWordWrap: Boolean;
	ALineMargin: Integer; AStateColors: TStateColors;
	ASelectedColor: TEditorColor;
	ALineColor, ASymbolColor, ANumberColor: COLORREF);

	function GetFontSize(Font: HFONT; const C: WideChar): TSize;
	var
		Org: THandle;
	begin
		Org := SelectObject(FDC, Font);
		GetTextExtentPoint32W(FDC, @C, 1, Result);
		SelectObject(FDC, Org);
	end;

	procedure SetSecondFont;
	var
		I: Integer;
		W: WideChar;
		S: TSize;
		Org: THandle;
	begin
		Org := SelectObject(FDC, FSecondFont.Handle);
		for I := Length(FSecondFontLetters) downto 1 do
		begin
			W := FSecondFontLetters[I];
			GetTextExtentPoint32W(FDC, @W, 1, S);
			FWidthCache[W] := S.cx;
		end;
		SelectObject(FDC, Org);
	end;

	procedure SetHalfWidthFont;
	var
		Y, H: Integer;
		I: WideChar;
		S: TSize;
		Org: THandle;
	begin
		repeat
			Y := GetFontSize(FHalfWidthFont.Handle, '|').cy;
			if Y <= FFontSize.cy + FLineMargin then Break;

			H := FHalfWidthFont.Height;
			asm
				mov eax, H
				sar eax, 31 {shr͎ZpVtgƂĎgȂ݂Ȃ̂}
				or eax, 1
				sub H, eax
			end;
			TFont(FHalfWidthFont).Height := H;
		until False;

		FHalfWidthFontShift := FFontSize.cy - Y;

		Org := SelectObject(FDC, FHalfWidthFont.Handle);
		for I := #$21 to #$ff do
		begin
			GetTextExtentPoint32W(FDC, @I, 1, S);
			FWidthCache[I] := S.cx;
		end;
		if FWidthCache['W'] = FWidthCache['l'] then
		begin
			{}
			if FWidthCache['W'] <> FFontSize.cx then
			begin
				for I := #$21 to #$ff do FWidthCache[I] := FFontSize.cx;
			end;
		end;
		SelectObject(FDC, Org);
	end;

begin
	Assert(AFont <> nil);

	FFont := TFont.Create;
	TFont(FFont).Assign(AFont);
	ReadyDC;
	FFontSize := GetFontSize(Font.Handle, 'x');
	FWidthCache[UnicodeBOM] := FFontSize.cy;

	FLineMargin := ALineMargin;

	if AHalfWidthFont <> nil then
	begin
		FHalfWidthFont := TFont.Create;
		TFont(FHalfWidthFont).Assign(AHalfWidthFont);
		TFont(FHalfWidthFont).Height := FFont.Height;
		SetHalfWidthFont;
	end;

	if (ASecondFont <> nil) and (ASecondFontLetters <> '') then
	begin
		FSecondFont := TFont.Create;
		TFont(FSecondFont).Assign(ASecondFont);
		FSecondFontLetters := ASecondFontLetters;
		SetSecondFont;
	end;

	FFixedFont := TFont.Create;
	if AFixedFont <> nil then
		TFont(FFixedFont).Assign(AFixedFont)
	else
		TFont(FFixedFont).Assign(AFont);
	TFont(FFixedFont).Pitch := FIXED_PITCH;

	FStateColors := AStateColors;
	FSelectedColor := ASelectedColor;
	FLineColor := ALineColor;
	FSymbolColor := ASymbolColor;
	FNumberColor := ANumberColor;

	FLineCount := ALineCount;
	FTabCount := ATabCount;

	inherited Create(ADocument, ADecoder,
		ALineCount * FFontSize.cx, ATabCount * FFontSize.cx,
		GetFontSize(FFixedFont.Handle, 'x').cx,
		FFontSize.cy + ALineMargin,
		AHanging, AExpelling, AIndenting, AWordWrap);
end;

function TEditorLayout.GetImeFont: TAnyFont;
begin
	Result := FFont
end;

function TEditorLayout.LetterWidth(Letter: WideChar): Integer;
var
	S: TSize;
begin
	if FWidthCache[Letter] <> 0 then
		Result := FWidthCache[Letter]
	else
	begin
		if FDC = 0 then ReadyDC;
		GetTextExtentPoint32W(FDC, @Letter, 1, S);
		FWidthCache[Letter] := S.cx;
		Result := S.cx
	end
end;

procedure TEditorLayout.ReadyDC;
begin
	if FDC = 0 then
	begin
		FDC := GetDC(0);
		SaveDC(FDC);
		SelectObject(FDC, Font.Handle);
	end;
end;

procedure TEditorLayout.Release;
begin
	if FDC <> 0 then
	begin
		RestoreDC(FDC, -1);
		ReleaseDC(0, FDC);
		FDC := 0;
	end;
end;

function TEditorLayout.StandardLetterWidth: Integer;
begin
	Result := FFontSize.cx
end;

{ TEditorPane }

procedure TEditorPane.AdjustForBottom;
begin
	if FScrollBase = sbTop then
		ScrollDown
end;

procedure TEditorPane.ApplyLayout;
begin
end;

procedure TEditorPane.AutoScroll(Dir: TScrollDirection; Repeating: Boolean);
const
	IDs: array[TScrollDirection] of DWORD = (AutoScrollUpTimerID, AutoScrollDownTimerID);
var
	ID: DWORD;
begin
	if FAutoScrolling = 0 then
	begin
		ID := IDs[Dir];
		if Repeating then
		begin
			SetTimer(Handle, ID, TimerElapse, nil);
			FAutoScrolling := ID;
		end
		else
		begin
			SendMessageW(WM_TIMER, ID, 0)
		end;
	end
end;

procedure TEditorPane.Backspace;
begin
	{L[ɒ铮Ăł͂ȂȂ}
	Layout.Document.CheckReadOnly;
	if Selected then
		DeleteSelected
	else
	begin
		FCaretPosition := View.DeleteSingleLetter(FCaretPosition, HexMode, -1);
		Scroll;
	end;
	RecordCommand(ercBackspace);
end;

procedure TEditorPane.BeginDrag;
begin
end;

function TEditorPane.CanPaste: Boolean;
begin
	Result := IsClipboardFormatAvailable(CF_UNICODETEXT) or
		IsClipboardFormatAvailable(RegisterClipboardFormatW(PowerWitchFormat))
end;

function TEditorPane.CanRedo: Boolean;
begin
	Result := View.Layout.Document.CanRedo
end;

function TEditorPane.CanUndo: Boolean;
begin
	Result := View.Layout.Document.CanUndo
end;

procedure TEditorPane.Changed(First, OldLast, NewLast, RepaintingFirst, RepaintingLast: Natural);
var
	R: TRect;
begin
	FCaretPreferableX := -1;

	if FCaretPosition > First then
	begin
		if FCaretPosition <= OldLast then
			FCaretPosition := First
		else
			Inc(FCaretPosition, NewLast - OldLast);

		SetCaretPositionInternal;
	end;

	if FAnchorPosition > First then
	begin
		if FAnchorPosition <= OldLast then
			FAnchorPosition := First
		else
			Inc(FAnchorPosition, NewLast - OldLast);
	end;

	if View is TTextView then
		TTextView(View).AdjustScrollOnChange(First, OldLast, NewLast, FScrollBase, FScrollFrom);

	FUpdateScrollBar := True;
	FAdjustStartLine := True;

	(Parent as TEditor).ReleaseMarker;

	if FScrollBase = sbBottom then
	begin
		R := ViewRect;
		R.Right := R.Left;
		R.Left := 0;
		Repaint(R);
	end;	

	RepaintArea(RepaintingFirst, RepaintingLast);
end;

procedure TEditorPane.ClearSelection;
begin
	if FSelected then
	begin
		FSelected := False;
		RepaintArea(FAnchorPosition, FCaretPosition);
		RecordCommand(ercClearSelection);
	end
end;

function TEditorPane.ClientToHexMode(const P: TPoint): Boolean;
var
	HexRect: TRect;
begin
	Result := View is TBinaryView;
	if Result then
	begin
		HexRect := ViewRect;
		HexRect.Right := HexRect.Left + HexWidth;
		Result := PtInRect(HexRect, P);
	end;
end;

function TEditorPane.ClientToPosition(const P: TPoint): Natural;
var
	LineHeight, X, Y: Integer;
	Line, Next: TLine;
	ViewRect: TRect;
begin
	ViewRect := Self.ViewRect;
	LineHeight := Layout.LineHeight;

	if FScrollBase = sbTop then
	begin
		Line := FScrollFrom;
		if P.Y >= ViewRect.Top then
		begin
			Y := ViewRect.Top;
			repeat
				Inc(Y, LineHeight);
				if Y > P.Y then	Break;
				Next := View.SuccLine(Line);
				if Next < 0 then Break;
				Line := Next;
			until False;
		end;
	end
	else
	begin
		Line := FScrollFrom;
		if P.Y <= ViewRect.Bottom then
		begin
			Y := ViewRect.Bottom;
			repeat
				Dec(Y, LineHeight);
				if Y <= P.Y then Break;
				Next := View.PredLine(Line);
				if Next < 0 then Break;
				Line := Next;
			until False;
		end;
	end;

	Assert(Line >= 0);
	{if Line < 0 then Result := Layout.Document.Size}

	if P.Y < ViewRect.Top then
		Result := Line
	else if P.Y >= ViewRect.Bottom then
	begin
		Next := View.SuccLine(Line);
		if Next < 0 then Next := Layout.Document.Size;
		Result := Next;
	end
	else if View is TBinaryView then
	begin
		if HexMode then
		begin
			X := (P.X - ViewRect.Left) div (Layout.FixedLetterWidth * 3);
			if X < 0 then
				X := 0
			else if X >= 16 then
				X := 15;
			Result := Line + X;
			if Result > Layout.Document.Size then
				Result := Layout.Document.Size;
		end
		else
			Result := View.ToPosition(Line, P.X - (ViewRect.Left + HexWidth));
	end
	else
		Result := View.ToPosition(Line, P.X - ViewRect.Left + FLeft);
end;

procedure TEditorPane.CopyToClipboard;
var
	Data: string;
	Text: WideString;
	Mem: THandle;
	P: Pointer;
	I: Integer;
begin
	if Selected then
	begin
		if OpenClipboard(Handle) then
		begin
			EmptyClipboard;

			{PowerWitch}
			Data := GetSelectedData;
			Mem := GlobalAlloc(GMEM_MOVEABLE, Length(Data) + 4);
			P := GlobalLock(Mem);
			PDWORD(P)^ := Length(Data);
			Move(Pointer(Data)^, Pointer(Cardinal(P) + 4)^, Length(Data));
			GlobalUnlock(Mem);
			SetClipboardData(RegisterClipboardFormatW(PowerWitchFormat), Mem);

			{Unicode}
			Text := GetSelectedText;
			for I := Length(Text) downto 1 do
			begin
				if Text[I] = UnicodeLineFeed then
				begin
					System.Delete(Text, I, 1);
					System.Insert(#13#10, Text, I);
				end;
			end;

			Mem := GlobalAlloc(GMEM_MOVEABLE, (Length(Text) + 1) * SizeOf(WideChar));
			P := GlobalLock(Mem);
			Move(PWideChar(Text)^, P^, (Length(Text) + 1) * SizeOf(WideChar));
			GlobalUnlock(Mem);
			SetClipboardData(CF_UNICODETEXT, Mem);

			CloseClipboard;
		end;
	end;
end;

constructor TEditorPane.Create(AParent: TAnyWindow; AHorz,
	AVert: TScrollBar; ALayout: TLayout; AViewClass: TViewClass);
begin
  FLayout := ALayout as TEditorLayout;
	FView := AViewClass.Create(ALayout, Self);
	FHexInputing := $ff;
  FInsertMode := True;
	inherited Create(AParent, AHorz, AVert);
end;

procedure TEditorPane.CursorDown;
var
	Old: Natural;
begin
	Old := FCaretPosition;
	FCaretPosition := View.DownCaretPosition(HexMode, FCaretPosition, FCaretPreferableX);
	if not FSemiFreeMoving then FCaretPreferableX := -1;
	SetCaretPositionInternal;
	Scroll;
	if FSelected then RepaintArea(Old, FCaretPosition);
	RecordCommand(ercCursorDown);
end;

procedure TEditorPane.CursorFirst;
var
	Old: Natural;
begin
	Old := FCaretPosition;
	FCaretPosition := 0;
	SetCaretPositionInternal;
	Scroll;
	if FSelected then RepaintArea(Old, FCaretPosition);
end;

procedure TEditorPane.CursorLast;
var
	Old: Natural;
begin
	Old := FCaretPosition;
	FCaretPosition := Layout.Document.Size;
	SetCaretPositionInternal;
	Scroll;
	if FSelected then RepaintArea(Old, FCaretPosition);
end;

procedure TEditorPane.CursorLeft;
var
	Old: Natural;
begin
	FCaretPreferableX := -1;
	Old := FCaretPosition;
	FCaretPosition := View.LeftCaretPosition(FHexMode, FCaretPosition);
	SetCaretPositionInternal;
	Scroll;
	if FSelected then RepaintArea(Old, FCaretPosition);
	RecordCommand(ercCursorLeft);
end;

procedure TEditorPane.CursorLineEnd;
var
	Old: Natural;
begin
	FCaretPreferableX := -1;
	Old := FCaretPosition;
	FCaretPosition := View.LineEndCaretPosition(FCaretPosition);
	SetCaretPositionInternal;
	Scroll;
	if FSelected then RepaintArea(Old, FCaretPosition);
	RecordCommand(ercCursorLineEnd);
end;

procedure TEditorPane.CursorLineTop(Smart: Boolean = True);
var
	Old: Natural;
begin
	FCaretPreferableX := -1;
	Old := FCaretPosition;
	FCaretPosition := View.LineTopCaretPosition(FCaretPosition, Smart);
	SetCaretPositionInternal;
	Scroll;
	if FSelected then RepaintArea(Old, FCaretPosition);
	RecordCommand(ercCursorLineTop);
end;

procedure TEditorPane.CursorRight;
var
	Old: Natural;
begin
	FCaretPreferableX := -1;
	Old := FCaretPosition;
	FCaretPosition := View.RightCaretPosition(FHexMode, FCaretPosition);
	SetCaretPositionInternal;
	Scroll;
	if FSelected then RepaintArea(Old, FCaretPosition);
	RecordCommand(ercCursorRight);
end;

procedure TEditorPane.CursorScrollDown;
var
	Y : Integer;
begin
	Y := View.BasedScrollDown(FScrollBase, FScrollFrom);
	RepaintScroll(0, Y);
	CursorDown;
end;

procedure TEditorPane.CursorScrollUp;
var
	Y : Integer;
begin
	Y := View.BasedScrollUp(FScrollBase, FScrollFrom);
	RepaintScroll(0, Y);
	CursorUp;
end;

procedure TEditorPane.CursorUp;
var
	Old: Natural;
begin
	Old := FCaretPosition;
	FCaretPosition := View.UpCaretPosition(HexMode, FCaretPosition, FCaretPreferableX);
	if not FSemiFreeMoving then FCaretPreferableX := -1;
	SetCaretPositionInternal;
	Scroll;
	if FSelected then RepaintArea(Old, FCaretPosition);
  RecordCommand(ercCursorUp);
end;

procedure TEditorPane.CutToClipboard;
begin
  Layout.Document.CheckReadOnly;
	CopyToClipboard;
	DeleteSelected;
end;

procedure TEditorPane.Delete;
begin
	{L[ɒ铮Ăł͂ȂȂ}
	Layout.Document.CheckReadOnly;
	if Selected then
		DeleteSelected
	else
	begin
		FCaretPosition := View.DeleteSingleLetter(FCaretPosition, HexMode, +1);
		Scroll;
	end;
	RecordCommand(ercDelete, #0, #0);
end;

procedure TEditorPane.DeleteSelected;
begin
	if Selected then
	begin
		FCaretPosition := View.InsertText(FCaretPosition, FAnchorPosition - FCaretPosition, '', HexMode);
		FSelected := False;
		Scroll;
	end;
end;

procedure TEditorPane.FeedForm;
begin
	InputText(UnicodeFormFeed)
end;

procedure TEditorPane.FeedLine;
var
	I: WideString;
begin
	I := '';
	if View is TTextView then I := (View as TTextView).AutoIndent(CaretPosition);
	InputText(LineFeedCode[FLineFeed] + I);
  RecordCommand(ercInputLineFeed);
end;

procedure TEditorPane.Finalize;
begin
	ReleaseView;
	inherited;
end;

function TEditorPane.GetSelected: Boolean;
begin
	Result := FSelected and (FAnchorPosition <> FCaretPosition)
end;

function TEditorPane.GetSelectedData: string;
begin
	Result := '';
	if Selected then
		Result := View.ReadData(SelectionFirstPosition, SelectionLastPosition);
end;

function TEditorPane.GetSelectedLength: Natural;
begin
	Result := 0;
	if Selected then
		Result := Abs(FAnchorPosition - FCaretPosition)
end;

function TEditorPane.GetSelectedText: WideString;
begin
	Result := '';
	if Selected then
		Result := View.ReadText(SelectionFirstPosition, SelectionLastPosition);
end;

function TEditorPane.GetSelectionFirstPosition: Natural;
begin
	Result := FCaretPosition;
	if FSelected and (FAnchorPosition < Result) then
		Result := FAnchorPosition;
end;

function TEditorPane.GetSelectionLastPosition: Natural;
begin
	Result := FCaretPosition;
	if FSelected and (FAnchorPosition > Result) then
		Result := FAnchorPosition;
end;

function TEditorPane.HexWidth: Integer;
begin
	Result := Layout.FixedLetterWidth * (3 * 16)
end;

procedure TEditorPane.IndentSelection;
var
	F, L: Natural;
begin
	Assert(View is TTextView);
	F := SelectionFirstPosition;
	L := SelectionLastPosition;
  FAnchorPosition := F;
	FCaretPosition := (View as TTextView).Indent(F, L);
end;

procedure TEditorPane.Initialize;
begin
	inherited;
	FCaretLine := 1;
  FCaretColumn := 1;
	FUpdateScrollBar := True;
	ApplyLayout;
end;

procedure TEditorPane.InputData(const Data: string);
begin
	FCaretPosition := View.InsertData(FCaretPosition, ReplaceLength(Length(Data)), Data, HexMode);
	ClearSelection;
	SetCaretPositionInternal;
	Scroll;
end;

procedure TEditorPane.InputText(const Text: WideString);
begin
	InputData(View.Layout.Decoder.EncodeString(Text))
end;

function TEditorPane.KeyDown(ShiftState: TShiftKeys; Code: Byte): Boolean;
var
	Dir: TScrollDirection;
begin
	Result := True;
	try
		case Code of
		VK_LEFT:
		begin
			if ssShift in ShiftState then StartSelection else ClearSelection;
			if ssControl in ShiftState then WordLeft else CursorLeft;
		end;
		VK_RIGHT:
		begin
			if ssShift in ShiftState then StartSelection else ClearSelection;
			if ssControl in ShiftState then WordRight else CursorRight;
		end;
		VK_UP:
		begin
			if ssShift in ShiftState then StartSelection else ClearSelection;
			if ssControl in ShiftState then ScrollUp else CursorUp;
		end;
		VK_DOWN:
		begin
			if ssShift in ShiftState then StartSelection else ClearSelection;
			if ssControl in ShiftState then ScrollDown else CursorDown;
		end;
		VK_HOME:
		begin
			if ssShift in ShiftState then StartSelection else ClearSelection;
			if ssControl in ShiftState then CursorFirst else CursorLineTop;
		end;
		VK_END:
		begin
			if ssShift in ShiftState then StartSelection else ClearSelection;
			if ssControl in ShiftState then CursorLast else CursorLineEnd;
		end;
		VK_PRIOR, VK_NEXT:
		begin
			if ssShift in ShiftState then StartSelection else ClearSelection;
			Dir := sdUp;
			if Code = VK_NEXT then Dir := sdDown;
			AutoScroll(Dir, not (ssControl in ShiftState));
		end;
		VK_TAB:
		begin
			if ssControl in ShiftState then
				Result := False
			else if View is TBinaryView then
			begin
				if (Parent as TEditor).FLostFocus and ((ssShift in ShiftState) = HexMode) then
					Result := False					
				else
					HexMode := not HexMode
			end
			else
			begin
				InputText(#9);
				RecordCommand(ercInputTab);
			end
		end;
		VK_RETURN:
			InputText(LineFeedCode[FLineFeed]);
		VK_BACK:
			Backspace;
		VK_DELETE:
			if ssControl in ShiftState then
				DeleteSelected
			else if ssShift in ShiftState then
				CutToClipboard
			else
				Delete;
		VK_INSERT:
			if ssControl in ShiftState then
				CopyToClipboard
			else if ssShift in ShiftState then
				PasteFromClipboard(False)
			else
				InsertMode := not InsertMode;
		Ord('X'):
			if ssControl in ShiftState then CutToClipboard;
		Ord('C'):
			if ssControl in ShiftState then CopyToClipboard;
		Ord('V'):
			if ssControl in ShiftState then PasteFromClipboard(False);
		else
			Result := inherited KeyDown(ShiftState, Code)
		end;
	except
		on EBufferReadOnly do Beep
	end
end;

function TEditorPane.KeyInput(Code: WideChar): Boolean;

	procedure HexInput(Code: Byte);
	var
		Inputing: Char;
	begin
		if FHexInputing and $f = 0 then
		begin
			Inputing := Chr(FHexInputing + Code);
			View.InsertData(FCaretPosition, 1, Inputing, True);
			RecordCommand(ercInputData, #0, Inputing);
			Inc(FCaretPosition);
			SetCaretPositionInternal;
		end
		else
		begin
			FHexInputing := Code * 16;
			View.InsertData(FCaretPosition, ReplaceLength(1), Chr(FHexInputing), True);
			RepaintCaret;
		end;
	end;

begin
	if Code = #$1b then
		Result := inherited KeyInput(Code)
	else if (Code >= #32) and (Code <> #$7f) then
	begin
		if View.Layout.Document.ReadOnly then
			Beep
		else
		begin
			if HexMode then
			begin
				case Code of
				'0'..'9': HexInput(Ord(Code) - Ord('0'));
				'a'..'f': HexInput(Ord(Code) - (Ord('a') - 10));
				'A'..'F': HexInput(Ord(Code) - (Ord('A') - 10));
				else Beep
				end;
			end
			else
			begin
				InputText(Code);
				RecordCommand(ercInputText, Code, #0);
			end;
		end;
		Result := True;
	end
	else
	begin
		Result := True;
	end
end;

function TEditorPane.KeyUp(ShiftState: TShiftKeys; Code: Byte): Boolean;
begin
	StopAutoScroll;
	Result := inherited KeyUp(ShiftState, Code);
end;

procedure TEditorPane.PaintWindow(DC: HDC);
var
	ViewRect: TRect;
	LineHeight: Integer;
	Layout: TEditorLayout;
	IsBinary: Boolean;
	CaretPos: TSmallPoint;
	CaretWidth: Integer;
	CaretPreferableX: Integer;
	Brush: TBrush;
	Pen: TPen;
	EmPen: TPen;
	SelStart, SelEnd: Natural;
	StandardLetterWidth: Integer;
	ViewRegion: HRGN;
	Marker: TBufferMarker;

	procedure ReadyCaret(X, Y: Integer; Width: Integer);
	var
		XForSet, YForSet: Integer;
	begin
		HideCaret(Handle);
		CreateCaret(Handle, 0, Width, LineHeight - Layout.LineMargin);
		XForSet := X;
		if X < ViewRect.Left then XForSet := -100;
		YForSet := Y + Layout.LineMargin;
		SetCaretPos(XForSet, YForSet);
		ShowCaret(Handle);
		SetImeComposition(Handle, Layout.ImeFont.Handle, X, YForSet);
	end;

	procedure PaintRuler;
	var
		R: TRect;
		X, TX, I, Y, N: Integer;
	begin
		Brush.Color := UnderLineColor;
		R.Left := CaretPos.X;
		R.Right := R.Left + CaretWidth;
		R.Top := 0;
		R.Bottom := ViewRect.Top - 1;
		FillRect(DC, R, Brush.Handle);

		SelectObject(DC, Pen.Handle);
		MoveToEx(DC, 0, ViewRect.Top - 1, nil);
		LineTo(DC, ViewRect.Right, ViewRect.Top - 1);
		if IsBinary then
		begin
			X := ViewRect.Left;
			TX := X + HexWidth;
			for I := 0 to 15 do begin
				if I > 0 then begin
					Y := 5;
					if I mod 8 = 0 then Y := 1
					else if I mod 4 = 0 then Y := 3;
					MoveToEx(DC, X, Y, nil);
					LineTo(DC, X, ViewRect.Top);
					MoveToEx(DC, TX, Y, nil);
					LineTo(DC, TX, ViewRect.Top);
				end;
				if esoRulerColmn in ShowOptions then begin
					PaintDigit(DC, X + StandardLetterWidth * 3 div 2, ViewRect.Top - 7, Layout.NumberColor, I);
					PaintDigit(DC, TX + StandardLetterWidth div 2, ViewRect.Top - 7, Layout.NumberColor, I);
				end;	
				Inc(X, StandardLetterWidth * 3);
				Inc(TX, StandardLetterWidth);
			end;
		end
		else
		begin
			X := ViewRect.Left - FLeft;
			N := Layout.LineCount;
			if N = 0 then N := High(Integer);
			for I := 1 to N do
			begin
				Inc(X, StandardLetterWidth);
				Y := 5;
				if I mod 10 = 0 then Y := 1
				else if I mod 5 = 0 then Y := 3;
				if X >= ViewRect.Right then Break;
				if X > ViewRect.Left then
				begin
					MoveToEx(DC, X, Y, nil);
					LineTo(DC, X, ViewRect.Top);
				end;
			end;
		end;
	end;

	procedure PaintIndent(X, Y: Integer);
	var
		MX, TY, BY: Integer;
	begin
		MX := X - 4;
		TY := Y + 3;
		BY := Y + LineHeight - 3;
		MoveToEx(DC, MX, BY, nil);
		LineTo(DC, MX, TY);
		LineTo(DC, MX - 3, TY + 3);
		MoveToEx(DC, MX, TY, nil);
		LineTo(DC, MX + 3, TY + 3);
	end;

	procedure PaintEOF(X, Y: Integer);
	var
		IX, IY: Integer;
	begin
		for IX := X + StandardLetterWidth - 1 downto X do
			for IY := Y + LineHeight - 1 downto Y do
				if (IX + IY) mod 2 = 0 then
					SetPixel(DC, IX, IY, Layout.SymbolColor);
	end;

	procedure PaintTAB(X, Y: Integer);
	var
		MY: Integer;
	begin
		MY := Y + LineHeight div 2;
		SetPixel(DC, X, MY, Layout.LineColor);
		SetPixel(DC, X + 2, MY, Layout.LineColor);
		MoveToEx(DC, X + 4, MY, nil);
		LineTo(DC, X + 7, MY);
		LineTo(DC, X + 4, MY - 3);
		MoveToEx(DC, X + 7, MY, nil);
		LineTo(DC, X + 4, MY + 3);
	end;

	procedure PaintNewLine(X, Y: Integer);
	var
		MX, MY: Integer;
		Org: HPEN;
	begin
		MX := X + StandardLetterWidth - 2;
		MY := Y + LineHeight - 5;
		Org := SelectObject(DC, EmPen.Handle);
		MoveToEx(DC, MX, Y + 3, nil);
		LineTo(DC, MX, MY - 1);
		LineTo(DC, MX - 1, MY);
		LineTo(DC, X + 1, MY);
		LineTo(DC, X + 4, MY - 3);
		MoveToEx(DC, X + 1, MY, nil);
		LineTo(DC, X + 4, MY + 3);
		SelectObject(DC, Org);
	end;

	procedure PaintNewLineCR(X, Y: Integer);
	var
		MY: Integer;
		Org: HPEN;
	begin
		MY := Y + LineHeight div 2;
		Org := SelectObject(DC, EmPen.Handle);
		MoveToEx(DC, X + StandardLetterWidth - 1, MY, nil);
		LineTo(DC, X + 1, MY);
		LineTo(DC, X + 4, MY - 3);
		MoveToEx(DC, X + 1, MY, nil);
		LineTo(DC, X + 4, MY + 3);
		SelectObject(DC, Org);
	end;

	procedure PaintNewLineLF(X, Y: Integer);
	var
		MX, BY: Integer;
		Org: HPEN;
	begin
		MX := X + StandardLetterWidth div 2;
		BY := Y + LineHeight - 3;
		Org := SelectObject(DC, EmPen.Handle);
		MoveToEx(DC, MX, Y + 3, nil);
		LineTo(DC, MX, BY);
		LineTo(DC, MX - 3, BY - 3);
		MoveToEx(DC, MX, BY, nil);
		LineTo(DC, MX + 3, BY - 3);
		SelectObject(DC, Org);
	end;

	procedure PaintNewPage(X, Y: Integer);
	var
		IX, IY: Integer;
	begin
		for IX := StandardLetterWidth - 1 downto 0 do begin
			for IY := LineHeight - 1 downto (StandardLetterWidth - IX - 1) * LineHeight div StandardLetterWidth do begin
				if (IX + IY) mod 2 = 0 then begin
					SetPixel(DC, X + IX, Y + IY, Layout.SymbolColor)
				end
			end
		end
	end;

	procedure PaintPinpoint(X, Y: Integer);
	var
  	Color: COLORREF;
	begin
		Color := Layout.StateColors[0].Foreground;
		SetPixel(DC, X, Y + LineHeight - 2, Color);
		SetPixel(DC, X + 1, Y + LineHeight - 1, Color);
		SetPixel(DC, X - 1, Y + LineHeight - 1, Color);
		SetPixel(DC, X + 2, Y + LineHeight, Color);
		SetPixel(DC, X - 2, Y + LineHeight, Color);
	end;

	procedure PaintLineNumber(const R: TRect; Number: Integer);
	const
		Modules: array[Boolean] of Integer = (10, 16);
	var
		I, V, X, Y, Module: Integer;
	begin
		X := R.Right - 4; {}
		Y := R.Bottom - 6;
		Module := Modules[IsBinary];
		for I := 1 to 8 do begin
			V := Number mod Module;
			Number := Number div Module;
			PaintDigit(DC, X, Y, Layout.NumberColor, V);
			if (Number = 0) and not IsBinary then Break;
			Dec(X, 4);
		end;
	end;

	procedure PaintLine(Y: Integer;
		var NewLine: Boolean;
		var LineNumber: Integer;
		var Indent: Integer; const Accessor: TBufferAccessor);
	var
		LineFirstPosition, LineLastPosition, ShowingNumber: Natural;
		Last, I, X, LX, X2, Bytes, Width, Over, Under, LetterY, CenterStep: Integer;
		InSelection, EndLine, Indenting: Boolean;
		R, SR: TRect;
		Info, State, BinaryCode: Byte;
		Image: array[0..1] of WideChar;
		UnicodeLetter: WideChar;
		SelectedFont: (ofFirst, ofSecond, ofHalfWidth);
		HexAccessor, TextAccessor, PreAccessor: TBufferAccessor;
    Pinpoint: Boolean;
	begin
		LineFirstPosition := Accessor.Index;
		LineLastPosition := View.LineEnd(LineFirstPosition);

		X := ViewRect.Left - FLeft;
		LX := 0;

		if Layout.Document.ModifiedRange(LineFirstPosition, LineLastPosition) then
		begin
			Brush.Color := ModifiedColor;
			R.Left := 0;
			R.Right := 3;
			R.Top := Y;
			R.Bottom := Y + LineHeight;
			FillRect(DC, R, Brush.Handle)
		end;

		if (IsBinary or (esoLineNumber in ShowOptions) and NewLine)
			and (LineFirstPosition < Layout.Document.Size) then
		begin
			R.Left := 3;
			R.Right := ViewRect.Left - 1;
			R.Top := Y;
			R.Bottom := Y + LineHeight;
			if IsBinary then
				ShowingNumber := LineFirstPosition
			else
				ShowingNumber := LineNumber;

			PaintLineNumber(R, ShowingNumber);
		end;

		if IsBinary then
		begin
			HexAccessor := Accessor;
			while HexAccessor.Index < LineLastPosition do
			begin
				SetTextColor(DC, Layout.StateColors[0].Foreground);
				if FSelected then
				begin
					InSelection := (SelStart <= HexAccessor.Index) and (HexAccessor.Index < SelEnd);
					if InSelection then
					begin
						R.Left := X;
						R.Top := Y;
						R.Right := R.Left + StandardLetterWidth * 3;
						R.Bottom := Y + LineHeight;
						Brush.Color := Layout.SelectedColor.Background;
						FillRect(DC, R, Brush.Handle);

						SetTextColor(DC, Layout.SelectedColor.Foreground);
					end;
				end;

				if Marker <> nil then
				begin
					if Marker.MarkedRange(HexAccessor.Index, HexAccessor.Index + 1, Pinpoint) then
						if Pinpoint then
              PaintPinpoint(X, Y)
						else
						begin
							for I := X + StandardLetterWidth * 3 - 1 downto X do
								if I mod 2 = 0 then
									SetPixel(DC, I, Y + LineHeight - 1, Layout.StateColors[0].Foreground);
						end;
				end;

				if FHexMode then
				begin
					if HexAccessor.Index = FCaretPosition then
					begin
						CaretPos.X := X;
						if FHexInputing <> $ff then Inc(CaretPos.X, StandardLetterWidth * 3 div 2);
						CaretPos.Y := Y;
					end;
				end;

				BinaryCode := HexAccessor.Take;

				Image[0] := WideChar(BinaryCode div 16 + Ord('0'));
				if Image[0] > '9' then Inc(Image[0], Ord('A') - Ord('9') - 1);
				Image[1] := WideChar(BinaryCode mod 16 + Ord('0'));
				if Image[1] > '9' then Inc(Image[1], Ord('A') - Ord('9') - 1);

				TextOutW(DC,
					X + (StandardLetterWidth + StandardLetterWidth div 2),
					Y + Layout.LineMargin, Image, 2);

				Inc(X, StandardLetterWidth * 3);
			end;
			X := ViewRect.Left + HexWidth;
		end;

		if ViewRegion <> 0 then
		begin
			SaveDC(DC);
			ExtSelectClipRgn(DC, ViewRegion, RGN_AND);
		end;

		if not NewLine and (Indent > 0) then
		begin
			Inc(X, Indent);
			PaintIndent(X, Y);
		end;

		Indenting := NewLine;

		TextAccessor := Accessor;
		SelectedFont := ofFirst;

		if IsBinary then
		begin
			if not TextAccessor.EndOfBuffer then
			begin
				while Layout.Sub[TextAccessor.Index] and SubLetter = 0 do
				begin
					Dec(X, StandardLetterWidth);
					Dec(TextAccessor.Index);
				end;
			end
		end;

		EndLine := False;

		while TextAccessor.Index < LineLastPosition do
		begin
			Info := Layout.Sub[TextAccessor.Index];
			State := Info and SubStateMask;

			PreAccessor := TextAccessor;
			UnicodeLetter := Layout.Decoder.Decode(TextAccessor);
			Bytes := TextAccessor.Index - PreAccessor.Index;

			if IsBinary then
			begin
				Width := Bytes * StandardLetterWidth;
				CenterStep := Width div 2;
			end
			else
			begin
				CenterStep := 0;
				EndLine := False;
				if UnicodeLetter = #9 then
					Width := (View as TTextView).TabWidth(LX)
				else if HasWideChar(LineFeedCodes, UnicodeLetter) then
				begin
					EndLine := True;
					Width := StandardLetterWidth
				end
				else
					Width := Layout.LetterWidth(UnicodeLetter);

				if Indenting then
				begin
					if HasWideChar(Layout.Indenting, UnicodeLetter) then
					begin
						Indenting := Indent + Width <= Layout.MaxIndentWidth;
						if Indenting then Inc(Indent, Width);
					end
					else
						Indenting := False;
				end;
			end;

			if (X + Width >= ViewRect.Left) and (X < ViewRect.Right) then
			begin
				InSelection := (SelStart < PreAccessor.Index + Bytes) and (PreAccessor.Index < SelEnd);
				if InSelection or (State <> 0) then
				begin
					if IsBinary then
					begin
						if InSelection then
						begin
							Over := SelStart - PreAccessor.Index;
							if Over < 0 then Over := 0;
							SR.Left := X + StandardLetterWidth * Over;
							Under := PreAccessor.Index + Bytes - SelEnd;
							if Under < 0 then Under := 0;
							SR.Right := X + StandardLetterWidth * (Bytes - Under);
						end
						else
						begin
							SR.Left := X;
							SR.Right := X + StandardLetterWidth * Bytes;
						end
					end
					else
					begin
						SR.Left := X;
						SR.Right := X + Width;
					end;
					SR.Top := Y;
					SR.Bottom := Y + LineHeight;
					if InSelection then
					begin
						Brush.Color := Layout.SelectedColor.Background;
						FillRect(DC, SR, Brush.Handle)
					end
					else if not EndLine then
					begin
						Brush.Color := Layout.StateColors[State].Background;
						FillRect(DC, SR, Brush.Handle)
					end;
				end;

				if Marker <> nil then
				begin
					if Marker.MarkedRange(PreAccessor.Index, TextAccessor.Index, Pinpoint) then
					begin
						if Pinpoint then
              PaintPinpoint(X, Y)
						else
						begin
							for I := X + Width - 1 downto X do
								if I mod 2 = 0 then
									SetPixel(DC, I, Y + LineHeight - 1, Layout.StateColors[0].Foreground);
						end
					end;
				end;

				case UnicodeLetter of
				#9:
					PaintTAB(X, Y);
				UnicodeLineFeed:
					PaintNewLine(X, Y);
				#13:
					PaintNewLineCR(X, Y);
				#10:
					PaintNewLineLF(X, Y);
				UnicodeFormFeed:
					PaintNewPage(X, Y);
				#$20:
					{ȂɂȂ};
				UnicodeFullWidthSpace:
				begin
					SelectObject(DC, GetStockObject(NULL_BRUSH));
					R.Left := X - 1;
					R.Bottom := Y + LineHeight;
					R.Right := X + Width + 1;
					Dec(R.Right, 3);
					Inc(R.Left, 3);
					Rectangle(DC, R.Left, R.Bottom - 5, R.Right, R.Bottom - 2);
				end;
				UnicodeBOM:
				begin
					R.Left := X;
					R.Top := Y;
					R.Right := X + Width;
					R.Bottom := Y + LineHeight;

					LetterY := Y + Layout.LineMargin;

					if SelectedFont <> ofFirst then
					begin
						RestoreDC(DC, -1);
						SelectedFont := ofFirst;
					end;

					SetTextColor(DC, EmPen.Color);
					ExtTextOutW(DC, X + CenterStep, LetterY, ETO_CLIPPED, nil, @BOMMark, 1, nil);
					if InSelection then
					begin
						ExtTextOutW(DC, X + CenterStep, LetterY, ETO_CLIPPED, @SR, @BOMMark, 1, nil);
					end;
				end;
				#0..#8, #11, #14..#$1f:
				begin
					Brush.Color := $ff;
					R.Left := X;
					R.Top := Y + 2;
					R.Right := X + Width;
					R.Bottom := Y + LineHeight - 2;
					Dec(R.Right, 2);
					Inc(R.Left, 2);
					FillRect(DC, R, Brush.Handle);
				end
				else
				begin
					R.Left := X;
					R.Top := Y;
					R.Right := X + Width;
					R.Bottom := Y + LineHeight;

					LetterY := Y + Layout.LineMargin;
					if (UnicodeLetter <= #$ff) and (Layout.HalfWidthFont <> nil) then
					begin
						if SelectedFont <> ofHalfWidth then
						begin
							if SelectedFont = ofFirst then SaveDC(DC);
							SelectObject(DC, Layout.HalfWidthFont.Handle);
							SelectedFont := ofHalfWidth;
						end;
						Inc(LetterY, Layout.HalfWidthFontShift);
					end
					else if (Layout.SecondFont <> nil) and
						HasWideChar(Layout.SecondFontLetters, UnicodeLetter) then
					begin
						if SelectedFont <> ofSecond then
						begin
							if SelectedFont = ofFirst then SaveDC(DC);
							SelectObject(DC, Layout.SecondFont.Handle);
							SelectedFont := ofSecond;
						end;
					end
					else
					begin
						if SelectedFont <> ofFirst then
						begin
							RestoreDC(DC, -1);
							SelectedFont := ofFirst;
						end;
					end;
					SetTextColor(DC, Layout.StateColors[State].Foreground);
					ExtTextOutW(DC, X + CenterStep, LetterY, ETO_CLIPPED, nil, @UnicodeLetter, 1, nil);
					if InSelection then
					begin
						SetTextColor(DC, Layout.SelectedColor.Foreground);
						ExtTextOutW(DC, X + CenterStep, LetterY, ETO_CLIPPED, @SR, @UnicodeLetter, 1, nil);
					end;
				end
				end;
			end;

			{XWi߂}
			if not FHexMode then begin
				if (PreAccessor.Index = FCaretPosition) and
					(PreAccessor.Index >= Accessor.Index) then
				begin
					CaretPos.X := X;
					CaretPos.Y := Y;
					CaretWidth := Width;
					X2 := X;
					if not EndLine then Inc(X2, Width);
					if not IsBinary and (TextAccessor.Index = LineLastPosition) and (CaretPreferableX >= X2) then begin
						CaretPos.X := CaretPreferableX;
					end;
				end;
			end;
			Inc(LX, Width);
			Inc(X, Width);
		end;

		if SelectedFont <> ofFirst then begin
			RestoreDC(DC, -1);
		end;

		if (LineLastPosition = Layout.Document.Size) and
			((IsBinary and (Layout.Document.Size mod 16 <> 0)) or
			(LineFirstPosition = LineLastPosition) or
			(Layout.Sub[LineLastPosition - 1] and SubLineEnd = 0)) then
		begin
			if LineLastPosition = FCaretPosition then begin
				CaretWidth := StandardLetterWidth;
				CaretPos.X := X;
				CaretPos.Y := Y;
				if not IsBinary and (CaretPreferableX > X) then begin
					CaretPos.X := CaretPreferableX;
				end;
				if HexMode then begin
					CaretPos.x := ViewRect.Left + (LineLastPosition - LineFirstPosition) * StandardLetterWidth * 3;
				end;
			end;
			PaintEOF(X, Y);
		end else if EndLine then begin
			Indent := 0;
			Inc(LineNumber)
		end else if not IsBinary then begin
			MoveToEx(DC, X, Y, nil);
			LineTo(DC, X, Y + LineHeight);
		end;

		if ViewRegion <> 0 then begin
			RestoreDC(DC, -1);
		end;

		NewLine := EndLine;
	end;

var
	X1, X2, Y, Indent: Integer;
	Next, CaretLine: TLine;
	LineRegion: HRGN;
	LineAccessor: TBufferAccessor;
	LineNumber: Integer;
	NewLine: Boolean;
begin
	ViewRect := Self.ViewRect;
	Layout := Self.Layout;
	LineHeight := Layout.LineHeight;
	IsBinary := View is TBinaryView;
	Marker := (Parent as TEditor).FMarker;

	SelStart := 0;
	SelEnd := 0;
	if Selected then
	begin
		if FAnchorPosition < FCaretPosition then
		begin
			SelStart := FAnchorPosition;
			SelEnd := FCaretPosition;
		end
		else
		begin
			SelStart := FCaretPosition;
			SelEnd := FAnchorPosition;
		end
	end;

	Brush := TBrush.Create;
	Brush.Owner := Garbage;

	Pen := TPen.Create;
	Pen.Owner := Garbage;
	Pen.Color := Layout.LineColor;

	EmPen := TPen.Create;
	EmPen.Owner := Garbage;
	EmPen.Color := Layout.SymbolColor;

	SelectObject(DC, Pen.Handle);
	MoveToEx(DC, ViewRect.Left - 1, 0, nil);
	LineTo(DC, ViewRect.Left - 1, ViewRect.Bottom);

	CaretPreferableX := FCaretPreferableX + ViewRect.Left - FLeft;

	if IsBinary then
	begin
		SelectObject(DC, Layout.FixedFont.Handle);
		StandardLetterWidth := Layout.FixedLetterWidth;
		X1 := ViewRect.Left + HexWidth;
		MoveToEx(DC, X1 - 1, 0, nil);
		LineTo(DC, X1 - 1, ViewRect.Bottom);
		X2 := X1 + StandardLetterWidth * 16;
		MoveToEx(DC, X2, 0, nil);
		LineTo(DC, X2, ViewRect.Bottom);
		SetTextAlign(DC, TA_TOP or TA_CENTER);
		ViewRegion := CreateRectRgn(X1, ViewRect.Top, X2, ViewRect.Bottom);
	end
	else
	begin
		StandardLetterWidth := Layout.StandardLetterWidth;
		SelectObject(DC, Layout.Font.Handle);
		X1 := ViewRect.Left;
		if Layout.LineWidth > 0 then
		begin
			X2 := ViewRect.Left + Layout.LineWidth - FLeft;
			MoveToEx(DC, X2, 0, nil);
			LineTo(DC, X2, ViewRect.Bottom);
			SetTextAlign(DC, TA_TOP or TA_LEFT);
		end;
		ViewRegion := CreateRectRgn(ViewRect.Left, ViewRect.Top, ViewRect.Right, ViewRect.Bottom);
	end;
	LineRegion := 0;
	if esoRuler in ShowOptions then
		LineRegion := CreateRectRgn(0, ViewRect.Top, ViewRect.Right, ViewRect.Bottom);

	SetBkMode(DC, TRANSPARENT);

	CaretPos.y := High(CaretPos.y);
	CaretWidth := 0;

	if LineRegion <> 0 then
	begin
		SaveDC(DC);
		ExtSelectClipRgn(DC, LineRegion, RGN_AND);
	end;

	if Layout.Document.Size > 0 then
	begin
		View.ReadyPaintLine(FScrollBase, FScrollFrom, Y, NewLine, LineNumber, Indent, LineAccessor);
		while Y < ViewRect.Bottom do
		begin
			PaintLine(Y, NewLine, LineNumber, Indent, LineAccessor);
			Next := View.SuccLine(LineAccessor.Index);
			if Next < 0 then Break;
			LineAccessor.Index := Next;
			Inc(Y, LineHeight);
		end
	end
	else
	begin
		Y := ViewRect.Top;
		PaintEOF(X1, Y);
		CaretPos.x := X1;
		if IsBinary and HexMode then CaretPos.x := ViewRect.Left;
		CaretPos.y := Y;
		CaretWidth := StandardLetterWidth;
	end;

	if CaretPos.y = High(CaretPos.y) then
	begin
		CaretLine := View.ToLine(CaretPosition, +1);
		CaretPos.X := View.ToX(CaretLine, CaretPosition);
		CaretPos.Y := -999;
		CaretWidth := 0;
		if IsBinary then
		begin
			if HexMode then
				CaretPos.X := CaretPos.X * 3
			else
				Inc(CaretPos.X, HexWidth);
		end;
		Inc(CaretPos.X, ViewRect.Left);
	end;

	DeleteObject(ViewRegion);

	if LineRegion <> 0 then
	begin
		RestoreDC(DC, -1);
		DeleteObject(LineRegion);
	end;

	if FInsertMode then
		CaretWidth := 2
	else if FHexInputing and $f = 0 then
		CaretWidth := StandardLetterWidth * 3 div 2
	else if FHexMode then
		CaretWidth := 3 * StandardLetterWidth;

	if esoRuler in ShowOptions then
		PaintRuler;

  FUnderLineY := -1;
	if Focused then
	begin
		ReadyCaret(CaretPos.X, CaretPos.Y, CaretWidth);
		FUnderLineY := CaretPos.Y + LineHeight;
		if (Parent as TEditor).UnderLine and (FUnderLineY > ViewRect.Top) then
		begin
			if not (esoDoubleBuffering in ShowOptions) then
				DC := GetDC(Handle);

			//Pen := TemporaryPen(RGBColor(UnderLineColor));
			Pen.Color := UnderLineColor;
			SelectObject(DC, Pen.Handle);
			MoveToEx(DC, 0, FUnderLineY, nil);
			LineTo(DC, 9999, FUnderLineY);

			if not (esoDoubleBuffering in ShowOptions) then
				ReleaseDC(Handle, DC);
		end;
	end;
end;

procedure TEditorPane.PasteFromClipboard(AsText: Boolean);
var
	Mem: THandle;
	P: Pointer;
	Data: string;
	AutoIndent, Text: WideString;
	I: Integer;
begin
	Layout.Document.CheckReadOnly;
	if OpenClipboard(Handle) then
	begin
		Mem := 0;
		if not AsText then
			Mem := GetClipboardData(RegisterClipboardFormatW(PowerWitchFormat));
		if Mem <> 0 then
		begin
			P := GlobalLock(Mem);
			SetString(Data, PChar(Cardinal(P) + 4), Cardinal(P^));
			InputData(Data);
			GlobalUnlock(Mem);
		end
		else
		begin
			Mem := GetClipboardData(CF_UNICODETEXT);
			Text := PWideChar(GlobalLock(Mem));
			//sR[h킹
			I := 1;
			while I <= Length(Text) do
			begin
				if (Text[I] = #13) or (Text[I] = #10) then
				begin
					if (I < Length(Text)) and (Text[I] = #13) and (Text[I + 1] = #10) then
						System.Delete(Text, I, 1);

					Text[I] := LineFeedCode[FLineFeed];
				end;
				Inc(I);
			end;
			//Cfg
			if AsText then
			begin
				AutoIndent := '';
				if View is TTextView then
					AutoIndent := (View as TTextView).AutoIndent(CaretPosition);

				Text := AdjustIndent(Text, AutoIndent, Layout.Indenting)
			end;
			InputText(Text);
			GlobalUnlock(Mem);
		end;
		CloseClipboard;
	end;
end;

procedure TEditorPane.RecordCommand(Command: TEditorRecordingCommand; Text: WideChar; Data: AnsiChar);
begin
end;

procedure TEditorPane.Redo;
begin
	if CanRedo then
	begin
		ClearSelection;
		if Layout.Document.RedoPosition <> CaretPosition then
			CaretPosition := Layout.Document.RedoPosition
		else
			CaretPosition := Layout.Document.Redo;

		Scroll;
	end;
end;

procedure TEditorPane.ReleaseView;
var
	Layout: TLayout;
begin
	Layout := FLayout;
  FLayout := nil;
	FView.Free;
	FView := nil;
	if (Layout <> nil) and (Layout.ViewCount = 0) then
		Layout.Free;
end;

procedure TEditorPane.RepaintArea(First, Last: Natural);
var
	Work: Natural;
	Y1, Y2: Integer;
	R: TRect;
begin
	if Last < First then
	begin
		Work := Last;
		Last := First;
		First := Work;
	end;
	View.AdjustRepaint(FScrollBase, FScrollFrom, First, Last, Y1, Y2);
	R.Left := 0;
	R.Top := Y1;
	R.Right := 9999;
	R.Bottom := Y2;
	Repaint(R);
end;
	
procedure TEditorPane.RepaintCaret;
var
	R: TRect;
	M: MSG;
begin
	if esoDoubleBuffering in ShowOptions then
		Repaint
	else
	begin
		R.Left := 0;
		R.Right := 9999;
		R.Top := FUnderLineY;
		R.Bottom := FUnderLineY + 1;
		InvalidateRect(Handle, @R, True);
		if esoRuler in ShowOptions then
		begin
			R.Top := 0;
			R.Bottom := 8;
			InvalidateRect(Handle, @R, True);
		end;
		if not PeekMessageW(M, Handle, WM_PAINT, WM_PAINT, PM_NOREMOVE) then
		begin
			GetCaretPos(R.TopLeft);
			R.Right := R.Left + 1;
			R.Bottom := R.Top + 1;
			InvalidateRect(Handle, @R, True);
			if not PeekMessageW(M, Handle, WM_PAINT, WM_PAINT, PM_NOREMOVE) then
			begin
				R.Left := 0;
				R.Top := 0;
				R.Right := 9999;
				R.Bottom := 1;
				InvalidateRect(Handle, @R, True);
				if not PeekMessageW(M, Handle, WM_PAINT, WM_PAINT, PM_NOREMOVE) then
					InvalidateRect(Handle, nil, True);
			end;
		end;
	end	
end;

procedure TEditorPane.RepaintScroll(X, Y: Integer);
var
	ViewRect: TRect;
	Right: Integer;
begin
	RepaintCaret;

	ViewRect := Self.ViewRect;
	Right := ViewRect.Right;

	{{}
	ScrollWindowEx(Handle, X, Y, @ViewRect, @ViewRect, 0, nil, SW_INVALIDATE or SW_ERASE);

	{sԍGA}
	ViewRect.Right := ViewRect.Left;
	ViewRect.Left := 0;

	ScrollWindowEx(Handle, 0, Y, @ViewRect, @ViewRect, 0, nil, SW_INVALIDATE or SW_ERASE);

	{Ruler}
	ViewRect.Bottom := ViewRect.Top;
	ViewRect.Top := 0;
	ViewRect.Left := ViewRect.Right;
	ViewRect.Right := Right;
	ScrollWindowEx(Handle, X, 0, @ViewRect, @ViewRect, 0, nil, SW_INVALIDATE or SW_ERASE);

	{XN[o[}
	FUpdateScrollBar := True;
end;

function TEditorPane.ReplaceLength(Bytes: Natural): Natural;
begin
	Result := 0;
	if Selected then
	begin
		FSelected := False;
		Result := FAnchorPosition - FCaretPosition;
	end
	else if not InsertMode then
	begin
		Result := Bytes;
		if FCaretPosition + Result > Layout.Document.Size then
			Result := Layout.Document.Size - FCaretPosition
	end;
end;

procedure TEditorPane.Scroll;
var
	ViewRect: TRect;
	PreLeft, X, Y, R, DiffY: Integer;
	Line, Next, PreFrom: TLine;
	LineHeight: Integer;
	PreBase: TScrollBase;
begin
	if Layout.Document.Size > 0 then
	begin
		ViewRect := Self.ViewRect;
		LineHeight := Layout.LineHeight;

		PreBase := FScrollBase;
		PreFrom := FScrollFrom;
		PreLeft := FLeft;

		if View is TBinaryView then
			FLeft := 0
		else
		begin
			Line := View.ToLine(FCaretPosition, +1);
			X := View.ToX(Line, FCaretPosition);
			if X < FLeft then
				FLeft := X
			else
			begin
				R := X + 4 * Layout.StandardLetterWidth - FLeft - (ViewRect.Right - ViewRect.Left);
				if R > 0 then
					Inc(FLeft, R);
			end;
		end;

		DiffY := 0;

		if PreBase = sbTop then
		begin
			if FCaretPosition < PreFrom then
			begin
				while FCaretPosition < FScrollFrom do
				begin
					FScrollFrom := View.PredLine(FScrollFrom);
					Inc(DiffY, LineHeight)
				end;
			end
			else
			begin
				Line := FScrollFrom;
				Y := 0;
				repeat
					Next := View.SuccLine(Line);
					if Next < 0 then Break;
					if Next > FCaretPosition then Break;
					Line := Next;
					Inc(Y, LineHeight);
					if Y + LineHeight > ViewRect.Bottom - ViewRect.Top then
					begin
						FScrollBase := sbBottom;
						FScrollFrom := Line;
					end
				until False;
			end
		end
		else
		begin
			Next := View.SuccLine(FScrollFrom);
			if Next < 0 then Next := Layout.Document.Size;
			if FCaretPosition >= Next then
			begin
				repeat
					Next := View.SuccLine(FScrollFrom);
					if Next < 0 then Break;
					if FCaretPosition < Next then Break;
					FScrollFrom := Next;
					Dec(DiffY, LineHeight)
				until False;
			end
			else
			begin
				Y := (ViewRect.Bottom - ViewRect.Top) - LineHeight;
				Line := FScrollFrom;
				while Line >= 0 do
				begin
					if Line <= FCaretPosition then Break;
					Line := View.PredLine(Line);
					Dec(Y, LineHeight);
					if Y < 0 then
					begin
						FScrollBase := sbTop;
						FScrollFrom := Line;
					end
				end;
			end
		end;

		if (DiffY <> 0) or (PreBase = FScrollBase) and (PreFrom = FScrollFrom) then
		begin
			{DiffYLȎ}
			RepaintScroll(PreLeft - FLeft, DiffY);
		end
		else if (PreLeft <> FLeft) or (PreBase <> FScrollBase) or (PreFrom <> FScrollFrom) then
		begin
			FUpdateScrollBar := True;
			Repaint;
		end
	end
end;

procedure TEditorPane.Scroll(const Position: TScrollingLinePosition);
var
	I, N: Integer;
begin
	if FKeepCaretYPosition then
	begin
		FScrollBase := Position.Base;
		FScrollFrom := View.LineTopCaretPosition(FCaretPosition, False);
		if Position.Base = sbTop then
		begin
			for I := 1 to Position.Gap do
			begin
				N := View.PredLine(FScrollFrom);
				if N < 0 then Break;
				FScrollFrom := N;
			end;
		end
		else
		begin
			for I := 1 to Position.Gap do
			begin
				N := View.SuccLine(FScrollFrom);
				if N < 0 then Break;
				FScrollFrom := N;
			end;
		end;
	end;
  FAdjustStartLine := True;
	Scroll;
end;

function TEditorPane.ScrollBase: TScrollBase;
begin
	Result := FScrollBase
end;

procedure TEditorPane.ScrollDown;
var
	Y: Integer;
begin
	Y := View.ScrollDown(FScrollBase, FScrollFrom);
	RepaintScroll(0, Y);
end;

function TEditorPane.ScrollingLinePositon: TScrollingLinePosition;
var
	I, J : Integer;
begin
	Result.Base := FScrollBase;
 	Result.Gap := 0;
	if Result.Base = sbTop then
	begin
		I := FScrollFrom;
		repeat
			J := View.LineEnd(I);
			if (J > FCaretPosition) or (J = Layout.Document.Size) then Break;
			Inc(Result.Gap);
			I := J;
		until False;
	end
	else
	begin
		I := FScrollFrom;
		repeat
			if I <= FCaretPosition then Break;
			Inc(Result.Gap);
			I := View.PredLine(I);
		until False;
	end
end;

procedure TEditorPane.ScrollUp;
var
	Y: Integer;
begin
	Y := View.ScrollUp(FScrollBase, FScrollFrom);
	RepaintScroll(0, Y);
end;

procedure TEditorPane.SelectAll;
begin
	ClearSelection;
	CaretPosition := 0;
	StartSelection;
	CaretPosition := Layout.Document.Size;
end;

procedure TEditorPane.SelectWord(Dir: Integer);
var
	N: TBufferAccessor;
	C: WideChar;
	Cur, First, Last, Temp: Natural;
begin
	if (SelectionFirstPosition <> SelectionLastPosition) then
	begin
		if Dir > 0 then WordRight else WordLeft
	end
	else
	begin
		N.Buffer := Layout.Document;
		Cur := CaretPosition;
		while Cur > 0 do
		begin
			N.Index := Layout.PredLetter(Cur);
			C := Layout.Decoder.Decode(N);
			case C of
			'A'..'Z', 'a'..'z', '0'..'9', '_':
				;
			else
				Break;
			end;
			Cur := Layout.PredLetter(Cur);
		end;
		First := Cur;
		Cur := CaretPosition;
		while Cur < Layout.Document.Size do
		begin
			N.Index := Cur;
			C := Layout.Decoder.Decode(N);
			case C of
			'A'..'Z', 'a'..'z', '0'..'9', '_':
				;
			else
				Break;
			end;
			Cur := N.Index;
		end;
		Last := Cur;
		if First < Last then
		begin
			if Dir < 0 then
			begin
				Temp := First;
				First := Last;
				Last := Temp;
			end;
			StartSelection;
			FAnchorPosition := First;
			CaretPosition := Last;
			Scroll;
		end
	end;
end;

procedure TEditorPane.SetCaretLine(const Value: Integer);
begin
	FCaretPosition := Layout.GetLogicalLineFromLineNumber(Value);
	FCaretPreferableX := -1;
	SetCaretPositionInternal;
end;

procedure TEditorPane.SetCaretPosition(const Value: Natural);
var
	Old: Natural;
begin
	Assert(Value <= Layout.Document.Size);
	FCaretPreferableX := -1;
	if Value <> FCaretPosition then
	begin
		Old := FCaretPosition;
		FCaretPosition := Value;
		SetCaretPositionInternal;
		if Selected then RepaintArea(Old, FCaretPosition);
	end;
end;

procedure TEditorPane.SetCaretPositionInternal;
var
	DummyScreenNumber, DummyScreenColumn: Integer;
begin
	FHexInputing := $ff;
	if View is TTextView then
		(View as TTextView).GetLocation(FCaretPosition, DummyScreenNumber, DummyScreenColumn, FCaretLine, FCaretColumn);
	RepaintCaret;
end;

procedure TEditorPane.SetHexMode(const Value: Boolean);
begin
	if View is TBinaryView then
	begin
		if Value <> FHexMode then
		begin
			FHexMode := Value;
			if not Value then
				FCaretPosition := View.FixCaretPosition(FCaretPosition);
			RepaintCaret;
		end;
	end
end;

procedure TEditorPane.SetInsertMode(const Value: Boolean);
begin
	FInsertMode := Value;
	RepaintCaret;
end;

procedure TEditorPane.SetLayout(ALayout: TLayout; AViewClass: TViewClass);
var
	NewView: TView;
begin
	NewView := AViewClass.Create(ALayout, Self);
	ReleaseView;
	FView := NewView;
	FLayout := ALayout as TEditorLayout;

	if NewView is TTextView then
	begin
		FAnchorPosition := View.FixCaretPosition(FAnchorPosition);
		FCaretPosition := View.FixCaretPosition(FCaretPosition);
		FHexMode := False;
		FHexInputing := $ff;
	end
	else
	begin
		FLeft := 0;
	end;

  FScrollFrom := View.ToLine(FScrollFrom, +1);

	FUpdateScrollBar := True;
  FAdjustStartLine := True;
	Scroll;

	if Handle <> 0 then
	begin
		ApplyLayout;
		Repaint
	end
end;

procedure TEditorPane.SetSelectedData(const Value: string);
var
	Anchor: Natural;
begin
	Anchor := SelectionFirstPosition;
	FCaretPosition := View.InsertData(FCaretPosition, ReplaceLength(Length(Value)), Value, HexMode);
  FAnchorPosition := Anchor;
	FSelected := True;
	SetCaretPositionInternal;
	Scroll;
end;

procedure TEditorPane.SetSelectedText(const Value: WideString);
begin
	SetSelectedData(Layout.Decoder.EncodeString(Value))
end;

procedure TEditorPane.StartSelection;
begin
	if not FSelected then
	begin
		FSelected := True;
		FAnchorPosition := FCaretPosition;
    RecordCommand(ercStartSelection);
	end
end;

procedure TEditorPane.StopAutoScroll;
begin
	if FAutoScrolling <> 0 then
	begin
		KillTimer(Handle, FAutoScrolling);
		FAutoScrolling := 0;
	end;
end;

procedure TEditorPane.Undo;
begin
	if CanUndo then
	begin
		ClearSelection;
		if Layout.Document.UndoPosition <> CaretPosition then
			CaretPosition := Layout.Document.UndoPosition
		else
			CaretPosition := Layout.Document.Undo;

		Scroll;
	end;
end;

procedure TEditorPane.UnindentSelection;
var
	F, L: Natural;
begin
	Assert(View is TTextView);
	F := SelectionFirstPosition;
	L := SelectionLastPosition;
	FAnchorPosition := F;
	FCaretPosition := (View as TTextView).Unindent(F, L);
end;

procedure TEditorPane.Update;

	procedure AdjustStartLine;
	var
		X, Y: Integer;
	begin
		FAdjustStartLine := False;
		View.AdjustScroll(FLeft, FScrollBase, FScrollFrom, X, Y);
    RepaintScroll(X, Y);
	end;

begin
	if FAdjustStartLine then AdjustStartLine;
	if FUpdateScrollBar then UpdateScrollBar;
	Layout.Release;
	inherited;
end;

procedure TEditorPane.UpdateScrollBar;
var
	Position, PageSize, Min, Max: Integer;
begin
	FUpdateScrollBar := False;

	View.VertScrollBar(FScrollFrom, Position, Min, Max, PageSize);
	VertScrollBar.SetParams(Position, Min, Max, PageSize);

	View.HorzScrollBar(FLeft, Position, Min, Max, PageSize);
	HorzScrollBar.SetParams(Position, Min, Max, PageSize);
end;

function TEditorPane.ViewRect: TRect;
begin
	Result := ClientRect;
	Inc(Result.Left, 4);
	if (esoLineNumber in ShowOptions) or (View is TBinaryView) then
		Inc(Result.Left, 8 * 4);
	if esoRuler in ShowOptions then
		Inc(Result.Top, RulerHeight);
end;

procedure TEditorPane.WMCopy(var Message: TWMCopy);
begin
	CopyToClipboard
end;

procedure TEditorPane.WMEraseBkgnd(var Message: TWMEraseBkgnd);
var
	Brush: TAnyBrush;
begin
	Message.Result := 0;
	if not (esoDoubleBuffering in ShowOptions) then
	begin
		Brush := TemporaryBrush(Color);
		FillRect(Message.DC, ClientRect, Brush.Handle);
		Message.Result := 1;
	end
end;

procedure TEditorPane.WMHScroll(var Message: TWMHScroll);
begin
	if View is TTextView then
	begin
		FLeft := HorzScrollBar.Value;
		Repaint;
	end;
	inherited;
end;

procedure TEditorPane.WMImeRequest(var Message: TMessage);
var
	S: WideString;
	P: PRECONVERTSTRING;
begin
	case Message.WParam of
	IMR_RECONVERTSTRING:
	begin
		Message.Result := SizeOf(RECONVERTSTRING) + (SelectedLength + 1) * SizeOf(WideChar);

		if Message.LParam <> 0 then
		begin
			S := SelectedText;

			P := PRECONVERTSTRING(Message.LParam);
			P^.dwSize := SizeOf(RECONVERTSTRING);
			P^.dwVersion := 0;
			P^.dwStrLen := Length(S);
			P^.dwStrOffset := SizeOf(RECONVERTSTRING);
			P^.dwCompStrLen := Length(S);
			P^.dwCompStrOffset := 0;
			P^.dwTargetStrLen := 0;
			P^.dwTargetStrOffset := 0;

			Move(PWideChar(S)^,
				Pointer(Message.LParam + SizeOf(RECONVERTSTRING))^,
				(Length(S) + 1) * SizeOf(WideChar));
		end
	end;
	else
		inherited
	end
end;

procedure TEditorPane.WMKillFocus(var Message: TWMKillFocus);
begin
	inherited;
  FDraggingState := edsNone;
	StopAutoScroll;
	HideCaret(Handle);
	DestroyCaret;
	RepaintCaret;
end;

procedure TEditorPane.WMLButtonDown(var Message: TWMLButtonDown);
var
	P: TPoint;
	NewPos: Natural;
begin
  SetFocus;

	P.X := Message.XPos;
	P.Y := Message.YPos;

	HexMode := ClientToHexMode(P);
	NewPos := ClientToPosition(P);

	if Selected and (SelectionFirstPosition <= NewPos) and (NewPos <= SelectionLastPosition) then
	begin
		FDraggingState := edsDragOrClick;
		FDraggingPosition := P;
	end
	else
	begin
		FDraggingState := edsDown;

		if MK_SHIFT and Message.Keys <> 0 then
			StartSelection
		else
			ClearSelection;

		CaretPosition := NewPos;

		if Selected then
			RepaintArea(FAnchorPosition, FCaretPosition);

		SetCapture(Handle);
		SetTimer(Handle, MouseScrollTimerID, TimerElapse, nil);
	end;

	Message.Result := 0;
end;

procedure TEditorPane.WMLButtonUp(var Message: TWMLButtonUp);
var
	P: TPoint;
begin
	if FDraggingState = edsDragOrClick then
	begin
		if MK_SHIFT and Message.Keys <> 0 then
			StartSelection
		else
			ClearSelection;

		P.X := Message.XPos;
		P.Y := Message.YPos;
		CaretPosition := ClientToPosition(P);
	end
	else if FDraggingState = edsDown then
	begin
		KillTimer(Handle, MouseScrollTimerID);
		ReleaseCapture;
	end;
	FDraggingState := edsNone;
	inherited;
end;

procedure TEditorPane.WMMouseMove(var Message: TWMMouseMove);
var
	P: TPoint;
	Old, New: Natural;
begin
	P.X := Message.XPos;
	P.Y := Message.YPos;

	if FDraggingState = edsDown then
	begin
		StartSelection;
		if (P.Y >= 0) and (P.Y < ClientSize.cy) then
		begin
			Old := CaretPosition;
			New := ClientToPosition(P);
			CaretPosition := New;
			RepaintArea(Old, New);
		end
	end
	else if (FDraggingState = edsDragOrClick) and DragDetect(FDraggingPosition, P) then
	begin
		FDraggingState := edsNone;
		BeginDrag;
	end;
	inherited;
end;

procedure TEditorPane.WMMouseWheel(var Message: TWMMouseWheel);
var
	M: TWMVScroll;
begin
	M.Msg := WM_VSCROLL;
	M.ScrollBar := SB_VERT;

	if Message.WheelDelta > 0 then
	begin
		M.ScrollCode := SB_LINEUP;
		Dispatch(M);
	end
	else if Message.WheelDelta < 0 then
	begin
		M.ScrollCode := SB_LINEDOWN;
		Dispatch(M);
	end;
end;

procedure TEditorPane.WMPaint(var Message: TWMPaint);
var
	MemoryBitmap: HBITMAP;
	MemoryDC: HDC;
	Size: TSize;
	Rect: TRect;
	PS: PAINTSTRUCT;
	Brush: TAnyBrush;
	DC: HDC;
begin
	if (Message.DC <> 0) or not (esoDoubleBuffering in ShowOptions) then
		inherited
	else
	begin
		DC := GetDC(Handle);
		Size := ClientSize;
		MemoryDC := CreateCompatibleDC(DC);
		MemoryBitmap := CreateCompatibleBitmap(DC, Size.cx, Size.cy);
    ReleaseDC(Handle, DC);
		SelectObject(MemoryDC, MemoryBitmap);
		Brush := TemporaryBrush(Color);
		Rect.Left := 0;
		Rect.Top := 0;
		Rect.Right := Size.cx;
		Rect.Bottom := Size.cy;
		FillRect(MemoryDC, ClientRect, Brush.Handle);
		PaintWindow(MemoryDC);
		BeginPaint(Handle, PS);
		BitBlt(PS.hdc, 0, 0, Size.cx, Size.cy, MemoryDC, 0, 0, SRCCOPY);
		EndPaint(Handle, PS);
		DeleteDC(MemoryDC);
		DeleteObject(MemoryBitmap);
	end;
end;

procedure TEditorPane.WMSetFocus(var Message: TWMSetFocus);
begin
	FUpdateScrollBar := True;
	inherited;
	RepaintCaret;
end;

procedure TEditorPane.WMSize(var Message: TWMSize);
begin
	inherited;
	FAdjustStartLine := True;
	FUpdateScrollBar := True;
end;

procedure TEditorPane.WMTimer(var Message: TWMTimer);
type
	T = procedure of object;

	procedure Roll(P: T);
	var
		I, L: Integer;
		F, C: Natural;
	begin
		if GetKeyState(VK_CONTROL) and $8000 = 0 then
			P
		else
		begin
			SendMessageW(WM_SETREDRAW, 0, 0);
			F := FScrollFrom;
			C := FCaretPosition;
			with ViewRect do L := (Bottom - Top) div Layout.LineHeight;
			for I := L downto 1 do P;
			SendMessageW(WM_SETREDRAW, 1, 0);
			if (F <> FScrollFrom) or (C <> FCaretPosition) then Repaint;
		end;
		UpdateScrollBar
	end;

var
	P: TPoint;
	R: TRect;
begin
	case Message.TimerID of
	AutoScrollUpTimerID:
		Roll(CursorScrollUp);
	AutoScrollDownTimerID:
		Roll(CursorScrollDown);
	MouseScrollTimerID:
	begin
		GetCursorPos(P);
		GetWindowRect(Handle, R);
		FMouseRepeating := not FMouseRepeating;
		if (P.Y < R.Top) and ((P.Y < R.Top - 50) or FMouseRepeating) then
			CursorUp
		else if (P.Y >= R.Bottom) and ((P.Y >= R.Bottom + 50) or FMouseRepeating) then
			CursorDown;
	end;
	else
		inherited
	end
end;

procedure TEditorPane.WMVScroll(var Message: TWMVScroll);
begin
	case Message.ScrollCode of
	SB_ENDSCROLL:
		;
	SB_LINEUP:
		ScrollUp;
	SB_LINEDOWN:
		ScrollDown;
	else
		View.VertScroll(FScrollBase, FScrollFrom, VertScrollBar.Value,
			VertScrollBar.Min, VertScrollBar.Max, VertScrollBar.PageSize);

		Repaint;
	end
end;

procedure TEditorPane.WordRight;
var
	N: TBufferAccessor;
	C: WideChar;
	Cur: Natural;
begin
	if FSelected and (SelectionFirstPosition = SelectionLastPosition) then
		SelectWord(+1)
	else
	begin
		N.Buffer := Layout.Document;
		Cur := CaretPosition;
		while Cur < Layout.Document.Size do
		begin
			N.Index := Cur;
			C := Layout.Decoder.Decode(N);
			case C of
			'A'..'Z', 'a'..'z', '0'..'9', '_':
				;
			else
				Break;
			end;
			Cur := N.Index;
		end;
		if Cur = CaretPosition then
		begin
			while Cur < Layout.Document.Size do
			begin
				N.Index := Cur;
				C := Layout.Decoder.Decode(N);
				case C of
				'A'..'Z', 'a'..'z', '0'..'9', '_':
					Break;
				end;
				Cur := N.Index;
			end;
		end;
		CaretPosition := Cur;
		Scroll;
	end;
	RecordCommand(ercWordRight);
end;

procedure TEditorPane.WordLeft;
var
	N: TBufferAccessor;
	C: WideChar;
	Cur: Natural;
begin
	if FSelected and (SelectionFirstPosition = SelectionLastPosition) then
		SelectWord(-1)
	else
	begin
		N.Buffer := Layout.Document;
		Cur := CaretPosition;
		while Cur > 0 do
		begin
			N.Index := Layout.PredLetter(Cur);
			C := Layout.Decoder.Decode(N);
			case C of
			'A'..'Z', 'a'..'z', '0'..'9', '_':
				;
			else
				Break;
			end;
			Cur := Layout.PredLetter(Cur);
		end;
		if Cur = CaretPosition then
		begin
			while Cur > 0 do
			begin
				N.Index := Layout.PredLetter(Cur);
				C := Layout.Decoder.Decode(N);
				case C of
				'A'..'Z', 'a'..'z', '0'..'9', '_':
					Break;
				end;
				Cur := Layout.PredLetter(Cur);
			end;
		end;
		CaretPosition := Cur;
		Scroll;
	end;
	RecordCommand(ercWordLeft);
end;

procedure TEditorPane.WMLButtonDblClk(var Message: TWMLButtonDblClk);
var
	Position: Natural;
begin
	Position := CaretPosition;
	StartSelection;
	SelectWord(+1);
	if CaretPosition = Position then begin
		ClearSelection;
		CursorLineTop(False);
		StartSelection;
    CursorLineEnd;
	end;
end;

{ TEditor }

constructor TEditor.Create(AParent: TAnyWindow;
	const APosition: TPoint; const ASize: TSize; ASplitable: TBarStyles);
begin
	inherited Create(AParent, APosition, ASize, ASplitable, [sbVertical, sbHorizontal], True);
	FDocument := TDocument.Create;
end;

function TEditor.CreateDefaultLayout: TEditorLayout;
const
	DefaultSelectedColor: TEditorColor = (Foreground: $ffffff; Background: 0);
var
	DefaultStateColors: TStateColors;
begin
	DefaultStateColors[0].Foreground := 0;
	DefaultStateColors[0].Background := $ffffff;
	Result := TEditorLayout.Create(Document, LocaleDecoder,
		ForClientFont, nil, '', nil, nil, 0, 4, '', '', '', False, 0,
		DefaultStateColors, DefaultSelectedColor, $0, $0, $0);
end;

function TEditor.CreatePane(AHorz, AVert: TScrollBar): T0SplitPane;
begin
	Result := TEditorPane.Create(Self, AHorz, AVert, ReadyLayout, ReadyViewClass)
end;

destructor TEditor.Destroy;
begin
	inherited;
	FMarker.Free;
	FDocument.Free;
end;

procedure TEditor.DoSplit(Dir: TBarStyle; Per: Real);
begin
	if (Dir = sbVertical) and not (sbVertical in Splited) then
		(Panes[0, 0] as TEditorPane).AdjustForBottom;

	inherited;
end;

function TEditor.GetMarked: Boolean;
begin
	Result := (FMarker <> nil) and FMarker.Marked
end;

function TEditor.GetMarker: TBufferMarker;
begin
	if FMarker = nil then
		FMarker := TBufferMarker.Create(Document);

	Result := FMarker;
end;

function TEditor.GetReferencePane: TEditorPane;
var
	I, J: Integer;
begin
	for I := Low(Panes) to High(Panes) do
		for J := Low(Panes[I]) to High(Panes[I]) do
			if Panes[I, J] <> nil then
			begin
				Result := Panes[I, J] as TEditorPane;
				Exit;
			end;

	Result := nil;
end;

function TEditor.ReadyLayout: TLayout;
var
	Ref: TEditorPane;
begin
	Ref := GetReferencePane as TEditorPane;
	if Ref = nil then
		Result := CreateDefaultLayout
	else
		Result := Ref.View.Layout
end;

function TEditor.ReadyViewClass: TViewClass;
var
	Ref: TEditorPane;
begin
	Ref := GetReferencePane as TEditorPane;
	if Ref = nil then
		Result := TTextView
	else
		Result := Ref.View.ViewType
end;

procedure TEditor.ReleaseMarker;
begin
	if FMarker <> nil then
	begin
		FMarker.Free;
		FMarker := nil;
		RepaintPanes
	end;
end;

procedure TEditor.RepaintPanes;
var
	I, J: Integer;
begin
	for I := Low(Panes) to High(Panes) do
		for J := Low(Panes[I]) to High(Panes[I]) do
			if Panes[I, J] <> nil then
				Panes[I, J].Repaint;
end;

procedure TEditor.SetUnderLine(const Value: Boolean);
var
	AC: T0Window;
begin
	FUnderLine := Value;
	AC := ActiveControl;
	if AC <> nil then AC.Repaint;
end;

procedure TEditor.Split;
begin
	BeginSpliting(sbVertical);
end;

procedure TEditor.Switch;

	function CheckIsPane(Wnd: TAnyWindow): Boolean; cdecl;
	begin
		Result:= Wnd is TEditorPane
	end;

var
	Wnd: TAnyWindow;
begin
	Wnd := FindNextWindow(ActiveControl, True, CCheckWnd(@CheckIsPane), GetContext);
	if Wnd <> nil then Wnd.SetFocus;
end;

end.

