unit LMDICtls;

interface

uses
	Windows, Messages, Types, CommCtrl, SysUtils,
	LObjects, LActions, LWindows, LMenus, LStdCtls, LCmmCtls;

const
	{Notification Code}
	MDICN_ACTIVATE = NM_COMMAND + 1;
	MDICN_CLOSE = NM_COMMAND + 2;

type
	TMDIChild = class(T0Window)
	private
		FClient: IMDIClient;
		procedure WMMDIActivate(var Message: TWMMDIActivate); message WM_MDIACTIVATE;
		procedure WMNCLButtonDblClk(var Message: TWMNCLButtonDblClk); message WM_NCLBUTTONDBLCLK;
		procedure WMNCRButtonUp(var Message: TWMNCRButtonUp); message WM_NCRBUTTONUP;
		procedure WMSysCommand(var Message: TWMSysCommand); message WM_SYSCOMMAND;
		procedure WMClose(var Message: TWMClose); message WM_CLOSE;
	protected
		class function GetDefaultSize: TSize; override;
		procedure MnemonicKey(var Message: TWMKey); override;
		procedure Initialize; override;
		procedure Finalize; override;
		procedure TrackSystemMenu;
		function CloseQuery: Boolean; virtual;
	public
		constructor Create(const AParent: IMDIClient;
			const AText: WideString; const APosition: TPoint; const ASize: TSize;
			Maximized: Boolean);
		procedure DefaultHandler(var Message); override;
		procedure Maximize;
		procedure Close;
		procedure Activate;
	end;

const
	WM_MDIRESTOREMENU = WM_USER - 1;

type
	TMDIClient = class(T0SubClassWindow, IMDIClient)
	private
		FDummyMenu: TMainMenu;
		{IMDIClient}
		function ClientWindow: T0Window;
		procedure OnCreate(Child: T0Window);
		procedure OnDestroy(Child: T0Window);
		procedure OnActivate(Child: T0Window);
		function GetActiveChild: TMDIChild;
		function GetActiveChildHandle: HWND;
		procedure WMMDIRestoreMenu(var Message: TMessage); message WM_MDIRESTOREMENU;
		procedure WMMDIMaximize(var Message: TMessage); message WM_MDIMAXIMIZE;
		procedure WMMDIActivate(var Message: TMessage); message WM_MDIACTIVATE;
		procedure WMMDIRestore(var Message: TMessage); message WM_MDIRESTORE;
		procedure WMSize(var Message: TMessage); message WM_SIZE;
	protected
		class function SubClassName: PWideChar; override;
		procedure PaintWindow(DC: HDC); override;
	public
		constructor Create(AParent: TAnyWindow);
		procedure Next;
		procedure PreserveMenu;
		procedure Activate(Child: T0Window);
		property ActiveChild: TMDIChild read GetActiveChild;
	end;

type
	CEachMDIChild = function(Child: TMDIChild; Context: Pointer): Boolean; cdecl;

	TTabMDIClient = class(TTabControl, IMDIClient)
	private
		FClient: TMDIClient;
		function ClientWindow: T0Window;
		procedure OnCreate(Child: T0Window);
		procedure OnDestroy(Child: T0Window);
		procedure OnActivate(Child: T0Window);
	protected
		procedure Initialize; override;
		procedure HandleNotification(Sender: TAnyWindow; Info: PNMHdr;
			var Result: Integer); override;
	public
		constructor Create(AParent: TAnyWindow; ASide: TAlignment);
		function DisplayRect: TRect; override;
		procedure PreserveMenu;
		procedure Realign; override;
		procedure Next;
		procedure Activate(Child: T0Window);
		function Enumerate(Proc: CEachMDIChild; Context: Pointer): TMDIChild;
		property Client: TMDIClient read FClient;
	end;

implementation

{ TMDIChild }

procedure TMDIChild.Activate;
begin
	FClient.Activate(Self);
end;

procedure TMDIChild.Close;
begin
	PostMessageW(WM_CLOSE, 0, 0);
end;

function TMDIChild.CloseQuery: Boolean;
begin
	Result := Notify(MDICN_CLOSE) = 0;
end;

constructor TMDIChild.Create(const AParent: IMDIClient; const AText: WideString;
	const APosition: TPoint; const ASize: TSize; Maximized: Boolean);
var
	CS: PCreateStructW;
begin
	FClient := AParent;
	FClient.PreserveMenu;
	inherited Create(AParent.ClientWindow, AText, APosition, ASize);
	CS := CreateStruct;
	CS^.style := MDIS_ALLCHILDSTYLES or WS_CLIPCHILDREN;
	if Maximized then CS^.style := CS^.style or WS_MAXIMIZE;
	CS^.dwExStyle := CS^.dwExStyle or (WS_EX_MDICHILD or WS_EX_WINDOWEDGE);
end;

procedure TMDIChild.DefaultHandler(var Message);
begin
	TMessage(Message).Result := DefMDIChildProcW(Handle, TMessage(Message).Msg,
		TMessage(Message).WParam, TMessage(Message).LParam);
end;

procedure TMDIChild.Finalize;
begin
	FClient.PreserveMenu;
	FClient.OnDestroy(Self);
	inherited;
end;

class function TMDIChild.GetDefaultSize: TSize;
begin
	Result := DefaultSize
end;

procedure TMDIChild.Initialize;
begin
	inherited;
	FClient.OnCreate(Self);
end;

procedure TMDIChild.Maximize;
begin
	FClient.ClientWindow.SendMessageW(WM_MDIMAXIMIZE, Handle, 0);
end;

procedure TMDIChild.MnemonicKey(var Message: TWMKey);
begin
	if Message.CharCode = VK_F6 then
	begin
    SendMessageW(WM_SYSCOMMAND, SC_NEXTWINDOW, 0);
		Message.Result := 1;
	end
	else
		inherited
end;

procedure TMDIChild.TrackSystemMenu;
var
	Menu: HMENU;
	P: TPoint;
	Cmd: DWORD;
begin
	Menu := GetSystemMenu(Handle, False);
	GetCursorPos(P);
	Cmd := DWORD(TrackPopupMenu(Menu,
			TPM_RETURNCMD or TPM_RIGHTBUTTON or TPM_LEFTALIGN or TPM_TOPALIGN,
			P.X, P.Y, 0, Handle, nil));
	SendMessageW(WM_SYSCOMMAND, Cmd, 0)
end;

procedure TMDIChild.WMClose(var Message: TWMClose);
begin
	if CloseQuery then
		inherited;
end;

procedure TMDIChild.WMMDIActivate(var Message: TWMMDIActivate);
begin
	if Message.ActiveWnd = Handle then
	begin
		FClient.OnActivate(Self);
		Notify(MDICN_ACTIVATE);
	end;
	inherited
end;

procedure TMDIChild.WMNCLButtonDblClk(var Message: TWMNCLButtonDblClk);
begin
	if Message.HitTest = HTCAPTION then FClient.PreserveMenu;
	inherited;
end;

procedure TMDIChild.WMNCRButtonUp(var Message: TWMNCRButtonUp);
begin
	if Message.HitTest = HTCAPTION then
		TrackSystemMenu
	else
		inherited
end;

procedure TMDIChild.WMSysCommand(var Message: TWMSysCommand);
begin
	case Message.CmdType of
	SC_MAXIMIZE,
	SC_NEXTWINDOW:
		FClient.PreserveMenu;
	end;
	inherited;
end;

{ TMDIClient }

procedure TMDIClient.Activate(Child: T0Window);
begin
	SendMessageW(WM_MDIACTIVATE, Child.Handle, 0)
end;

function TMDIClient.ClientWindow: T0Window;
begin
	Result := Self
end;

constructor TMDIClient.Create(AParent: TAnyWindow);
const
	Struct: CLIENTCREATESTRUCT = (hWindowMenu: 0; idFirstChild: 0);
var
	CS: PCreateStructW;
begin
	inherited Create(AParent, '', ZeroPoint, ZeroSize);
	Alignment := alCenter;
	CS := CreateStruct;
	CS^.lpCreateParams := @Struct;
	FDummyMenu := TMainMenu.Create;
	FDummyMenu.Owner := Self;
end;

function TMDIClient.GetActiveChild: TMDIChild;
begin
	Result := GetLWnd(GetActiveChildHandle) as TMDIChild;
end;

function TMDIClient.GetActiveChildHandle: HWND;
begin
	Result := SendMessageW(WM_MDIGETACTIVE, 0, 0);
end;

procedure TMDIClient.Next;
begin
	SendMessageW(WM_MDINEXT, GetActiveChildHandle, 0);
end;

procedure TMDIClient.OnActivate(Child: T0Window);
begin
end;

procedure TMDIClient.OnCreate(Child: T0Window);
begin
end;

procedure TMDIClient.OnDestroy(Child: T0Window);
begin
end;

procedure TMDIClient.PaintWindow(DC: HDC);
var
	R: TRect;
begin
	R := ClientRect;
	DrawEdge(DC, R, EDGE_SUNKEN, BF_RECT);
end;

procedure TMDIClient.PreserveMenu;
begin
	while FDummyMenu.Count > 0 do
	begin
		RemoveMenu(FDummyMenu.Handle, 0, MF_BYPOSITION);
	end;
	SendMessageW(WM_MDISETMENU, FDummyMenu.Handle, 0);
	PostMessageW(WM_MDIRESTOREMENU, 0, 0);
end;

class function TMDIClient.SubClassName: PWideChar;
begin
	Result := 'MDICLIENT'
end;

procedure TMDIClient.WMMDIActivate(var Message: TMessage);
begin
	PreserveMenu;
	inherited;
end;

procedure TMDIClient.WMMDIMaximize(var Message: TMessage);
begin
	PreserveMenu;
	inherited;
end;

procedure TMDIClient.WMMDIRestore(var Message: TMessage);
begin
	GetLWnd(Message.WParam).WindowStyle[WS_SYSMENU] := True;
	inherited;
end;

procedure TMDIClient.WMMDIRestoreMenu(var Message: TMessage);
var
	Window: T0TopWindow;
begin
	Window := TopParent;
	if Window.Menu <> nil then
		SetMenu(Window.Handle, Window.Menu.Handle);
end;

procedure TMDIClient.WMSize(var Message: TMessage);
begin
	inherited;
  Repaint;
end;

{ TTabMDIClient }

procedure TTabMDIClient.Activate(Child: T0Window);
begin
  FClient.Activate(Child);
end;

function TTabMDIClient.ClientWindow: T0Window;
begin
	Result := Client
end;

constructor TTabMDIClient.Create(AParent: TAnyWindow; ASide: TAlignment);
begin
	inherited Create(AParent, ZeroPoint, ZeroSize, ASide, [tcsMultiLine, tcsNoFocus]);
	Alignment := alCenter;
end;

function TTabMDIClient.DisplayRect: TRect;
begin
	Result := ClientRect;
	Result.Top := (inherited DisplayRect).Top;
end;

function TTabMDIClient.Enumerate(Proc: CEachMDIChild; Context: Pointer): TMDIChild;
var
	I: Integer;
begin
	for I := 0 to Count - 1 do
	begin
		Result := At[I].Data;
		if Proc(Result, Context) then Exit;
	end;
	Result := nil;
end;

procedure TTabMDIClient.HandleNotification(Sender: TAnyWindow;
	Info: PNMHdr; var Result: Integer);
var
	ActiveChild: TMDIChild;
begin
	if Sender = Self then
	begin
		case Info.code of
		TCN_SELCHANGE:
			if Count > 0 then
			begin
				ActiveChild := TMDIChild(At[SelectedIndex].Data);
				if IsIconic(ActiveChild.Handle) then
					Client.SendMessageW(WM_MDIRESTORE, ActiveChild.Handle, 0);
				Client.SendMessageW(WM_MDIACTIVATE, ActiveChild.Handle, 0);
			end;
		TCN_MCLICK:
			if Count > 0 then
			begin
				ActiveChild := Client.ActiveChild;
				FClient.Next;
				ActiveChild.Close;
			end;
		TCN_DBLCLK:
		begin
			ActiveChild := Client.ActiveChild;
			if ActiveChild <> nil then
			begin
				if ActiveChild.WindowStyle[WS_MAXIMIZE] then
					Client.SendMessageW(WM_MDIRESTORE, ActiveChild.Handle, 0)
				else
					ActiveChild.Maximize;
			end;
		end;
		TCN_RCLICK:
		begin
			if Count > 0 then
			begin
				ActiveChild := TMDIChild(At[SelectedIndex].Data);
				ActiveChild.TrackSystemMenu;
			end;
		end;
		end
	end;
	inherited;
end;

procedure TTabMDIClient.Initialize;
begin
	inherited;
	FClient := TMDIClient.Create(TopParent);
	FClient.Alignment := alNone;
end;

procedure TTabMDIClient.Next;
begin
	FClient.Next
end;

procedure TTabMDIClient.OnActivate(Child: T0Window);
var
	I: Integer;
	Item: TTabControl.TItem;
begin
	for I := Count - 1 downto 0 do
	begin
		Item := At[I];
		if Item.Data = Child then
		begin
			SelectedIndex := I;
			Break;
		end;
	end;
end;

procedure TTabMDIClient.OnCreate(Child: T0Window);
var
	Item: TTabControl.TItem;
begin
	Item := Add;
	Item.Text := Child.Text;
	Item.Data := Child;
	Realign
end;

procedure TTabMDIClient.OnDestroy(Child: T0Window);
var
	I: Integer;
	Item: TTabControl.TItem;
begin
	for I := Count - 1 downto 0 do
	begin
		Item := At[I];
		if Item.Data = Child then
		begin
			Remove(Item);
			Break;
		end;
	end;
	Realign
end;

procedure TTabMDIClient.PreserveMenu;
begin
	FClient.PreserveMenu
end;

procedure TTabMDIClient.Realign;
var
	W, R: TRect;
begin
	W := BoundsRect;
	R := DisplayRect;
	OffsetRect(R, W.Left, W.Top);
	FClient.BoundsRect := R;
	FClient.BringToFront(Self);
end;

end.
