unit LMenus;

{*******************************************************}
{                                                       }
{       Light Windows                                   }
{       Menu Interface                                  }
{                                                       }
{       1998..2004 YT                                   }
{                                                       }
{*******************************************************}

interface

uses
	Windows, Messages, SysUtils, SysConst,
	SysUtilsW, SysHacks, AlgUtils, WinTips,
	LConsts, LObjects, LActions;

type
	TAnyMenuItem = class;
	TAnyMenu = class;

	ICommandManager = interface(IInterface)
		['{66441436-41C2-4762-946C-5236B83787CB}']
		function RegisterCommand(const AItem: TObject): UINT;
		procedure UnregisterCommand(const ACommand: UINT);
		function ActiveKeyManager: IKeyManager;
	end;

	CEachMenuItem = function(Item: TAnyMenuItem; Context: Pointer): Boolean cdecl;

	TAnyMenuItem = class(T0Object)
	private
		FMenu: TAnyMenu;
		FID: UINT;
		function GetText: WideString;
		procedure SetText(const Value: WideString);
		procedure SetEnabled(const Value: Boolean);
		function GetEnabled: Boolean;
		function GetChecked: Boolean;
		procedure SetChecked(const Value: Boolean);
		function GetShortCuts: WideString;
		procedure SetShortCuts(const Value: WideString);
		function GetSeparator: Boolean;
		procedure SetSeparator(const Value: Boolean);
		function GetRadioCheck: Boolean;
		procedure SetRadioCheck(const Value: Boolean);
		function GetOwnerDraw: Boolean;
		procedure SetOwnerDraw(const Value: Boolean);
	protected
		function GetFullText: WideString; virtual;
		procedure SetFullText(const Value: WideString); virtual;
		function GetType: DWORD;
		function GetState(Flag: DWORD): Boolean;
		procedure SetState(Flag: DWORD; Value: Boolean);
		procedure SetType(ClearFlags, SetFlags: DWORD);
		function GetSub: TAnyMenu;
		procedure SetSub(const Value: TAnyMenu);
		procedure Modify;
		procedure Update; virtual;
		procedure SetAccelerator(const Value: WideChar); virtual;
		function GetAccelerator: WideChar; virtual;
	public
		constructor Create(AMenu: TAnyMenu; AID: UINT);
		procedure Execute; virtual;
		procedure Selected; virtual;
		property Separator: Boolean read GetSeparator write SetSeparator;
		property Menu: TAnyMenu read FMenu; //ej[
		property ID: UINT read FID;
		property FullText: WideString read GetFullText write SetFullText;
		property Text: WideString read GetText write SetText;
		property Accelerator: WideChar read GetAccelerator write SetAccelerator;
		property ShortCuts: WideString read GetShortCuts write SetShortCuts;
		property Enabled: Boolean read GetEnabled write SetEnabled;
		property Checked: Boolean read GetChecked write SetChecked;
		property RadioCheck: Boolean read GetRadioCheck write SetRadioCheck;
		property OwnerDraw: Boolean read GetOwnerDraw write SetOwnerDraw;
		property Sub: TAnyMenu read GetSub write SetSub;
	end;

	TAnyMenu = class(T0Object)
	private
		FPredItems: TAnyMenu;
		FSuccItems: TAnyMenu;
		FHandle: HMENU;
		function GetAt(AIndex: Integer): TAnyMenuItem;
		function GetCount: Integer;
	protected
		procedure Modify; virtual;
	public
		constructor Create(AHandle: THandle);
		procedure AfterConstruction; override;
		procedure BeforeDestruction; override;
		function IndexOf(AItem: TAnyMenuItem): Integer;
		procedure Update; virtual;
		property Handle: HMENU read FHandle;
		property At[AIndex: Integer]: TAnyMenuItem read GetAt; default;
		function EnumerateItems(Proc: CEachMenuItem; Context: Pointer): TAnyMenuItem;
		function Find(ID: UINT): TAnyMenuItem;
		property Count: Integer read GetCount;
	end;

	{LWj[}
	TMenuItem = class(TAnyMenuItem)
	private
		FFullText: WideString;
	protected
		function GetFullText: WideString; override;
		procedure SetFullText(const Value: WideString); override;
	public
		constructor Create(AMenu: TAnyMenu);
		procedure AfterConstruction; override;
		procedure BeforeDestruction; override;
	end;

type
	TActionMenuItem = class(TMenuItem)
	private
		FAction: IAction;
		FAccelerator: WideChar;
		FNeedCreateSubItems: Boolean;
	protected
		procedure Update; override;
		function MakeShortCuts: WideString;
		procedure SetAccelerator(const Value: WideChar); override;
		function GetAccelerator: WideChar; override;
	public
		constructor Create(AMenu: TAnyMenu; AAction: IAction);
		procedure Execute; override;
		procedure Selected; override;
		property Action: IAction read FAction;
	end;

type
	TPopupMenu = class(TAnyMenu)
	public
		constructor Create;
		procedure BeforeDestruction; override;
	end;

type
	TMainMenu = class(TAnyMenu)
	private
		FModified: Boolean;
	protected
		procedure Modify; override;
	public
		constructor Create;
		procedure BeforeDestruction; override;
		procedure Update; override;
	end;

//L[}bv
type
	TKeyAssignment = class(T0Object, IKeyAssignment, IKeyManager)
	private
		FTable: array[0..7, Byte] of IAction;
		FAttached: array of IKeyAssignment;
		FConnection: TKeyAssignment;
		function GetTable(const AShift: TShiftKeys; AKey: Integer): IAction;
		procedure SetTable(const AShift: TShiftKeys; AKey: Integer;
						const Value: IAction);
	public
		procedure Attach(const AAssignment: IKeyAssignment);
		procedure Detach(const AAssignment: IKeyAssignment);
		function FindAction(const AShift: TShiftKeys; AKey: Integer): IAction;
		property Table[const Shift: TShiftKeys; Key: Integer]: IAction
						read GetTable write SetTable; default;
		property Connection: TKeyAssignment read FConnection write FConnection;
	end;

{j[nh烁j[𓾂}

function GetLMenu(Menu: HMENU): TAnyMenu;
function GetMenu(Menu: HMENU): TAnyMenu;

{j[eLXg}

function AssembleMenuText(const AText: WideString;
				AAccelerator: WideChar; const AShortCuts: WideString): WideString;

procedure DisassembleMenuText(const Value: WideString; var AText: WideString;
				var AAccelerator: WideChar; var AShortCuts: WideString);

function SelectActionItemsPopup(const AItems: IActionItems; const ScreenPos: TPoint): IActionItem;

{ Utilities }

var
	InitialMenuOwnerDraw: Boolean;

function CreateSeparator(Menu: TAnyMenu): TMenuItem;
function CreateMenuItem(Menu: TAnyMenu; const Text: WideString; Radio: Boolean = False): TMenuItem; overload;
function CreateMenuItem(Menu: TAnyMenu; const Action: IAction; Accelerator: WideChar): TMenuItem; overload;
function CreateSubMenu(Item: TAnyMenuItem): TAnyMenu;

implementation

type
	TActionSubMenuItem = class(TMenuItem)
	private
		FActionItem: Pointer; {weak}
		FNeedCreateSubItems: Boolean;
	public
		constructor Create(AMenu: TAnyMenu; const AActionItem: IActionItem);
		procedure AfterConstruction; override;
		procedure Execute; override;
		procedure Selected; override;
	end;

procedure AddActionItems(const ActionItems: IActionItems; Menu: TAnyMenu; AOwnerDraw: Boolean);
var
	R: Boolean;

	function EnumProc(const ActionItem: IActionItem): Boolean; cdecl;
	var
		M: TActionSubMenuItem;
	begin
		M := TActionSubMenuItem.Create(Menu, ActionItem);
		M.Text := ActionItem.Text;
		M.Checked := ActionItem.Checked;
		M.OwnerDraw := AOwnerDraw;
		R := True;
		Result := False;
	end;

begin
	R := False;
	ActionItems.Enumerate(CEachActionItem(@EnumProc), GetContext);
	if not R then
		with TActionSubMenuItem.Create(Menu, nil) do
		begin
			Text := SNothing;
			Enabled := False;
			OwnerDraw := AOwnerDraw
		end
end;

procedure TActionSubMenuItem.AfterConstruction;
var
	Items: TPopupMenu;
begin
	inherited;

	if (FActionItem <> nil) and (IActionItem(FActionItem).Items <> nil) then
	begin
		FNeedCreateSubItems := True;

		Items := TPopupMenu.Create;
		Items.Owner := Self;
		Sub := Items;
	end;
end;

constructor TActionSubMenuItem.
				Create(AMenu: TAnyMenu; const AActionItem: IActionItem);
begin
	inherited Create(AMenu);
	FActionItem := Pointer(AActionItem);
(*
	weak reference ̗R:
	IActionItemQƃJEgŔĵ͂ނ(^^)Ȃ̂
	Ă܂ꍇ_ReleaseĂԂƗ邽
*)
end;

procedure TActionSubMenuItem.Execute;
begin
	if FActionItem <> nil then
		IActionItem(FActionItem).Execute
end;

procedure TActionSubMenuItem.Selected;
var
	Items: IActionItems;
begin
	if FNeedCreateSubItems then
	begin
		FNeedCreateSubItems := False;
		if FActionItem <> nil then {Thanks to zzz.}
		begin
			Items := IActionItem(FActionItem).Items; {Ô}
			if Items <> nil then
			begin
				//Items.Update;
				AddActionItems(Items, Sub, OwnerDraw);
			end
		end
	end
end;

threadvar
	MenuFirst, MenuLast: TAnyMenu;

function GetLMenu(Menu: HMENU): TAnyMenu;
var
	I: TAnyMenu;
begin
	if not IsMenu(Menu) then
	begin
		Result := nil;
		Exit;
	end;

	I := MenuFirst;
	while I <> nil do
	begin
		if I.Handle = Menu then
		begin
			Result := I;
			Exit;
		end;

		I := I.FSuccItems;
	end;

	Result := nil;
end;

function GetMenu(Menu: HMENU): TAnyMenu;
begin
	if not IsMenu(Menu) then
		Result := nil
	else
	begin
		Result := GetLMenu(Menu);
		if Result = nil then
		begin
			Result := TAnyMenu.Create(Menu);
			Result.Owner := Garbage;
		end;
	end;
end;

function AssembleMenuText(const AText: WideString;
				AAccelerator: WideChar; const AShortCuts: WideString): WideString;
label
	Processed; {ǂȂcqR[hTCY100oCgȏ㌸}
var
	Temp: WideString;
	Index: Integer;
begin
	Temp := AText;

	if AAccelerator <> #0 then
	begin
		if ByteToCharLen(Temp, MaxInt) = Length(Temp) then
		begin
			Index := Pos(AAccelerator, Temp);
			if Index = 0 then Index := Pos(WideLoCase(AAccelerator), Temp);
			if Index > 0 then
			begin
				Insert('&', Temp, Index);
				goto Processed;
			end;
		end;

		Index := Length(Temp) + 1;
		while (Index > 2) and (Temp[Index - 1] = '.') do
			Dec(Index);
		Insert('(&' + AAccelerator + ')', Temp, Index);

Processed:
	end;

	if AShortCuts <> '' then
		Temp := Temp + #9 + AShortCuts;

	Result := Temp;
end;

procedure DisassembleMenuText(const Value: WideString; var AText: WideString;
				var AAccelerator: WideChar; var AShortCuts: WideString);
var
	I: Integer;
	Temp: WideString;
begin
	Temp := Value;

	AShortCuts := '';
	I := Pos(#9, Temp);
	if I > 0 then
	begin
		AShortCuts := Copy(Temp, I + 1, MaxInt);
		Delete(Temp, I, MaxInt);
	end;

	AAccelerator := #0;
	I := Pos('&', Temp); {Shift-JISȂv}
	if I > 0 then
	begin
		AAccelerator := Temp[I + 1];

		if (I > 1) and (Temp[I - 1] = '(') then
			Delete(Temp, I - 1, 4)
		else
			Delete(Temp, I, 1);
	end;
	AText := Temp;
end;

function CreateSubMenu(Item: TAnyMenuItem): TAnyMenu;
begin
	Result := TPopupMenu.Create;
	Result.Owner := Item;
	Item.Sub := Result
end;

function CreateSeparator(Menu: TAnyMenu): TMenuItem;
begin
	Result := TMenuItem.Create(Menu);
	Result.Separator := True;
	Result.OwnerDraw := InitialMenuOwnerDraw;
end;

function CreateMenuItem(Menu: TAnyMenu; const Text: WideString; Radio: Boolean = False): TMenuItem;
begin
	Result := TMenuItem.Create(Menu);
	Result.RadioCheck := Radio;
	Result.FullText := Text;
	Result.OwnerDraw := InitialMenuOwnerDraw;
end;

function CreateMenuItem(Menu: TAnyMenu; const Action: IAction; Accelerator: WideChar): TMenuItem; overload;
begin
	Result := TActionMenuItem.Create(Menu, Action);
	Result.Accelerator := Accelerator;
	Result.OwnerDraw := InitialMenuOwnerDraw;
end;

function SelectActionItemsPopup(const AItems: IActionItems; const ScreenPos: TPoint): IActionItem;
var
	PopupMenu: TPopupMenu;
	Cmd: UINT;
	Window: HWND;
	I: Integer;
	Msg: TMsg;
begin
	//AItems.Update;

	PopupMenu := TPopupMenu.Create;
	PopupMenu.Owner := Garbage;
	AddActionItems(AItems, PopupMenu, False);

	Window := GetFocus;
	PeekMessageW(Msg, Window, WM_CHAR, WM_CHAR, PM_REMOVE);

	Cmd := UINT(TrackPopupMenu(PopupMenu.Handle,
		TPM_RETURNCMD or TPM_LEFTBUTTON, ScreenPos.X, ScreenPos.Y, 0, Window, nil));

	Result := nil;
	for I := 0 to PopupMenu.Count - 1 do
		if PopupMenu.At[I].ID = Cmd then
		begin
			Result := IActionItem((PopupMenu.At[I] as TActionSubMenuItem).FActionItem);
			Break;
		end;
end;

{ TAnyMenuItem }

constructor TAnyMenuItem.Create(AMenu: TAnyMenu; AID: UINT);
begin
	inherited Create;
	FMenu:= AMenu;
	FID:= AID;
end;

procedure TAnyMenuItem.Execute;
begin
end;

function TAnyMenuItem.GetFullText: WideString;
var
	Info: MENUITEMINFOW;
	Buffer: array[0..1023] of WideChar;
begin
	Info.cbSize:= SizeOf(Info); //Win95ł44
	Info.fMask := MIIM_TYPE;

	Info.dwTypeData := Buffer;
	Info.cch := Length(Buffer);

	GetMenuItemInfoW(FMenu.Handle, ID, False, Info);
	if Info.fType and MFT_SEPARATOR = 0 then
		SetString(Result, Info.dwTypeData, Info.cch div SizeOf(WideChar));
end;

function TAnyMenuItem.GetSeparator: Boolean;
begin
	Result := GetType and MFT_SEPARATOR <> 0;
end;

function TAnyMenuItem.GetState(Flag: DWORD): Boolean;
var
	Info: MENUITEMINFOW;
begin
	Info.cbSize:= 44;
	Info.fMask := MIIM_STATE;
	GetMenuItemInfoW(FMenu.Handle, ID, False, Info);
	Result := Info.fState and Flag <> 0;
end;

function TAnyMenuItem.GetEnabled: Boolean;
begin
	Result := not GetState(MFS_DISABLED)
end;

function TAnyMenuItem.GetChecked: Boolean;
begin
	Result := GetState(MFS_CHECKED)
end;

function TAnyMenuItem.GetType: DWORD;
var
	Info: MENUITEMINFOW;
begin
	Info.cbSize:= SizeOf(Info);
	Info.fMask := MIIM_TYPE;
	Info.dwTypeData := nil;
	Info.cch := 0;

	GetMenuItemInfoW(FMenu.Handle, ID, False, Info);
	Result := Info.fType;
end;

function TAnyMenuItem.GetSub: TAnyMenu;
var
	Info: MENUITEMINFOW;
begin
	Info.cbSize := SizeOf(Info);
	Info.fMask := MIIM_SUBMENU;
	GetMenuItemInfoW(FMenu.Handle, ID, False, Info);
	Result := GetMenu(Info.hSubMenu);
end;

function TAnyMenuItem.GetAccelerator: WideChar;
var
	T, S: Widestring;
begin
	DisassembleMenuText(FullText, T, Result, S);
end;

function TAnyMenuItem.GetOwnerDraw: Boolean;
begin
	Result := GetType and MFT_OWNERDRAW <> 0;
end;

function TAnyMenuItem.GetRadioCheck: Boolean;
begin
	Result := GetType and MFT_RADIOCHECK <> 0;
end;

function TAnyMenuItem.GetShortCuts: WideString;
var
	T: WideString;
	A: WideChar;
begin
	DisassembleMenuText(FullText, T, A, Result);
end;

function TAnyMenuItem.GetText: WideString;
var
	S: WideString;
	A: WideChar;
begin
	DisassembleMenuText(FullText, Result, A, S);
end;

procedure TAnyMenuItem.Modify;
begin
	FMenu.Modify;
end;

procedure TAnyMenuItem.Selected;
begin
end;

procedure TAnyMenuItem.SetAccelerator(const Value: WideChar);
var
	T, S: WideString;
	A: WideChar;
begin
	DisassembleMenuText(FullText, T, A, S);
	FullText := AssembleMenuText(T, Value, S);
end;

procedure TAnyMenuItem.SetChecked(const Value: Boolean);
begin
	SetState(MFS_CHECKED, Value)
end;

procedure TAnyMenuItem.SetEnabled(const Value: Boolean);
begin
	SetState(MFS_DISABLED, not Value);
end;

procedure TAnyMenuItem.SetFullText(const Value: WideString);
var
	Info: MENUITEMINFOW;
begin
	if GetFullText <> Value then
	begin
		Info.cbSize:= SizeOf(Info); //Win95ł44
		Info.fMask := MIIM_TYPE;
		Info.fType := GetType;
		Info.dwTypeData := PWideChar(Value);
		Info.cch := Length(Value) * SizeOf(WideChar);
		SetMenuItemInfoW(FMenu.Handle, ID, False, Info);
		Modify;
	end
end;

procedure TAnyMenuItem.SetOwnerDraw(const Value: Boolean);
begin
	if GetOwnerDraw <> Value then
		if Value then
			SetType(0, MFT_OWNERDRAW)
		else
			SetType(MFT_OWNERDRAW, 0);
end;

procedure TAnyMenuItem.SetRadioCheck(const Value: Boolean);
begin
	if GetRadioCheck <> Value then
		if Value then
			SetType(0, MFT_RADIOCHECK)
		else
			SetType(MFT_RADIOCHECK, 0)
end;

procedure TAnyMenuItem.SetSeparator(const Value: Boolean);
begin
	if GetSeparator <> Value then
		if Value then
			SetType(MFT_STRING, MFT_SEPARATOR)
		else
			SetType(MFT_SEPARATOR, MFT_STRING)
end;

procedure TAnyMenuItem.SetShortCuts(const Value: WideString);
var
	T, S: WideString;
	A: WideChar;
begin
	DisassembleMenuText(FullText, T, A, S);
	FullText := AssembleMenuText(T, A, Value);
end;

procedure TAnyMenuItem.SetState(Flag: DWORD; Value: Boolean);
var
	Info: MENUITEMINFOW;
begin
	Info.cbSize:= 44;
	Info.fMask := MIIM_STATE;
	GetMenuItemInfoW(FMenu.Handle, ID, False, Info);

	if Value then
		Info.fState := Info.fState or Flag
	else
		Info.fState := Info.fState and not Flag;

	SetMenuItemInfoW(FMenu.Handle, ID, False, Info);
	Modify;
end;

procedure TAnyMenuItem.SetSub(const Value: TAnyMenu);
var
	Info: MENUITEMINFOW;
begin
	if Value <> GetSub then
	begin
		Info.cbSize := 44;
		Info.fMask := MIIM_SUBMENU;
		if Value <> nil then
			Info.hSubMenu := Value.Handle
		else
			Info.hSubMenu := 0;

		SetMenuItemInfoW(FMenu.Handle, ID, False, Info);
		Modify;
	end
end;

procedure TAnyMenuItem.SetText(const Value: WideString);
var
	T, S: WideString;
	A: WideChar;
begin
	DisassembleMenuText(FullText, T, A, S);
	FullText := AssembleMenuText(Value, A, S);
end;

procedure TAnyMenuItem.SetType(ClearFlags, SetFlags: DWORD);
var
	Info: MENUITEMINFOW;
	T: WideString;
begin
	Info.cbSize := SizeOf(Info);
	Info.fMask := MIIM_TYPE;

	Info.fType := GetType and not ClearFlags or SetFlags;

	T := FullText;
	Info.dwTypeData := PWideChar(T);
	Info.cch := Length(T) * SizeOf(WideChar);

	SetMenuItemInfoW(FMenu.Handle, ID, False, Info);
	Modify;
end;

procedure TAnyMenuItem.Update;
begin
end;

{ TAnyMenu }

procedure TAnyMenu.AfterConstruction;
begin
	inherited;
	AddToLinkList(MenuFirst, MenuLast, Self, FPredItems, FSuccItems);
end;

procedure TAnyMenu.BeforeDestruction;
var
	I: Integer;
begin
	//Lj[ڂ|
	for I := Count - 1 downto 0 do
		if At[I] is TMenuItem then
			At[I].Free;

	RemoveFromLinkList(MenuFirst, MenuLast, Self, FPredItems, FSuccItems);
	inherited;
end;

function TAnyMenu.EnumerateItems(Proc: CEachMenuItem; Context: Pointer): TAnyMenuItem;
var
	I: Integer;
begin
	for I := 0 to Count - 1 do
	begin
		Result := At[I];
		if Proc(Result, Context) then Exit;
	end;
	Result := nil
end;

constructor TAnyMenu.Create(AHandle: THandle);
begin
	inherited Create;
	FHandle:= AHandle;
end;

function TAnyMenu.Find(ID: UINT): TAnyMenuItem;
var
	I: Integer;
	Sub: TAnyMenu;
begin
	for I := Count - 1 downto 0 do
	begin
		Result := At[I];
		if Result.ID = ID then Exit;
		Sub := Result.Sub;
		if Sub <> nil then
		begin
			Result := Sub.Find(ID);
			if Result <> nil then Exit;
		end;
	end;
	Result := nil;
end;

function TAnyMenu.GetAt(AIndex: Integer): TAnyMenuItem;
var
	Info: MENUITEMINFOW;
	ValidPtr: Boolean;
	OrdExceptClass: TClass;

	function FindProc(Obj: T0Object): Boolean cdecl;
	begin
		Result := (Obj is TAnyMenuItem) and
						(TAnyMenuItem(Obj).Menu = Self) and
						(TAnyMenuItem(Obj).ID = Info.wID)
	end;

begin
	Info.cbSize := 44;
	Info.fMask := MIIM_DATA or MIIM_ID;
	Info.dwItemData := 0;

	if not Windows.GetMenuItemInfoW(FHandle, AIndex, True, Info) then
		RaiseLastOSError;

	Result := TAnyMenuItem( Info.dwItemData );

	{}
	OrdExceptClass := ExceptionClass;
	ExceptionClass := EAbort;
	try
		ValidPtr := TObject(Result) is TAnyMenuItem;
	except
		ValidPtr := False
	end;
	ExceptionClass := OrdExceptClass;

	{IuWFNgȂExternalMenuItem}
	if not ValidPtr then
	begin
		{S~}
		Result := TAnyMenuItem(Garbage.EnumerateObjects(CEachObject(@FindProc), GetContext));
		if Result <> nil then Exit;

		{΍쐬}
		Result := TAnyMenuItem.Create(Self, Info.wID);
		Result.Owner:= Garbage;
	end;
end;

function TAnyMenu.GetCount: Integer;
begin
	Result:= Windows.GetMenuItemCount(FHandle);
end;

function TAnyMenu.IndexOf(AItem: TAnyMenuItem): Integer;
var
	I: Integer;
begin
	for I:= 0 to Count - 1 do
		if Windows.GetMenuItemID(FHandle, I) = AItem.ID then
		begin
			Result:= I;
			Exit;
		end;

	Result:= -1;
end;

procedure TAnyMenu.Modify;
begin
end;

procedure TAnyMenu.Update;
var
	I: Integer;
begin
	for I := Count - 1 downto 0 do
		At[I].Update;
end;

{ TMenuItem }

constructor TMenuItem.Create(AMenu: TAnyMenu);
begin
	Assert(AMenu <> nil);
	inherited Create(AMenu, (CurrentThread as ICommandManager).RegisterCommand(Self));
end;

procedure TMenuItem.AfterConstruction; {override}
const
	MenuInfoTemplate: MENUITEMINFOW = (
					cbSize: SizeOf(MENUITEMINFOW);
					fMask: MIIM_CHECKMARKS or MIIM_DATA or MIIM_ID or MIIM_STATE or MIIM_SUBMENU or MIIM_TYPE;
					fType: MFT_STRING;
					fState: MFS_ENABLED or MFS_UNCHECKED;
					dwTypeData: '';
					cch: 0{SizeOf(WideChar)});
var
	NewPos: Integer;
	Info: MENUITEMINFOW;
{$IFOPT C+}
	R: Boolean;
{$ENDIF}
begin
	inherited;

	Info := MenuInfoTemplate;
	Info.wID := ID;
	Info.dwItemData:= Cardinal(Self);

	NewPos:= Windows.GetMenuItemCount(Menu.Handle);

	{$IFOPT C+} R := {$ENDIF} InsertMenuItemW(Menu.Handle, NewPos, True, Info);
	{$IFOPT C+} Assert(R); {$ENDIF}

	Modify;
end;

procedure TMenuItem.BeforeDestruction; {override}
begin
{$IFOPT C+}
	if not RemoveMenu(Menu.Handle, ID, MF_BYCOMMAND) then
		OutputDebugStringW('LW: RemoveMenu API Failed');
{$ELSE}
	RemoveMenu(Menu.Handle, ID, MF_BYCOMMAND);
{$ENDIF}

	Modify;

	(CurrentThread as ICommandManager).UnregisterCommand(ID);
	inherited;
end;

function TMenuItem.GetFullText: WideString;
begin
	Result := FFullText
end;

procedure TMenuItem.SetFullText(const Value: WideString);
begin
	inherited SetFullText(Value);
	FFullText := Value;
end;

{ TActionMenuItem }

constructor TActionMenuItem.Create(AMenu: TAnyMenu; AAction: IAction);
begin
	inherited Create(AMenu);
	FAction := AAction;
end;

procedure TActionMenuItem.Execute;
begin
	FAction.Execute;
end;

procedure TActionMenuItem.Update;
var
	SubMenu: TAnyMenu;
	SubItem: TAnyMenuItem;
	I: Integer;
begin
	//Text

	FullText := AssembleMenuText(FAction.Text, FAccelerator, MakeShortCuts);

	// Enabled

	if Enabled <> FAction.Enabled then
		Enabled := FAction.Enabled;

	// Checked

	if Checked <> FAction.Checked then
		Checked := FAction.Checked;

	//Sub
	SubMenu := Sub;

	if SubMenu <> nil then
		for I := SubMenu.Count - 1 downto 0 do
		begin
			SubItem := SubMenu.At[I];
			if SubItem is TActionSubMenuItem then
				SubItem.Free;
		end;

	if FAction.Items <> nil then
	begin
		FNeedCreateSubItems := True;
		if SubMenu = nil then
		begin
			SubMenu := TPopupMenu.Create;
			SubMenu.Owner := Self;
			Sub := SubMenu;
		end;
	end
	else if SubMenu <> nil then
	begin
		if (SubMenu.Count = 0) and (SubMenu.Owner = Self) then
		begin
			Sub := nil;
			SubMenu.Free;
		end;
	end;

	inherited Update;
end;

procedure TActionMenuItem.Selected;
var
	S: TAnyMenu;
begin
	if Enabled and FNeedCreateSubItems then
	begin
		FNeedCreateSubItems := False;

		//FAction.Items.Update;

		S := Sub;

		if S.Count > 0 then
		begin
			with TActionSubMenuItem.Create(S, nil) do
			begin
				Separator := True;
        OwnerDraw := Self.OwnerDraw
			end
		end;

		AddActionItems(FAction.Items, S, OwnerDraw);
	end;

	inherited Selected;
end;

function TActionMenuItem.MakeShortCuts: WideString;
var
	SI, KI: Byte;
	KM: IKeyManager;
begin
	KM := (CurrentThread as ICommandManager).ActiveKeyManager;
	Result := '';
	if KM <> nil then
		for SI := 0 to 7 do
			for KI := 0 to 255 do
				if KM.FindAction(TShiftKeys(SI), KI) = FAction then
				begin
					if Result <> '' then
						Result := Result + #32;

					if ssAlt in TShiftKeys(SI) then
						Result := Result + KeyNameText(VK_MENU) + '+';
					if ssControl in TShiftKeys(SI) then
						Result := Result + KeyNameText(VK_CONTROL) + '+';
					if ssShift in TShiftKeys(SI) then
						Result := Result + KeyNameText(VK_SHIFT) + '+';

					Result := Result + KeyNameText(KI);
				end;
end;

function TActionMenuItem.GetAccelerator: WideChar;
begin
	Result := FAccelerator
end;

procedure TActionMenuItem.SetAccelerator(const Value: WideChar);
begin
	FAccelerator := Value
end;

{ TPopupMenu }

constructor TPopupMenu.Create;
begin
	inherited Create(Windows.CreatePopupMenu);
end;

procedure TPopupMenu.BeforeDestruction;
begin
	inherited;
	Windows.DestroyMenu(Handle);
end;

{ TMainMenu }

constructor TMainMenu.Create;
begin
	inherited Create(Windows.CreateMenu);
end;

procedure TMainMenu.BeforeDestruction;
begin
	inherited;
	Windows.DestroyMenu(Handle);
end;

procedure TMainMenu.Modify;
begin
	//ύXL^Ă
	FModified:= True;
end;

function EnumWindowsProc_MenuUpdate(Wnd: HWND; Menu: HMENU): BOOL; stdcall;
{
	j[gpĂEBhẼXgĂ
	ĝh낤.
	邱Ƃ LMenu.pas  LWindows.pas zQƂȂčς.
}
begin
	if Windows.GetMenu(Wnd) = Menu then
		DrawMenuBar(Wnd);

	Result:= True;
end;

procedure TMainMenu.Update;
begin
	inherited Update;

	//XV
	if FModified then
	begin
		EnumWindows(@EnumWindowsProc_MenuUpdate, Handle);
		FModified := False;
	end;
end;

{ TKeyAssignment }

function TKeyAssignment.
				FindAction(const AShift: TShiftKeys; AKey: Integer): IAction;
var
	Current: TKeyAssignment;
	I: Integer;
begin
	Current := Self;
	repeat
		for I := Low(FAttached) to High(FAttached) do
		begin
			Result := FAttached[I].Table[AShift, AKey];
			if Result <> nil then
				Exit;
		end;

		Result := Table[AShift, AKey];
		if Result <> nil then
			Exit;

		Current := Current.Connection;
	until Current = nil;
end;

function TKeyAssignment.
				GetTable(const AShift: TShiftKeys; AKey: Integer): IAction;
begin
	Result := FTable[Byte(AShift), AKey];
end;

procedure TKeyAssignment.
				SetTable(const AShift: TShiftKeys; AKey: Integer; const Value: IAction);
begin
	FTable[Byte(AShift), AKey] := Value
end;

procedure TKeyAssignment.Attach(const AAssignment: IKeyAssignment);
begin
	SetLength(FAttached, Length(FAttached) + 1);
	FAttached[High(FAttached)] := AAssignment;
end;

procedure TKeyAssignment.Detach(const AAssignment: IKeyAssignment);
var
	I, J: Integer;
begin
	for I := Low(FAttached) to High(FAttached) do
		if FAttached[I] = AAssignment then
		begin
			for J := I to High(FAttached) - 1 do
				FAttached[J] := FAttached[J + 1];

			SetLength(FAttached, Length(FAttached) - 1);
			Break;
		end
end;

end.

