unit LMenusXP;

{*******************************************************}
{                                                       }
{       Light Windows                                   }
{       Menu Interface Windows XP - like                }
{                                                       }
{       2002..2004 YT                                   }
{                                                       }
{*******************************************************}

interface

uses
	Windows, Messages,
	SysUtilsW, AlgUtils, WinTips,
	LMenus, LGraphics, LActions, LWindows;

procedure ProcessMeasureItemXP(var Message: TWMMeasureItem);
procedure ProcessDrawItemXP(var Message: TWMDrawItem; Width: Integer = 0);
procedure ProcessMenuCharXP(var Message: TWMMenuChar);

var
	UseTheme: Boolean;

implementation

procedure ProcessMeasureItemXP(var Message: TWMMeasureItem);
var
	MenuItem: TMenuItem;
	Size: TSize;
	FT: WideString;
	I: Integer;
begin
	if Message.MeasureItemStruct.CtlType = ODT_MENU then
	begin
		MenuItem := GetLMenuItem(Message.MeasureItemStruct.itemID);
		if MenuItem <> nil then
		begin
			if MenuItem.Separator then
			begin
				Message.MeasureItemStruct.itemWidth := 1;
				Message.MeasureItemStruct.itemHeight := 5;
			end
			else
			begin
				FT := MenuItem.FullText;
				I := Pos('&', FT);
				if I > 0 then Delete(FT, I, 1);
				Size := GetFontSize(MenuFont.Handle, FT);
				if MenuItem.Menu is TPopupMenu then
				begin
					Inc(Size.cx, 24 + 16);
					Inc(Size.cy, 8);
					if Size.cy < 20 then
						Size.cy := 20;
				end;
				Message.MeasureItemStruct.itemWidth := Size.cx;
				Message.MeasureItemStruct.itemHeight := Size.cy;
			end;
			Message.Result := 1;
		end
	end;
end;

procedure ProcessDrawItemXP(var Message: TWMDrawItem; Width: Integer);
type
	TCheck = array[0..5] of TPoint;
const
	CheckX = 4;
	CheckY = 8;
	CheckSize = 2;
	CheckTemplate: TCheck = (
					(X: CheckX                 + 2; Y: CheckY                 + 2),
					(X: CheckX + CheckSize     + 2; Y: CheckY + CheckSize     + 2),
					(X: CheckX + CheckSize * 3 + 2; Y: CheckY - CheckSize     + 2),
					(X: CheckX + CheckSize * 3 + 2; Y: CheckY - CheckSize - 1 + 2),
					(X: CheckX + CheckSize     + 2; Y: CheckY + CheckSize - 1 + 2),
					(X: CheckX - 1             + 2; Y: CheckY - 2             + 2));

	BackColorKey: array[Boolean] of Integer = (COLOR_MENU, COLOR_MENUBAR);
var
	MenuItem: TAnyMenuItem;
	R, Space: TRect;
	T: WideString;
	DC: HDC;
	Pen: TPen;
	I, TabPos, VC: Integer;
	Check: TCheck;
	Menu: TAnyMenu;
	IsPopupMenu: Boolean;
	BorderColor, SelectColor, SelectColorDark: COLORREF;
	TextColor: array[Boolean] of COLORREF;
	BackBrush: array[Boolean] of HBRUSH;
	Brush: TBrush;
begin
	if Message.DrawItemStruct.CtlType = ODT_MENU then
	begin
		MenuItem := GetLMenuItem(Message.DrawItemStruct.itemID);
		if MenuItem <> nil then
		begin
			BorderColor := GetSysColor(COLOR_HIGHLIGHT);
			SelectColor := (BorderColor and $fcfcfc) div 4 + $bfbfbf;
			SelectColorDark := (SelectColor and $fefefe) div 2 + $3f3f3f;

			TextColor[False] := GetSysColor(COLOR_GRAYTEXT);
			TextColor[True] := GetSysColor(COLOR_MENUTEXT);

			BackBrush[False] := GetSysColorBrush(BackColorKey[UseTheme]);
			BackBrush[True] := GetSysColorBrush(COLOR_WINDOW);

			DC := Message.DrawItemStruct.hDC;
			SaveDC(DC);

			Brush := TBrush.Create;
			Brush.Owner := Garbage;

			Pen := TPen.Create;
			Pen.Owner := Garbage;

			Menu := MenuItem.Menu;
			IsPopupMenu := Menu is TPopupMenu;

			R := Message.DrawItemStruct.rcItem;

			if not IsPopupMenu and (Width > 0) and (Menu[Menu.Count - 1] = MenuItem) then
			begin
				Space := R;
				Space.Right := Width;
				FillRect(DC, Space, BackBrush[True]);
			end;

			if Message.DrawItemStruct.itemState and (ODS_SELECTED or ODS_FOCUS or ODS_HOTLIGHT) <> 0 then
			begin
				SelectObject(DC, BackBrush[True]);
				Pen.Color := TextColor[False];
				if IsPopupMenu or (Message.DrawItemStruct.itemState and ODS_HOTLIGHT <> 0) then
				begin
					Pen.Color := BorderColor;
					if MenuItem.Enabled then
					begin
						Brush.Color := SelectColor;
						SelectObject(DC, Brush.Handle);
					end	
				end;
				SelectObject(DC, Pen.Handle);
				Rectangle(DC, R.Left, R.Top, R.Right, R.Bottom);
			end
			else
			begin
				if IsPopupMenu then
				begin
					Space := R;
					Space.Right := Space.Left + 20;
					FillRect(DC, Space, BackBrush[False]);

					Inc(Space.Left, 20);
					Space.Right := R.Right;
					FillRect(DC, Space, BackBrush[True]);
				end
				else
				begin
					FillRect(DC, R, BackBrush[Width > 0]);
				end
			end;

			if MenuItem.Separator then
			begin
				Pen.Color := TextColor[False];
				SelectObject(DC, Pen.Handle);

				VC := (R.Top + R.Bottom) div 2;

				MoveToEx(DC, R.Left + (24 + 4), VC, nil);
				LineTo(DC, R.Right - 16, VC);
			end
			else
			begin
				SetBkMode(DC, TRANSPARENT);
				SetTextColor(DC, TextColor[MenuItem.Enabled]);

				T := MenuItem.FullText;

				if IsPopupMenu then
				begin
					Inc(R.Left, 24);
					Dec(R.Right, 16);

					TabPos := Pos(#9, T);
					if TabPos = 0 then TabPos := Length(T) + 1;

					DrawTextW(DC, Pointer(T), TabPos - 1, R, DT_LEFT or DT_VCENTER or DT_SINGLELINE);
					if TabPos <= Length(T) then
						DrawTextW(DC, PWideChar(Pointer(T)) + TabPos, Length(T) - TabPos, R, DT_RIGHT or DT_VCENTER or DT_SINGLELINE);

					if MenuItem.RadioCheck then
					begin
						R := Message.DrawItemStruct.rcItem;
						Inc(R.Left, 2);
						Inc(R.Top, 2);
						R.Right := R.Left + 16;
						R.Bottom := R.Top + 16;

						Brush.Color := SelectColorDark;
						if Message.DrawItemStruct.itemState and (ODS_SELECTED or ODS_FOCUS or ODS_HOTLIGHT) <> 0 then
						begin
							Pen.Color := BorderColor;
							SelectObject(DC, Brush.Handle);
							SelectObject(DC, Pen.Handle);
							Rectangle(DC, R.Left, R.Top, R.Right, R.Bottom)
						end
						else
							FillRect(DC, R, Brush.Handle);
					end;

					if MenuItem.Checked then
					begin
						Check := CheckTemplate;
						for I := High(Check) downto Low(Check) do Inc(Check[I].Y, R.Top);
						Pen.Color := 0;
						SelectObject(DC, Pen.Handle);
						Polyline(DC, Check, Length(Check));
					end;
				end
				else
					DrawTextW(DC, Pointer(T), Length(T), R, DT_CENTER or DT_VCENTER or DT_SINGLELINE);
			end;
			RestoreDC(DC, -1);
			Message.Result := 1;
		end
	end;
end;

procedure ProcessMenuCharXP(var Message: TWMMenuChar);
var
	Letter: WideChar;
	Menu: TAnyMenu;
	Item: TAnyMenuItem;
	I: Integer;
begin
	Menu := GetLMenu(Message.Menu);
	if Menu <> nil then
	begin
		Letter := WideUpCase(PWideChar(@Message.User)^);
		for I := Menu.Count - 1 downto 0 do
		begin
			Item := Menu.At[I];
			if WideUpCase(Item.Accelerator) = Letter then
			begin
				Message.Result := I or (2 shl 16);
				Break
			end;
		end;
	end;
end;

procedure SetUseTheme;
const
	ThemeKey = 'Software\Microsoft\Windows\CurrentVersion\ThemeManager';
	ThemeActive = 'ThemeActive';
var
	Key: HKEY;
	DataSize: DWORD;
	Data: array[0..7] of Char;
	R: Integer;
begin
	if RegOpenKeyEx(HKEY_CURRENT_USER, ThemeKey, 0, KEY_QUERY_VALUE, Key) = ERROR_SUCCESS then
	begin
		DataSize := SizeOf(Data);
		if RegQueryValueEx(Key, ThemeActive, nil,  nil, PByte(@Data), @DataSize) = ERROR_SUCCESS then
		begin
			R := DataSize - 2;
			UseTheme := (R > 0) or ((R = 0) and (Data[0] <> '0'));
		end;
		RegCloseKey(Key);
	end;
end;

initialization

	SetUseTheme

end.

