unit LStdCtls;

{*******************************************************}
{                                                       }
{       LightWindows                                    }
{       Standard Controls                               }
{                                                       }
{       1998..2004 YT                                   }
{                                                       }
{*******************************************************}

interface

uses
	Windows, Messages, SysUtils, Types,
	SysHacks, WinTips,
	LObjects, LActions, LWindows, LGraphics;

type
	{ListBox, ComboBox, TreeView, etc.. }
	T0ControlItem = class(T0Object)
	protected
		function GetData: Pointer; virtual; abstract;
		function GetText: WideString; virtual; abstract;
		procedure SetData(const Value: Pointer); virtual; abstract;
		procedure SetText(const Value: WideString); virtual; abstract;
	public
		property Text: WideString read GetText write SetText;
		property Data: Pointer read GetData write SetData;
	end;

	CEachItem = function(Item: T0ControlItem; Context: Pointer): Boolean cdecl;

type
	{Rg[荞. OƂ͋tɎgZ@̓X[p[NX}
	T0SubClassWindow = class(T0Window)
	private
		FOrgWndProc: Pointer;
		FBrush: TBrush;
	protected
		class procedure CreateClass(var WndClass: WNDCLASSEXW); override;
		class function SubClassName: PWideChar; virtual; abstract;
		class function WindowClassName: string; override;
		procedure ReleaseBrush;
		procedure UpdateBrush;
		procedure ColorChanged; override;
		procedure ProcessCtlColor(var Message: TWMCtlColor); override;
		procedure Initialize; override;
		procedure Finalize; override;
		procedure PaintWindow(DC: HDC); override;
	public
		procedure AfterConstruction; override;
		procedure DefaultHandler(var Message); override;
	end;

{BUTTON}
type
	T0ButtonControl = class(T0SubClassWindow)
	protected
		class function SubClassName: PWideChar; override;
		class function GetDefaultSize: TSize; override;
		procedure MnemonicKey(var Message: TWMKey); override;
		procedure Initialize; override;
		function GetChecked: Boolean;
		procedure SetChecked(const Value: Boolean);
		property Checked: Boolean read GetChecked write SetChecked;
	end;

type
	TButton = class(T0ButtonControl)
	private
		FIsDefault: Boolean;
	protected
		procedure Update; override;
	public
		procedure MnemonicKey(var Message: TWMKey); override;
		property IsDefault: Boolean read FIsDefault write FIsDefault;
		function OnClick(Event: PNMHDR): Boolean; inline;
	end;

	TCloseButton = class(TButton)
	protected
		procedure HandleNotification(Sender: TAnyWindow; Info: PNMHDR; var Result: LRESULT); override;
	end;

type
	TCheckBox = class(T0ButtonControl)
	private
		procedure WMSetText(var Message: TWMSetText); message WM_SETTEXT;
	protected
		class function GetDefaultSize: TSize; override;
		procedure Resize;
		procedure Initialize; override;
	public
		constructor Create(AParent: TAnyWindow;
						const AText: WideString; const APosition: TPoint; const ASize: TSize);
		property Checked;
	end;

type
	TRadioButton = class(T0ButtonControl)
	protected
		function KeyDown(ShiftState: TShiftKeys; Code: Byte): Boolean; override;
		class function GetDefaultSize: TSize; override;
	public
		constructor Create(AParent: TAnyWindow;
						const AText: WideString; const APosition: TPoint; const ASize: TSize);
		function CanFocus: Boolean; override;
		property Checked;
	end;

type
	TGroupBox = class(T0ButtonControl)
	private
		procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
	public
		constructor Create(AParent: TAnyWindow;
						const AText: WideString; const APosition: TPoint; const ASize: TSize);
	end;

{EDIT}
type
	T0EditControl = class(T0SubClassWindow)
	private
		function GetSelectedLength: Integer;
		procedure SetSelectedLength(const Value: Integer);
		procedure SetCaretIndex(const Value: Integer);
		function GetCaretIndex: Integer;
	protected
		class function SubClassName: PWideChar; override;
		class procedure CreateClass(var WndClass: WNDCLASSEXW); override;
		procedure Initialize; override;
		procedure SetReadOnly(AFlag: Boolean);
		function GetReadOnly: Boolean;
	public
		constructor Create(AParent: TAnyWindow;
						const AText: WideString; const APosition: TPoint; const ASize: TSize);
		procedure SelectAll;
		property ReadOnly: Boolean read GetReadOnly write SetReadOnly;
		property CaretIndex: Integer read GetCaretIndex write SetCaretIndex;
		property SelectedLength: Integer
						read GetSelectedLength write SetSelectedLength;
		property Color;
	end;

type
	TLineField = class(T0EditControl)
	private
		procedure WMKeyDown(var Message: TWMKeyDown); message WM_KEYDOWN;
	protected
		class function GetDefaultSize: TSize; override;
	end;

type
	TTextField = class(T0EditControl)
	private
		procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
	public
		constructor Create(AParent: TAnyWindow;
						const AText: WideString; const APosition: TPoint; const ASize: TSize);
	end;

{LISTBOX}

type
	TListBoxStyles = set of (
		lbsOwnerDraw  =  4 {LBS_OWNERDRAWFIXED = $10} ,
		lbsHasStrings =  6 {LBS_HASSTRINGS     = $40} ,
		_lbsHigh = 31);

	TListBoxItem = class;

	T0ListBoxControl = class(T0SubClassWindow)
	private
		procedure WMKeyDown(var Message: TWMKeyDown); message WM_KEYDOWN;
		function GetAt(AIndex: Integer): TListBoxItem;
		function GetCount: Integer;
		function GetSelectedIndex: Integer;
		procedure SetSelectedIndex(const Value: Integer);
		function GetCaretIndex: Integer;
		procedure SetCaretIndex(const Value: Integer);
	protected
		class function SubClassName: PWideChar; override;
		procedure Initialize; override;
		property SelectedIndex: Integer read GetSelectedIndex write SetSelectedIndex;
		property CaretIndex: Integer read GetCaretIndex write SetCaretIndex;
	public
		constructor Create(AParent: TAnyWindow; const APosition: TPoint; const ASize: TSize;
			const Styles: TListBoxStyles = []);
		function Enumerate(Proc: CEachItem; Context: Pointer): T0ControlItem;
		procedure Clear;
		procedure Remove(Item: T0ControlItem);
		function Add: T0ControlItem;
		function Insert(const Next: T0ControlItem): T0ControlItem;
		property At[AIndex: Integer]: TListBoxItem read GetAt;
		property Count: Integer read GetCount;
	end;

	TListBoxItem = class(T0ControlItem)
	private
		FListBox: T0ListBoxControl;
		FIndex: Integer;
	protected
		function GetData: Pointer; override;
		function GetText: WideString; override;
		procedure SetData(const Value: Pointer); override;
		procedure SetText(const Value: WideString); override;
	public
		constructor Create(AListBox: T0ListBoxControl; AIndex: Integer);
		property Index: Integer read FIndex;
	end;

type
	TListBox = class(T0ListBoxControl)
	public type
		TItem = TListBoxItem;
	private
		function GetItemHeight: Integer;
		function GetSelected: T0ControlItem;
		procedure SetSelected(const Value: T0ControlItem);
	public
		property SelectedIndex;
		property CaretIndex;
		property ItemHeight: Integer read GetItemHeight;
		property Selected: T0ControlItem read GetSelected write SetSelected;
	end;

{COMBOBOX}
type
	T0ComboBoxControl = class(T0SubClassWindow)
	private
		FEdit: HWND;
		FOrgEditProc: Pointer;
		FDropDownCount: Integer;
		function GetDroppedDown: Boolean;
		procedure SetDroppedDown(const Value: Boolean);
		function GetAt(AIndex: Integer): T0ControlItem;
		function GetSelectedIndex: Integer;
		procedure SetSelectedIndex(const Value: Integer);
		function GetSelected: T0ControlItem;
		procedure SetSelected(const Value: T0ControlItem);
	protected
		class function SubClassName: PWideChar; override;
		class function GetDefaultSize: TSize; override;
		procedure HandleNotification(Sender: TAnyWindow; Info: PNMHDR; var Result: LRESULT); override;
		procedure AdjustDropDown;
		procedure DoDropDown; virtual;
		procedure DoCloseUp; virtual;
		procedure DoSelected; virtual;
		function EditProc(Wnd: HWND; Msg: UINT; WP: WPARAM; LP: LPARAM): LRESULT; virtual;
		property EditHandle: HWND read FEdit;
		property SelectedIndex: Integer read GetSelectedIndex write SetSelectedIndex;
		procedure Initialize; override;
	public
		constructor Create(AParent: TAnyWindow;
			const AText: WideString; const APosition: TPoint; const ASize: TSize);
		function Enumerate(Proc: CEachItem; Context: Pointer): T0ControlItem;
		function ItemHeight: Integer;
		procedure Clear;
		function Count: Integer;
		function Add: T0ControlItem;
		function Insert(const Next: T0ControlItem): T0ControlItem;
		property At[AIndex: Integer]: T0ControlItem read GetAt;
		property DroppedDown: Boolean read GetDroppedDown write SetDroppedDown;
		property DropDownCount: Integer read FDropDownCount write FDropDownCount;
		property Selected: T0ControlItem read GetSelected write SetSelected;
	end;

	TComboBoxItem = class(T0ControlItem)
	private
		FComboBox: T0ComboBoxControl;
		FIndex: Integer;
	protected
		function GetData: Pointer; override;
		procedure SetData(const Value: Pointer); override;
		function GetText: WideString; override;
		procedure SetText(const Value: WideString); override;
	public
		constructor Create(AComboBox: T0ComboBoxControl; AIndex: Integer);
		property Index: Integer read FIndex;
	end;

type
	TComboBox = class(T0ComboBoxControl)
	private
		function GetCaretIndex: Integer;
		function GetSelectedLength: Integer;
		procedure SetCaretIndex(const Value: Integer);
		procedure SetSelectedLength(const Value: Integer);
		function GetSelectedText: WideString;
		procedure SetSelectedText(const Value: WideString);
	public
		constructor Create(AParent: TAnyWindow;
						const AText: WideString; const APosition: TPoint; const ASize: TSize);
		procedure SelectAll;
		property CaretIndex: Integer read GetCaretIndex write SetCaretIndex;
		property SelectedLength: Integer
						read GetSelectedLength write SetSelectedLength;
		property SelectedText: WideString read GetSelectedText write SetSelectedText;
	end;

type
	TDropDownListBox = class(T0ComboBoxControl)
	public
		constructor Create(AParent: TAnyWindow; const APosition: TPoint; const ASize: TSize);
		property SelectedIndex;
	end;

{STATIC}
type
	T0StaticControl = class(T0SubClassWindow)
	protected
		class function SubClassName: PWideChar; override;
	end;

	TStaticText = class(T0StaticControl)
	private
		procedure WMSetText(var Message: TWMSetText); message WM_SETTEXT;
	protected
		class function GetDefaultSize: TSize; override;
		procedure Initialize; override;
		procedure Resize;
	public
		constructor Create(AParent: TAnyWindow;
			const AText: WideString; const APosition: TPoint; const ASize: TSize);
	end;

{SCROLLBAR}
type
	TScrollBar = class(T0SubClassWindow)
	private
		FScroll: SCROLLINFO;
		FManager: PScrollBarManager;
		procedure WMVScroll(var Message: TWMScroll); message WM_VSCROLL;
		procedure WMHScroll(var Message: TWMScroll); message WM_HSCROLL;
		procedure SetMax(const Value: Integer);
		procedure SetMin(const Value: Integer);
		procedure SetPageSize(const Value: Cardinal);
		procedure SetValue(const Value: Integer);
	protected
		class function GetDefaultSize: TSize; override;
		class function SubClassName: PWideChar; override;
		property Manager: PScrollBarManager read FManager;
    procedure Initialize; override;
	public
		constructor Create(AParent: TAnyWindow;
						const APosition: TPoint; const ASize: TSize; AStyle: TBarStyle);
		procedure SetParams(Position, Min, Max, PageSize: Integer);
		property Min: Integer read FScroll.nMin write SetMin;
		property Max: Integer read FScroll.nMax write SetMax;
		property PageSize: Cardinal read FScroll.nPage write SetPageSize;
		property Value: Integer read FScroll.nPos write SetValue;
	end;

const
	DefaultColor: TColor = (SysColor: High(Integer));

function OneLineDefaultSize: TSize;

implementation

function OneLineDefaultSize: TSize;
begin
	Result.cx := 100;
	Result.cy := Abs(ForClientFont.Height) + 2 * GetSystemMetrics(SM_CYEDGE) + 4;
end;

{ T0SubClassWindow }

procedure T0SubClassWindow.AfterConstruction; {override}
var
	WndClass: WNDCLASSEXW;
begin
	Color := DefaultColor;

	GetClassInfoExW(0, SubClassName, WndClass);
	FOrgWndProc := WndClass.lpfnWndProc;

	inherited;
end;

procedure T0SubClassWindow.Initialize; {override}
begin
	inherited;
	UpdateBrush;
end;

procedure T0SubClassWindow.Finalize;
begin
	ReleaseBrush;
	inherited;
end;

class procedure T0SubClassWindow.CreateClass(var WndClass: WNDCLASSEXW);
var
	OrgClass: WNDCLASSEXW;
begin
	//ۑ
	OrgClass := WndClass;
	//{̃NX̏𓾂
	GetClassInfoExW(0, SubClassName, WndClass);
	//㏑
	WndClass.lpfnWndProc  := OrgClass.lpfnWndProc;
	WndClass.hInstance    := OrgClass.hInstance;
	WndClass.lpszClassName:= OrgClass.lpszClassName;
end;

procedure T0SubClassWindow.PaintWindow(DC: HDC); {override}
begin
	CallWindowProcW(FOrgWndProc, Handle, WM_PAINT, DC, 0);
end;

procedure T0SubClassWindow.ReleaseBrush;
begin
	if FBrush <> nil then
	begin
		FBrush.Free;
		FBrush := nil;
	end;
end;

procedure T0SubClassWindow.UpdateBrush;
var
	LB: TLogBrush;
begin
	ReleaseBrush;

	if Color.SysColor < 0 then
	begin
		LB.lbStyle := BS_SOLID;
		LB.lbColor := Color.Value;
		FBrush := TBrush.Create;
		FBrush.Assign(LB);
	end;
end;

procedure T0SubClassWindow.ProcessCtlColor(var Message: TWMCtlColor); {override}
begin
	if (Message.ChildWnd <> Handle) or (Color.SysColor = High(Integer)) then
		{eNX͎󂯂Ɛeɓ邽inheritedƖ[v}
		Message.Result := DefWindowProcW(Handle, Message.Msg,
						TMessage(Message).WParam, TMessage(Message).LParam)
	else
	begin
		SetBkColor(Message.ChildDC, Color.Value);

		if Color.SysColor < 0 then
			Message.Result := FBrush.Handle
		else
			Message.Result := GetSysColorBrush(Color.SysColor);
	end;	
end;

procedure T0SubClassWindow.DefaultHandler(var Message); {override}
begin
	TMessage(Message).Result:= CallWindowProcW(FOrgWndProc, Handle,
					TMessage(Message).Msg,
					TMessage(Message).WParam, TMessage(Message).LParam);
end;

procedure T0SubClassWindow.ColorChanged;
begin
	UpdateBrush;
end;

class function T0SubClassWindow.WindowClassName: String;
begin
	Result := inherited WindowClassName + ':' + SubClassName
end;

{ T0ButtonControl }

procedure T0ButtonControl.Initialize;
begin
	inherited;
	Font := ForLabelFont;
end;

class function T0ButtonControl.SubClassName: PWideChar;
begin
	Result := 'BUTTON';
end;

class function T0ButtonControl.GetDefaultSize: TSize;
begin
	Result.cx:= 75;
	Result.cy:= 25;
end;

procedure T0ButtonControl.MnemonicKey(var Message: TWMKey); {override}
begin
	if Enabled and Showing and (GetMnemonic(Text) = Message.CharCode) then
	begin
		SendMessageW(BM_CLICK, 0, 0);
		Message.Result := 1;
	end;
	{{^ChildƂ͖̂inherited͕sv}
end;

function T0ButtonControl.GetChecked: Boolean;
begin
	Result := SendMessageW(BM_GETCHECK, 0, 0) <> 0;
end;

procedure T0ButtonControl.SetChecked(const Value: Boolean);
begin
	{BST_UNCHECKED = 0, BST_CHECKED = 1 𗘗p}
	SendMessageW(BM_SETCHECK, Ord(Value), 0);
end;

{ TButton }

procedure TButton.MnemonicKey(var Message: TWMKey);
var
	Focused: HWND;
begin
	if (Message.CharCode = 13) and WindowStyle[BS_DEFPUSHBUTTON] and Enabled and Showing then
	begin
		Focused := GetFocus;
		SendMessageW(BM_CLICK, 0, 0);
		Windows.SetFocus(Focused);
		Message.Result := 1;
	end
	else
		inherited;
end;

function TButton.OnClick(Event: PNMHDR): Boolean;
begin
	Result := (Self <> nil) and (Event.hwndFrom = Handle)
  	and (Event.code = BN_CLICKED + NM_COMMAND)
end;

procedure TButton.Update;
var
	P: T0Window;
	D: Boolean;
begin
	P := Self;
	while not (P is T0TopWindow) and (P.Parent is T0Window) do
		P := T0Window(P.Parent);

	D := Focused or IsDefault and ((P.ActiveControl = nil) or
					(P.ActiveControl.SendMessageW(WM_GETDLGCODE,0,0) and
					DLGC_DEFPUSHBUTTON = 0));

	if D <> WindowStyle[BS_DEFPUSHBUTTON] then
	begin
		WindowStyle[BS_DEFPUSHBUTTON] := D;
		Repaint;
	end;

	inherited;
end;

{ TCloseButton }

procedure TCloseButton.HandleNotification(Sender: TAnyWindow; Info: PNMHDR; var Result: LRESULT);
begin
	inherited;
	if {(Sender = Self) and} (Info.code = NM_COMMAND + BN_CLICKED) then
	begin
    TopParent.Close;
	end;
end;

{ TCheckBox }

constructor TCheckBox.Create(AParent: TAnyWindow; const AText: WideString;
				const APosition: TPoint; const ASize: TSize);
var
	CS: PCREATESTRUCTW;
begin
	inherited Create(AParent, AText, APosition, ASize);

	CS := CreateStruct;
	CS^.style := CS^.style or BS_AUTOCHECKBOX;
end;

class function TCheckBox.GetDefaultSize: TSize;
begin
	Result := OneLineDefaultSize
end;

procedure TCheckBox.Initialize;
begin
  inherited;
	Resize;
end;

procedure TCheckBox.Resize;
var
	S: TSize;
	R: TRect;
begin
	R := BoundsRect;
	S := GetFontSize(Font.Handle, Text);
	R.Right := R.Left + S.cx + 32;
	BoundsRect := R;
end;

procedure TCheckBox.WMSetText(var Message: TWMSetText);
begin
	inherited;
	Resize
end;

{ TRadioButton }

function TRadioButton.CanFocus: Boolean;
begin
	Result := inherited CanFocus and Checked;
end;

constructor TRadioButton.Create(AParent: TAnyWindow; const AText: WideString;
	const APosition: TPoint; const ASize: TSize);
var
	CS: PCREATESTRUCTW;
begin
	inherited Create(AParent, AText, APosition, ASize);

	CS := CreateStruct;
	CS^.style := CS^.style or BS_AUTORADIOBUTTON;
end;

class function TRadioButton.GetDefaultSize: TSize;
begin
	Result := OneLineDefaultSize
end;

function TRadioButton.KeyDown(ShiftState: TShiftKeys; Code: Byte): Boolean;
var
	W: HWND;
	Wnd: TAnyWindow;
begin
	case Code of
	VK_UP:
		W := GetWindow(Handle, GW_HWNDPREV);
	VK_DOWN:
		W := GetWindow(Handle, GW_HWNDNEXT);
	else
		Result := inherited KeyDown(ShiftState, Code);
		Exit;
	end;
	Wnd := GetLWnd(W);
	if Wnd is TRadioButton then
	begin
		Checked := False;
		TRadioButton(Wnd).Checked := True;
		TRadioButton(Wnd).SetFocus;
	end;
	Result := True;
end;

{ TGroupBox }

constructor TGroupBox.Create(AParent: TAnyWindow; const AText: WideString;
	const APosition: TPoint; const ASize: TSize);
var
	CS: PCREATESTRUCTW;
begin
	inherited Create(AParent, AText, APosition, ASize);

	CS := CreateStruct;
	CS^.style := CS^.style and not WS_TABSTOP or BS_GROUPBOX;
	CS^.dwExStyle := CS^.dwExStyle or WS_EX_CONTROLPARENT;
end;

procedure TGroupBox.WMEraseBkgnd(var Message: TWMEraseBkgnd);
var
	Brush: HBRUSH;
begin
	Brush := SendMessageW(WM_CTLCOLORBTN, Message.DC, Handle);
	FillRect(Message.DC, ClientRect, Brush);
end;

{ T0EditControl }

constructor T0EditControl.Create(AParent: TAnyWindow;
				const AText: WideString; const APosition: TPoint; const ASize: TSize);
var
	CS: PCreateStructW;
begin
	inherited;

	CS := CreateStruct;
	CS.style    := CS.style     or (ES_AUTOHSCROLL or ES_LEFT);
	CS.dwExStyle:= CS.dwExStyle or WS_EX_CLIENTEDGE;
end;

class function T0EditControl.SubClassName: PWideChar;
begin
	Result:= 'EDIT';
end;

class procedure T0EditControl.CreateClass(var WndClass: WNDCLASSEXW);
begin
	inherited;
	WndClass.hCursor := LoadCursorW(0, PWideChar(IDC_IBEAM));
end;

procedure T0EditControl.SetReadOnly(AFlag: Boolean);
begin
	SendMessageW(EM_SETREADONLY, Ord(AFlag), 0);
end;

function T0EditControl.GetReadOnly: Boolean;
begin
	Result:= WindowStyle[ES_READONLY] 
end;

procedure T0EditControl.SelectAll;
begin
	SendMessageW(EM_SETSEL, 0, -1);
end;

function T0EditControl.GetSelectedLength: Integer;
var
	S, E: Integer;
begin
	SendMessageW(EM_GETSEL, WPARAM(@S), LPARAM(@E));
	Result := E - S;
end;

procedure T0EditControl.SetSelectedLength(const Value: Integer);
var
	C: Integer;
begin
	C := CaretIndex;
	SendMessageW(EM_SETSEL, C, C + Value);
end;

procedure T0EditControl.SetCaretIndex(const Value: Integer);
begin
	SendMessageW(EM_SETSEL, Value, Value);
end;

function T0EditControl.GetCaretIndex: Integer;
begin
	SendMessageW(EM_GETSEL, WPARAM(@Result), LPARAM(nil));
end;

{ TEditField }

class function TLineField.GetDefaultSize: TSize;
begin
	Result := OneLineDefaultSize
end;

procedure TLineField.WMKeyDown(var Message: TWMKeyDown); {message}
begin
	case Message.CharCode of
	VK_UP, VK_DOWN:
	begin
		Message.Result := 1;
		DialogKey(Message)
	end
	else
		inherited;
	end;
end;

procedure T0EditControl.Initialize;
begin
	inherited;
	Font := ForClientFont;
end;

{ TTextField }

constructor TTextField.Create(AParent: TAnyWindow; const AText: WideString;
	const APosition: TPoint; const ASize: TSize);
var
	CS: PCreateStructW;
begin
	inherited Create(AParent, AText, APosition, ASize);

	CS:= CreateStruct;
	CS.style := CS.style or (ES_MULTILINE or ES_WANTRETURN or
		WS_HSCROLL or WS_VSCROLL or ES_AUTOHSCROLL or ES_AUTOVSCROLL);
end;

procedure TTextField.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
	inherited;
	if not ReadOnly then
		Message.Result := Message.Result or DLGC_WANTTAB;
end;

{ T0ListBoxControl }

constructor T0ListBoxControl.Create(AParent: TAnyWindow;
	const APosition: TPoint; const ASize: TSize; const Styles: TListBoxStyles);
var
	CS: PCreateStructW;
begin
	inherited Create(AParent, '', APosition, ASize);

	CS:= CreateStruct;
	DWORD(CS.style) := DWORD(CS.style) or
		(WS_VSCROLL or LBS_USETABSTOPS or LBS_NOINTEGRALHEIGHT or LBS_NOTIFY) or
		DWORD(Styles);
	CS.dwExStyle:= CS.dwExStyle or WS_EX_CLIENTEDGE;
end;

procedure T0ListBoxControl.Initialize;
begin
	inherited;
	Font := ForClientFont;
end;

class function T0ListBoxControl.SubClassName: PWideChar;
begin
	Result:= 'LISTBOX';
end;

function T0ListBoxControl.GetAt(AIndex: Integer): TListBoxItem;

	function FindProc(Obj: T0Object): Boolean cdecl;
	begin
		Result := (Obj is TListBoxItem) and
						(TListBoxItem(Obj).FListBox = Self) and
						(TListBoxItem(Obj).FIndex = AIndex)
	end;

begin
	Assert((AIndex >= 0) and (AIndex < Count));

	{S~}
	Result := TListBoxItem(Garbage.EnumerateObjects(CEachObject(@FindProc), GetContext));
	if Result = nil then
	begin
		{S~ɖ΍쐬}
		Result := TListBoxItem.Create(Self, AIndex);
		Result.Owner := Garbage;
	end
end;

function T0ListBoxControl.Add: T0ControlItem;
begin
	Result := Insert(nil)
end;

procedure T0ListBoxControl.Clear;
begin
	SendMessageW(LB_RESETCONTENT, 0, 0)
end;

function T0ListBoxControl.Enumerate(Proc: CEachItem; Context: Pointer): T0ControlItem;
var
	I: Integer;
begin
	for I := 0 to Count - 1 do
	begin
		Result := GetAt(I);
		if Proc(Result, Context) then Exit;
	end;
	Result := nil;
end;

function T0ListBoxControl.GetCount: Integer;
begin
	Result:= SendMessageW(LB_GETCOUNT, 0, 0);
end;

function T0ListBoxControl.Insert(const Next: T0ControlItem): T0ControlItem;
const
	P: PWideChar = #0#0;
var
	Idx: Integer;
begin
	if Next = nil then
		Result := At[SendMessageW(LB_ADDSTRING, 0, LPARAM(P))]
	else
	begin
		Assert(Next is TListBoxItem);
		Assert(TListBoxItem(Next).FListBox = Self);

		Idx := TListBoxItem(Next).FIndex;
		Result := At[SendMessageW(LB_INSERTSTRING, Idx, LPARAM(P))];
	end
end;

procedure T0ListBoxControl.WMKeyDown(var Message: TWMKeyDown);
begin
	case Message.CharCode of
	VK_LEFT, VK_RIGHT:
	begin
		Message.Result := 1;
		DialogKey(Message);
	end
	else
		inherited;
	end;
end;

procedure T0ListBoxControl.Remove(Item: T0ControlItem);
begin
	Assert(Item is TListBoxItem);
	SendMessageW(LB_DELETESTRING, TListBoxItem(Item).FIndex, 0);
end;

function T0ListBoxControl.GetSelectedIndex: Integer;
begin
	Result := SendMessageW(LB_GETCURSEL, 0, 0); {LB_ERR = -1}
end;

procedure T0ListBoxControl.SetSelectedIndex(const Value: Integer);
begin
	SendMessageW(LB_SETCURSEL, Value, 0);
end;

function T0ListBoxControl.GetCaretIndex: Integer;
begin
	Result := SendMessageW(LB_GETCARETINDEX, 0, 0); {LB_ERR = -1}
end;

procedure T0ListBoxControl.SetCaretIndex(const Value: Integer);
begin
	SendMessageW(LB_SETCARETINDEX, Value, 0);
end;

{ TListBoxItem }

constructor TListBoxItem.Create(AListBox: T0ListBoxControl; AIndex: Integer);
begin
	inherited Create;
	FListBox := AListBox;
	FIndex := AIndex;
end;

function TListBoxItem.GetData: Pointer;
begin
	Result:= Pointer( FListBox.SendMessageW(LB_GETITEMDATA, FIndex, 0) );
end;

procedure TListBoxItem.SetData(const Value: Pointer);
begin
	FListBox.SendMessageW(LB_SETITEMDATA, FIndex, LPARAM(Value));
end;

function TListBoxItem.GetText: WideString;
var
	Len: Integer;
begin
	Len := FListBox.SendMessageW(LB_GETTEXTLEN, FIndex, 0);
	SetLength(Result, Len);
	if Len = 0 then Exit; { 0 O邱ƂŁPCharoRȂOK}
	FListBox.SendMessageW(LB_GETTEXT, FIndex, LPARAM(Result));
end;

procedure TListBoxItem.SetText(const Value: WideString);
var
	SaveData: Pointer;
	Selected, Caret: Boolean;
begin
	Caret := FListBox.CaretIndex = FIndex;
	Selected := FListBox.SelectedIndex = FIndex;

	SaveData := GetData;
	FListBox.SendMessageW(LB_DELETESTRING, FIndex, 0);
	FListBox.SendMessageW(LB_INSERTSTRING, FIndex, LPARAM(PWideChar(Value)));
	SetData(SaveData);

	if Selected then FListBox.SelectedIndex := FIndex;
	if Caret then FListBox.CaretIndex := FIndex;
end;

{ TListBox }

function TListBox.GetItemHeight: Integer;
begin
	Result := SendMessageW(LB_GETITEMHEIGHT, 0, 0)
end;

function TListBox.GetSelected: T0ControlItem;
var
	S: Integer;
begin
	S := SelectedIndex;
	if S >= 0 then
		Result := At[S]
	else
		Result := nil
end;

procedure TListBox.SetSelected(const Value: T0ControlItem);
begin
	if Value = nil then
		SelectedIndex := -1
	else
	begin
		Assert(Value is TListBoxItem);
		SelectedIndex := TListBoxItem(Value).Index;
	end
end;

{ T0ComboBoxControl }

function HookEditWindowProc(Wnd: HWND;
				Msg: UINT; WP: WParam; LP: LParam): LRESULT; stdcall;
var
	ParentWnd: T0Window;
begin
	ParentWnd := GetLWnd(GetParent(Wnd));
	Assert(ParentWnd is T0ComboBoxControl);
	Result := T0ComboBoxControl(ParentWnd).EditProc(Wnd, Msg, WP, LP)
end;

constructor T0ComboBoxControl.Create(AParent: TAnyWindow; const AText: WideString;
	const APosition: TPoint; const ASize: TSize);
var
	CS: PCreateStructW;
begin
	inherited Create(AParent, AText, APosition, ASize);

	FDropDownCount := 8;

	CS := CreateStruct;
	CS.style := CS.style or (CBS_AUTOHSCROLL or CBS_HASSTRINGS or WS_VSCROLL);
end;

function T0ComboBoxControl.EditProc(Wnd: HWND; Msg: UINT; WP: WPARAM; LP: LPARAM): LRESULT;
begin
	case Msg of
	WM_KEYDOWN, WM_KEYUP:
		if WP in [VK_RETURN, VK_ESCAPE, VK_TAB] then
		begin
			Result := 0;
			if not DroppedDown then Result := SendMessageW(Msg, WP, LP);
			Exit;
		end;
	WM_CHAR:
		if WP in [9, 13, 27] then
		begin
			if DroppedDown then
			begin
				DroppedDown := False;
				Result := 0;
			end
			else
				Result := SendMessageW(Msg, WP, LP);

			Exit;
		end;
	WM_SYSKEYDOWN, WM_SYSKEYUP, WM_SYSCHAR:
		if not DroppedDown then
		begin
			Result := SendMessageW(Msg, WP, LP);
			Exit;
		end;
	WM_SETFOCUS:
	begin
		SendMessageW(Msg, WP, LP);
	end;
	end;
	Result := CallWindowProcW(FOrgEditProc, Wnd, Msg, WP, LP);
end;

function T0ComboBoxControl.Count: Integer;
begin
	Result := SendMessageW(CB_GETCOUNT, 0, 0)
end;

function T0ComboBoxControl.Add: T0ControlItem;
const
	W: PWideChar = '';
begin
	Result := At[SendMessageW(CB_ADDSTRING, 0, LPARAM(W))];
end;

procedure T0ComboBoxControl.Clear;
begin
	while SendMessageW(CB_DELETESTRING, 0, 0) > 0 do
end;

function T0ComboBoxControl.GetAt(AIndex: Integer): T0ControlItem;

	function FindProc(Obj: T0Object): Boolean cdecl;
	begin
		Result := (Obj is TComboBoxItem) and
						(TComboBoxItem(Obj).FComboBox = Self) and
						(TComboBoxItem(Obj).FIndex = AIndex)
	end;

begin
	Assert((AIndex >= 0) and (AIndex < Count));

	{S~}
	Result := TComboBoxItem(Garbage.EnumerateObjects(CEachObject(@FindProc), GetContext));
	if Result <> nil then Exit;

	{S~ɖ΍쐬}
	Result := TComboBoxItem.Create(Self, AIndex);
	Result.Owner := Garbage;
end;

procedure T0ComboBoxControl.Initialize;
begin
  inherited;
	Font := ForClientFont;

	{EDITTuNX}
	FEdit := FindWindowExW(Handle, 0, 'EDIT', nil);
	FOrgEditProc := Pointer(GetWindowLongW(FEdit, GWL_WNDPROC));
	SetWindowLongW(FEdit, GWL_WNDPROC, LongInt(@HookEditWindowProc));
end;

function T0ComboBoxControl.Insert(const Next: T0ControlItem): T0ControlItem;
const
	W: PWideChar = '';
var
	Idx: Integer;
begin
	if Next = nil then
		Result := Add
	else
	begin
		Assert(Next is TComboBoxItem);
		Assert(TComboBoxItem(Next).FComboBox = Self);
		Idx := TComboBoxItem(Next).FIndex;
		Result := At[SendMessageW(CB_INSERTSTRING, Idx, LPARAM(W))];
	end
end;

function T0ComboBoxControl.ItemHeight: Integer;
begin
	Result := SendMessageW(CB_GETITEMHEIGHT, 0, 0)
end;

class function T0ComboBoxControl.SubClassName: PWideChar;
begin
	Result := 'COMBOBOX'
end;

function T0ComboBoxControl.GetDroppedDown: Boolean;
begin
	Result := LongBool(SendMessageW(CB_GETDROPPEDSTATE, 0, 0))
end;

class function T0ComboBoxControl.GetDefaultSize: TSize;
begin
	Result := OneLineDefaultSize
end;

procedure T0ComboBoxControl.SetDroppedDown(const Value: Boolean);
begin
	SendMessageW(CB_SHOWDROPDOWN, Ord(Value), 0);
end;

procedure T0ComboBoxControl.AdjustDropDown;
var
	ItemCount: Integer;
begin
	ItemCount := Count;
	if ItemCount > DropDownCount then ItemCount := DropDownCount;
	if ItemCount < 1 then ItemCount := 1;
//	FDroppingDown := True;
//	try
		SetWindowPos(Handle, 0, 0, 0,
						BoundsSize.cx, ItemHeight * ItemCount + BoundsSize.cy + 2,
						SWP_NOMOVE or SWP_NOZORDER or SWP_NOACTIVATE or
						SWP_NOREDRAW or SWP_HIDEWINDOW);
//	finally
//		FDroppingDown := False;
//	end;
	SetWindowPos(Handle, 0, 0, 0, 0, 0,
					SWP_NOMOVE or SWP_NOSIZE or SWP_NOZORDER or SWP_NOACTIVATE or
					SWP_NOREDRAW or SWP_SHOWWINDOW);
end;
{
function T0ComboBoxControl.Enumerator: IControlItemEnumerator;
begin
	Result := TControlItemEnumerator.Create(GetAt, Count)
end;
}
procedure T0ComboBoxControl.DoCloseUp;
begin
end;

procedure T0ComboBoxControl.DoDropDown;
begin
end;

procedure T0ComboBoxControl.DoSelected;
begin
end;

function T0ComboBoxControl.GetSelectedIndex: Integer;
begin
	Result := SendMessageW(CB_GETCURSEL, 0, 0)
end;

procedure T0ComboBoxControl.SetSelectedIndex(const Value: Integer);
begin
	SendMessageW(CB_SETCURSEL, Value, 0)
end;

procedure T0ComboBoxControl.HandleNotification(Sender: TAnyWindow; Info: PNMHDR; var Result: LRESULT);
begin
	case Info.code of
	CBN_DROPDOWN + NM_COMMAND:
		begin
			DoDropDown;
			AdjustDropDown;
			Exit;
		end;
	CBN_CLOSEUP + NM_COMMAND:
		DoCloseUp;
	CBN_SELCHANGE + NM_COMMAND:
		DoSelected;
	end;

	inherited;
end;

function T0ComboBoxControl.Enumerate(Proc: CEachItem; Context: Pointer): T0ControlItem;
var
	I: Integer;
begin
	for I := 0 to Count - 1 do
	begin
		Result := GetAt(I);
		if Proc(Result, Context) then Exit;
	end;
	Result := nil;
end;

function T0ComboBoxControl.GetSelected: T0ControlItem;
var
	S: Integer;
begin
	S := SelectedIndex;
	if S >= 0 then
		Result := At[S]
	else
		Result := nil
end;

procedure T0ComboBoxControl.SetSelected(const Value: T0ControlItem);
begin
	if Value = nil then
		SelectedIndex := -1
	else
	begin
		Assert(Value is TListBoxItem);
		SelectedIndex := TListBoxItem(Value).Index;
	end
end;

{ TComboBoxItem }

constructor TComboBoxItem.Create(AComboBox: T0ComboBoxControl; AIndex: Integer);
begin
	inherited Create;
	FComboBox := AComboBox;
	FIndex := AIndex;
end;

function TComboBoxItem.GetData: Pointer;
begin
	Result := Pointer(FComboBox.SendMessageW(CB_GETITEMDATA, FIndex, 0))
end;

function TComboBoxItem.GetText: WideString;
var
	Len: Integer;
begin
(*
	Len:= FComboBox.SendMessageA(CB_GETLBTEXTLEN, FIndex, 0);
	SetLength(Result, Len);
	if Len = 0 then Exit; { 0 O邱ƂŁPCharoRȂOK}
	FComboBox.SendMessageA(CB_GETLBTEXT, FIndex, LPARAM(Result));
*)
	Len:= FComboBox.SendMessageW(CB_GETLBTEXTLEN, FIndex, 0);
	SetLength(Result, Len);
	if Len = 0 then Exit; { 0 O邱ƂŁPWideCharoRȂOK}
	FComboBox.SendMessageW(CB_GETLBTEXT, FIndex, LPARAM(Result));
end;

procedure TComboBoxItem.SetData(const Value: Pointer);
begin
	FComboBox.SendMessageW(CB_SETITEMDATA, FIndex, LPARAM(Value))
end;

procedure TComboBoxItem.SetText(const Value: WideString);
var
	SaveData: Pointer;
begin
(*
	FComboBox.SendMessageA(CB_DELETESTRING, FIndex, 0);
	FComboBox.SendMessageA(CB_INSERTSTRING, FIndex, LPARAM(PChar(Value)));
*)
	SaveData := GetData;
	FComboBox.SendMessageW(CB_DELETESTRING, FIndex, 0);
	FComboBox.SendMessageW(CB_INSERTSTRING, FIndex, LPARAM(PWideChar(Value)));
	SetData(SaveData)
end;

{ TComboBox }

constructor TComboBox.Create(AParent: TAnyWindow; const AText: WideString;
				const APosition: TPoint; const ASize: TSize);
var
	CS: PCreateStructW;
begin
	inherited Create(AParent, AText, APosition, ASize);

	CS := CreateStruct;
	CS.style := CS.style or CBS_DROPDOWN;
end;

procedure TComboBox.SelectAll;
begin
	SendMessageW(CB_SETEDITSEL, 0, MakeLParam(0, $ffff))
end;

function TComboBox.GetCaretIndex: Integer;
begin
	SendMessageW(CB_GETEDITSEL, WPARAM(@Result), LPARAM(nil));
end;

function TComboBox.GetSelectedLength: Integer;
var
	S, E: Integer;
begin
	SendMessageW(CB_GETEDITSEL, WPARAM(@S), LPARAM(@E));
	Result := E - S;
end;

procedure TComboBox.SetCaretIndex(const Value: Integer);
var
	L: LPARAM;
begin
	LongRec(L).Lo := Value;
	LongRec(L).Hi := Value;
	SendMessageW(CB_SETEDITSEL, 0, L)
end;

procedure TComboBox.SetSelectedLength(const Value: Integer);
var
	L: LPARAM;
begin
	LongRec(L).Lo := CaretIndex;
	LongRec(L).Hi := LongRec(L).Lo + Value;
	SendMessageW(CB_SETEDITSEL, 0, L);
end;

function TComboBox.GetSelectedText: WideString;
var
	S, E: Integer;
begin
	SendMessageW(CB_GETEDITSEL, WPARAM(@S), LPARAM(@E));
	Result := Copy(Text, S + 1, E - S);
end;

procedure TComboBox.SetSelectedText(const Value: WideString);
var
	Edit: HWND;
begin
	Edit := FindWindowExW(Handle, 0, 'EDIT', nil);
	Windows.SendMessageW(Edit, EM_REPLACESEL, 1, LPARAM(PWideChar(Value)));
end;

{ TDropDownListBox }

constructor TDropDownListBox.Create(AParent: TAnyWindow;
				const APosition: TPoint; const ASize: TSize);
var
	CS: PCreateStructW;
begin
	inherited Create(AParent, '', APosition, ASize);

	CS := CreateStruct;
	CS.style := CS.style or CBS_DROPDOWNLIST;
end;

{ T0StaticControl }

class function T0StaticControl.SubClassName: PWideChar;
begin
	Result := 'STATIC'
end;

{ TStaticText }

constructor TStaticText.Create(AParent: TAnyWindow; const AText: WideString;
	const APosition: TPoint; const ASize: TSize);
var
	CS: PCreateStructW;
begin
	inherited Create(AParent, AText, APosition, ASize);
	CS := CreateStruct;
	CS^.style := CS^.style and not WS_TABSTOP;
end;

class function TStaticText.GetDefaultSize: TSize;
begin
	Result := OneLineDefaultSize
end;

procedure TStaticText.Initialize;
begin
	inherited;
	Font := ForLabelFont;
	Resize;
end;

procedure TStaticText.Resize;
var
	S: TSize;
	R: TRect;
begin
	R := BoundsRect;
	S := GetFontSize(Font.Handle, Text);
	R.Top := (R.Top + R.Bottom - S.cy) div 2;
	R.Right := R.Left + S.cx;
	R.Bottom := R.Top + S.cy;
	BoundsRect := R;
end;

procedure TStaticText.WMSetText(var Message: TWMSetText);
begin
	inherited;
	Resize;
end;

{ TScrollBar }

constructor TScrollBar.Create(AParent: TAnyWindow;
				const APosition: TPoint; const ASize: TSize; AStyle: TBarStyle);
const
	Styles: array[TBarStyle] of Integer = (SBS_HORZ, SBS_VERT);
var
	CS: PCreateStructW;
begin
	inherited Create(AParent, '', APosition, ASize);
	CS := CreateStruct;
	CS^.style := CS^.Style and not WS_TABSTOP or Styles[AStyle];

	FScroll.cbSize:= SizeOf(FScroll);
	FScroll.fMask := SIF_DISABLENOSCROLL or SIF_ALL;

	FManager := DefaultScrollBarManager;
end;

class function TScrollBar.GetDefaultSize: TSize;
begin
	Result.cx := GetSystemMetrics(SM_CXVSCROLL);
	Result.cy := GetSystemMetrics(SM_CYHSCROLL);
end;

procedure TScrollBar.Initialize;
begin
	inherited;
	Manager.Initialize(Handle)
end;

procedure TScrollBar.SetMax(const Value: Integer);
begin
	FScroll.nMax := Value;
	Manager.SetScrollInfo(Handle, SB_CTL, FScroll, True)
end;

procedure TScrollBar.SetMin(const Value: Integer);
begin
	FScroll.nMin := Value;
	Manager.SetScrollInfo(Handle, SB_CTL, FScroll, True)
end;

procedure TScrollBar.SetPageSize(const Value: Cardinal);
begin
	FScroll.nPage := Value;
	Manager.SetScrollInfo(Handle, SB_CTL, FScroll, True)
end;

procedure TScrollBar.SetParams(Position, Min, Max, PageSize: Integer);
begin
	FScroll.nPos := Position;
	FScroll.nMin := Min;
	FScroll.nMax := Max;
	FScroll.nPage := PageSize;
	Manager.SetScrollInfo(Handle, SB_CTL, FScroll, True)
end;

procedure TScrollBar.SetValue(const Value: Integer);
begin
	FScroll.nPos := Value;
	Manager.SetScrollInfo(Handle, SB_CTL, FScroll, True)
end;

class function TScrollBar.SubClassName: PWideChar;
begin
	Result := 'SCROLLBAR';
end;

procedure TScrollBar.WMHScroll(var Message: TWMScroll);
begin
	ProcessScroll(Message, Manager, SB_CTL, FScroll);
end;

procedure TScrollBar.WMVScroll(var Message: TWMScroll);
begin
	ProcessScroll(Message, Manager, SB_CTL, FScroll);
end;

end.

