unit LStreams;

{*******************************************************}
{                                                       }
{       Light Windows                                   }
{       Streams                                         }
{                                                       }
{       2000..2004 YT                                   }
{                                                       }
{*******************************************************}

{$WARN SYMBOL_PLATFORM OFF}

interface

uses
	Windows, SysUtils, SysConst,
	AlgUtils,
	LObjects, LConsts;

type
	TFileMode = Cardinal;

const
	fmCreate = High(TFileMode);

	soFromBeginning = 0;
	soFromCurrent   = 1;
	soFromEnd       = 2;

type
	T0Stream = class(T0Object)
	protected
		function GetReadable: Boolean; virtual;
		function GetWritable: Boolean; virtual;
		function GetSeekable: Boolean; virtual;
		function GetPosition: Cardinal; virtual;
		procedure SetPosition(Value: Cardinal); virtual;
		function GetSize: Cardinal; virtual;
		procedure SetSize(Value: Cardinal); virtual;
		function GetRest: Cardinal; virtual;
	public
		function Read(var Buffer; Count: Cardinal): Cardinal; virtual;
		function Refer(var Buffer; Count: Cardinal): Cardinal; virtual;
		function Write(const Buffer; Count: Cardinal): Cardinal; virtual;

		function CopyFrom(Source: T0Stream; Count: Cardinal = High(Cardinal)): Cardinal;

		property Readable: Boolean read GetReadable;
		property Writable: Boolean read GetWritable;
		property Seekable: Boolean read GetSeekable;

		property Position: Cardinal read GetPosition write SetPosition;
		property Size: Cardinal read GetSize write SetSize;
		property Rest: Cardinal read GetRest;
	end;

type
	T0MemoryStream = class(T0Stream)
	private
		FMemory: Pointer;
		FSize: Cardinal;
		FPosition: Cardinal;
	protected
		function GetReadable: Boolean; override;
		function GetSeekable: Boolean; override;
		function GetPosition: Cardinal; override;
		procedure SetPosition(Value: Cardinal); override;
		function GetSize: Cardinal; override;
		function GetRest: Cardinal; override;

		procedure SetPointer(AMemory: Pointer; ASize: Cardinal);
	public
		function Read(var Buffer; Count: Cardinal): Cardinal; override;
		function Refer(var Buffer; Count: Cardinal): Cardinal; override;

		property Memory: Pointer read FMemory;
	end;

//obt@
type
	TMemoryStream = class(T0MemoryStream)
	private
		FCapacity: Cardinal;
		procedure SetCapacity(Value: Cardinal);
	protected
		function GetWritable: Boolean; override;
		procedure SetSize(Value: Cardinal); override;
	public
		destructor Destroy; override;

		function Write(const Buffer; Count: Cardinal): Cardinal; override;

		property Capacity: Cardinal read FCapacity write SetCapacity;
	end;

//̃f[^Xg[ANZX
type
	TMemoryScanStream = class(T0MemoryStream)
	protected
		function GetWritable: Boolean; override;
		procedure SetSize(Value: Cardinal); override;
	public
		constructor Create(AMemory: Pointer; ASize: Cardinal);

		function Write(const Buffer; Count: Cardinal): Cardinal; override;
	end;

//\[X
type
	TResourceStream = class(T0MemoryStream)
	public
		constructor Create(Instance: HINST; const ResName: WideString; ResType: PWideChar);
	end;

//OƂ̂Ƃ
type
	T0IOStream = class(T0Stream)
	private
		FBuffer: Pointer;
		FBufferSize: Cardinal;
		FBufferRest: Cardinal;
	protected
		function GetPosition: Cardinal; override;
		procedure SetPosition(Value: Cardinal); override;

		//3̊֐͓pȂ̂ŊOĂ΂Ȃ
		function Input(var Buffer; Count: Cardinal): Cardinal; virtual; abstract;
		function Output(const Buffer; Count: Cardinal): Cardinal; virtual; abstract;
		function Seek(Count: Cardinal; Mode: Integer): Cardinal; virtual;
	public
		procedure BeforeDestruction; override;
		function Read(var Buffer; Count: Cardinal): Cardinal; override;
		function Refer(var Buffer; Count: Cardinal): Cardinal; override;
		function Write(const Buffer; Count: Cardinal): Cardinal; override;
	end;

type
	T0HandleStream = class(T0IOStream)
	private
		FHandle: THandle;
	protected
		function Input(var Buffer; Count: Cardinal): Cardinal; override;
		function Output(const Buffer; Count: Cardinal): Cardinal; override;
		procedure SetHandle(AHandle: THandle);
	public
		property Handle: THandle read FHandle;
	end;

type
	TFileStream = class(T0HandleStream)
	private
		FAccessMode: LongWord;
		function GetLastWrite: FILETIME;
    function GetAttribute: DWORD;
	protected
		function GetReadable: Boolean; override;
		function GetWritable: Boolean; override;
		function GetSeekable: Boolean; override;
		function GetSize: Cardinal; override;
		procedure SetSize(Value: Cardinal); override;
		function Seek(Count: Cardinal; Mode: Integer): Cardinal; override;
	public
		constructor Create(const FileName: WideString; Mode: TFileMode); overload;
		constructor Create(const FileName: WideString; DesiredAccess, SharedMode, CreateMode: DWORD); overload;
		procedure BeforeDestruction; override;
		property LastWrite: FILETIME read GetLastWrite;
		property Attribute: DWORD read GetAttribute;
	end;

type
	EStreamError = class(Exception);
	EReadError = class(EStreamError);
	EWriteError = class(EStreamError);
	ESeekError = class(EStreamError);

procedure WriteLineToStream(S: T0Stream; const T: UTF8String = '');
function ReadLineFromStream(S: T0Stream): UTF8String;

implementation

procedure WriteLineToStream(S: T0Stream; const T: UTF8String);
begin
	S.Write(Pointer(T)^, Length(T));
	S.Write(SLineBreak, Length(SLineBreak));
end;

function ReadLineFromStream(S: T0Stream): UTF8String;
const
	BufSize = 4096;
var
	L, R: Integer;
	U: Cardinal;
begin
	U := 0;
	repeat
		SetLength(Result, U + BufSize);
		L := S.Refer(Pointer(Result)^, BufSize);
		if L = 0 then
		begin
			SetLength(Result, U);
			Break;
		end;
		R := Pos(SLineBreak, Result) + Length(SLineBreak) - 1;
		if R < 0 then
		begin
			R := L - Length(SLineBreak) + 1;
			S.Read(Pointer(Cardinal(Result) + U)^, R);
			Inc(U, R);
		end
		else
		begin
			S.Read(Pointer(Cardinal(Result) + U)^, R);
			Inc(U, R - Length(SLineBreak));
			SetLength(Result, U);
			Break;
		end;
	until False;
end;

{ T0Stream }

function T0Stream.GetReadable: Boolean;
begin
	Result:= False;
end;

function T0Stream.GetWritable: Boolean;
begin
	Result:= False;
end;

function T0Stream.GetSeekable: Boolean;
begin
	Result:= False;
end;

function T0Stream.GetPosition: Cardinal;
asm
	call System.@AbstractError
end;

procedure T0Stream.SetPosition(Value: Cardinal);
asm
	call System.@AbstractError
end;

function T0Stream.GetSize: Cardinal;
asm
	call System.@AbstractError
end;

procedure T0Stream.SetSize(Value: Cardinal);
asm
	call System.@AbstractError
end;

function T0Stream.GetRest: Cardinal;
begin
	Result:= Size - Position;
end;

function T0Stream.Read(var Buffer; Count: Cardinal): Cardinal;
asm
	call System.@AbstractError
end;

function T0Stream.Refer(var Buffer; Count: Cardinal): Cardinal;
asm
	call System.@AbstractError
end;

function T0Stream.Write(const Buffer; Count: Cardinal): Cardinal;
asm
	call System.@AbstractError
end;

function T0Stream.CopyFrom(Source: T0Stream; Count: Cardinal): Cardinal;
const
	MaxBufSize = $10000; {64 K Byte}
var
	MoveSize: Cardinal;
	BufSize: Cardinal;
	Buffer: PChar;
begin
	if Count > MaxBufSize then {Count = MaxInt ̏ꍇ܂}
		BufSize := MaxBufSize
	else
		BufSize := Count;

	GetMem(Buffer, BufSize);
	try
		Result := 0;
		repeat
			MoveSize := Source.Read(Buffer^, BufSize);
			if MoveSize = 0 then Break;
			if Write(Buffer^, MoveSize) <> MoveSize then
				raise EWriteError.Create(SWriteFailed);
			Inc(Result, MoveSize);
			if (Count <> High(Cardinal)) and (Count - Result < BufSize) then
			begin
				BufSize := Count - Result;
				if BufSize = 0 then Break;
			end;
		until False;
	finally
		FreeMem(Buffer);
	end;
end;

{ T0MemoryStream }

procedure T0MemoryStream.SetPointer(AMemory: Pointer; ASize: Cardinal);
begin
	FMemory:= AMemory;
	FSize:= ASize;
end;

function T0MemoryStream.GetReadable: Boolean; {override}
begin
	Result:= True;
end;

function T0MemoryStream.GetSeekable: Boolean; {override}
begin
	Result:= True;
end;

function T0MemoryStream.GetPosition: Cardinal; {override}
begin
	Result:= FPosition;
end;

procedure T0MemoryStream.SetPosition(Value: Cardinal); {override}
begin
	// FPosition = Length ̈ʒu EOF Ȃ̂ŋ
	if Value > FSize then
		raise ESeekError.Create(SSeekFailed);

	FPosition:= Value;
end;

function T0MemoryStream.GetSize: Cardinal; {override}
begin
	Result:= FSize;
end;

function T0MemoryStream.GetRest: Cardinal; {override}
begin
	Result:= FSize - FPosition;
end;

function T0MemoryStream.Read(var Buffer; Count: Cardinal): Cardinal; {override}
begin
	Result := Refer(Buffer, Count); {蔲}
	Inc(FPosition, Result);
end;

function T0MemoryStream.Refer(var Buffer; Count: Cardinal): Cardinal; {override}
begin
	Assert(FPosition <= FSize);

	Result:= FSize - FPosition;
	if Result > Count then Result:= Count;

	Move(Pointer(Cardinal(FMemory) + FPosition)^, Buffer, Result);
end;

{ TMemoryStream }

destructor TMemoryStream.Destroy; {override}
begin
	if FMemory <> Nil then
		GlobalFreePtr(FMemory);

	inherited;
end;

function TMemoryStream.GetWritable: Boolean; {override}
begin
	Result:= True;
end;

procedure TMemoryStream.SetSize(Value: Cardinal); {override}
begin
	if Value > FCapacity then
		try
			SetCapacity(Value);
		except
			raise EWriteError.Create(SSetEofFailed); //ϊ
		end;

	FSize:= Value;
	if FPosition > Value then FPosition:= Value;
end;

procedure TMemoryStream.SetCapacity(Value: Cardinal);
const
	MemoryDelta = $2000; { Must be a power of 2 }
var
	NewMem: Pointer;
begin
	Assert(Value >= FSize); //OقȁH

	//Jグ...
	Value:= (Value + (MemoryDelta - 1)) and not (MemoryDelta - 1);

	if Value <> FCapacity then
	begin
		if Value = 0 then
			begin GlobalFreePtr(Memory); NewMem:= nil; end
		else
		begin
			if FMemory = nil then
				NewMem:= GlobalAllocPtr(HeapAllocFlags, Value)
			else
				NewMem:= GlobalReallocPtr(FMemory, Value, HeapAllocFlags);

			if NewMem = nil then RunError(203);
		end;
		FMemory:= NewMem;
	end;
	FCapacity:= Value;
end;

function TMemoryStream.Write(const Buffer; Count: Cardinal): Cardinal; {override}
var
	NewPos: Cardinal;
begin
	NewPos:= FPosition + Count;
	if NewPos > FSize then SetSize(NewPos);

	Move(Buffer, Pointer(Cardinal(FMemory) + FPosition)^, Count);

	FPosition:= NewPos;
	Result:= Count;
end;

{ TMemoryScanStream }

constructor TMemoryScanStream.Create(AMemory: Pointer; ASize: Cardinal);
begin
	inherited Create;
	SetPointer(AMemory, ASize);
end;

function TMemoryScanStream.GetWritable: Boolean;
begin
	Result:= True;
end;

procedure TMemoryScanStream.SetSize(Value: Cardinal);
begin
	FSize:= Value;
end;

function TMemoryScanStream.Write(const Buffer; Count: Cardinal): Cardinal;
{ Size ̎g͂Ȃ }
begin
	Result:= FSize - FPosition;
	if Count < Result then Result:= Count;

	Move(Buffer, Pointer(Cardinal(FMemory) + FPosition)^, Result);
	Inc(FPosition, Result);
end;

{ TResourceStream }

constructor TResourceStream.Create(Instance: HINST;
	const ResName: WideString; ResType: PWideChar);
var
	ResID: HRSRC;
	ResGlobal: HGLOBAL;
	Memory: Pointer;
begin
	inherited Create;
	{Obtains the resource ID}
	ResID := FindResourceW(hInstance, PWideChar(ResName), ResType);
	if ResID = 0 then RaiseLastOSError;
	{Obtains memory and size}
	ResGlobal := LoadResource(hInstance, ResID);
	Memory := LockResource(ResGlobal);
	if (ResGlobal = 0) or (Memory = nil) then RaiseLastOSError;
	SetPointer(Memory, SizeOfResource(hInstance, ResID));
end;

{ T0IOStream }

procedure T0IOStream.BeforeDestruction;
begin
	if FBuffer <> nil then FreeMem(FBuffer);
	inherited;
end;

function T0IOStream.GetPosition: Cardinal;
begin
	Result := Seek(0, soFromCurrent) - FBufferRest;
end;

procedure T0IOStream.SetPosition(Value: Cardinal); {override}
begin
	Seek(Value, soFromBeginning);
	FBufferSize:= 0;
	FBufferRest:= 0;
end;

function T0IOStream.Read(var Buffer; Count: Cardinal): Cardinal; {override}
begin
	Assert(Readable);

	//Result:= Min(FBufferRest, Count)
	Result:= FBufferRest;
	if Result > Count then Result:= Count;

	//obt@Rs[
	Move(Pointer(Cardinal(FBuffer) + FBufferSize - FBufferRest)^, Buffer, Result);
	Dec(FBufferRest, Result);

	//cۂɓǂ
	Dec(Count, Result);
	if Count > 0 then
		Inc(Result, Input(Pointer(Cardinal(@Buffer) + Result)^, Count));
end;

function T0IOStream.Refer(var Buffer; Count: Cardinal): Cardinal; {override}
const
	NextSize = 2048; //must be 2^X
var
	AddSize, NewBufferSize, ReadSize: Cardinal;
begin
	Assert(Readable);

	Result:= Count;

	if Result > FBufferRest then
	begin
		//obt@łΏk
		if FBufferSize - FBufferRest > NextSize * 2 then
		begin
			Move(Pointer(Cardinal(FBuffer) + FBufferSize - FBufferRest)^,
							FBuffer^, FBufferRest);
			FBufferSize:= FBufferRest;
		end;

		//ǉvZ(Vǂݍޕ 2^X ̔{ɂȂ悤ɂ)
		AddSize:= (Count - FBufferRest + NextSize) and not (NextSize - 1);
		NewBufferSize:= FBufferSize + AddSize;

		//[ăAP[g
		ReallocMem(FBuffer, NewBufferSize);

		//ǉǂݍ
		ReadSize:= Input(Pointer(Cardinal(FBuffer) + FBufferSize - FBufferRest)^,
						AddSize);

		Inc(FBufferSize, ReadSize);
		Inc(FBufferRest, ReadSize);

		//[ǂݐ؂Ȃꍇ΍
		if Result >= FBufferRest then Result:= FBufferRest;
	end;

	Move(Pointer(Cardinal(FBuffer) + FBufferSize - FBufferRest)^, Buffer, Result);
end;

function T0IOStream.Write(const Buffer; Count: Cardinal): Cardinal; {override}
begin
	Assert(Writable);

	if FBufferRest <> 0 then
	begin
		Seek(-FBufferRest, soFromCurrent);
		FBufferSize:= 0;
		FBufferRest:= 0;
	end;

	Result:= Output(Buffer, Count);
end;

function T0IOStream.Seek(Count: Cardinal; Mode: Integer): Cardinal;
asm
	call System.@AbstractError
end;

{ T0HandleStream }

procedure T0HandleStream.SetHandle(AHandle: THandle);
begin
	FHandle:= AHandle;
end;

function T0HandleStream.Input(var Buffer; Count: Cardinal): Cardinal; {override}
begin
	if not ReadFile(FHandle, Buffer, Count, DWORD(Result), Nil) then
		raise EReadError.Create(SReadFailed);
end;

function T0HandleStream.Output(const Buffer; Count: Cardinal): Cardinal; {override}
begin
	if not WriteFile(FHandle, Buffer, Count, DWORD(Result), Nil) then
		raise EReadError.Create(SWriteFailed);
end;

{ TFileStream }

procedure TFileStream.BeforeDestruction;
begin
	if Handle <> INVALID_HANDLE_VALUE then
		CloseHandle(Handle);

	inherited;
end;

constructor TFileStream.Create(const FileName: WideString; Mode: TFileMode);
const
	AccessModes: array[0..3] of LongWord = (
					{fmOpenRead}      GENERIC_READ,
					{fmOpenWrite}     GENERIC_WRITE,
					{fmOpenReadWrite} GENERIC_READ or GENERIC_WRITE,
					{fmCreate}        GENERIC_READ or GENERIC_WRITE);
var
	DesiredAccess, SharedMode, CreateMode: LongWord;
begin
	{ANZX[h}
	DesiredAccess := AccessModes[Mode and 3];

	{L[h. fmSharedDenyWrite ݒ肷̂ʓ|Ȃ̂ŏɂ}
	SharedMode := 0;
	if FAccessMode and GENERIC_WRITE = 0 then SharedMode := FILE_SHARE_READ;

	{쐬[h}
	CreateMode := OPEN_EXISTING;
	if Mode = fmCreate then CreateMode := CREATE_ALWAYS;

  Create(FileName, DesiredAccess, SharedMode, CreateMode);
end;

constructor TFileStream.Create(const FileName: WideString; DesiredAccess, SharedMode, CreateMode: DWORD);
var
	H: THandle;
	E: EInOutError;
begin
	inherited Create;

	FAccessMode := DesiredAccess;

	{Xg[V[PVXL悩낤
	 Ȃ̌オ]߂ƂƏĂĂ邵}
	H := CreateFileW(PWideChar(FileName), DesiredAccess, SharedMode, nil, CreateMode,
					FILE_ATTRIBUTE_NORMAL or FILE_FLAG_SEQUENTIAL_SCAN, 0);

	SetHandle(H);
	if H = INVALID_HANDLE_VALUE then begin
		E := EInOutError.CreateFmt(SFileOpenFailed, [FileName]);
		E.ErrorCode:= GetLastError;
		raise E;
	end;
end;

function TFileStream.GetAttribute: DWORD;
var
	Info: BY_HANDLE_FILE_INFORMATION;
begin
	GetFileInformationByHandle(Handle, Info);
	Result := Info.dwFileAttributes;
end;

function TFileStream.GetLastWrite: FILETIME;
begin
	GetFileTime(Handle, nil, nil, @Result)
end;

function TFileStream.GetReadable: Boolean;
begin
	Result:= FAccessMode and GENERIC_READ <> 0;
end;

function TFileStream.GetSeekable: Boolean;
begin
	Result:= True;
end;

function TFileStream.GetSize: Cardinal;
begin
	Result:= GetFileSize(Handle, Nil);
end;

function TFileStream.GetWritable: Boolean;
begin
	Result:= FAccessMode and GENERIC_WRITE <> 0;
end;

function TFileStream.Seek(Count: Cardinal; Mode: Integer): Cardinal; {override}
begin
	Result := SetFilePointer(Handle, Count, nil, Mode);
	if Integer(Result) < 0 then
		raise ESeekError.Create(SSeekFailed);
end;

procedure TFileStream.SetSize(Value: Cardinal);
label
	Error;
var
	OrgPos: Cardinal;
begin
	OrgPos:= Position;

	//VCL ɂƂ̂Qsł炵
	if SetFilePointer(Handle, Value, Nil, FILE_BEGIN)= $ffffffff then goto Error;
	if not SetEndOfFile(Handle) then goto Error;

	if OrgPos > Size then OrgPos:= Size;
	Position:= OrgPos;

Exit;
Error:
	raise EWriteError.Create(SSetEofFailed);
end;

end.
