unit LWindows;

{*******************************************************}
{                                                       }
{       LightWindows                                    }
{       Windows GUI Application Framework               }
{                                                       }
{       1998..2006 YT                                   }
{                                                       }
{*******************************************************}

{$WARN SYMBOL_PLATFORM OFF}

interface

uses
	Windows, Messages, Types, SysUtils, Imm,
	SysUtilsW, SysHacks, AlgUtils, WinTips,
	LConsts, LObjects, LGraphics, LActions, LMenus;

type
	TAlignment = (alNone, alLeft, alCenter, alRight, alTop, alBottom);
	THorzAlignment = alLeft..alRight;
	TBevel = (bvNone, bvRaised, bvLowered);
	TBarStyle = (sbHorizontal, sbVertical);
	TBarStyles = set of TBarStyle;

	TApplicationSynchronizer = TAtomicMultiReadExclusiveWriteSynchronizer;
	TApplicationSynchronizerEntry = TApplicationSynchronizer.TEntry;
	PApplicationSynchronizerEntry = ^TApplicationSynchronizerEntry;

const
	UseDefault = Integer(CW_USEDEFAULT);
	ZeroPoint: TPoint = (X: 0; Y: 0);
	ZeroSize: TSize = (CX: 0; CY: 0);
	ZeroRect: TRect = (Left: 0; Top: 0; Right: 0; Bottom: 0);
	DefaultPoint: TPoint = (X: UseDefault; Y: UseDefault);
	DefaultSize: TSize = (CX: UseDefault; CY: UseDefault);
	WM_LIGHTWINDOWS = $BD00;
	LW_NULL     = 0;
	LW_ACTIVATE = 1; {T0IsolatedWindow}
	LW_READY    = 2; {TGraphicsScreen}
	LW_NCPAINT  = 3; {T0SplitWindow}
	WM_LW_UPDATEGUI = WM_LIGHTWINDOWS + 1;
	NM_COMMAND = High(Integer) - High(Word); {WM_COMMAND  WM_NOTIFY Mapping}

type
	TAnyWindow = class;

	CCheckWnd = function(Wnd: TAnyWindow; Context: Pointer): Boolean cdecl;

	TAnyWindow = class(T0Object)
	private
		FHandle: HWND;
		function GetText: WideString;
		procedure SetText(const Value: WideString);
		function GetEnabled: Boolean; inline;
		procedure SetEnabled(Value: Boolean);
		function GetFont: TAnyFont;
		procedure SetFont(Value: TAnyFont);
		function GetTabStop: Boolean; inline;
		function GetVisible: Boolean; inline;
		procedure SetVisible(const Value: Boolean);
		function GetParent: TAnyWindow;
		function GetClientRect: TRect;
		function GetClientSize: TSize;
		function GetPosition: TPoint;
		function GetBoundsRect: TRect;
		function GetBoundsSize: TSize;
		procedure SetBoundsRect(const Value: TRect);
		procedure SetBoundsSize(const Value: TSize);
		procedure SetClientSize(const Value: TSize);
		procedure SetPosition(const Value: TPoint);
		function GetWindowLong(AIndex: Integer): LongWord; inline;
		procedure SetWindowLong(AIndex: Integer; const Value: LongWord);
		function GetWindowStyle(AStyle: DWORD): Boolean; inline;
		procedure SetWindowStyle(AStyle: DWORD; const Value: Boolean);
		function GetWindowExStyle(AStyle: DWORD): Boolean; inline;
		procedure SetWindowExStyle(AStyle: DWORD; const Value: Boolean);
	protected
		function FindNextWindow(Current: TAnyWindow; GoForward: Boolean; Check: CCheckWnd; Context: Pointer): TAnyWindow;
	public
		constructor Create(AWnd: HWND);
		function CanFocus: Boolean; dynamic;
		function Focused: Boolean;
		procedure SetFocus;
		procedure Show;
		procedure Hide;
		function Showing: Boolean;
		function ScreenToClient(const P: TPoint): TPoint;
		function ClientToScreen(const P: TPoint): TPoint;
		procedure BringToFront(Wnd: TAnyWindow = nil);
		procedure SetBounds(X, Y, CX, CY: Integer);
		procedure Repaint; overload;
		procedure Repaint(const Rect: TRect); overload;
		procedure RepaintFrame;
		function SendMessageA(Msg: UINT; WP: WPARAM; LP: LPARAM): LRESULT; deprecated;
		function PostMessageA(Msg: UINT; WP: WPARAM; LP: LPARAM): Boolean; deprecated;
		function SendMessageW(Msg: UINT; WP: WPARAM; LP: LPARAM): LRESULT;
		function PostMessageW(Msg: UINT; WP: WPARAM; LP: LPARAM): Boolean;
		function PostedMessage(Msg: UINT): Boolean;
		property Handle: HWND read FHandle;
		property Parent: TAnyWindow read GetParent;
		property Text: WideString read GetText write SetText;
		property Enabled: Boolean read GetEnabled write SetEnabled;
		property Font: TAnyFont read GetFont write SetFont;
		property TabStop: Boolean read GetTabStop;
		property Visible: Boolean read GetVisible write SetVisible;
		property Position: TPoint read GetPosition write SetPosition;
		property BoundsSize: TSize read GetBoundsSize write SetBoundsSize;
		property BoundsRect: TRect read GetBoundsRect write SetBoundsRect;
		property ClientSize: TSize read GetClientSize write SetClientSize;
		property ClientRect: TRect read GetClientRect;
		property WindowLong[AIndex: Integer]: LongWord read GetWindowLong write SetWindowLong;
		property WindowStyle[AStyle: DWORD]: Boolean read GetWindowStyle write SetWindowStyle;
		property WindowExStyle[AStyle: DWORD]: Boolean read GetWindowExStyle write SetWindowExStyle;
	end;

{LWEBhE}

type
	T0LeadWindow = class;
	T0TopWindow = class;

	PWindowStruct = ^TWindowStruct;
	TWindowStruct = record
		CreateStruct: CREATESTRUCTW;
		Text: WideString; {lpszName ̎}
	end;

	TWindowOperating = set of (opInMenuLoop, opInMoving, opInTracking);

	T0Window = class(TAnyWindow)
	private
		FWindowStruct: PWindowStruct;
		FActiveChild: HWND;
		FFont: TFont;
		FColor: TColor;
		FDestroying: Boolean;
		FAlignment: TAlignment;
		procedure WMCreate(var Message: TWMCreate); message WM_CREATE;
		procedure WMDestroy(var Message: TWMDestroy); message WM_DESTROY;
		procedure WMNCDestroy(var Message: TWMNCDestroy); message WM_NCDESTROY;
		procedure WMCommand(var Message: TWMCommand); message WM_COMMAND;
		procedure WMNotify(var Message: TWMNotify); message WM_NOTIFY;
		procedure WMParentNotify(var Message: TWMParentNotify); message WM_PARENTNOTIFY;
		procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
		procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
		procedure WMSize(var Message: TWMSize); message WM_SIZE;
		procedure WMSetFont(var Message: TWMSetFont); message WM_SETFONT;
		procedure WMGetFont(var Message: TWMGetFont); message WM_GETFONT;
		procedure WMKeyDown(var Message: TWMKey); message WM_KEYDOWN;
		procedure WMSysKeyDown(var Message: TWMSysKeyDown); message WM_SYSKEYDOWN;
		procedure WMChar(var Message: TWMChar); message WM_CHAR;
		procedure WMSysChar(var Message: TWMChar); message WM_SYSCHAR;
		procedure WMKeyUp(var Message: TWMKey); message WM_KEYUP;
		procedure WMSysKeyUp(var Message: TWMSysKeyDown); message WM_SYSKEYUP;
		procedure WMVScroll(var Message: TWMScroll); message WM_VSCROLL;
		procedure WMHScroll(var Message: TWMScroll); message WM_HSCROLL;
		procedure WMSysColorChange(var Message: TWMSysColorChange); message WM_SYSCOLORCHANGE;
		procedure WMCtlColorStatic(var Message: TWMCtlColorStatic); message WM_CTLCOLORSTATIC;
		procedure WMCtlColorEdit(var Message: TWMCtlColorEdit); message WM_CTLCOLOREDIT;
		procedure WMCtlColorBtn(var Message: TWMCtlColorBtn); message WM_CTLCOLORBTN;
		procedure WMCtlColorListBox(var Message: TWMCtlColorListBox); message WM_CTLCOLORLISTBOX;
		procedure WMInitMenuPopup(var Message: TWMInitMenuPopup); message WM_INITMENUPOPUP;
		function GetActiveControl: T0Window;
		procedure SetActiveControl(const Value: T0Window);
		function GetFont: TAnyFont;
		procedure SetColor(const Value: TColor);
		function GetLParent: T0Window;
	protected
		procedure Initialize; virtual;
		procedure Finalize; virtual;
		function CreateStruct: PCreateStructW; inline;
		class procedure RegisterClass;
		class procedure CreateClass(var WndClass: WNDCLASSEXW); virtual;
		class function WindowClassName: string; dynamic;
		class function GetDefaultSize: TSize; virtual;
		function WindowProc(Msg: UINT; WP: WPARAM; LP: LPARAM): LRESULT;
		procedure Update; virtual;
		procedure HandleNotification(Sender: TAnyWindow; Info: PNMHDR; var Result: LRESULT); virtual;
		function Notify(Code: Integer): LRESULT;
		procedure FocusChanged; dynamic;
		procedure ColorChanged; dynamic;
		procedure DialogKey(var Message: TWMKey); dynamic;
		procedure MnemonicKey(var Message: TWMKey); dynamic;
		function KeyDown(ShiftState: TShiftKeys; Code: Byte): Boolean; virtual;
		function KeyUp(ShiftState: TShiftKeys; Code: Byte): Boolean; virtual;
		function KeyInput(Code: WideChar): Boolean; virtual;
		procedure PaintWindow(DC: HDC); virtual;
		procedure FontChanged; dynamic;
		procedure ProcessCtlColor(var Message: TWMCtlColor); dynamic;
		function ProcessScroll(var Message: TWMScroll; Manager: PScrollBarManager; Kind: Integer; var SI: SCROLLINFO): Integer;
		property Destroying: Boolean read FDestroying;
		property Color: TColor read FColor write SetColor;
		property LParent: T0Window read GetLParent;
	public
		constructor Create(AParent: TAnyWindow;
			const AText: WideString; const APosition: TPoint; const ASize: TSize);
		procedure AfterConstruction; override;
		procedure BeforeDestruction; override;
		procedure DefaultHandler(var Message); override;
		procedure Realign; virtual;
		function TopParent: T0TopWindow;
		property Alignment: TAlignment read FAlignment write FAlignment;
		property Font read GetFont;
		property ActiveControl: T0Window read GetActiveControl write SetActiveControl;
	end;

	{for MDI}
	IMDIClient = interface
		['{3EE69A28-E915-4501-9ACA-8EB4797FAE3D}']
		function ClientWindow: T0Window;
		procedure OnCreate(Child: T0Window);
		procedure OnDestroy(Child: T0Window);
		procedure OnActivate(Child: T0Window);
		procedure Next;
		procedure PreserveMenu;
		procedure Activate(Child: T0Window);
	end;

	{ڏoEBhE}

	T0TopWindow = class(T0Window)
	private
		FLeader: T0LeadWindow;
		FHint: WideString;
		FMenu: TMainMenu;
		FSelectedMenuItemID: UINT;
		FOperating: TWindowOperating;
		FTextLength: Integer; { MDI only }
		procedure WMClose(var Message: TWMClose); message WM_CLOSE;
		procedure WMQueryEndSession(var Message: TWMQueryEndSession); message WM_QUERYENDSESSION;
		procedure WMMenuSelect(var Message: TWMMenuSelect); message WM_MENUSELECT;
		procedure WMEnterMenuLoop(var Message); message WM_ENTERMENULOOP;
		procedure WMExitMenuLoop(var Message); message WM_EXITMENULOOP;
		procedure WMEnterSizeMove(var Message); message WM_ENTERSIZEMOVE;
		procedure WMExitSizeMove(var Message); message WM_EXITSIZEMOVE;
		procedure SetMenu(Value: TMainMenu);
		function GetStayOnTop: Boolean;
		procedure SetStayOnTop(const Value: Boolean);
	protected
		function CloseQuery: Boolean; virtual;
		procedure Update; override;
		function ReadyHint: WideString; virtual;
		procedure DialogKey(var Message: TWMKey); override;
		procedure Finalize; override;
		procedure DefaultMDIHandler(const Client: IMDIClient; var Message);
	public
		constructor Create(AParent: T0TopWindow;
			const AText: WideString; const APosition: TPoint; const ASize: TSize);
		procedure Close;
		property Leader: T0LeadWindow read FLeader;
		property Hint: WideString read FHint;
		function SelectedMenuItem: TAnyMenuItem;
		function IsForeground: Boolean; inline;
		procedure Foreground;
		procedure BeginTracking;
		procedure EndTracking;
		property Menu: TMainMenu read FMenu write SetMenu;
		property Operating: TWindowOperating read FOperating;
		property StayOnTop: Boolean read GetStayOnTop write SetStayOnTop;
	end;

	T0PopupWindow = class(T0TopWindow)
	public
		constructor Create(AParent: T0TopWindow;
			const AText: WideString; const APosition: TPoint; const ASize: TSize);
	end;

	T0IsolatedWindow = class(T0TopWindow)
	private
		procedure WMActivate(var Message: TWMActivate); message WM_ACTIVATE;
		procedure WMSetCursor(var Message: TWMSetCursor); message WM_SETCURSOR;
		procedure WMLightWindows(var Message: TMessage); message WM_LIGHTWINDOWS;
	protected
		class function GetDefaultSize: TSize; override;
	public
		constructor Create(AParent: T0TopWindow;
			const AText: WideString; const APosition: TPoint; const ASize: TSize);
	end;

	T0LeadWindow = class(T0TopWindow)
	private
		FTerminated: Boolean;
		procedure WMClose(var Message: TWMClose); message WM_CLOSE;
		procedure WMQuit(var Message: TWMQuit); message WM_QUIT;
		procedure WMQueryEndSession(var Message: TWMQueryEndSession);	message WM_QUERYENDSESSION;
		procedure WMEndSession(var Message: TWMEndSession); message WM_ENDSESSION;
		procedure WMEnable(var Message: TWMEnable); message WM_ENABLE;
	protected
		procedure ProcessIdle;
		procedure Setup; virtual;
	public
		procedure MessageLoop;
		procedure Update; override;
		property Terminated: Boolean read FTerminated;
	end;

	CQuery = function(Context: Pointer): Boolean cdecl;

	T0DialogWindow = class(T0LeadWindow)
	private
		FDialogResult: Integer;
		FContext: Pointer;
		FQuery: CQuery;
		FForegroundWindow: HWND; {êƂ̐e}
		FImeAllowed: Boolean;
		procedure WMCreate(var Message: TWMCreate); message WM_CREATE;
		procedure WMActivate(var Message: TWMActivate); message WM_ACTIVATE;
    procedure WMNCHitTest(var Message: TWMNCXHitMessage); message WM_NCHITTEST;
	protected
		class procedure CreateClass(var WndClass: WNDCLASSEXW); override;
		procedure Setup; override;
		property ImeAllowed: Boolean read FImeAllowed write FImeAllowed;
		procedure MnemonicKey(var Message: TWMKey); override;
		property OnQuery: CQuery read FQuery write FQuery;
		function Valid: Boolean; virtual;
		procedure PaintWindow(DC: HDC); override;
	public
		constructor Create(AParent: T0TopWindow; const AText: WideString);
		function Execute: Boolean; virtual;
		property DialogResult: Integer read FDialogResult write FDialogResult;
		property Context: Pointer read FContext write FContext;
	end;

	T0MainWindow = class(T0LeadWindow)
	protected
		class procedure CreateClass(var WndClass: WNDCLASSEXW); override;
		class function GetDefaultSize: TSize; override;
		procedure Setup; override;
	public
		constructor Create(const AText: WideString; const APosition: TPoint; const ASize: TSize);
	end;

type
	T0Application = class;

	T0Thread = class(T0Object, IInterface, IObservable)
	private
		FPredThread: T0Thread;
		FSuccThread: T0Thread;
		FObservableObject: TObservable;
		FObservable: IObservable;
		FId: DWORD;
		FHandle: THandle;
		FLeadWindow: T0LeadWindow;
		FInitializationEntry: PApplicationSynchronizerEntry;
	protected
		procedure Initialize; virtual;
		procedure Finalize; virtual;
		procedure Execute; virtual; abstract;
		procedure Run;
		procedure ProcessIdle;
		class procedure SetName(const Name: string);
	public
		constructor Create; overload;
		constructor Create(Application: T0Application); overload;
		procedure BeforeDestruction; override;
		property Id: DWORD read FId;
		property Handle: THandle read FHandle;
		property AsObservable: IObservable read FObservable implements IObservable;
		property LeadWindow: T0LeadWindow read FLeadWindow;
	end;

	T0GUIThread = class(T0Thread, ICommandManager)
	private
		FMainWindow: T0MainWindow;
		FCommands: array of TObject;
		FSystemFont: TAnyFont;
		FDefaultGUIFont: TAnyFont;
		FFixedFont: TAnyFont;
		FMenuFont: TAnyFont;
		FMessageFont: TAnyFont;
		FStatusFont: TAnyFont;
		FRendezvousWindow: HWND;
		{ICommandManager}
		function RegisterCommand(const AItem: TObject): UINT;
		procedure UnregisterCommand(const ACommand: UINT);
		function ActiveKeyManager: IKeyManager;
	protected
		procedure Execute; override;
		function Fire: DWORD; virtual;
		procedure WaitMessage(var Msg: TMsg);
		procedure ReadyRendezvous;
	public
		function PostMessageW(Msg: UINT; WP: WPARAM; LP: LPARAM): Boolean;
		procedure Rendezvous(Entry: PApplicationSynchronizerEntry);
		property MainWindow: T0MainWindow read FMainWindow;
	end;

	CEachThread = function(Thread: T0Thread; Context: Pointer): Boolean cdecl;

	T0Application = class(T0Object, IInterface, IObservable,
		IReadWriteSync, IReadWriteSyncEx)
	private
		FUpdateSynchronizer: TApplicationSynchronizer;
		FThreadMonitorSynchronizer: TCriticalSection;
		FThreadMonitorEvent: THandle;
		FMainThread: T0Thread;
		FFirstThread: T0Thread;
		FLastThread: T0Thread;
		FTitle: WideString;
		FObservableObject: TObservable;
		FObservable: IObservable;
	protected
		procedure RequestAllThreads;
		procedure ProcessIdle;
		procedure Initialize(AMainThread: T0Thread); overload;
		procedure MonitorThreads;
		procedure WaitLoop;
		property ThreadMonitorSynchronizer: TCriticalSection read FThreadMonitorSynchronizer;
	public
		procedure AfterConstruction; override;
		procedure BeforeDestruction; override;
		procedure Initialize; overload; virtual; abstract;
		procedure Run;
		function ThreadIdToObject(ThreadId: DWORD): T0Thread;
		function EnumerateThreads(Block: CEachThread; Context: Pointer): T0Thread;
		procedure BeginRead;
		procedure EndRead;
		function TryBeginWrite: Boolean;
		function BeginWrite: Boolean;
		procedure EndWrite;
		procedure Async;
		procedure RestoreAsync;
		property Title: WideString read FTitle write FTitle;
		property AsObservable: IObservable read FObservable implements IObservable;
		class function ExeName: WideString;
		class function ExePath: WideString;
	end;

var
	Application: T0Application;
	LWPropName: PWideChar;

threadvar
	CurrentThread: T0Thread;

function GetWnd(Wnd: HWND): TAnyWindow;
function GetLWnd(Wnd: HWND): T0Window;
function GetLWndThread(Wnd: HWND): T0GUIThread;
function GetLMenuItem(Command: UINT): TMenuItem;
function GetObjectFromCommand(Command: UINT): TObject;

function SystemFont: TAnyFont;
function DefaultGUIFont: TAnyFont;
function FixedFont: TAnyFont;
function MenuFont: TAnyFont;
function MessageFont: TAnyFont;
function StatusFont: TAnyFont;

type TFontFunc = function: TAnyFont;
var ForClientFont: TFontFunc = MessageFont;
var ForLabelFont: TFontFunc = StatusFont;

{ Utilities }

function Size(CX, CY: Integer): TSize; inline;
function InMainThread: Boolean; inline;
procedure EnableIme(Wnd: HWND; APermission: Boolean);
function SetImeComposition(Window: HWND; Font: HFONT; X, Y: Integer): Boolean;
function GetShiftState: TShiftKeys;
function GetMnemonic(const AText: WideString): DWORD;
procedure Realign(Parent: THandle; var R: TRect);
function TemporaryPen(const Color: TColor): TAnyPen;
function TemporaryBrush(const Color: TColor): TAnyBrush;
function LoadImage(const FileName: WideString): HBITMAP;
function SendWMCopyData(Window: HWND; MagicNumber: DWORD; Handle: THandle; Data: Pointer; Size: Cardinal): LRESULT; overload;
function SendWMCopyData(Window: HWND; MagicNumber: DWORD; Handle: THandle; const Data: WideString): LRESULT; overload;

implementation

threadvar
	CreationWindow: T0Window;

function WindowProc(Wnd: HWND; Msg: UINT; WP: WParam; LP: LParam): LRESULT; stdcall;
var
	Window: T0Window;
begin
	Window := Pointer(GetPropW(Wnd, LWPropName));
	Result := Window.WindowProc(Msg, WP, LP);
end;

function WindowFirstProc(Wnd: HWND; Msg: UINT; WP: WParam; LP: LParam): LRESULT; stdcall;
var
	Window: T0Window;
begin
	Window := CreationWindow;
	Window.FHandle := Wnd;

	SetWindowLongW(Wnd, GWL_WNDPROC, Integer(@WindowProc));
	SetPropW(Wnd, LWPropName, Integer(Window));

	Result := Window.WindowProc(Msg, WP, LP);
end;

{ TAnyWindow }

function TAnyWindow.ScreenToClient(const P: TPoint): TPoint;
begin
	Result := P;
	Windows.ScreenToClient(FHandle, Result);
end;

function TAnyWindow.ClientToScreen(const P: TPoint): TPoint;
begin
	Result := P;
	Windows.ClientToScreen(FHandle, Result);
end;

function TAnyWindow.PostedMessage(Msg: UINT): Boolean;
var
	M: Windows.MSG;
begin
	Result := PeekMessageW(M, Handle, Msg, Msg, PM_NOREMOVE);
end;

function TAnyWindow.PostMessageA(Msg: UINT; WP: WPARAM; LP: LPARAM): Boolean;
begin
	{$WARN SYMBOL_DEPRECATED OFF}
	Result := Windows.PostMessageA(Handle, Msg, WP, LP)
	{$WARN SYMBOL_DEPRECATED ON}
end;

function TAnyWindow.SendMessageA(Msg: UINT; WP: WPARAM; LP: LPARAM): LRESULT;
begin
	{$WARN SYMBOL_DEPRECATED OFF}
	Result := Windows.SendMessageA(Handle, Msg, WP, LP)
	{$WARN SYMBOL_DEPRECATED ON}
end;

function TAnyWindow.PostMessageW(Msg: UINT; WP: WPARAM; LP: LPARAM): Boolean;
begin
	Result := Windows.PostMessageW(Handle, Msg, WP, LP)
end;

function TAnyWindow.SendMessageW(Msg: UINT; WP: WPARAM; LP: LPARAM): LRESULT;
begin
	Result := Windows.SendMessageW(Handle, Msg, WP, LP)
end;

function TAnyWindow.GetText: WideString;
var
	Len: Integer;
begin
	Len:= GetWindowTextLengthW(Handle);
	SetLength(Result, Len);
	if Len > 0 then
		Windows.GetWindowTextW(Handle, Pointer(Result), Len + 1);
end;

procedure TAnyWindow.SetText(const Value: WideString);
begin
	Windows.SetWindowTextW(Handle, PWideChar(Value));
end;

function TAnyWindow.GetEnabled: Boolean;
begin
	Result:= IsWindowEnabled(Handle);
end;

procedure TAnyWindow.SetEnabled(Value: Boolean);
begin
	EnableWindow(Handle, Value);
end;

function TAnyWindow.GetFont: TAnyFont;
var
	F: HFONT;
begin
	F := SendMessageW(WM_GETFONT, 0, 0);
	if F = 0 then
		Result := SystemFont
	else
	begin
		Result := TAnyFont.Create(F);
		Result.Owner := Garbage;
	end;
end;

procedure TAnyWindow.SetFont(Value: TAnyFont);
var
	F: HFONT;
begin
	F := 0;
	if (Value <> nil) and (Value <> SystemFont) then
		F := Value.Handle;

	SendMessageW(WM_SETFONT, F, 0);
end;

function TAnyWindow.FindNextWindow(Current: TAnyWindow; GoForward: Boolean; Check: CCheckWnd; Context: Pointer): TAnyWindow;
var
	CurrentWnd, Target: HWND;
	Passed: Boolean;

	function EnumChildProc_FindNext(Wnd: HWND): Boolean; cdecl;
	begin
		Result := False;
		if GoForward then
		begin
			if CurrentWnd = Wnd then
				Passed:= True
			else if not Passed and (Target <> 0) then
				begin end
			else if CurrentWnd = Windows.GetParent(Wnd) then
				begin end
			else if Check(GetWnd(Wnd), Context) then
			begin
				Target:= Wnd;
				if Passed then Result := True;
			end;
		end
		else
		begin
			if CurrentWnd = Wnd then
			begin
				Passed:= True;
				if Target <> 0 then Result := True;
			end
			else if CurrentWnd = Windows.GetParent(Wnd) then
				begin end
			else if Check(GetWnd(Wnd), Context) then
				Target:= Wnd;
		end;
	end;

begin
	CurrentWnd := Handle;
	if Current <> nil then
		CurrentWnd := Current.Handle;

	Target    := 0;
	Passed    := False;

	EnumerateChildWindows(Handle, CEachWindow(@EnumChildProc_FindNext), GetContext);

	Result := GetWnd(Target);
end;

function TAnyWindow.CanFocus: Boolean;
begin
	Result := Showing and Enabled and TabStop;
end;

function TAnyWindow.Focused: Boolean;
var
	FW: HWND;
begin
	{ LightWindows ł͎qtH[JXĂ True }
	FW := GetFocus;
	repeat
		if FW = FHandle then
			begin Result := True; Exit; end;

		if FW = 0 then
			begin Result := False; Exit; end;

		FW := Windows.GetParent(FW);
	until False;
end;

procedure TAnyWindow.SetFocus;
begin
	if CanFocus then
		Windows.SetFocus(Handle)
end;

function TAnyWindow.Showing: Boolean;
var
	W: HWND;
begin
	Result := True;
	W := Handle;
	repeat
		Result := Result and IsWindowVisible(W);
		if Windows.GetWindowLongW(W, GWL_STYLE) and WS_POPUP <> 0 then Break;
		W := Windows.GetParent(W);
	until W = 0;
end;

function TAnyWindow.GetTabStop: Boolean;
begin
	Result := WindowStyle[WS_TABSTOP];
end;

function TAnyWindow.GetVisible: Boolean;
begin
	Result := IsWindowVisible(Handle);
end;

procedure TAnyWindow.Repaint;
begin
	Windows.InvalidateRect(Handle, nil, True);
end;

procedure TAnyWindow.Repaint(const Rect: TRect);
begin
	Windows.InvalidateRect(Handle, @Rect, True);
end;

procedure TAnyWindow.BringToFront(Wnd: TAnyWindow);
var
	W: HWND;
begin
	W := HWND_BOTTOM;
	if Wnd <> nil then
	begin
		W := Windows.GetWindow(Wnd.Handle, GW_HWNDPREV);
		if W = 0 then W := HWND_TOP;
	end;
	SetWindowPos(Handle, W, 0, 0, 0, 0,
		SWP_NOSIZE or SWP_NOMOVE or SWP_NOREDRAW or SWP_NOACTIVATE);
end;

procedure TAnyWindow.SetBounds(X, Y, CX, CY: Integer);
begin
	Windows.SetWindowPos(Handle, 0, X, Y, CX, CY, SWP_NOZORDER or SWP_NOACTIVATE)
end;

function TAnyWindow.GetParent: TAnyWindow;
begin
	Result:= GetWnd(Windows.GetParent(Handle));
end;

function TAnyWindow.GetClientSize: TSize;
var
	R: TRect;
begin
	Windows.GetClientRect(Handle, R);
	Result.CX := R.Right  - R.Left;
	Result.CY := R.Bottom - R.Top;
end;

procedure TAnyWindow.SetClientSize(const Value: TSize);
var
	HasMenu: Boolean;
	WR, CR, AR: TRect;
begin
	HasMenu := Windows.GetMenu(Handle) <> 0;

	AR.TopLeft := ZeroPoint;
	AR.BottomRight := TPoint(Value);
	AdjustWindowRectEx(AR,
					Windows.GetWindowLongW(Handle, GWL_STYLE),
					HasMenu,
					Windows.GetWindowLongW(Handle, GWL_EXSTYLE));

	SetWindowPos(Handle, 0, 0, 0, AR.Right - AR.Left, AR.Bottom - AR.Top,
					SWP_NOACTIVATE or SWP_NOMOVE or SWP_NOZORDER);

	if HasMenu then
	begin
		GetWindowRect(Handle, WR);
		Windows.GetClientRect(Handle, CR);
		SetWindowPos(Handle, 0, 0, 0,
						WR.Right - WR.Left,
						Value.CY + (WR.Bottom - WR.Top) - (CR.Bottom - CR.Top),
						SWP_NOACTIVATE or SWP_NOMOVE or SWP_NOZORDER);
	end;
end;

function TAnyWindow.GetClientRect: TRect;
begin
	Windows.GetClientRect(Handle, Result);
end;

function TAnyWindow.GetBoundsRect: TRect;
var
	ParentWnd: HWND;
begin
	GetWindowRect(FHandle, Result);

	ParentWnd:= Windows.GetParent(Handle);
	if ParentWnd <> 0 then
	begin
		Dec(Result.Right, Result.Left);
		Dec(Result.Bottom, Result.Top);
		Windows.ScreenToClient(ParentWnd, Result.TopLeft);
		Inc(Result.Right, Result.Left);
		Inc(Result.Bottom, Result.Top);
	end;
end;

procedure TAnyWindow.SetBoundsRect(const Value: TRect);
begin
	SetWindowPos(Handle, 0, Value.Left, Value.Top,
					Value.Right - Value.Left, Value.Bottom - Value.Top,
					SWP_NOACTIVATE or SWP_NOZORDER);
end;

function TAnyWindow.GetPosition: TPoint;
var
	R: TRect;
	ParentWnd: HWND;
begin
	GetWindowRect(FHandle, R);
	Result:= R.TopLeft;

	ParentWnd:= Windows.GetParent(Handle);
	if ParentWnd <> 0 then
		Windows.ScreenToClient(ParentWnd, Result);
end;

procedure TAnyWindow.SetPosition(const Value: TPoint);
begin
	SetWindowPos(Handle, 0, Value.X, Value.Y, 0, 0, SWP_NOACTIVATE or SWP_NOSIZE or SWP_NOZORDER);
end;

function TAnyWindow.GetBoundsSize: TSize;
var
	R: TRect;
begin
	GetWindowRect(Handle, R);
	Result.CX := R.Right - R.Left;
	Result.CY := R.Bottom- R.Top;
end;

procedure TAnyWindow.SetBoundsSize(const Value: TSize);
begin
	SetWindowPos(Handle, 0, 0, 0, Value.CX, Value.CY,
					SWP_NOACTIVATE or SWP_NOMOVE or SWP_NOZORDER);
end;

procedure TAnyWindow.Hide;
begin
	ShowWindow(Handle, SW_HIDE)
end;

function TAnyWindow.GetWindowLong(AIndex: Integer): LongWord;
begin
	Result:= Windows.GetWindowLongW(Handle, AIndex);
end;

procedure TAnyWindow.SetWindowLong(AIndex: Integer; const Value: LongWord);
begin
	Windows.SetWindowLongW(Handle, AIndex, Value);
end;

function TAnyWindow.GetWindowStyle(AStyle: DWORD): Boolean;
begin
	Result := GetWindowLongW(Handle, GWL_STYLE) and AStyle <> 0
end;

procedure TAnyWindow.SetWindowStyle(AStyle: DWORD; const Value: Boolean);
var
	S, N: DWORD;
begin
	S := Windows.GetWindowLongW(Handle, GWL_STYLE);
	N := S xor (S xor not DWORD(Ord(Value) - 1)) and AStyle;
	if N <> S then Windows.SetWindowLongW(Handle, GWL_STYLE, N);
end;

function TAnyWindow.GetWindowExStyle(AStyle: DWORD): Boolean;
begin
	Result := WindowLong[GWL_EXSTYLE] and AStyle <> 0
end;

procedure TAnyWindow.SetWindowExStyle(AStyle: DWORD; const Value: Boolean);
var
	S, N: DWORD;
begin
	S := Windows.GetWindowLongW(Handle, GWL_EXSTYLE);
	N := S xor (S xor not DWORD(Ord(Value) - 1)) and AStyle;
	if N <> S then Windows.SetWindowLongW(Handle, GWL_EXSTYLE, N);
end;

procedure TAnyWindow.RepaintFrame;
begin
	SendMessageW(WM_NCPAINT, 1, 0);
end;

procedure TAnyWindow.Show;
begin
	ShowWindow(Handle, SW_SHOWNA)
end;

procedure TAnyWindow.SetVisible(const Value: Boolean);
begin
	{if Value <> GetVisible then}
	if Value then Show else Hide
end;

constructor TAnyWindow.Create(AWnd: HWND);
begin
	FHandle := AWnd;
	inherited Create;
end;

{ T0Window }

constructor T0Window.Create(AParent: TAnyWindow; const AText: WideString;
				const APosition: TPoint; const ASize: TSize);
var
	CS: PCreateStructW;
	DSize: TSize;
begin
	inherited Create(0);

	{NXo^}
	RegisterClass;

	{EBhEp[^}
	New(FWindowStruct);

	FWindowStruct^.Text := AText;

	CS := @FWindowStruct^.CreateStruct;

	CS^.lpCreateParams := nil;
	CS^.hInstance:= MainInstance;
	CS^.hMenu:= 0;
	CS^.hwndParent:= 0;
	if AParent <> nil then
		CS^.hwndParent := AParent.Handle;
	CS^.cy:= ASize.cY;
	CS^.cx:= ASize.cX;
	if (ASize.cX = UseDefault) or (ASize.cY = UseDefault) then
	begin
		DSize:= GetDefaultSize;
		if ASize.cY = UseDefault then CS^.cy:= DSize.cy;
		if ASize.cX = UseDefault then CS^.cx:= DSize.cx;
	end;
	CS^.y := APosition.Y;
	CS^.x := APosition.X;
	CS^.style:= WS_CLIPSIBLINGS or WS_CLIPCHILDREN
					or WS_VISIBLE or WS_CHILD or WS_TABSTOP;
	CS^.lpszName := PWideChar(FWindowStruct^.Text);
	CS^.lpszClass := nil;
	CS^.dwExStyle := 0;
end;

procedure T0Window.AfterConstruction; {override}
var
	Struct: PCREATESTRUCTW;
	C: WideString;
{$IFOPT C+}
	S: Boolean;
{$ENDIF}
begin
	inherited;

	{쐬邼Ɠ`}
	CreationWindow := Self;

	{NX}
	C := WindowClassName;

	{쐬}
	Struct := @FWindowStruct^.CreateStruct;
	{$IFOPT C+} S := 0 <> {$ENDIF} Windows.CreateWindowExW(
					Struct.dwExStyle,
					Pointer(C),
					Struct.lpszName,
					Struct.style,
					Struct.x,
					Struct.y,
					Struct.cx,
					Struct.cy,
					Struct.hwndParent,
					Struct.hMenu,
					Struct.hInstance,
					Struct.lpCreateParams);

{$IFOPT C+}
	Assert(S);
{$ENDIF}

	{CreateStruct ͂vȂ}
	Dispose(FWindowStruct);
	FWindowStruct := nil;
end;

procedure T0Window.BeforeDestruction; {override}
begin
	{RXgN^̗OƂ邩}
	if (FHandle <> 0) and not Destroying then
	begin
		FDestroying := True;
		Windows.DestroyWindow(Handle);
		FHandle := 0;
	end;

	if FWindowStruct <> nil then
		Dispose(FWindowStruct);

	inherited;
end;

procedure T0Window.Finalize;
begin
	FFont.Free;
	FFont := nil;
end;

procedure T0Window.Initialize;
begin
end;

function T0Window.CreateStruct: PCreateStructW;
begin
	Result := @FWindowStruct^.CreateStruct
end;

class function T0Window.WindowClassName: string;
var
	PName: PShortString;
begin
	PName := PShortString(PPointer(Integer(Self) + vmtClassName)^);
	SetString(Result, PChar(PName) + 2, Ord(PName^[0]) - 1);
end;

class procedure T0Window.CreateClass(var WndClass: WNDCLASSEXW);
begin
	WndClass.style         := CS_HREDRAW or CS_VREDRAW or CS_DBLCLKS;
	WndClass.cbClsExtra    := 0;
	WndClass.cbWndExtra    := 0;
	WndClass.hIcon         := 0;
	WndClass.hCursor       := LoadCursorW(0, PWideChar(IDC_ARROW));
	WndClass.hbrBackground := GetSysColorBrush(COLOR_3DFACE); {W}
	WndClass.lpszMenuName  := nil;
	WndClass.hIconSm       := 0;
end;

class function T0Window.GetDefaultSize: TSize;
begin
	Assert(False)
end;

class procedure T0Window.RegisterClass;
var
	WndClass: WNDCLASSEXW;
	C: WideString;
	P: PWideChar;
begin
	{NX}
	C := WindowClassName;
	P := PWideChar(C);
	{̎擾ɐɓo^ς݁A̎AȂ}
	WndClass.cbSize := SizeOf(WndClass);
	if not GetClassInfoExW(MainInstance, P, WndClass) then
	begin
		{ݒ}
		WndClass.cbSize        := SizeOf(WndClass);
		WndClass.lpfnWndProc   := @WindowFirstProc;
		WndClass.hInstance     := MainInstance;
		WndClass.lpszClassName := P; 
		CreateClass(WndClass);
		{o^}
		{$IFDEF C+} OSCheck( {$ENDIF} Windows.RegisterClassExW(WndClass) {$IFDEF C+} } {$ENDIF};
	end
end;

function EnumChildProc_Update(WND: HWND; LP: LPARAM): BOOL; stdcall;
var
	WO: T0Window;
begin
	WO:= GetLWnd(WND);
	if WO <> nil then
		WO.Update;

	Result:= True;
end;

procedure T0Window.Update;
begin
	EnumChildWindows(FHandle, @EnumChildProc_Update, 0);
end;

function T0Window.WindowProc(Msg: UINT; WP: WPARAM; LP: LPARAM): LRESULT;
var
	M: TMessage;
begin
	M.Msg := Msg;
	M.WParam := WP;
	M.LParam := LP;
	M.Result := 0;
	try
		Application.BeginRead;
		try
			Dispatch(M);
		finally
			Application.EndRead;
		end;
	except
		ShowException(ExceptObject, ExceptAddr);
	end;
	Result := M.Result;
end;

procedure T0Window.DefaultHandler(var Message);
begin
	TMessage(Message).Result := DefWindowProcW(FHandle,
		TMessage(Message).Msg,
		TMessage(Message).WParam,
		TMessage(Message).LParam);
end;

procedure T0Window.HandleNotification(Sender: TAnyWindow; Info: PNMHDR; var Result: LRESULT);
var
	P: T0Window;
begin
	P := LParent;
	if P <> nil then
		P.HandleNotification(Sender, Info, Result);
end;

procedure T0Window.Realign;
var
	R: TRect;
begin
	R := ClientRect;
	LWindows.Realign(Handle, R)
end;

procedure T0Window.WMCreate(var Message: TWMCreate);
begin
	inherited;
	Initialize;
end;

procedure T0Window.WMDestroy(var Message: TWMDestroy);
begin
	inherited;
	Finalize;
end;

function PropEnumProc_Clear(Wnd: HWND;
				Name: PWideChar; Data: THandle): BOOL; stdcall;
begin
	RemovePropW(Wnd, Name);
	Result:= True;
end;

procedure T0Window.WMNCDestroy(var Message: TWMNCDestroy);
begin
	EnumPropsW(Handle, @PropEnumProc_Clear); {TFarProc = Pointer}
	inherited;
	if not Destroying then
	begin
		FDestroying := True;
		Free;
	end;
end;

procedure T0Window.WMCommand(var Message: TWMCommand);
var
	Item: TAnyMenuItem;
	Info: TNMHDR;
	Wnd: TAnyWindow;
	Receiver: T0Window;
	Dummy: LRESULT;
begin
	if Message.Ctl = 0 then
		case Message.NotifyCode of
		0: {Menu}
		begin
			{//TortoiseCVSexplorer̍XV(0xA220=41504)𑗂ė₪c
			if WindowLong[GWL_WNDPROC] <> DWORD(@LWindows.WindowProc) then
			begin
				MessageBoxW(Handle, 'sWM_COMMAND', 'x', MB_ICONWARNING);
			end;}
			
			Item := GetLMenuItem(Message.ItemID);
			if Item <> nil then
				Item.Execute;
		end;
(*
		1: {Accelerator - LW ł͎gȂ}
			;
*)
		end
	else
	begin
		Wnd := GetWnd(Message.Ctl);

		Info.hwndFrom := Message.Ctl;
		Info.idFrom := SHORT(Message.ItemID); {Signed}
		Info.code := NM_COMMAND + Message.NotifyCode;

		Receiver := Self;
		if (Message.Ctl <> Handle) and (Wnd is T0Window) then
		begin
			Receiver := T0Window(Wnd); {lȂ瓖l}
		end;

		Receiver.HandleNotification(Wnd, @Info, Dummy);
	end;

	inherited;
end;

procedure T0Window.WMNotify(var Message: TWMNotify);
var
	Wnd: TAnyWindow;
	Receiver: T0Window;
begin
	Wnd := GetWnd(Message.NMHDR^.hwndFrom);

	Receiver := Self;
	if (Message.NMHDR^.hwndFrom <> Handle) and (Wnd is T0Window) then
	begin
		Receiver := T0Window(Wnd); {lȂ瓖l}
	end;

	inherited;
	Receiver.HandleNotification(Wnd, Message.NMHDR, Message.Result);
end;

function T0Window.GetActiveControl: T0Window;
var
	A: T0Window;
begin
	Result := GetLWnd(FActiveChild);
	while Result <> nil do
	begin
		A := Result.GetActiveControl;
		if (A = nil) or (Result.WindowLong[GWL_EXSTYLE] and WS_EX_CONTROLPARENT = 0) then
			Break;

		Result := A;
	end;
end;

procedure T0Window.SetActiveControl(const Value: T0Window);
var
	C, P: T0Window;
begin
	if Value = nil then
		FActiveChild := 0
	else
	begin
		if Focused then
			Value.SetFocus
		else
		begin
			C := Value;
			P := Value.LParent;
			while P <> nil do
			begin
				P.FActiveChild := C.Handle;
				if P = Self then Break;
				C := P;
				P := P.LParent;
			end;
		end
	end
end;

procedure T0Window.FocusChanged;
begin
end;

procedure T0Window.DialogKey(var Message: TWMKey);
var
	P: T0Window;
begin
	P := LParent;
	if P <> nil then
		P.DialogKey(Message);
end;

function EnumChildProc_Mnemonic(Wnd: HWND; LP: LPARAM): BOOL; stdcall;
var
	W: T0Window;
begin
	W := GetLWnd(Wnd);
	if W <> nil then
		W.MnemonicKey(TWMKey(Pointer(LP)^));
	Result := TWMKey(Pointer(LP)^).Result = 0;
end;

procedure T0Window.MnemonicKey(var Message: TWMKey);
begin
	EnumChildWindows(Handle, @EnumChildProc_Mnemonic, LPARAM(@Message));
end;

function T0Window.Notify(Code: Integer): LRESULT;
var
	Hdr: TNMHdr;
begin
	Hdr.hwndFrom := Handle;
	Hdr.idFrom := 0;
	Hdr.code := Code;
	Result := 0;
  HandleNotification(Self, @Hdr, Result);
end;

function T0Window.KeyDown(ShiftState: TShiftKeys; Code: Byte): Boolean;
begin
	Result := False
end;

function T0Window.KeyInput(Code: WideChar): Boolean;
begin
	Result := False
end;

function T0Window.KeyUp(ShiftState: TShiftKeys; Code: Byte): Boolean;
begin
	Result := False
end;

procedure T0Window.PaintWindow(DC: HDC);
begin
end;

procedure T0Window.ProcessCtlColor(var Message: TWMCtlColor);
var
	Wnd: T0Window;
begin
	if Message.ChildWnd <> FHandle then
	begin
		Wnd := GetLWnd(Message.ChildWnd);
		if Wnd <> nil then
		begin
			Wnd.Dispatch(Message);
			Exit;
		end;
	end;
	DefaultHandler(Message);
end;

function T0Window.ProcessScroll(var Message: TWMScroll;
	Manager: PScrollBarManager; Kind: Integer; var SI: SCROLLINFO): Integer;
var
	Old, Max: Integer;
	TmpSI: TScrollInfo;
begin
	Message.Result:= 0;

	Old:= SI.nPos;

	case Message.ScrollCode of
	SB_LineUp:				Dec(SI.nPos);
	SB_LineDown:			Inc(SI.nPos);
	SB_PageUp:				Dec(SI.nPos, SI.nPage);
	SB_PageDown:			Inc(SI.nPos, SI.nPage);
	SB_ThumbPosition: SI.nPos:= GetScrollPos(Handle, Kind);
	SB_ThumbTrack:		begin
											TmpSI.cbSize:= SizeOf(TScrollInfo);
											TmpSI.fMask:= SIF_TRACKPOS;
											Manager.GetScrollInfo(Handle, Kind, TmpSI);
											SI.nPos:= TmpSI.nTrackPos;
										end;
	SB_Top:						SI.nPos:= SI.nMin;
	SB_Bottom:				SI.nPos:= SI.nMax;
	SB_EndScroll:			;
	end;

	if SI.nPos < SI.nMin then
		SI.nPos:= SI.nMin
	else
	begin
		Max:= SI.nMax;
		if SI.nPage > 1 then
		begin
			Dec(Max, SI.nPage);
			Inc(Max);
			if Max < SI.nMin then Max := SI.nMin; {Disabled̏ꍇ}
		end;

		if SI.nPos > Max then
			SI.nPos:= Max;
	end;

	if Old <> SI.nPos then
		Manager.SetScrollPos(Handle, Kind, SI.nPos, True);

	Result:= SI.nPos - Old;
end;

procedure T0Window.WMCtlColorBtn(var Message: TWMCtlColorBtn);
begin
	ProcessCtlColor(Message);
end;

procedure T0Window.WMCtlColorEdit(var Message: TWMCtlColorEdit);
begin
	ProcessCtlColor(Message);
end;

procedure T0Window.WMCtlColorListBox(var Message: TWMCtlColorListBox);
begin
	ProcessCtlColor(Message);
end;

procedure T0Window.WMCtlColorStatic(var Message: TWMCtlColorStatic);
begin
	ProcessCtlColor(Message);
end;

procedure T0Window.WMPaint(var Message: TWMPaint);
var
	DC: HDC;
	PS: PAINTSTRUCT;
	SaveKey: Integer;
begin
	DC := Message.DC;
	if Message.DC = 0 then
		DC := BeginPaint(Handle, PS);

	SaveKey := SaveDC(DC);
	PaintWindow(DC);
	RestoreDC(DC, SaveKey);

	if Message.DC = 0 then
		EndPaint(Handle, PS);
end;

procedure T0Window.WMParentNotify(var Message: TWMParentNotify);
begin
	inherited;

	if (Message.Event = WM_DESTROY) and (Message.ChildWnd = FActiveChild) then
		FActiveChild := 0;
end;

procedure T0Window.WMSetFocus(var Message: TWMSetFocus);
var
	P, C: T0Window;
begin
	inherited;

	{ǂŕKqEBhEEBhE͂͂}

	if FActiveChild <> 0 then
		Windows.SetFocus(FActiveChild)
	else
	begin
		FocusChanged;
		C := Self;
		P := LParent;
		while (P <> nil) and C.WindowStyle[WS_CHILD] do
		begin
			P.FActiveChild := C.Handle;
			P.FocusChanged;
			C := P;
			P := P.LParent;
		end;
	end
end;

procedure T0Window.WMSize(var Message: TWMSize);
begin
	inherited;
	Realign;
end;

procedure T0Window.WMChar(var Message: TWMChar);
var
	I, Mask: Integer;
	P: Boolean;
	HandleCache: HWND;
begin
	if (Message.CharCode = 32) and (GetKeyState(VK_CONTROL) and $8000 <> 0) then
		Message.CharCode := 0;

	if LongRec(Message.KeyData).Lo = 0 then
		Inc(LongRec(Message.KeyData).Lo);

	HandleCache := Handle;
	P := False;
	for I := LongRec(Message.KeyData).Lo downto 1 do
	begin
		P := KeyInput(WideChar(Message.CharCode)) or P;
		if not IsWindow(HandleCache) then Exit;
	end;
	if P then Exit;

	case Message.CharCode of
	9:
		Mask := DLGC_WANTTAB;
	$0D, $1b:
		Mask := DLGC_WANTALLKEYS;
	$20:
		Mask := DLGC_WANTCHARS or DLGC_BUTTON;
	else
		Mask := DLGC_WANTCHARS;
	end;

	if SendMessageW(WM_GETDLGCODE, 0, 0) and Mask = 0 then
	begin
		DialogKey(Message);
		if (Message.Result = 0) and (Mask = DLGC_WANTCHARS) then Beep;
		Exit;
	end;

	inherited
end;

procedure T0Window.WMSysChar(var Message: TWMChar);
begin
	DialogKey(Message);
	if (Message.Result = 0) and (Message.CharCode > 31) then
		inherited
end;

procedure T0Window.WMKeyDown(var Message: TWMKey);
label
	ToInherited; //R[hGA10oCgx̍œKɂȂ
var
	I, Mask: Integer;
	P: Boolean;
	HandleCache: HWND;
	ShiftState: TShiftKeys;
begin
	if LongRec(Message.KeyData).Lo = 0 then
		Inc(LongRec(Message.KeyData).Lo);

	ShiftState := GetShiftState;
	HandleCache := Handle;

	P := False;
	for I := LongRec(Message.KeyData).Lo downto 1 do
	begin
		P := KeyDown(ShiftState, Message.CharCode) or P;
		if not IsWindow(HandleCache) then Exit;
	end;
	if P then Exit;

	case Message.CharCode of
	VK_TAB:                                      Mask:= DLGC_WANTTAB;
	VK_LEFT, VK_RIGHT, VK_UP, VK_DOWN:           Mask:= DLGC_WANTARROWS;
	VK_RETURN, VK_EXECUTE, VK_ESCAPE, VK_CANCEL: Mask:= DLGC_WANTALLKEYS;
	VK_F6:
		if ssControl in ShiftState then
			Mask := DLGC_WANTALLKEYS
		else
			goto ToInherited;
	else
		goto ToInherited;
	end;

	if SendMessageW(WM_GETDLGCODE, 0, 0) and Mask = 0 then
	begin
		DialogKey(Message);
		Exit;
	end;

ToInherited:
	inherited
end;

procedure T0Window.WMSysKeyDown(var Message: TWMSysKeyDown);
begin
	if (Message.CharCode = VK_MENU) or not KeyDown(GetShiftState, Message.CharCode) then
	begin
		DialogKey(Message);
		if Message.Result = 0 then
			inherited
	end
end;

procedure T0Window.WMKeyUp(var Message: TWMKey);
begin
	if not KeyUp(GetShiftState, Message.CharCode) then inherited
end;

procedure T0Window.WMSysKeyUp(var Message: TWMSysKeyDown);
begin
	if not KeyUp(GetShiftState, Message.CharCode) then
	begin
		DialogKey(Message);
		if Message.Result = 0 then
			inherited
	end
end;

procedure T0Window.WMGetFont(var Message: TWMGetFont);
begin
	if FFont <> nil then
		Message.Result := FFont.Handle
	else
		inherited;
end;

procedure T0Window.WMSetFont(var Message: TWMSetFont);
label
	Changed, Clear, ExitProc;
begin
	if Message.Font = 0 then
		if FFont = nil then
			goto ExitProc
		else
			goto Clear;

	if FFont <> nil then
	begin
		if FFont.Handle = Message.Font then
			goto ExitProc;
	end
	else
	begin
		FFont := TFont.Create;
		FFont.Owner := Self;
	end;
	FFont.Assign(Message.Font);
	if FFont.Handle = SystemFont.Handle then
		goto Clear;

	goto Changed;

Clear:
	FFont.Free;
	FFont := nil;

Changed:
	FontChanged;

ExitProc:
	inherited;
end;

procedure T0Window.FontChanged;
begin
end;

function T0Window.GetFont: TAnyFont;
begin
	if FFont <> nil then
		Result := FFont
	else
		Result := inherited GetFont
end;

function T0Window.TopParent: T0TopWindow;
var
	W: TAnyWindow;
begin
	W := Self;
	while (W <> nil) and not (W is T0TopWindow) do
		W := W.Parent;

	Result := T0TopWindow(W);
end;

procedure T0Window.ColorChanged;
begin
end;

procedure T0Window.SetColor(const Value: TColor);
begin
	FColor := Value;
	ColorChanged;
end;

procedure T0Window.WMSysColorChange(var Message: TWMSysColorChange);
begin
	inherited;

	if FColor.SysColor >= 0 then
		FColor.Value := GetSysColor(FColor.SysColor);

	ColorChanged;
end;

procedure T0Window.WMHScroll(var Message: TWMScroll);
var
	Wnd: T0Window;
begin
	inherited;
	if (Message.ScrollBar <> 0) and (Message.ScrollBar <> Handle) then
	begin
		Wnd := GetLWnd(Message.ScrollBar);
		if Wnd <> nil then
			Wnd.Dispatch(Message);
	end;
end;

procedure T0Window.WMVScroll(var Message: TWMScroll);
var
	Wnd: T0Window;
begin
	inherited;
	if (Message.ScrollBar <> 0) and (Message.ScrollBar <> Handle) then
	begin
		Wnd := GetLWnd(Message.ScrollBar);
		if Wnd <> nil then
			Wnd.Dispatch(Message);
	end;
end;

procedure T0Window.WMInitMenuPopup(var Message: TWMInitMenuPopup);
var
	M: TAnyMenu;
begin
	inherited;
	M := GetLMenu(Message.MenuPopup);
	if M <> nil then M.Update
end;

function T0Window.GetLParent: T0Window;
begin
	Result := GetLWnd(Windows.GetParent(Handle))
end;

{ T0TopWindow }

constructor T0TopWindow.Create(AParent: T0TopWindow;
				const AText: WideString; const APosition: TPoint; const ASize: TSize);
var
	CS: PCreateStructW;
begin
	inherited Create(AParent, AText, APosition, ASize);

	if AParent is T0LeadWindow then
		FLeader := T0LeadWindow(AParent)
	else if AParent <> nil then
		FLeader := AParent.Leader;

	CS := CreateStruct;
	CS.style := CS.style and not (WS_TABSTOP or WS_CHILD) or WS_OVERLAPPEDWINDOW;
end;

procedure T0TopWindow.Close;
begin
	Windows.PostMessageW(Handle, WM_CLOSE, 0, 0);
end;

function T0TopWindow.CloseQuery: Boolean;
begin
	Result:= True;
end;

function CheckTabOrder(Wnd: TAnyWindow; Context: Pointer): Boolean; cdecl;
begin
	Result:= Wnd.CanFocus;
end;

procedure T0TopWindow.DialogKey(var Message: TWMKey); {override}
var
	Next: Boolean;
	Wnd: TAnyWindow;
	OrgCharCode: DWORD;
begin
	case Message.Msg of
	WM_KEYDOWN, WM_SYSKEYDOWN:
	begin
		case Message.CharCode of
		VK_F6:
		begin
			MnemonicKey(Message);
			Exit;
		end;
		VK_TAB:            Next := GetKeyState(VK_SHIFT) and $80 = 0;
		VK_LEFT, VK_UP:    Next := False;
		VK_RIGHT, VK_DOWN: Next := True;
		else
			Exit;
		end;
		if Message.Msg = WM_KEYDOWN then
		begin
			Message.Result := 1;

			Wnd := FindNextWindow(ActiveControl, Next, CheckTabOrder, GetContext);
			if Wnd <> nil then
			begin
				if Wnd is T0Window then
					T0Window(Wnd).FActiveChild := 0;
				Windows.SetFocus(Wnd.Handle);
			end;
		end
	end;
	WM_CHAR, WM_SYSCHAR:
	begin
		OrgCharCode := Message.CharCode;
		case OrgCharCode of
		Ord('a')..Ord('z'): Message.CharCode := OrgCharCode - (Ord('a') - Ord('A'));
		end;
		MnemonicKey(Message);
		Message.CharCode := OrgCharCode;
	end;
	end
end;

procedure T0TopWindow.WMClose(var Message: TWMClose);
begin
	if CloseQuery then
		inherited;
end;

procedure T0TopWindow.WMQueryEndSession(var Message: TWMQueryEndSession);
begin
	Message.Result := Ord(CloseQuery and CallTerminateProcs);
end;

procedure T0TopWindow.Update;
begin
	{inherited}
	FHint := ReadyHint;

	inherited Update;

	{Tuj[Ă܂ƂȂ̂Ńj[[v͍XVȂ}
	if not (opInMenuLoop in FOperating) and (FMenu <> nil) then
		FMenu.Update
end;

function T0TopWindow.ReadyHint: WideString;
begin
	if opInMenuLoop in FOperating then
		Result := ''
	else
		Result := SReady;
end;

function T0TopWindow.SelectedMenuItem: TAnyMenuItem;
begin
	Result := GetLMenuItem(FSelectedMenuItemID);
end;

procedure T0TopWindow.WMMenuSelect(var Message: TWMMenuSelect);
var
	M: TAnyMenuItem;
begin
	inherited;

	if Message.MenuFlag = $ffff then
		FSelectedMenuItemID := 0
	else if Message.MenuFlag and MF_SYSMENU <> 0 then
		FSelectedMenuItemID := 0
	else if Message.MenuFlag and MF_POPUP <> 0 then
	begin
		M := GetMenu(Message.Menu).At[Message.IDItem];
		if M <> nil then
		begin
			M.Selected;
			FSelectedMenuItemID := M.ID
		end;
	end
	else
		FSelectedMenuItemID := Message.IDItem;

	Update;
end;

procedure T0TopWindow.WMEnterMenuLoop(var Message);
begin
	Include(FOperating, opInMenuLoop);
	inherited;
end;

procedure T0TopWindow.WMExitMenuLoop(var Message);
begin
	Exclude(FOperating, opInMenuLoop);
	inherited;
end;

procedure T0TopWindow.SetMenu(Value: TMainMenu);
begin
	if Value = FMenu then Exit;

	Windows.SetMenu(Handle, 0);
	FMenu := Value;

	if FMenu <> nil then
	begin
		FMenu.Update;
		Windows.SetMenu(Handle, FMenu.Handle);
	end;
end;

function T0TopWindow.IsForeground: Boolean;
begin
	Result := GetForegroundWindow = Handle
end;

procedure T0TopWindow.WMEnterSizeMove(var Message);
begin
	Include(FOperating, opInMoving);
	inherited;
end;

procedure T0TopWindow.WMExitSizeMove(var Message);
begin
	Exclude(FOperating, opInMoving);
	inherited;
end;

procedure T0TopWindow.BeginTracking;
begin
	Include(FOperating, opInTracking)
end;

procedure T0TopWindow.EndTracking;
begin
	Exclude(FOperating, opInTracking)
end;

function T0TopWindow.GetStayOnTop: Boolean;
begin
	Result := WindowExStyle[WS_EX_TOPMOST]
end;

procedure T0TopWindow.SetStayOnTop(const Value: Boolean);
const
	H: array[Boolean] of HWND = (HWND_NOTOPMOST, HWND_TOPMOST);
begin
	SetWindowPos(Handle, H[Value], 0, 0, 0, 0, SWP_NOSIZE or SWP_NOMOVE)
end;

procedure T0TopWindow.Finalize;
begin
	Windows.SetMenu(Handle, 0);
	inherited;
end;

procedure T0TopWindow.Foreground;
{$IFDEF FORCEFOREGROUND}
var
	ForeWindow: HWND;
	ForeThread: DWORD;
	CurrentThread: DWORD;
{$ENDIF}
begin
{$IFDEF FORCEFOREGROUND}
	ForeWindow := GetForegroundWindow;
	if ForeWindow <> 0 then
	begin
		ForeThread := GetWindowThreadProcessId(ForeWindow);
		CurrentThread := LWindows.CurrentThread.ID;
		if (ForeThread <> 0) and (ForeThread <> CurrentThread) then
		begin
			SetActiveWindow(Handle);
			AttachThreadInput(CurrentThread, ForeThread, True);
			SetForegroundWindow(ForeWindow);
			SetForegroundWindow(Handle);
			AttachThreadInput(CurrentThread, ForeThread, False);
		end
		else
			SetForegroundWindow(Handle);
	end;
{$ELSE}
	SetForegroundWindow(Handle);
{$ENDIF}
end;

procedure T0TopWindow.DefaultMDIHandler(const Client: IMDIClient; var Message);
var
	ClientHandle: HWND;
begin
	case TMessage(Message).Msg of
	WM_SIZE, WM_COMMAND:
		inherited DefaultHandler(Message);
	WM_GETTEXTLENGTH:
		TMessage(Message).Result := FTextLength;
	else
		if TMessage(Message).Msg = WM_SETTEXT then
		begin
			FTextLength := lstrlenW(PWideChar(TMessage(Message).LParam));
		end;

		ClientHandle := 0;
		if Client <> nil then
			ClientHandle := Client.ClientWindow.Handle;

		TMessage(Message).Result := DefFrameProcW(FHandle, ClientHandle,
			TMessage(Message).Msg,
			TMessage(Message).WParam,
			TMessage(Message).LParam);

		if TMessage(Message).Msg = WM_GETTEXT then
		begin
			if TMessage(Message).Result > FTextLength then
			begin
				TMessage(Message).Result := FTextLength;
				PWideChar(TMessage(Message).LParam)[FTextLength] := #0;
      end;	
    end;	
	end;
end;

{ T0PopupWindow }

constructor T0PopupWindow.Create(AParent: T0TopWindow; const AText: WideString;
	const APosition: TPoint; const ASize: TSize);
var
	CS: PCreateStructW;
begin
	Assert(AParent <> nil);

	inherited;

	CS := CreateStruct;
	CS.style := CS.style and not (WS_MINIMIZEBOX or WS_MAXIMIZEBOX) or Integer(WS_POPUP);
end;

{ T0IsolatedWindow }

constructor T0IsolatedWindow.Create(AParent: T0TopWindow;
	const AText: WideString; const APosition: TPoint; const ASize: TSize);
begin
	Assert(AParent <> nil);

	inherited;
	CreateStruct^.hwndParent := 0;
end;

class function T0IsolatedWindow.GetDefaultSize: TSize;
begin
	Result := DefaultSize
end;

procedure T0IsolatedWindow.WMActivate(var Message: TWMActivate);
begin
	if (Message.Active <> 0) and not Enabled then
		PostMessageW(WM_LIGHTWINDOWS, LW_ACTIVATE, 0);

	inherited;
end;

procedure T0IsolatedWindow.WMSetCursor(var Message: TWMSetCursor);
begin
	if (Message.MouseMsg = WM_LBUTTONDOWN) and not Enabled then
		PostMessageW(WM_LIGHTWINDOWS, LW_ACTIVATE, 0);

	inherited;
end;

procedure T0IsolatedWindow.WMLightWindows(var Message: TMessage);
begin
	case Message.WParam of
	LW_ACTIVATE:
		SetActiveWindow(T0GUIThread(CurrentThread).LeadWindow.Handle);
	end;
end;

{ T0LeadWindow }

function EnumWindowsProc_Update(Wnd: HWND; LP: LPARAM): BOOL; stdcall;
var
	WO: T0Window;
begin
	WO := GetLWnd(Wnd);

	if WO is T0TopWindow then
		if T0TopWindow(WO).Leader = T0LeadWindow(LP) then
			WO.Update;

	Result := True;
end;

procedure T0LeadWindow.Update;
begin
	inherited Update;
	EnumWindows(@EnumWindowsProc_Update, LPARAM(Self));
end;

procedure T0LeadWindow.MessageLoop;
var
	CurrentGUIThread: T0GUIThread;
	Store: T0Object;
	OldLead: T0LeadWindow;
	Msg: TMsg;
begin
	FTerminated := False;

	Assert(CurrentThread is T0GUIThread);
	CurrentGUIThread := T0GUIThread(CurrentThread);

	OldLead := CurrentGUIThread.LeadWindow;
	CurrentGUIThread.FLeadWindow := Self;
	Store := Garbage.Protect; {GChL̈ʒuL}
	try
		{bZ[W[v}
		while not Terminated do
		begin
			if not PeekMessageW(Msg, 0, 0, 0, PM_REMOVE) then
				CurrentGUIThread.WaitMessage(Msg);

			if Msg.message = WM_QUIT then
				Msg.hwnd := FHandle;

			TranslateMessage(Msg); {TranslateMessageAWp炵}
			DispatchMessageW(Msg);
		end;
	finally
		Garbage.Unprotect(Store); {GChL̈ʒu߂}
		CurrentGUIThread.FLeadWindow := OldLead;
	end;
end;

procedure T0LeadWindow.ProcessIdle;
begin
	try
		Update;
	except
		ShowException(ExceptObject, ExceptAddr)
	end
end;

procedure T0LeadWindow.Setup;
begin
end;

procedure T0LeadWindow.WMQuit(var Message: TWMQuit); {message}
begin
	FTerminated := True;
end;

procedure T0LeadWindow.WMClose(var Message: TWMClose); {message}
begin
	{DefWindowProcDestroyWindow Ăł܂̂inherited͍sȂ}
	FTerminated := CloseQuery;
end;

procedure T0LeadWindow.WMEndSession(var Message: TWMEndSession); {message}
begin
	FTerminated := True;
	inherited;
end;

procedure T0LeadWindow.WMEnable(var Message: TWMEnable); {message}

	function EnumWindowsProc_Enable(Wnd: HWND): BOOL; cdecl;
	var
		W: TAnyWindow;
	begin
		W := GetLWnd(WND);
		if W is T0TopWindow then
		begin
			if T0TopWindow(W).Leader = Self then
				EnableWindow(WND, Message.Enabled);
		end;

		Result := False;
	end;

begin
	EnumerateWindows(CEachWindow(@EnumWindowsProc_Enable), GetContext);
	inherited;
end;

procedure T0LeadWindow.WMQueryEndSession(var Message: TWMQueryEndSession);
begin
	Message.Result := Ord(CloseQuery)
end;

{ T0DialogWindow }

constructor T0DialogWindow.Create(AParent: T0TopWindow; const AText: WideString);
var
	CS: PCreateStructW;
begin
	FContext := Context;
	//if AParent = nil then AParent := CurrentThread.LeadWindow;
	inherited Create(AParent, AText, DefaultPoint, ZeroSize);

	CS:= CreateStruct;
	CS.Style:= CS.Style and not
					(WS_VISIBLE or WS_MAXIMIZEBOX or WS_MINIMIZEBOX or WS_THICKFRAME)
					or (WS_DLGFRAME or Integer(WS_POPUP));
	CS.dwExStyle:= CS.dwExStyle or (WS_EX_DLGMODALFRAME or WS_EX_WINDOWEDGE);

	FImeAllowed := True;
  FForegroundWindow := GetForegroundWindow;
end;

function T0DialogWindow.Execute: Boolean;
var
	ParWnd: T0LeadWindow;
begin
	FDialogResult := ID_CANCEL;

	ParWnd := Leader; {ParentPopupIsolatedȂ炻̃[_[}
	if ParWnd <> nil then
		ParWnd.Enabled := False;

	Enabled := True;

	Setup;
	ShowWindow(Handle, SW_SHOW);
	MessageLoop;

	if ParWnd <> nil then
		ParWnd.Enabled := True;

	ShowWindow(Handle, SW_HIDE);

	Result := FDialogResult <> ID_CANCEL;
end;

class procedure T0DialogWindow.CreateClass(var WndClass: WNDCLASSEXW);
begin
	inherited;
	WndClass.style := WndClass.Style or (CS_DBLCLKS or CS_SAVEBITS or CS_BYTEALIGNWINDOW);
end;

procedure T0DialogWindow.WMCreate(var Message: TWMCreate);
var
	SysMenu: HMENU;
begin
	inherited;

	SysMenu := GetSystemMenu(Handle, False);
	DeleteMenu(SysMenu, SC_TASKLIST, MF_BYCOMMAND);
	DeleteMenu(SysMenu, 7, MF_BYPOSITION);
	DeleteMenu(SysMenu, 5, MF_BYPOSITION);
	DeleteMenu(SysMenu, SC_MAXIMIZE, MF_BYCOMMAND);
	DeleteMenu(SysMenu, SC_MINIMIZE, MF_BYCOMMAND);

	if not WindowStyle[WS_THICKFRAME] then begin
		DeleteMenu(SysMenu, SC_SIZE, MF_BYCOMMAND);
	end;

	DeleteMenu(SysMenu, SC_RESTORE, MF_BYCOMMAND);
end;

procedure T0DialogWindow.WMNCHitTest(var Message: TWMNCXHitMessage);
var
	R: TRect;
begin
	inherited;
	if WindowStyle[WS_THICKFRAME] then begin
		R := ClientRect;
		R.Left := R.Right - GetSystemMetrics(SM_CXVSCROLL);
		R.Top := R.Bottom - GetSystemMetrics(SM_CYHSCROLL);
		if PtInRect(R, ScreenToClient(Point(Message.XCursor, Message.YCursor))) then begin
			Message.Result := HTBOTTOMRIGHT;
		end
	end;
end;

procedure T0DialogWindow.Setup;
var
	P: HWND;
	R, ParentRect, ScreenRect: TRect;
	X, Y: Integer;
begin
	GetWindowRect(Handle, R);

	P := Windows.GetWindow(Handle, GW_OWNER);
	if (P <> 0) and IsWindowVisible(P) then begin
		GetWindowRect(P, ParentRect);
		ScreenRect := GetActiveMonitorRect(P);
	end
	else begin
		ScreenRect := GetActiveMonitorRect(FForegroundWindow);
		ParentRect := ScreenRect;
	end;

	X := ((ParentRect.Right  - ParentRect.Left) - (R.Right - R.Left)) div 2 + ParentRect.Left;
	Y := ((ParentRect.Bottom - ParentRect.Top ) - (R.Bottom - R.Top)) div 2 + ParentRect.Top;

	if X < ScreenRect.Left then X := ScreenRect.Left;
	if Y < ScreenRect.Top then Y := ScreenRect.Top;

	SetWindowPos(Handle, 0, X, Y, 0, 0, SWP_NOACTIVATE or SWP_NOSIZE or SWP_NOZORDER);
end;

procedure T0DialogWindow.WMActivate(var Message: TWMActivate);
begin
	inherited;
	if not ImeAllowed then begin
		EnableIme(Handle, Message.Active = 0);
	end
end;

procedure T0DialogWindow.MnemonicKey(var Message: TWMKey);
begin
	if Message.CharCode = 27 then begin
		SendMessageW(WM_SYSCOMMAND, SC_CLOSE, 0);
		Message.Result := 1;
	end
	else inherited;
end;

procedure T0DialogWindow.PaintWindow(DC: HDC);
var
	R: TRect;
begin
	if WindowStyle[WS_THICKFRAME] then begin
		R := ClientRect;
		R.Left := R.Right - GetSystemMetrics(SM_CXVSCROLL);
		R.Top := R.Bottom - GetSystemMetrics(SM_CYHSCROLL);
		DrawFrameControl(DC, R, DFC_SCROLL, DFCS_SCROLLSIZEGRIP);
	end;
	inherited;
end;

function T0DialogWindow.Valid: Boolean;
begin
	Result := (@FQuery = nil) or FQuery(FContext)
end;

{ T0MainWindow }

constructor T0MainWindow.Create(
				const AText: WideString; const APosition: TPoint; const ASize: TSize);
var
	CS: PCreateStructW;
begin
	inherited Create(nil, AText, APosition, ASize);

	CS := CreateStruct;
	CS.style := CS.style and not (WS_VISIBLE or WS_POPUP);

	Assert(Application <> nil);
	Assert(CurrentThread <> nil);

	T0GUIThread(CurrentThread).FMainWindow := Self;
end;

procedure T0MainWindow.Setup;
begin
	if Text = '' then
		Text := Application.Title;
	ShowWindow(Handle, CmdShow)
end;

class procedure T0MainWindow.CreateClass(var WndClass: WNDCLASSEXW);
begin
	inherited;
	WndClass.hIcon := LoadIconW(MainInstance, 'MAINICON');
end;

class function T0MainWindow.GetDefaultSize: TSize; {override}
begin
	Result := DefaultSize;
end;

{ T0Thread }

procedure ThreadInitialize(Parameter: Pointer); cdecl;
var
	Thread: T0Thread;
begin
	Thread := Parameter;
	Thread.Initialize;
end;

function ThreadProc(Parameter: Pointer): Integer;
var
	Thread: T0Thread;
begin
	Thread := Parameter;
	Application.FUpdateSynchronizer.Accept(Thread.FInitializationEntry^, ThreadInitialize, Thread);
	try
		Thread.Execute;
	except
	on EAbort do
		;
	else
		ShowException(ExceptObject, ExceptAddr);
	end;
	Thread.Finalize;

	{E͂ACXbhŊĎĂ炤}
	SetEvent(Application.FThreadMonitorEvent);

	Result := 0;
end;

procedure T0Thread.BeforeDestruction;
begin
	{ɃCXbhŎs}
	if Application.FMainThread <> Self then
	begin
		Application.ThreadMonitorSynchronizer.BeginWrite;
		RemoveFromLinkList(Application.FFirstThread, Application.FLastThread,
					Self, Self.FPredThread, Self.FSuccThread);
		Application.ThreadMonitorSynchronizer.EndWrite;
	end;

	CloseHandle(Handle);
	inherited;
end;

constructor T0Thread.Create;
begin
	inherited Create;
	Run;
end;

constructor T0Thread.Create(Application: T0Application);
begin
	inherited Create;
	Initialize;
end;

procedure T0Thread.Finalize;
begin
	FreeObjects;
	FObservable := nil;
	FObservableObject.Free;
	CurrentThread := nil;
	LObjects.CurrentThread := nil;
end;

procedure T0Thread.Run;
var
	Id: DWORD;
	Entry: TApplicationSynchronizerEntry;
	ThreadHandle: THandle;
begin
	FInitializationEntry := @Entry;
	Entry.Initialize;
	ThreadHandle := BeginThread(nil, 0, ThreadProc, Self, 0, ID);
	Application.FUpdateSynchronizer.Rendezvous(ThreadHandle, Entry);
	Entry.Finalize;
	CloseHandle(ThreadHandle);

	Application.ThreadMonitorSynchronizer.BeginWrite;
	AddToLinkList(Application.FFirstThread, Application.FLastThread,
					Self, FPredThread, FSuccThread);
	Application.ThreadMonitorSynchronizer.EndWrite;
end;

type
	TThreadNameInfo = record
		FType: LongWord;
		FName: PChar;
		FThreadID: LongWord;
		FFlags: LongWord;
	end;

class procedure T0Thread.SetName(const Name: string);
var
	ThreadNameInfo: TThreadNameInfo;
begin
	ThreadNameInfo.FType := $1000;
	ThreadNameInfo.FName := PChar(Name);
	ThreadNameInfo.FThreadID := $FFFFFFFF;
	ThreadNameInfo.FFlags := 0;
	try
  	{BDS̃fobKɒʒm}
		RaiseException($406D1388, 0, sizeof(ThreadNameInfo) div sizeof(LongWord), PDWORD(@ThreadNameInfo));
	except
	end;
end;

procedure T0Thread.ProcessIdle;
begin
	{ANVȂǂXV}
	try
		FObservableObject.Notify;
	except
		ShowException(ExceptObject, ExceptAddr);
	end;
end;

procedure T0Thread.Initialize;
begin
	FID := GetCurrentThreadId;
	FHandle := DuplicateHandle(GetCurrentThread);

	CurrentThread := Self;
	LObjects.CurrentThread := Self;

	FObservableObject := TObservable.Create;
	FObservable := FObservableObject;

	Garbage := TGarbage.Create;
	Garbage.Owner := FObservableObject;
end;

{ T0GUIThread }

const
	WM_LW_RENDEZVOUS = WM_USER + 2;

function RendezvousWindowProc(Wnd: HWND; Msg: UINT; WP: WParam; LP: LParam): LRESULT; stdcall;
begin
	case Msg of
	WM_LW_RENDEZVOUS:
	begin
		Application.FUpdateSynchronizer.Accept(PApplicationSynchronizerEntry(LP)^);
		Result := 0;
	end;
	else
		Result := DefWindowProcW(Wnd, Msg, WP, LP);
	end;
end;

function T0GUIThread.ActiveKeyManager: IKeyManager;
begin
	{ĂяoƂGUIɌ܂Ă邵Aj[LăEBhEȂ}
	Assert(CurrentThread is T0GUIThread);

	if T0GUIThread(CurrentThread).LeadWindow <> nil then
		Supports(T0GUIThread(CurrentThread).LeadWindow, IKeyManager, Result)
end;

procedure T0GUIThread.ReadyRendezvous;
const
	ClassName: array[0..10] of WIDECHAR = ('R', 'E', 'N', 'D', 'E', 'Z', 'V', 'O', 'U', 'S', #0);
var
	WindowClass: TWndClassExW;
begin
	WindowClass.cbSize := SizeOf(WindowClass);
	WindowClass.style := 0;
	WindowClass.lpfnWndProc := @RendezvousWindowProc;
	WindowClass.cbClsExtra := 0;
	WindowClass.cbWndExtra := 0;
	WindowClass.hInstance := MainInstance;
	WindowClass.hIcon := 0;
	WindowClass.hCursor := 0;
	WindowClass.hbrBackground := 0;
	WindowClass.lpszMenuName := nil;
	WindowClass.lpszClassName := ClassName;
	WindowClass.hIconSm := 0;
	RegisterClassExW(WindowClass);
	FRendezvousWindow := CreateWindowExW(0, ClassName, '', 0, 0, 0, 0, 0, 0, 0, MainInstance, nil);
	Assert(FRendezvousWindow <> 0);
end;

function T0GUIThread.RegisterCommand(const AItem: TObject): UINT;
var
	I: Integer;
begin
	for I := Low(FCommands) to High(FCommands) do
		if FCommands[I] = nil then
		begin
			FCommands[I] := AItem;
			Result := I + 1024;
			Exit;
		end;

	SetLength(FCommands, Length(FCommands) + 1);
	FCommands[High(FCommands)] := AItem;
	Result := High(FCommands) + 1024;
end;

procedure T0GUIThread.Rendezvous(Entry: PApplicationSynchronizerEntry);
begin
	Assert(Self <> CurrentThread);
	Assert(FRendezvousWindow <> 0);
	Windows.PostMessageW(FRendezvousWindow, WM_LW_RENDEZVOUS, 0, LPARAM(Entry));
	Application.FUpdateSynchronizer.Rendezvous(Handle, Entry^);
end;

procedure T0GUIThread.UnregisterCommand(const ACommand: UINT);
begin
	Assert(FCommands[ACommand - 1024] <> nil);

	FCommands[ACommand - 1024] := nil;
end;

procedure T0GUIThread.Execute;
begin
	if MainWindow <> nil then
	begin
		Application.BeginRead;
		try
			MainWindow.Setup;
			MainWindow.MessageLoop;
		finally
			Application.EndRead;
		end
	end;
end;

function T0GUIThread.Fire: DWORD;
begin
	Result := INFINITE
end;

procedure T0GUIThread.WaitMessage(var Msg: TMsg);
var
	WaitTime: DWORD;
	Msg2: TMsg;
	Signed: THandle;
begin
	Application.Async;
	Application.ProcessIdle;
	repeat
		WaitTime := Fire;
		case WaitFor([], Signed, WaitTime, [wfInputAvailable], QS_ALLINPUT) of
		wrInput:
		begin
			GetMessageW(Msg, 0, 0, 0);
			case Msg.message of
			WM_LW_UPDATEGUI:
				if not PeekMessageW(Msg2, 0, 0, 0, PM_NOREMOVE) then
				begin
					Application.BeginRead;
					ProcessIdle;
					LeadWindow.ProcessIdle;
					Garbage.Collect;
					Application.EndRead;
				end;
			else
				Break;
			end;
		end
		else
			Application.ProcessIdle;
		end;
	until False;
	Application.RestoreAsync;
end;

function T0GUIThread.PostMessageW(Msg: UINT; WP: WPARAM; LP: LPARAM): Boolean;
begin
	Result := PostThreadMessageW(ID, Msg, WP, LP)
end;

{ T0Application }

procedure T0Application.AfterConstruction;
begin
	inherited;

	FThreadMonitorSynchronizer.Initialize;
	FUpdateSynchronizer.Initialize;

	FThreadMonitorEvent := CreateManualEvent(False);

	FObservableObject := TObservable.Create;
	FObservable := FObservableObject;

	LObjects.Application := Self;
	Application := Self;
end;

procedure T0Application.Async;
begin
	FUpdateSynchronizer.Async
end;

procedure T0Application.BeforeDestruction;
begin
	if FMainThread <> nil then
	begin
		FMainThread.Finalize;
		FMainThread.Free;
		FMainThread := nil;
	end;

	FreeObjects;
	FObservable := nil;
  FObservableObject.Free;

	FThreadMonitorSynchronizer.Finalize;
	FUpdateSynchronizer.Finalize;

	CloseHandle(FThreadMonitorEvent);

	LObjects.Application := nil;
	Application := nil;
	inherited;
end;

procedure T0Application.BeginRead;
begin
	FUpdateSynchronizer.BeginRead
end;

function T0Application.BeginWrite: Boolean;
begin
	Result := FUpdateSynchronizer.BeginWrite
end;

procedure T0Application.EndRead;
begin
	FUpdateSynchronizer.EndRead
end;

procedure T0Application.EndWrite;
begin
	FUpdateSynchronizer.EndWrite
end;

function T0Application.EnumerateThreads(Block: CEachThread;
	Context: Pointer): T0Thread;
var
	Found: Boolean;
begin
	ThreadMonitorSynchronizer.BeginRead;
	Result := FFirstThread;
	while Result <> nil do
	begin
		ThreadMonitorSynchronizer.EndRead;
		Found := Block(Result, Context); {EndReadĂ̂ŗONĂOK}
		ThreadMonitorSynchronizer.BeginRead;
		if Found then Break;
		Result := Result.FSuccThread;
	end;
	ThreadMonitorSynchronizer.EndRead;
end;

class function T0Application.ExeName: WideString;
var
	Buffer: array[0..MAX_PATH] of WideChar;
begin
	SetString(Result, Buffer, GetModuleFileNameW(0, Buffer, Length(Buffer)))
end;

procedure T0Application.Initialize(AMainThread: T0Thread);
begin
	FMainThread := AMainThread;
  T0Thread.SetName('Program');
end;

procedure T0Application.MonitorThreads;
var
	I, Next: T0Thread;
begin
	Assert(MainThreadID = GetCurrentThreadId);
	ThreadMonitorSynchronizer.BeginWrite;
  ResetEvent(FThreadMonitorEvent);
	I := FFirstThread;
	while I <> nil do
	begin
		Next := I.FSuccThread;
		if WaitForSingleObject(I.Handle, 0) = WAIT_OBJECT_0 then I.Free;
		I := Next;
	end;
	ThreadMonitorSynchronizer.EndWrite;
end;

procedure T0Application.ProcessIdle;
begin
	if FUpdateSynchronizer.TryBeginWrite then
	begin
		try
			FObservableObject.Notify;
			RequestAllThreads;
		except
			ShowException(ExceptObject, ExceptAddr);
		end;
		if MainThreadID = GetCurrentThreadId then
		begin
			if WaitFor(FThreadMonitorEvent, 0) = wrSignaled then MonitorThreads
		end;
		FUpdateSynchronizer.EndWrite;
	end
end;

procedure T0Application.RequestAllThreads;
var
	I: T0Thread;
begin
	ThreadMonitorSynchronizer.BeginRead;
	I := FFirstThread;
	while I <> nil do
	begin
		if I is T0GUIThread then
			T0GUIThread(I).PostMessageW(WM_LW_UPDATEGUI, 0, 0);
		I := I.FSuccThread;
	end;
	ThreadMonitorSynchronizer.EndRead;
end;

procedure T0Application.RestoreAsync;
begin
	FUpdateSynchronizer.RestoreAsync
end;

procedure T0Application.Run;
begin
	if FMainThread <> nil then FMainThread.Execute;
	WaitLoop;
end;

function T0Application.ThreadIdToObject(ThreadId: DWORD): T0Thread;
var
	I: T0Thread;
begin
	ThreadMonitorSynchronizer.BeginRead;
	Result := nil;
	I := FFirstThread;
	while I <> nil do
	begin
		if I.ID = ThreadId then
		begin
			Result := I;
			Break;
		end;
		I := I.FSuccThread;
	end;
	ThreadMonitorSynchronizer.EndRead;
end;

function T0Application.TryBeginWrite: Boolean;
begin
	Result := FUpdateSynchronizer.TryBeginWrite
end;

procedure T0Application.WaitLoop;
var
	Events: array[0..1] of THandle;

	procedure SetEvents;
	begin
		ThreadMonitorSynchronizer.BeginRead;
		Events[0] := 0;
		if FFirstThread <> nil then Events[0] := FFirstThread.Handle;
		Events[1] := FThreadMonitorEvent;
		ThreadMonitorSynchronizer.EndRead;
	end;

var
	Msg: TMsg;
	Signed: THandle;
begin
  SetEvents;
	while Events[0] <> 0 do
	begin
		case WaitFor(Events, Signed, INFINITE, [wfInputAvailable], QS_ALLINPUT) of
		wrSignaled:
		begin
			if Application.TryBeginWrite then
			begin
				MonitorThreads;
				Application.EndWrite;
			end;
			if Garbage <> nil then
			begin
				Application.BeginRead;
				Garbage.Collect;
				Application.EndRead;
			end;
      SetEvents;
		end;
		wrInput:
		begin
			{ꉞbZ[W[v񂵂Ă}
			while PeekMessageW(Msg, 0, 0, 0, PM_REMOVE) do
			begin
				TranslateMessage(Msg);
				DispatchMessageW(Msg);
			end;
		end;
		end;
	end;
end;

class function T0Application.ExePath: WideString;
begin
	Result := ExtractFilePath(ExeName)
end;

{ Supports }

function GetWnd(Wnd: HWND): TAnyWindow;

	function FindProc(Obj: T0Object): Boolean cdecl;
	begin
		Result := (Obj is TAnyWindow) and (TAnyWindow(Obj).Handle = Wnd)
	end;

begin
	if not IsWindow(Wnd) then
		Result := nil
	else
	begin
		Result:= GetLWnd(Wnd);
		if Result = nil then
		begin
			Result := TAnyWindow(Garbage.EnumerateObjects(CEachObject(@FindProc), GetContext));
			if Result = nil then
			begin
				Result := TAnyWindow.Create(Wnd);
				Result.Owner := Garbage;
			end
		end
	end
end;

function GetLWnd(Wnd: HWND): T0Window;
begin
	if not IsWindow(Wnd) or (GetWindowThreadProcessId(Wnd) <> GetCurrentThreadId) then
		Result := nil
	else
		Result := T0Window(GetPropW(Wnd, LWPropName))
end;

function GetLWndThread(Wnd: HWND): T0GUIThread;
var
	ThreadId, ProcessId: DWORD;
begin
	if IsWindow(Wnd) then
	begin
		ThreadId := GetWindowThreadProcessId(Wnd, ProcessId);
		if ProcessId = GetCurrentProcessId then
		begin
			if ThreadId = CurrentThread.Id then
				Result := CurrentThread as T0GUIThread
			else
				Result := Application.ThreadIdToObject(ThreadId) as T0GUIThread;

			Exit;
		end;
	end;

	Result := nil;
end;

function GetLMenuItem(Command: UINT): TMenuItem;
begin
	if Command = 0 then
		Result := nil
	else
		Result := GetObjectFromCommand(Command) as TMenuItem;
end;

function GetObjectFromCommand(Command: UINT): TObject;
var
	Thread: T0GUIThread;
{$IFOPT C+}
	S: array[0..255] of WideChar;
{$ENDIF}
begin
	if (Command < $400) or (Command >= $f000) then
		Result := nil
	else
	begin
		Thread := T0GUIThread(CurrentThread);
		Dec(Command, $400);
{$IFOPT C+}
		if Command > UINT(High(Thread.FCommands)) then
		begin
			wsprintfW(S, 'Bad command id %d in thread %d.', Command, Thread.ID);
			OutputDebugStringW(S);
			Result := nil;
		end
		else
{$ENDIF}
		begin
			Result := Thread.FCommands[Command];
		end	
	end;	
end;

function SystemFont: TAnyFont;
var
	T: T0GUIThread;
	Font: TAnyFont;
begin
	Assert(CurrentThread is T0GUIThread);
	T := T0GUIThread(CurrentThread);
	Font := T.FSystemFont;
	if Font = nil then
	begin
		Font := TAnyFont.Create(GetStockObject(SYSTEM_FONT));
		Font.Owner := T;
		T.FSystemFont := Font;
	end;
	Result := Font
end;

function FixedFont: TAnyFont;
var
	T: T0GUIThread;
	Font: TAnyFont;
begin
	Assert(CurrentThread is T0GUIThread);
	T := T0GUIThread(CurrentThread);
	Font := T.FFixedFont;
	if Font = nil then
	begin
		Font := TAnyFont.Create(GetStockObject(SYSTEM_FIXED_FONT));
		Font.Owner := T;
		T.FFixedFont := Font;
	end;
	Result := Font
end;

function DefaultGUIFont: TAnyFont;
var
	T: T0GUIThread;
	Font: TAnyFont;
begin
	Assert(CurrentThread is T0GUIThread);
	T := T0GUIThread(CurrentThread);
	Font := T.FDefaultGUIFont;
	if Font = nil then
	begin
		Font := TAnyFont.Create(GetStockObject(DEFAULT_GUI_FONT));
		Font.Owner := T;
		T.FDefaultGUIFont := Font;
	end;
	Result := Font
end;

function MenuFont: TAnyFont;
var
	T: T0GUIThread;
	Metrics: NONCLIENTMETRICSW;
	Font: TAnyFont;
begin
	Assert(CurrentThread is T0GUIThread);
	T := T0GUIThread(CurrentThread);
	Font := T.FMenuFont;
	if Font = nil then
	begin
		Metrics.cbSize := SizeOf(Metrics);
		SystemParametersInfoW(SPI_GETNONCLIENTMETRICS, SizeOf(Metrics), @Metrics, 0);
		Font := TFont.Create;
		TFont(Font).Assign(Metrics.lfMenuFont);
		Font.Owner := T;
		T.FMenuFont := Font;
	end;
	Result := Font;
end;

function MessageFont: TAnyFont;
var
	T: T0GUIThread;
	Metrics: NONCLIENTMETRICSW;
	Font: TAnyFont;
begin
	Assert(CurrentThread is T0GUIThread);
	T := T0GUIThread(CurrentThread);
	Font := T.FMessageFont;
	if Font = nil then
	begin
		Metrics.cbSize := SizeOf(Metrics);
		SystemParametersInfoW(SPI_GETNONCLIENTMETRICS, SizeOf(Metrics), @Metrics, 0);
		Font := TFont.Create;
		TFont(Font).Assign(Metrics.lfMessageFont);
		Font.Owner := T;
		T.FMessageFont := Font;
	end;
	Result := Font;
end;

function StatusFont: TAnyFont;
var
	T: T0GUIThread;
	Metrics: NONCLIENTMETRICSW;
	Font: TAnyFont;
begin
	Assert(CurrentThread is T0GUIThread);
	T := T0GUIThread(CurrentThread);
	Font := T.FStatusFont;
	if Font = nil then
	begin
		Metrics.cbSize := SizeOf(Metrics);
		SystemParametersInfoW(SPI_GETNONCLIENTMETRICS, SizeOf(Metrics), @Metrics, 0);
		Font := TFont.Create;
		TFont(Font).Assign(Metrics.lfStatusFont);
		Font.Owner := T;
		T.FStatusFont := Font;
	end;
	Result := Font;
end;

{ Utilities }

function WINNLSEnableIME(Wnd: HWND; ABool: BOOL): BOOL; stdcall;
				external User32;

procedure EnableIme(Wnd: HWND; APermission: Boolean);
var
	IMC: HIMC;
begin
	if not APermission then
	begin
		IMC := ImmGetContext(Wnd);
		if IMC <> 0 then
		begin
			ImmSetOpenStatus(IMC, False);
			ImmReleaseContext(Wnd, IMC);
		end;
	end;
	WinNLSEnableIME(Wnd, APermission);
end;

function SetImeComposition(Window: HWND; Font: HFONT; X, Y: Integer): Boolean;
var
	H: HIMC;
	CForm: TCompositionForm;
	LFont: TLogFontW;
begin
	Result := False;
	H := ImmGetContext(Window);
	if H <> 0 then
	begin
		with CForm do
		begin
			dwStyle := CFS_POINT;
			ptCurrentPos.x := X;
			ptCurrentPos.y := Y;
		end;
		ImmSetCompositionWindow(H, @CForm);
		if Font <> 0 then
		begin
			GetObjectW(Font, SizeOf(TLogFont), @LFont);
			ImmSetCompositionFontW(H, @LFont);
		end;
		ImmReleaseContext(Window, H);
		Result := True;
	end;
end;

function GetShiftState: TShiftKeys;
var
	S: TKeyboardState;
begin
	GetKeyboardState(S);

	Result := [];
	if S[VK_SHIFT]   and $80 <> 0 then Include(Result, ssShift);
	if S[VK_CONTROL] and $80 <> 0 then Include(Result, ssControl);
	if S[VK_MENU]    and $80 <> 0 then Include(Result, ssAlt);
end;

function Size(CX, CY: Integer): TSize;
begin
	Result.cx := CX;
	Result.cy := CY
end;

function GetMnemonic(const AText: WideString): DWORD;
var
	P: PWideChar;
begin
	P:= PWideChar(AText);
	repeat
		P:= WideStrScan(P, '&');
		if P = nil then begin Result := 0; Exit; end;
		Inc(P);
	until P^ <> '&';
	Result := Ord(WideUpCase(P^));
end;

function InMainThread: Boolean;
begin
	Result := GetCurrentThreadId = MainThreadID
end;

function TemporaryPen(const Color: TColor): TAnyPen;
begin
	Result := TPen.Create;
	TPen(Result).Color := Color.Value;
	Result.Owner := Garbage;
end;	

function TemporaryBrush(const Color: TColor): TAnyBrush;
begin
	if Color.SysColor < 0 then
	begin
		Result := TBrush.Create;
		TBrush(Result).Color := Color.Value;
	end
	else
	begin
		Result := TAnyBrush.Create(GetSysColorBrush(Color.SysColor))
	end;
	Result.Owner := Garbage;
end;

function LoadImage(const FileName: WideString): HBITMAP;
begin
	Result := LoadImageW(0, PWideChar(FileName), IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE);
end;

function SendWMCopyData(Window: HWND; MagicNumber: DWORD; Handle: THandle; Data: Pointer; Size: Cardinal): LRESULT;
var
	Struct: COPYDATASTRUCT;
begin
	Struct.dwData := MagicNumber;
	Struct.cbData := Size;
	Struct.lpData := Data;
  Result := SendMessageW(Window, WM_COPYDATA, Handle, LPARAM(@Struct))
end;	

function SendWMCopyData(Window: HWND; MagicNumber: DWORD; Handle: THandle; const Data: WideString): LRESULT; 
var
	Struct: COPYDATASTRUCT;
begin
	Struct.dwData := MagicNumber;
	Struct.cbData := Length(Data) * SizeOf(WideChar);
	Struct.lpData := Pointer(Data);
  Result := SendMessageW(Window, WM_COPYDATA, Handle, LPARAM(@Struct))
end;	

procedure Realign(Parent: THandle; var R: TRect);

	function EnumChildProc_Realign_Vert(Wnd: HWND): Boolean; cdecl;
	label
		BreakRutine;
	var
		WR: TRect;
		W: T0Window;
		H: Integer;
	begin
		if GetParent(Wnd) = Parent then
		begin
			W := GetLWnd(WND);
			if (W <> nil) and W.WindowStyle[WS_VISIBLE] then
			begin
				WR:= R;
				H:= W.BoundsSize.CY;
				case W.Alignment of
				alTop:    begin WR.Bottom:= WR.Top + H; R.Top:= WR.Bottom; end;
				alBottom: begin WR.Top:= WR.Bottom - H; R.Bottom:= WR.Top; end;
				else      goto BreakRutine;
				end;
				W.BoundsRect := WR;
			end;
		end;
	BreakRutine:
		Result := False;
	end;

	function EnumChildProc_Realign_Horz(Wnd: HWND): Boolean; cdecl;
	label
		BreakRutine;
	var
		WR: TRect;
		W: T0Window;
		H: Integer;
	begin
		if GetParent(Wnd) = Parent then
		begin
			W := GetLWnd(WND);
			if (W <> nil) and W.WindowStyle[WS_VISIBLE] then
			begin
				WR:= R;
				H:= W.BoundsSize.CX;
				case W.Alignment of
				alLeft:   begin WR.Right:= WR.Left + H; R.Left:= WR.Right; end;
				alRight:  begin WR.Left:= WR.Right - H; R.Right:= WR.Left; end;
				else      goto BreakRutine;
				end;
				W.BoundsRect:= WR;
			end;
		end;
	BreakRutine:
		Result := False;
	end;

	function EnumChildProc_Realign_Client(Wnd: HWND): Boolean; cdecl;
	var
		W: T0Window;
	begin
		if GetParent(Wnd) = Parent then
		begin
			W := GetLWnd(WND);
			if (W <> nil) and W.WindowStyle[WS_VISIBLE] then
				if W.Alignment = alCenter then W.BoundsRect:= R;
		end;
		Result := False;
	end;

var
	Context: Pointer;
begin
	Context := GetContext;
	EnumerateChildWindows(Parent, CEachWindow(@EnumChildProc_Realign_Vert)  , Context);
	EnumerateChildWindows(Parent, CEachWindow(@EnumChildProc_Realign_Horz)  , Context);
	EnumerateChildWindows(Parent, CEachWindow(@EnumChildProc_Realign_Client), Context);
end;

function CreatePropName: PWideChar;
var
	W: array[0..10] of WideChar;
begin
	W[0] := 'L';
	W[1] := 'W';
	wsprintfW(W + 2, '%.8lx', GetCurrentProcessID);
	Result := PWideChar(AddAtomW(W));
end;

initialization

	LWPropName := CreatePropName;

finalization

	Application.Free;
	DeleteAtom(ATOM(LWPropName));

end.

