unit LZipStreams;

interface

uses
	ZLibPas,
	LStreams;

type
	TZipStream = class(T0Stream)
	private
		FZBuffer: array[Word] of Char;
		FZStreamRec: z_stream;
		FDataStream: T0Stream;
	protected
		function GetWritable: Boolean; override;
	public
		constructor Create(ADataStream: T0Stream; ALevel: Integer = Z_DEFAULT_COMPRESSION);
		property DataStream: T0Stream read FDataStream;
		function Write(const Buffer; Count: Cardinal): Cardinal; override;
		procedure BeforeDestruction; override;
	end;

	TUnzipStream = class(T0Stream)
	private
		FZBuffer: array[Word] of Char;
		FZStreamRec: z_stream;
		FDataStream: T0Stream;
		FBuffer: Pointer;
		FBufferSize: Cardinal;
		FBufferRest: Cardinal;
	protected
		function GetReadable: Boolean; override;
		function Input(var Buffer; Count: Cardinal): Cardinal;
	public
		constructor Create(ADataStream: T0Stream);
		property DataStream: T0Stream read FDataStream;
		procedure BeforeDestruction; override;
		function Read(var Buffer; Count: Cardinal): Cardinal; override;
		function Refer(var Buffer; Count: Cardinal): Cardinal; override;
	end;

implementation

function ZCheck(Code: Integer; var Clear: Boolean): Integer;
begin
	Result := Code;
	Clear := Code >= 0;
end;

{ TZipStream }

procedure TZipStream.BeforeDestruction;
var
	Check: Boolean;
begin
	FZStreamRec.next_in := nil;
	FZStreamRec.avail_in := 0;
	while (ZCheck(deflate(FZStreamRec, Z_FINISH), Check) <> Z_STREAM_END) and (FZStreamRec.avail_out = 0) do
	begin
		if not Check then Break;
		if FDataStream.Write(FZBuffer, SizeOf(FZBuffer)) <> SizeOf(FZBuffer) then Break;
		FZStreamRec.next_out := FZBuffer;
		FZStreamRec.avail_out := SizeOf(FZBuffer);
	end;
	if FZStreamRec.avail_out < SizeOf(FZBuffer) then
		FDataStream.Write(FZBuffer, SizeOf(FZBuffer) - FZStreamRec.avail_out);

	deflateEnd(FZStreamRec);

	inherited;
end;

constructor TZipStream.Create(ADataStream: T0Stream; ALevel: Integer = Z_DEFAULT_COMPRESSION);
var
	Inited: Boolean;
begin
	inherited Create;

	FDataStream := ADataStream;

	FZStreamRec.next_out := FZBuffer;
	FZStreamRec.avail_out := SizeOf(FZBuffer);
	ZCheck(deflateInit_(FZStreamRec, ALevel, ZLib_Version, SizeOf(FZStreamRec)), Inited);
	Assert(Inited);
end;

function TZipStream.GetWritable: Boolean;
begin
	Result := True
end;

function TZipStream.Write(const Buffer; Count: Cardinal): Cardinal;
var
	Check: Boolean;
begin
	FZStreamRec.next_in := @Buffer;
	FZStreamRec.avail_in := Count;
	while FZStreamRec.avail_in > 0 do
	begin
		ZCheck(deflate(FZStreamRec, 0), Check);
		if not Check then
		begin
			Result := 0;
			Exit;
		end;
		if FZStreamRec.avail_out = 0 then
		begin
			if FDataStream.Write(FZBuffer, SizeOf(FZBuffer)) <> SizeOf(FZBuffer) then
			begin
				Result := 0;
				Exit;
			end;
			FZStreamRec.next_out := FZBuffer;
			FZStreamRec.avail_out := SizeOf(FZBuffer);
		end;
	end;
	Result := Count;
end;

{ TUnzipStream }

procedure TUnzipStream.BeforeDestruction;
begin
	InflateEnd(FZStreamRec);
	if FBuffer <> nil then FreeMem(FBuffer);
	inherited;
end;

constructor TUnzipStream.Create(ADataStream: T0Stream);
var
	Inited: Boolean;
begin
	inherited Create;

	FDataStream := ADataStream;

	FZStreamRec.next_out := FZBuffer;
	FZStreamRec.avail_out := SizeOf(FZBuffer);
	ZCheck(inflateInit_(FZStreamRec, ZLib_Version, SizeOf(FZStreamRec)), Inited);
	Assert(Inited);
end;

function TUnzipStream.GetReadable: Boolean;
begin
	Result := True
end;

function TUnzipStream.Input(var Buffer; Count: Cardinal): Cardinal;
var
	Check: Boolean;
begin
	FZStreamRec.next_out := @Buffer;
	FZStreamRec.avail_out := Count;

	while FZStreamRec.avail_out > 0 do
	begin
		if FZStreamRec.avail_in = 0 then
		begin
			FZStreamRec.avail_in := FDataStream.Read(FZBuffer, SizeOf(FZBuffer));
			if FZStreamRec.avail_in = 0 then
			begin
				Result := Count - Cardinal(FZStreamRec.avail_out);
				Exit;
			end;
			FZStreamRec.next_in := FZBuffer;
		end;
		ZCheck(Inflate(FZStreamRec, 0), Check);
		if not Check then
		begin
			Result := 0;
			Exit;
		end
	end;
	Result := Count;
end;

function TUnzipStream.Read(var Buffer; Count: Cardinal): Cardinal;
begin
	Assert(Readable);

	//Result:= Min(FBufferRest, Count)
	Result:= FBufferRest;
	if Result > Count then Result:= Count;

	//obt@Rs[
	Move(Pointer(Cardinal(FBuffer) + FBufferSize - FBufferRest)^, Buffer, Result);
	Dec(FBufferRest, Result);

	//cۂɓǂ
	Dec(Count, Result);
	if Count > 0 then
		Inc(Result, Input(Pointer(Cardinal(@Buffer) + Result)^, Count));
end;

function TUnzipStream.Refer(var Buffer; Count: Cardinal): Cardinal;
const
	NextSize = 2048; //must be 2^X
var
	AddSize, NewBufferSize, ReadSize: Cardinal;
begin
	Assert(Readable);

	Result:= Count;

	if Result > FBufferRest then
	begin
		//obt@łΏk
		if FBufferSize - FBufferRest > NextSize * 2 then
		begin
			Move(Pointer(Cardinal(FBuffer) + FBufferSize - FBufferRest)^,
							FBuffer^, FBufferRest);
			FBufferSize:= FBufferRest;
		end;

		//ǉvZ(Vǂݍޕ 2^X ̔{ɂȂ悤ɂ)
		AddSize:= (Count - FBufferRest + NextSize) and not (NextSize - 1);
		NewBufferSize:= FBufferSize + AddSize;

		//[ăAP[g
		ReallocMem(FBuffer, NewBufferSize);

		//ǉǂݍ
		ReadSize:= Input(Pointer(Cardinal(FBuffer) + FBufferSize - FBufferRest)^, AddSize);

		Inc(FBufferSize, ReadSize);
		Inc(FBufferRest, ReadSize);

		//[ǂݐ؂Ȃꍇ΍
		if Result >= FBufferRest then Result:= FBufferRest;
	end;

	Move(Pointer(Cardinal(FBuffer) + FBufferSize - FBufferRest)^, Buffer, Result);
end;

end.
